Type: Package
Package: campsismod
Title: Generic Implementation of a PK/PD Model
Version: 1.3.2
Authors@R: 
    person("Nicolas", "Luyckx", , "nicolas.luyckx@calvagone.com", role = c("aut", "cre"))
Description: A generic, easy-to-use and expandable implementation of a
    pharmacokinetic (PK) / pharmacodynamic (PD) model based on the S4
    class system. This package allows the user to read and write
    pharmacometric models from and to files, including a JSON-based
    interface to import Campsis models defined using a formal JSON schema
    distributed with the package. Models can be adapted further on the fly
    in the R environment using an intuitive API to add, modify or delete
    equations, ordinary differential equations (ODEs), model parameters or
    compartment properties (such as infusion duration or rate,
    bioavailability and initial values). The package also provides export
    facilities for use with the simulation packages 'rxode2' and
    'mrgsolve'. The package itself is licensed under the GPL (>= 3); the
    JSON schema file shipped in inst/extdata is licensed separately under
    the Creative Commons Attribution 4.0 International (CC BY 4.0).  This
    package is designed and intended to be used with the package
    'campsis', a PK/PD simulation platform built on top of 'rxode2' and
    'mrgsolve'.
License: GPL (>= 3)
URL: https://github.com/Calvagone/campsismod,
        https://calvagone.github.io/,
        https://calvagone.github.io/campsismod.doc/
BugReports: https://github.com/Calvagone/campsismod/issues
Depends: R (>= 4.0.0)
Imports: assertthat, dplyr, ggplot2, jsonlite, jsonvalidate,
        LaplacesDemon, magrittr, MASS, methods, purrr, readr, rlang,
        tibble, tidyr, utils
Suggests: devtools, knitr, pkgdown, rmarkdown, roxygen2, testthat, xfun
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
LazyData: true
RoxygenNote: 7.3.3
Collate: 'global.R' 'data.R' 'utilities.R' 'special_operators.R'
        'check.R' 'json_element.R' 'generic.R' 'generic_element_list.R'
        'generic_element_position.R' 'generic_list.R' 'pattern.R'
        'model_statement.R' 'model_statements.R'
        'model_unknown_statement.R' 'model_comment.R'
        'model_line_break.R' 'model_equation.R' 'model_ode.R'
        'model_if_statement.R' 'compartment.R' 'compartment_property.R'
        'compartment_properties.R' 'compartment_bioavailability.R'
        'compartment_lag_time.R' 'compartment_infusion_duration.R'
        'compartment_infusion_rate.R' 'compartment_initial_condition.R'
        'compartments.R' 'parameter.R' 'parameters.R' 'code_record.R'
        'code_records.R' 'model_parser.R' 'campsis_model.R'
        'omega_block.R' 'omega_blocks.R'
        'parameter_uncertainty_utils.R' 'replication_settings.R'
        'replicated_campsis_model.R' 'model_add_suffix.R'
        'rxode_model.R' 'rxode_conversion.R' 'mrgsolve_model.R'
        'mrgsolve_conversion.R' 'json_interface.R'
NeedsCompilation: no
Packaged: 2026-05-07 12:31:51 UTC; nicolas.luyckx
Author: Nicolas Luyckx [aut, cre]
Maintainer: Nicolas Luyckx <nicolas.luyckx@calvagone.com>
Repository: CRAN
Date/Publication: 2026-05-07 14:00:48 UTC
Built: R 4.6.0; ; 2026-05-07 17:18:28 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(":=")
export(AutoReplicationSettings)
export(Bioavailability)
export(CampsisModel)
export(CodeRecords)
export(Comment)
export(Compartment)
export(Compartments)
export(Equation)
export(ErrorRecord)
export(IfStatement)
export(InfusionDuration)
export(InfusionRate)
export(InitialCondition)
export(JSONElement)
export(LagTime)
export(LineBreak)
export(MainRecord)
export(ManualReplicationSettings)
export(ModelStatements)
export(Ode)
export(OdeRecord)
export(Omega)
export(OmegaBlock)
export(OmegaBlocks)
export(Parameters)
export(Pattern)
export(Position)
export(Sigma)
export(Theta)
export(UnknownStatement)
export(VariablePattern)
export(add)
export(addRSE)
export(addSuffix)
export(allNa)
export(as.data.frame)
export(assertSingleCharacterString)
export(autoDetectNONMEM)
export(contains)
export(default)
export(delete)
export(disable)
export(export)
export(exportToJSON)
export(extractLhs)
export(extractRhs)
export(extractTextBetweenBrackets)
export(find)
export(fixOmega)
export(getByIndex)
export(getByName)
export(getCampsismodOption)
export(getCampsismodOptions)
export(getCompartmentIndex)
export(getNONMEMName)
export(getName)
export(getNameInModel)
export(getNames)
export(getOmegaBlock)
export(getOmegaIndexes)
export(getPrefix)
export(getRecordDelimiter)
export(getRecordEquationNames)
export(getRecordName)
export(getUncertainty)
export(getVarCov)
export(hasComment)
export(hasExactMethod)
export(hasOffDiagonalOmegas)
export(indexOf)
export(isComment)
export(isDiag)
export(isEmptyLine)
export(isEquation)
export(isIfStatement)
export(isMatrixPositiveDefinite)
export(isODE)
export(isRecordDelimiter)
export(isRxODE)
export(isStrictRecordDelimiter)
export(jsonToParameter)
export(loadFromJSON)
export(mapJSONPropertiesToS4Slots)
export(mapS4SlotsToJSONProperties)
export(maxIndex)
export(minIndex)
export(move)
export(mrgsolveBlock)
export(mrgsolveCapture)
export(mrgsolveCompartment)
export(mrgsolveMain)
export(mrgsolveMatrix)
export(mrgsolveOde)
export(mrgsolveParam)
export(mrgsolveTable)
export(processExtraArg)
export(read)
export(read.allparameters)
export(read.campsis)
export(read.model)
export(read.parameters)
export(read.pmxmod)
export(read.varcov)
export(removeNaColumn)
export(replace)
export(replaceAll)
export(replicate)
export(rxodeCode)
export(rxodeMatrix)
export(rxodeParams)
export(select)
export(setMinMax)
export(shiftOmegaIndexes)
export(sort)
export(standardise)
export(toString)
export(trim)
export(updateCompartments)
export(write)
exportClasses(auto_replication_settings)
exportClasses(campsis_model)
exportClasses(comment)
exportClasses(compartment)
exportClasses(compartment_bioavailability)
exportClasses(compartment_infusion_duration)
exportClasses(compartment_infusion_rate)
exportClasses(compartment_initial_condition)
exportClasses(compartment_lag_time)
exportClasses(compartment_properties)
exportClasses(compartment_property)
exportClasses(compartments)
exportClasses(double_array_parameter)
exportClasses(duration_record)
exportClasses(equation)
exportClasses(error_record)
exportClasses(export_type)
exportClasses(f_record)
exportClasses(if_statement)
exportClasses(init_record)
exportClasses(json_element)
exportClasses(lag_record)
exportClasses(line_break)
exportClasses(main_record)
exportClasses(manual_replication_settings)
exportClasses(model_statement)
exportClasses(model_statements)
exportClasses(mrgsolve_type)
exportClasses(ode)
exportClasses(ode_record)
exportClasses(omega)
exportClasses(parameter)
exportClasses(parameters)
exportClasses(pattern)
exportClasses(pmx_element)
exportClasses(pmx_list)
exportClasses(pmx_position)
exportClasses(pmx_position_by_element)
exportClasses(pmx_position_by_index)
exportClasses(rate_record)
exportClasses(replicated_campsis_model)
exportClasses(replication_settings)
exportClasses(rxode_type)
exportClasses(sigma)
exportClasses(single_array_parameter)
exportClasses(statements_record)
exportClasses(theta)
exportClasses(unknown_statement)
exportClasses(variable_pattern)
importFrom(LaplacesDemon,rinvchisq)
importFrom(LaplacesDemon,rinvwishart)
importFrom(MASS,mvrnorm)
importFrom(assertthat,are_equal)
importFrom(assertthat,assert_that)
importFrom(dplyr,across)
importFrom(dplyr,any_of)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,group_split)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,starts_with)
importFrom(dplyr,where)
importFrom(ggplot2,aes)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_density)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,theme_bw)
importFrom(jsonlite,parse_json)
importFrom(jsonvalidate,json_schema)
importFrom(magrittr,`%>%`)
importFrom(methods,"slot<-")
importFrom(methods,callNextMethod)
importFrom(methods,findMethods)
importFrom(methods,getGeneric)
importFrom(methods,is)
importFrom(methods,isGeneric)
importFrom(methods,new)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(methods,validObject)
importFrom(purrr,discard)
importFrom(purrr,flatten_chr)
importFrom(purrr,flatten_int)
importFrom(purrr,imap)
importFrom(purrr,keep)
importFrom(purrr,map)
importFrom(purrr,map_chr)
importFrom(purrr,map_dbl)
importFrom(purrr,map_df)
importFrom(purrr,map_lgl)
importFrom(readr,read_delim)
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,`:=`)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,pivot_longer)
importFrom(utils,capture.output)
importFrom(utils,hasName)
importFrom(utils,read.csv)
importFrom(utils,write.csv)
importFrom(utils,write.table)

# campsismod <img src='man/figures/logo.png' align="right" alt="" width="120" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/Calvagone/campsismod/workflows/R-CMD-check/badge.svg)](https://github.com/Calvagone/campsismod/actions)
[![Codecov test
coverage](https://codecov.io/gh/Calvagone/campsismod/graph/badge.svg)](https://app.codecov.io/gh/Calvagone/campsismod)
[![CRAN
version](http://www.r-pkg.org/badges/version/campsismod)](https://cran.r-project.org/package=campsismod)
[![CRAN total
downloads](https://cranlogs.r-pkg.org/badges/grand-total/campsismod)](https://cran.r-project.org/package=campsismod)
[![CRAN total
downloads](https://cranlogs.r-pkg.org/badges/campsismod)](https://cran.r-project.org/package=campsismod)
[![campsismod status
badge](https://calvagone.r-universe.dev/campsismod/badges/version)](https://calvagone.r-universe.dev/campsismod)
<!-- badges: end -->

## Installation

You can install the released version of `campsismod` from CRAN with:

``` r
install.packages("campsismod")
```

Alternatively, the package can also be installed with `devtools`:

``` r
devtools::install_github("Calvagone/campsismod")
```

## Basic examples

### Load example from model library

Load 2-compartment PK model from built-in model library:

``` r
library(campsismod)
model <- model_suite$pk$`2cpt_fo`
```

### Write Campsis model

The model can be exported to files using `write`.

``` r
model %>% write(file="path_to_model_folder")
```

In this case, the model code will be contained in the `model.campsis`
file. Parameters (THETA, OMEGA and SIGMA) will be stored in their
respective CSV file.

``` r
list.files("path_to_model_folder")
#> [1] "model.campsis" "omega.csv"     "sigma.csv"     "theta.csv"
```

Alternatively, the model can also be exported in JSON format into a
single file:

``` r
model %>% write(file="my_model.json")
```

### Read and show Campsis model

The model can be loaded from the previously created folder:

``` r
model <- read.campsis(file="path_to_model_folder")
```

Or, from the previously created JSON file:

``` r
model <- read.campsis(file="my_model.json")
```

The model can then be output in the console using `show`:

``` r
show(model)
#> [MAIN]
#> TVBIO=THETA_BIO
#> TVKA=THETA_KA
#> TVVC=THETA_VC
#> TVVP=THETA_VP
#> TVQ=THETA_Q
#> TVCL=THETA_CL
#> 
#> BIO=TVBIO
#> KA=TVKA * exp(ETA_KA)
#> VC=TVVC * exp(ETA_VC)
#> VP=TVVP * exp(ETA_VP)
#> Q=TVQ * exp(ETA_Q)
#> CL=TVCL * exp(ETA_CL)
#> 
#> [ODE]
#> d/dt(A_ABS)=-KA*A_ABS
#> d/dt(A_CENTRAL)=KA*A_ABS + Q/VP*A_PERIPHERAL - Q/VC*A_CENTRAL - CL/VC*A_CENTRAL
#> d/dt(A_PERIPHERAL)=Q/VC*A_CENTRAL - Q/VP*A_PERIPHERAL
#> 
#> [F]
#> A_ABS=BIO
#> 
#> [ERROR]
#> CONC=A_CENTRAL/VC
#> if (CONC <= 0.001) CONC=0.001
#> CONC_ERR=CONC*(1 + EPS_PROP_RUV)
#> 
#> 
#> THETA's:
#>   name index value   fix                            label unit
#> 1  BIO     1     1 FALSE                  Bioavailability <NA>
#> 2   KA     2     1 FALSE                  Absorption rate  1/h
#> 3   VC     3    10 FALSE    Volume of central compartment    L
#> 4   VP     4    40 FALSE Volume of peripheral compartment    L
#> 5    Q     5    20 FALSE           Inter-compartment flow  L/h
#> 6   CL     6     3 FALSE                        Clearance  L/h
#> OMEGA's:
#>   name index index2 value   fix type
#> 1   KA     1      1    25 FALSE  cv%
#> 2   VC     2      2    25 FALSE  cv%
#> 3   VP     3      3    25 FALSE  cv%
#> 4    Q     4      4    25 FALSE  cv%
#> 5   CL     5      5    25 FALSE  cv%
#> SIGMA's:
#>       name index index2 value   fix type
#> 1 PROP_RUV     1      1   0.1 FALSE   sd
#> No variance-covariance matrix
#> 
#> Compartments:
#> A_ABS (CMT=1)
#> A_CENTRAL (CMT=2)
#> A_PERIPHERAL (CMT=3)
```

### Simulate with rxode2 or mrgsolve

``` r
library(campsis)

dataset <- Dataset(5) %>%
  add(Bolus(time=0, amount=1000, ii=12, addl=2)) %>%
  add(Observations(times=0:36))

rxode <- model %>% simulate(dataset=dataset, dest="rxode2", seed=0)
mrgsolve <- model %>% simulate(dataset=dataset, dest="mrgsolve", seed=0)
```

``` r
spaghettiPlot(rxode, "CONC")
```

<figure>
<img src="vignettes/resources/results_rxode.png"
alt="rxode2 simulation results" />
<figcaption aria-hidden="true">rxode2 simulation results</figcaption>
</figure>

``` r
spaghettiPlot(mrgsolve, "CONC")
```

<figure>
<img src="vignettes/resources/results_mrgsolve.png"
alt="mrgsolve simulation results" />
<figcaption aria-hidden="true">mrgsolve simulation results</figcaption>
</figure>

#_______________________________________________________________________________
#----                          parameters class                             ----
#_______________________________________________________________________________

validateParametersByType <- function(object, type, emptyParameter) {
  params <- object %>% select(type)
  if (params %>% length() == 0) {
    return(character())
  }
  maxIndex <- params %>% maxIndex()
  minIndex <- params %>% minIndex()
  if (is.na(minIndex)) {
    return(paste0("At least one ", type %>% toupper(), " index is NA"))
  }
  if (minIndex != 1) {
    return(paste0("First ", type %>% toupper(), " index is different than 1"))
  }
  for (i in seq_len(maxIndex)) {
    search <- emptyParameter
    if (is(emptyParameter, "double_array_parameter")) {
      search@index <- i
      search@index2 <- i
    } else {
      search@index <- i
    }
    param <- params %>% getByIndex(search)
    if (length(param) == 0) {
      return(paste0("No ", type %>% toupper(), " with index ", i))
    }
    if (is.na(param@value)) {
      return(paste0(type %>% toupper(), " with index ", i, " has NA value"))
    }
  }
  return(character())
}

validateParameters <- function(object) {
  check1 <- validateParametersByType(object, "theta", Theta())
  check2 <- validateParametersByType(object, "omega", Omega())
  check3 <- validateParametersByType(object, "sigma", Sigma())
  return(c(check1, check2, check3))
}

#' 
#' Parameters class.
#' 
#' @slot varcov associated variance-covariance matrix
#' @export
setClass(
  "parameters",
  representation(
    varcov = "matrix"
  ),
  contains = "pmx_list",
  prototype = prototype(type="parameter", varcov=matrix(numeric(0), nrow=0, ncol=0)),
  validity = validateParameters
)

#' 
#' Create a list of parameters.
#' 
#' @return an empty list of parameters  
#' @export
Parameters <- function() {
  return(new("parameters"))
}

#_______________________________________________________________________________
#----                              add                                      ----
#_______________________________________________________________________________

#' @rdname add
setMethod("add", signature=c("parameters", "single_array_parameter"), definition=function(object, x) {
  if (is.na(x@index)) {
    maxIndex <- object %>% select(as.character(class(x))) %>% maxIndex()
    x@index <- as.integer(maxIndex + 1)
  }
  return(methods::callNextMethod(object, x))
})

#' @rdname add
setMethod("add", signature=c("parameters", "double_array_parameter"), definition=function(object, x) {
  if (is.na(x@index) && is.na(x@index2)) {
    maxIndex <- object %>% select(as.character(class(x))) %>% maxIndex()
    x@index <- as.integer(maxIndex + 1)
    x@index2 <- as.integer(maxIndex + 1)
  }
  return(methods::callNextMethod(object, x))
})

#' @rdname add
setMethod("add", signature=c("parameters", "parameters"), definition=function(object, x) {
  return(object %>% appendParameters(x))
})

#' Append parameters.
#' 
#' @param params1 base set of parameters
#' @param params2 extra set of parameters to be appended
#' @return the resulting set of parameters
#' @importFrom purrr discard map_chr
#' @importFrom assertthat are_equal
#' @keywords internal
#' 
appendParameters <- function(params1, params2) {
  getParameterNamesInModel <- function(parameters) {
    retValue <- parameters@list %>%
      purrr::map_chr(.f=function(parameter) {
        if (is(parameter, "double_array_parameter") && !isDiag(parameter)) {
          return(NA)
        } else {
          return(parameter %>% getNameInModel())
        }
      }) %>%
      purrr::discard(~is.na(.x))
    return(retValue)
  }
  paramNames1 <- getParameterNamesInModel(params1)
  paramNames2 <- getParameterNamesInModel(params2)
  
  checkCollisionOnParams <- paramNames1 %in% paramNames2
  if (any(checkCollisionOnParams)) {
    stop(paste0("Model can't be appended because of duplicate parameter name(s): ", paste0(paramNames1[checkCollisionOnParams], collapse=", ")))
  }
  
  thetaMax <- params1 %>% select("theta") %>% maxIndex()
  omegaMax <- params1 %>% select("omega") %>% maxIndex()
  sigmaMax <- params1 %>% select("sigma") %>% maxIndex()

  for (theta in (params2 %>% select("theta"))@list) {
    theta@index <- theta@index + thetaMax
    params1 <- params1 %>% add(theta)
  }
  for (omega in (params2 %>% select("omega"))@list) {
    omega@index <- omega@index + omegaMax
    omega@index2 <- omega@index2 + omegaMax
    params1 <- params1 %>% add(omega)
  }
  for (sigma in (params2 %>% select("sigma"))@list) {
    sigma@index <- sigma@index + sigmaMax
    sigma@index2 <- sigma@index2 + sigmaMax
    params1 <- params1 %>% add(sigma)
  }

  # Merge variance-covariance matrices
  varcov1 <- params1@varcov
  varcov2 <- params2@varcov
  params1@varcov <- appendVarcov(varcov1, varcov2)

  return(params1 %>% sort())
}

appendVarcov <- function(varcov1, varcov2) {
  if (length(varcov1) == 0 && length(varcov2) > 0) {
    return(varcov2)
  
  } else if (length(varcov2) == 0 && length(varcov1) > 0) {
    return(varcov1)
  
  } else if (length(varcov1) == 0 && length(varcov2) == 0) {
    return(matrix(numeric(0), nrow=0, ncol=0))
  
  } else {
    dimnames1 <- dimnames(varcov1)
    assertthat::are_equal(dimnames1[[1]], dimnames1[[2]])
    colnames1 <- dimnames1[[1]]
    
    dimnames2 <- dimnames(varcov2)
    assertthat::are_equal(dimnames2[[1]], dimnames2[[2]])
    colnames2 <- dimnames2[[1]]
    
    colnames <- c(colnames1, colnames2)
    assertthat::assert_that(!any(duplicated(colnames)), msg="Duplicate parameter names in resulting variance-covariance matrix")
    totalDim <- length(colnames)
    
    varcov <- matrix(numeric(length(totalDim)^2), nrow=totalDim, ncol=totalDim)
    dimnames(varcov) <- list(colnames, colnames)
    
    varcov[seq_along(colnames1), seq_along(colnames1)] <- varcov1
    varcov[length(colnames1) + seq_along(colnames2), length(colnames1) + seq_along(colnames2)] <- varcov2
    return(varcov)
  }
}

#_______________________________________________________________________________
#----                             addRSE                                    ----
#_______________________________________________________________________________

#' @rdname addRSE
setMethod("addRSE", signature=c("parameters", "parameter", "numeric"), definition=function(object, parameter, value, ...) {
  parameter_ <- object %>%
    find(parameter)
  
  if (is.null(parameter_)) {
    stop("Parameter ", parameter %>% getName(), " not found in model")
  }
  
  # Define variance-covariance matrix (single value)
  varcov <- matrix((value/100*abs(parameter_@value))^2, nrow=1, ncol=1)
  name <- parameter_ %>% getName()
  dimnames(varcov) <- list(name, name)
  
  # Remove last value if it exists
  colnames <- colnames(object@varcov)
  if (name %in% colnames) {
    object@varcov <- object@varcov[-which(name==colnames), -which(name==colnames), drop=FALSE]
  }
  
  # Update variance-covariance matrix
  object@varcov <- appendVarcov(object@varcov, varcov)
  
  return(object)
})

#_______________________________________________________________________________
#----                             delete                                    ----
#_______________________________________________________________________________

#' Match single array parameter from list based on the name instead of the index.
#' If a match is found, its index is automatically copied.
#' 
#' @param object parameters
#' @param x single array parameter to match
#' @return the same parameter is no match was found or the same parameter with updated index if a match was found
#' @keywords internal
matchSingleArrayParameter <- function(object, x) {
  # If index is NA, index will be the index of the replaced parameter
  if (is.na(x@index) && !is.na(x@name)) {
    existingParam <- object %>% getByName(x %>% getName())
    if (existingParam %>% length() == 1) {
      x@index <- existingParam@index   # Copy index!
    }
  }
  return(x)
}

#' Match double array parameter from list based on the name instead of the index.
#' If a match is found, its indexes are automatically copied.
#' 
#' @param object parameters
#' @param x double array parameter to match
#' @return the same parameter is no match was found or the same parameter with updated indexes if a match was found
#' @keywords internal
matchDoubleArrayParameter <- function(object, x) {
  # If index is NA, index will be the index of the replaced parameter
  if (is.na(x@index) && is.na(x@index2) && !is.na(x@name)) {
    existingParam <- object %>% getByName(x %>% getName())
    if (existingParam %>% length() == 1) {
      x@index <- existingParam@index   # Copy index!
      x@index2 <- existingParam@index2 # Copy index2!
    }
  }
  return(x)
}

#' @rdname delete
setMethod("delete", signature=c("parameters", "single_array_parameter"), definition=function(object, x) {
  x <- matchSingleArrayParameter(object, x)
  return(methods::callNextMethod(object, x))
})

#' @rdname delete
setMethod("delete", signature=c("parameters", "double_array_parameter"), definition=function(object, x) {
  x <- matchDoubleArrayParameter(object, x)
  return(methods::callNextMethod(object, x))
})

#_______________________________________________________________________________
#----                              disable                                  ----
#_______________________________________________________________________________

#' @rdname disable
#' @importFrom purrr discard keep map
setMethod("disable", signature=c("parameters", "character"), definition=function(object, x, ...) {
  variabilities <- c("IIV", "IOV", "RUV", "VARCOV", "VARCOV_OMEGA", "VARCOV_SIGMA")
  msg <- paste0("Only these variabilities can be disabled: ", paste0("'", variabilities, "'", collapse=", "))
  assertthat::assert_that(list(...) %>% length()==0, msg="Extra arguments are not accepted")
  assertthat::assert_that(all(x %in% variabilities), msg=msg)
  
  # Disable IIV
  if ("IIV" %in% x) {
    omega_ <- (object %>% select("omega"))@list %>%
      purrr::map(.f=function(param) {
          param@value <- 0
          return(param)
        })
    object <- object %>% replace(omega_)
  }
  
  # Disable IOV (note that IOV is a subset of IIV)
  if ("IOV" %in% x) {
    omega_ <- (object %>% select("omega"))@list %>%
      purrr::discard(.p=~is.na(.x@same)) %>%
      purrr::map(.f=function(param) {
          param@value <- 0
          return(param)
        })
    object <- object %>% replace(omega_)
  }
  
  # Disable RUV
  if ("RUV" %in% x) {
    sigma_ <- (object%>% select("sigma"))@list %>%
      purrr::map(.f=function(param) {
        param@value <- 0
        return(param)
      })
    object <- object %>% replace(sigma_)
  }
  
  # Disable VARCOV (variance covariance matrix)
  if ("VARCOV" %in% x) {
    object@varcov <- matrix(numeric(0), ncol=0, nrow=0)
  }
  
  # Disable all omegas or sigmas in varcov
  varcovOmega <- "VARCOV_OMEGA" %in% x
  varcovSigma <- "VARCOV_SIGMA" %in% x
  
  if (varcovOmega || varcovSigma) {
    # Retrieve varcov parameters to remove
    varcovParams <- colnames(object@varcov) %>%
      purrr::map(.f=function(.x) {
        return(object %>% getByName(.x))
      }) %>%
      purrr::keep(.p=~(is(.x, "omega") && varcovOmega) ||
                    (is(.x, "sigma") && varcovSigma))
    
    # Retrieve the corresponding indexes in the matrix
    indexesToRemove <- varcovParams %>%
      purrr::map_int(.f=~which(colnames(object@varcov) == .x %>% getName()))
    
    # Update variance-covariance matrix
    if (length(indexesToRemove) > 0) {
      object@varcov <- object@varcov[-indexesToRemove, -indexesToRemove, drop=FALSE]
    }
  }
  
  return(object)
})

#_______________________________________________________________________________
#----                           exportToJSON                                ----
#_______________________________________________________________________________

omegaSigmaJsonIndexFix <- function(json, parameters, type) {
  index <- json$index
  index2 <- json$index2
  if (index != index2) {
    p1 <- parameters %>% campsismod::getByIndex(new(type, index=index, index2=index))
    p2 <- parameters %>% campsismod::getByIndex(new(type, index=index2, index2=index2))
    json$name <- p1@name
    json$name2 <- p2@name
  }
  json$index <- NULL
  json$index2 <- NULL
  return(json)
}

toJSONParamReference <- function(param, parameters) {
  json <- list()
  if (is(param, "theta")) {
    json$type <- "theta_ref"
    json$name <- param@name
  } else if (is(param, "double_array_parameter")) {
    if (is(param, "omega")) {
      json$type <- "omega_ref"
      emptyParam1 <- Omega()
      emptyParam2 <- Omega()
    } else if (is(param, "sigma")) {
      json$type <- "sigma_ref"
      emptyParam1 <- Sigma()
      emptyParam2 <- Sigma()
    } else {
      stop("Either omega or sigma")
    }
    if (param %>% isDiag()) {
      json$name <- param@name
    } else {
      emptyParam1@index <- param@index
      emptyParam1@index2 <- param@index
      refParam1 <- parameters %>% getByIndex(emptyParam1)
      json$name <- refParam1@name
      emptyParam2@index <- param@index2
      emptyParam2@index2 <- param@index2
      refParam2 <- parameters %>% getByIndex(emptyParam2)
      json$name2 <- refParam2@name
    }
  } else {
    stop("Unknown parameter type")
  }
  return(json)
}

varcovToJSON <- function(parameters) {
  varcov <- parameters@varcov
  assertthat::assert_that(nrow(varcov)==ncol(varcov))
  
  colNames <- dimnames(varcov)[[1]]
  rowNames <- dimnames(varcov)[[2]]
  assertthat::assert_that(all(colNames==rowNames))
  assertthat::assert_that(length(colNames)==nrow(varcov))
  
  parametersList <- parameters@list
  parameterNames <- parametersList %>%
    purrr::map_chr(~.x %>% getName())
  
  json <- list()
  
  for (i in seq_len(nrow(varcov))) {
    for (j in seq_len(nrow(varcov))) {
      if (j > i) {
        next
      }
      covValue <- varcov[i, j]
      if (covValue==0) {
        next
      } 
      rowName <- rowNames[i]
      colName <- colNames[j]
      rowParamIndex <- which(parameterNames==rowName)
      colParamIndex <- which(parameterNames==colName)
      assertthat::assert_that(length(rowParamIndex)==1,
                              msg=sprintf("No parameter found for variance matrix entry '%s'", rowName))
      assertthat::assert_that(length(colParamIndex)==1,
                              msg=sprintf("No parameter found for variance matrix entry '%s'", colName))
      rowParam <- parametersList[[rowParamIndex]]
      columnParam <- parametersList[[colParamIndex]]
      varcovEntry <- list()
      varcovEntry$type <- "varcov_entry"
      varcovEntry$ref1 <- toJSONParamReference(param=rowParam, parameters=parameters)
      varcovEntry$ref2 <- toJSONParamReference(param=columnParam, parameters=parameters)
      varcovEntry$cov <- covValue
      json[[length(json) + 1]] <- varcovEntry
    }
  }
  return(json)
}

#' @rdname exportToJSON
setMethod("exportToJSON", signature=c("parameters"), definition=function(object, ...) {
  object <- object %>%
    campsismod::sort()

  json <- object@list %>%
    purrr::map(function(x) {
        pJson <- exportToJSON(x)@data
        if (pJson$type=="theta") {
          pJson$index <- NULL
        } else if (pJson$type=="omega" || pJson$type=="sigma") {
          pJson <- omegaSigmaJsonIndexFix(json=pJson, parameters=object, type=pJson$type)
        } else {
          print(pJson$type)
          stop("Should never occur")
        }
        return(pJson)
      })
  
  return(JSONElement(json))
})

#_______________________________________________________________________________
#----                             fixOmega                                  ----
#_______________________________________________________________________________

#' Fix omega matrix for SAME OMEGA parameters that have NA values due to imperfections in Pharmpy import.
#' 
#' @param object generic object
#' @return the parameter that matches
#' @export
#' @rdname fixOmega
fixOmega <- function(object) {
  stop("No default function is provided")
}

setGeneric("fixOmega", function(object) {
  standardGeneric("fixOmega")
})

#' @rdname fixOmega
setMethod("fixOmega", signature=c("parameters"), definition=function(object) {
  
  # First order parameters
  object <- object %>% sort()

  # We need at least to elements
  if (length(object@list) < 2) {
    return(object)
  }
  
  # Select omega's only
  omegas <- object %>% select("omega")
  
  # Checking all 'same' are NA's
  sameVector <- omegas@list %>% purrr::map_lgl(.f=~.x@same)
  assertthat::assert_that(all(is.na(sameVector)), msg="all 'same' must be NA")

  # New list of omega's, add first omega into the list 
  omegas_ <- Parameters()
  omegas_ <- omegas_ %>% add(omegas@list[[1]])
  
  # Fix NA problems
  # .x is the accumulated results or initial value (a 'parameters' object here)
  # .y next value in sequence (an omega here)
  returned_omega_ <- purrr::accumulate(.x=omegas@list[2:length(omegas@list)], .f=function(.x, .y) {
    lastOmega <- .x@list[[.x@list %>% length()]]
    currentOmega <- .y
    
    # Is my current omega SAME as previous?
    if (is.na(currentOmega@value)) {
      currentOmega@value <- lastOmega@value
      currentOmega@same <- TRUE
      if (is.na(lastOmega@same)) {
        lastOmega@same <- FALSE
        # Update first SAME omega
        .x <- .x %>% replace(lastOmega) 
      }
    }
    # Update slot 'fix' based on last omega
    if (is.na(currentOmega@fix)) {
      currentOmega@fix <- lastOmega@fix
    }
    
    # Accumulate here
    .x <- .x %>% add(currentOmega)
    
    return(.x)
  }, .init=omegas_)
  
  # Replace all previous omega's by new ones
  object <- object %>% replace(returned_omega_)

  return(object)
})

#_______________________________________________________________________________
#----                             getByIndex                              ----
#_______________________________________________________________________________

#' @rdname getByIndex
setMethod("getByIndex", signature=c("parameters", "parameter"), definition=function(object, x) {
  subList <- object %>% select(as.character(class(x)))
  if (is(x, "theta")) {
    retValue <- subList@list %>% purrr::keep(~(.x@index==x@index))
  } else {
    retValue <- subList@list %>% purrr::keep(~(.x@index==x@index)&(.x@index2==x@index2))
  }
  if (length(retValue) > 0) {
    retValue <- retValue[[1]]
  }
  return(retValue)
})

#_______________________________________________________________________________
#----                          getUncertainty                               ----
#_______________________________________________________________________________

#' @importFrom tibble tibble
#' @rdname getUncertainty
setMethod("getUncertainty", signature=c("parameters"), definition=function(object, ...) {
  varcov <- object %>% getVarCov()
  if (is.null(varcov)) {
    return(tibble::tibble(name=character(0), se=numeric(0), "rse%"=numeric(0)))
  } else {
    return(object@list %>%
             purrr::map_df(.f=~getUncertainty(object=.x, varcov=varcov, parameters=object)))
  }
})

#_______________________________________________________________________________
#----                             getVarCov                                 ----
#_______________________________________________________________________________

#' @rdname getVarCov
setMethod("getVarCov", signature=c("parameters"), definition=function(object) {
  varcov <- object@varcov
  if (varcov %>% length() == 0) {
    return(NULL)
  } else {
    return(varcov)
  }
})

#_______________________________________________________________________________
#----                                minIndex                               ----
#_______________________________________________________________________________

#' Min index.
#' 
#' @param object generic object
#' @return min index
#' @export
#' @rdname minIndex
minIndex <- function(object) {
  stop("No default function is provided")
}

setGeneric("minIndex", function(object) {
  standardGeneric("minIndex")
})

#' @rdname minIndex
setMethod("minIndex", signature=c("parameters"), definition=function(object) {
  if (object %>% length() == 0) {
    return(0)
  }
  return(object@list %>% purrr::map_int(.f=function(.x) {
    if (is(.x, "double_array_parameter")) {
      return(min(c(.x@index, .x@index2)))
    } else {
      return(.x@index)
    }
  }) %>% min())
})

#_______________________________________________________________________________
#----                                maxIndex                               ----
#_______________________________________________________________________________

#' Max index.
#' 
#' @param object generic object
#' @return max index
#' @export
#' @rdname maxIndex
maxIndex <- function(object) {
  stop("No default function is provided")
}

setGeneric("maxIndex", function(object) {
  standardGeneric("maxIndex")
})

#' @rdname maxIndex
setMethod("maxIndex", signature=c("parameters"), definition=function(object) {
  if (object %>% length() == 0) {
    return(as.integer(0))
  }
  return(object@list %>% purrr::map_int(.f=function(.x) {
    if (is(.x, "double_array_parameter")) {
      return(max(c(.x@index, .x@index2)))
    } else {
      return(.x@index)
    }
  }) %>% max())
})

#_______________________________________________________________________________
#----                                 read                                  ----
#_______________________________________________________________________________

dataframeToParameter <- function(row, type) {
  name <- ifelse(is.null(row$name), NA, row$name) # Optional
  label <- ifelse(is.null(row$label), as.character(NA), row$label) # Optional
  unit <- ifelse(is.null(row$unit), as.character(NA), row$unit) # Optional
  comment <- ifelse(is.null(row$comment), as.character(NA), row$comment) # Optional
  min <- ifelse(is.null(row$min), as.numeric(NA), as.numeric(row$min)) # Optional
  max <- ifelse(is.null(row$max), as.numeric(NA), as.numeric(row$max)) # Optional
  
  if (type=="theta") {
    param <- Theta(name=name, index=row$index, value=row$value, min=min, max=max,
                   fix=row$fix, label=label, unit=unit, comment=comment)
  } else if(type=="omega") {
    same <- ifelse(is.null(row$same), NA, row$same) # Optional
    param <- Omega(name=name, index=row$index, index2=row$index2, value=row$value, min=min, max=max,
                   fix=row$fix, type=row$type, same=same, label=label, comment=comment)
  } else if(type=="sigma") {
    param <- Sigma(name=name, index=row$index, index2=row$index2, value=row$value, min=min, max=max,
                   fix=row$fix, type=row$type, label=label, comment=comment)
  } else {
    stop(paste0("type must be one of: theta, omega or sigma"))
  }
  return(param)
}

#' Read parameters file.
#' 
#' @param file path to CSV file
#' @param type parameter type: 'theta', 'omega' or 'sigma'
#' @return parameters sub list
#' @importFrom readr read_delim
#' @importFrom dplyr across group_split
#' @importFrom purrr map
#' @export
read.parameters <- function(file, type) {
  assertthat::assert_that(type %in% c("theta", "omega", "sigma"),
                          msg="Type must be one of these: 'theta', 'omega' or 'sigma'")
  df <- readr::read_delim(file=file, lazy=FALSE, show_col_types=FALSE, progress=FALSE) %>%
    dplyr::mutate(ROWID=dplyr::row_number())
  list <- df %>%
    dplyr::group_split(dplyr::across("ROWID")) %>%
    purrr::map(~dataframeToParameter(as.list(.x), type=type))
  attributes(list) <- NULL
  return(new("parameters", list=list))
}

#' Read variance-covariance file.
#' 
#' @param file path to CSV file
#' @return variance-covariance matrix
#' @importFrom assertthat assert_that
#' @importFrom utils read.csv
#' @export
read.varcov <- function(file) {
  dataframe <- utils::read.csv(file=file)
  row.names(dataframe) <- dataframe[,1] # First column contains parameter names
  matrix <- dataframe[,-1] %>% as.matrix()
  assertthat::assert_that(all(rownames(matrix)==colnames(matrix)), 
      msg="Row names are different than column names in variance-covariance matrix")
  return(matrix)
}

#' Read all parameters files at once.
#' 
#' @param folder path to folder or path to zipped project
#' @return parameters object
#' @export
read.allparameters <- function(folder) {
  thetaPath <- file.path(folder, "theta.csv")
  omegaPath <- file.path(folder, "omega.csv")
  sigmaPath <- file.path(folder, "sigma.csv")
  varcovPath <- file.path(folder, "varcov.csv")
  
  if (file.exists(thetaPath)) {
    theta <- read.parameters(file=thetaPath, type="theta")
  } else {
    theta <- Parameters()
    warning(paste0("No file 'theta.csv' could be found."))
  }
  if (file.exists(omegaPath)) {
    omega <- read.parameters(file=omegaPath, type="omega")
  } else {
    omega <- Parameters()
    warning(paste0("No file 'omega.csv' could be found."))
  }
  if (file.exists(sigmaPath)) {
    sigma <- read.parameters(file=sigmaPath, type="sigma")
  } else {
    sigma <- Parameters()
    warning(paste0("No file 'sigma.csv' could be found."))
  }

  parameters <-  Parameters() %>%
    add(theta) %>%
    add(omega) %>%
    add(sigma)

  if (file.exists(varcovPath)) {
    varcov <- read.varcov(varcovPath)
    parameters@varcov <- varcov
  }
  return(parameters)
}

#_______________________________________________________________________________
#----                             replace                                   ----
#_______________________________________________________________________________

#' @rdname replace
setMethod("replace", signature=c("parameters", "single_array_parameter"), definition=function(object, x) {
  x <- matchSingleArrayParameter(object, x)
  return(methods::callNextMethod(object, x))
})

#' @rdname replace
setMethod("replace", signature=c("parameters", "double_array_parameter"), definition=function(object, x) {
  x <- matchDoubleArrayParameter(object, x)
  return(methods::callNextMethod(object, x))
})

#_______________________________________________________________________________
#----                                 select                                ----
#_______________________________________________________________________________

#' @rdname select
setMethod("select", signature=c("parameters"), definition=function(object, ...) {
  args <- list(...)
  msg <- "Please select one of those parameter types: 'theta', 'omega' or 'sigma'"
  assertthat::assert_that(length(args) > 0, msg=msg)
  type <- args[[1]]
  assertthat::assert_that(type %in% c("theta", "omega", "sigma"), msg=msg)
  
  object@list <- object@list %>% purrr::keep(~as.character(class(.x))==type)
  return(object)
})

#_______________________________________________________________________________
#----                            setMinMax                                  ----
#_______________________________________________________________________________

#' @rdname setMinMax
setMethod("setMinMax", signature=c("parameters", "parameter", "numeric", "numeric"), definition=function(object, parameter, min, max, ...) {
  parameter_ <- object %>%
    find(parameter)
  
  if (is.null(parameter_)) {
    stop("Parameter ", parameter %>% getNameInModel(), " not found in model")
  }
  
  # Replace old values
  parameter_@min <- min
  parameter_@max <- max
  
  # Replace old parameter
  object <- object %>%
    replace(parameter_)
  
  return(object)
})

#' @rdname setMinMax
setMethod("setMinMax", signature=c("parameters", "character", "numeric", "numeric"), definition=function(object, parameter, min, max, ...) {
  assertthat::assert_that(parameter %in% c("theta", "omega", "sigma"), msg="Parameter must be one of: 'theta', 'omega' or 'sigma'")
  object@list <- object@list %>%
    purrr::map(.f=function(x) {
      if (is(x, parameter)) {
        x@min <- min
        x@max <- max
        
        # Special case for covariance
        if (is(x, "double_array_parameter") && !(x %>% isDiag()) && min >= 0) {
          x@min <- -max
          cat(sprintf("Info: min value of %s (covariance) set to -max value\n", x %>% getName()))
        }
      }
      return(x)
    })
  return(object)
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

showUncertaintyOnParameters <- function(parameters, discard_na_columns=NULL) {
  retValue <- purrr::map_df(parameters@list, .f=as.data.frame, row.names=character(), optional=FALSE) %>%
    removeNaColumn(discard_na_columns)
  
  if (parameters %>% length() > 0) {
    uncertainty <- parameters %>% getUncertainty()
    # Show uncertainty if at least one parameter has uncertainty
    if (any(!is.na(uncertainty$se))) {
      retValue <- dplyr::bind_cols(retValue, uncertainty %>% dplyr::select(-"name")) 
    }
  }
  return(retValue)
}

setMethod("show", signature=c("parameters"), definition=function(object) {
  cat("THETA's:\n")
  print(showUncertaintyOnParameters(object %>% select("theta"), discard_na_columns=c("min", "max", "label", "unit", "comment")))
  cat("OMEGA's:\n")
  print(showUncertaintyOnParameters(object %>% select("omega"), discard_na_columns=c("min", "max", "same", "label", "comment")))
  cat("SIGMA's:\n")
  print(showUncertaintyOnParameters(object %>% select("sigma"), discard_na_columns=c("min", "max", "label", "comment")))
  if (is.null(object %>% getVarCov())) {
    cat("No variance-covariance matrix\n")
  } else {
    cat("Variance-covariance matrix available (see ?getVarCov)\n")
  }
})

#_______________________________________________________________________________
#----                                  sort                                 ----
#_______________________________________________________________________________

#' @rdname sort
setMethod("sort", signature=c("parameters"), definition=function(x, decreasing=FALSE, ...) {
  types <- x@list %>% purrr::map_chr(~as.character(class(.x)))
  indexes1 <- x@list %>% purrr::map_int(~.x@index)
  indexes2 <- x@list %>% purrr::map_int(.f=function(.x){
    if("index2" %in% methods::slotNames(.x)) {
      return(.x@index2)
    } else {
      return(as.integer(0))
    }
  })
  
  # Reorder
  types <- factor(types, levels=c("theta", "omega", "sigma"), labels=c("theta", "omega", "sigma"))
  order <- order(types, indexes1, indexes2)
  
  # Apply result to original list
  x@list <- x@list[order]
  return(x)
})

#_______________________________________________________________________________
#----                            standardise                                ----
#_______________________________________________________________________________

#' @rdname standardise
setMethod("standardise", signature=c("parameters"), definition=function(object, ...) {
  list <- object@list %>% purrr::map(.f=function(param) {
    return(param %>% standardise(parameters=object))
  })
  retValue <- Parameters()
  retValue@list <- list
  retValue@varcov <- object@varcov
  return(retValue)
})

#_______________________________________________________________________________
#----                                 write                                 ----
#_______________________________________________________________________________

#' Write subset of parameters (theta, omega or sigma).
#' 
#' @param object subset of parameters
#' @param file filename
#' @param ... extra arguments, like \code{defaultDf} for empty parameters list
#' @return TRUE if success
#' @importFrom dplyr any_of select where 
#' @importFrom utils write.csv
writeParameters <- function(object, file, ...) {
  df <- purrr::map_df(object@list, .f=as.data.frame, row.names=character(), optional=FALSE)
  
  # Get rid of specific columns if all NA
  naColumns <- c("min", "max", "fix", "same", "label", "unit", "comment")
  df <- df %>% removeNaColumn(naColumns)
  
  if (nrow(df)==0) {
    df <- processExtraArg(args=list(...), name="defaultDf", mandatory=TRUE)
  }
  utils::write.csv(df, file=file, row.names=FALSE)
  return(TRUE)
}



#' Write variance-covariance matrix.
#' 
#' @param object matrix
#' @param file filename
#' @return TRUE if success
#' @importFrom utils write.csv
writeVarcov <- function(object, file) {
  utils::write.csv(object, file=file)
  return(TRUE)
}

#' @rdname write
setMethod("write", signature=c("parameters", "character"), definition=function(object, file, ...) {
  theta <- object %>% select("theta")
  omega <- object %>% select("omega")
  sigma <- object %>% select("sigma")
  varcov <- object@varcov
  
  theta %>% writeParameters(file=file.path(file, "theta.csv"),
                  defaultDf=data.frame(name=character(), index=integer(), value=numeric(), fix=logical()))
  omega %>% writeParameters(file=file.path(file, "omega.csv"),
                  defaultDf=data.frame(name=character(), index=integer(), index2=integer(), value=numeric(), fix=logical(), type=character()))
  sigma %>% writeParameters(file=file.path(file, "sigma.csv"),
                  defaultDf=data.frame(name=character(), index=integer(), index2=integer(), value=numeric(), fix=logical(), type=character()))
  
  if (length(varcov) > 0) {
    varcov %>% writeVarcov(file=file.path(file, "varcov.csv"))
  }
})

#_______________________________________________________________________________
#----                           pattern class                               ----
#_______________________________________________________________________________

#' 
#' Pattern class.
#' 
#' @export
setClass(
  "pattern",
  representation(
  ),
  contains="character",
  validity=function(object) {
    if (object %>% as.character() %>% length()!=1) {
      return("Invalid pattern. Pattern must be a single character string.")
    }
    return(TRUE)
  }
)

#' 
#' Create a pattern.
#' 
#' @param x regular expression
#' @return a pattern
#' @export
Pattern <- function(x) {
  return(new("pattern", x))
}

#_______________________________________________________________________________
#----                         variable_pattern class                        ----
#_______________________________________________________________________________

#' 
#' Variable pattern class.
#' 
#' @export
setClass(
  "variable_pattern",
  representation(
  ),
  contains="pattern"
)

#' 
#' Create a variable pattern.
#' 
#' @param x variable name
#' @return a variable pattern
#' @export
VariablePattern <- function(x) {
  return(new("variable_pattern", x))
}

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("character", "variable_pattern", "character"), definition=function(object, pattern, replacement, ...) {
  return(gsub(paste0("([^a-zA-Z0-9_]|^)(", pattern, ")([^a-zA-Z0-9_]|$)"), replacement=paste0("\\1", replacement, "\\3"), x=object))
})

#' @rdname replaceAll
setMethod("replaceAll", signature=c("character", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  return(gsub(pattern, replacement=replacement, x=object))
})

library(testthat)

context("Test the different patterns that can be created to replace occurrences")

test_that("Pattern object works as expected", {
  str <- "ETA_KA + THETA_KA + ETA_KA + ETA_KA3 + ETA_KA"
  pattern <- Pattern("ETA_KA")
  expect_equal(pattern %>% as.character(), "ETA_KA")
  expect_equal(str %>% replaceAll(pattern, "ETA_KA2"), "ETA_KA2 + THETA_KA2 + ETA_KA2 + ETA_KA23 + ETA_KA2")
  
  str <- "HELLO"
  expect_equal(str %>% replaceAll(pattern, ""), "HELLO")
})

test_that("Variable pattern object works as expected", {
  str <- "ETA_KA + THETA_KA + ETA_KA + ETA_KA3 + ETA_KA"
  pattern <- VariablePattern("ETA_KA")
  expect_equal(pattern %>% as.character(), "ETA_KA")
  expect_equal(str %>% replaceAll(pattern, "ETA_KA2"), "ETA_KA2 + THETA_KA + ETA_KA2 + ETA_KA3 + ETA_KA2")
  
  str <- "HELLO"
  expect_equal(str %>% replaceAll(pattern, ""), "HELLO")
})

test_that("Replace occurrences in model works as expected", {

  model <- model_suite$testing$nonmem$advan1_trans1 %>%
    add(LineBreak()) %>%
    add(Comment("Check replacement also works in IF-statement")) %>%
    add(IfStatement("K==1", Equation("XX", "K*10"))) %>%
    add(UnknownStatement("THIS IS AN UNKNOWN STATEMENT THAT CONTAINS THE VARIABLE K"))
  model <- model %>% replaceAll("K", "K2")
  
  expect_equal(model %>% find(Equation("K2")), Equation("K2", "THETA_K*exp(ETA_K)"))
  expect_equal(model %>% find(Ode("A_CENTRAL")), Ode("A_CENTRAL", "-K2*A_CENTRAL"))
  expect_equal(model %>% find(Ode("A_OUTPUT")), Ode("A_OUTPUT", "K2*A_CENTRAL"))
  expect_equal(model %>% find(IfStatement("K2==1", Equation("XX"))), IfStatement("K2==1", Equation("XX", "K2*10")))

  main <- model %>% find(MainRecord())
  lastStatement <- main@statements@list[[length(main@statements@list)]]
  expect_equal(lastStatement, UnknownStatement("THIS IS AN UNKNOWN STATEMENT THAT CONTAINS THE VARIABLE K2"))
})

test_that("Function replaceAll also replaces occurrences in compartment properties", {
  model <- model_suite$pk$'1cpt_fo'
  
  model <- model %>% 
    replaceAll("BIO", "BIO2")
  
  # Make sure BIO2 equation is there
  expect_equal(model %>% campsismod::find(Equation("BIO2")), Equation("BIO2", "TVBIO"))
  
  # Make sure compartment property was updated well
  expect_equal(model %>% find(Bioavailability(compartment=1)), Bioavailability(compartment=1, rhs="BIO2"))  
})

#_______________________________________________________________________________
#----                          parameter class                              ----
#_______________________________________________________________________________

#' 
#' Parameter class. Any parameter in a pharmacometric model.
#' 
#' @slot name parameter name, optional (although recommended)
#' @slot index parameter index, integer
#' @slot value parameter value (e.g. the estimated value from a modelling tool)
#' @slot min minimum value for this parameter when parameter uncertainty is enabled
#' @slot max maximum value for this parameter when parameter uncertainty is enabled
#' @slot fix logical value, say if parameter was fixed in the modelling phase
#' @slot label parameter label, any string
#' @slot comment any comment on this parameter, any string
#' @export
setClass(
  "parameter",
  representation(
    name = "character",
    index = "integer",
    value = "numeric",
    min = "numeric",
    max = "numeric",
    fix = "logical",
    label = "character",
    comment = "character"
  ),
  contains = "pmx_element",
  prototype = prototype(name=as.character(NA), index=as.integer(NA),
                        value=as.numeric(NA), min=as.numeric(NA), max=as.numeric(NA), fix=FALSE,
                        label=as.character(NA), unit=as.character(NA), comment=as.character(NA)),
  validity = function(object) {
    check <- expectOneForAll(object, c("name", "index", "fix", "value", "label", "comment"))
    return(check)
  }
)

#_______________________________________________________________________________
#----                   single_array_parameter class                        ----
#_______________________________________________________________________________

#' 
#' Single-array parameter class. This parameter has a single index value.
#' 
#' @export
setClass(
  "single_array_parameter",
  representation(
  ),
  contains = "parameter"
)

#_______________________________________________________________________________
#----                   double_array_parameter class                        ----
#_______________________________________________________________________________

#' 
#' Double-array parameter class. This parameter has 2 indexes. 
#' It can thus be used to define correlations between parameters.
#' 
#' @export
setClass(
  "double_array_parameter",
  representation(
    index2 = "integer",
    type = "character"
  ),
  contains = "single_array_parameter",
  prototype = prototype(type="var"),
  validity = function(object) {
    check1 <- expectOne(object, "type")
    check2 <-
      if (object@type %in% c("var", "sd", "covar", "cor", "cv", "cv%")) {
        character()
      } else {
        "Type should be one of: 'var', 'sd', 'covar', 'cor', 'cv' or 'cv%'"
      }
    check3 <- 
      if (is.na(object@index) && is.na(object@index2)) {
        character() # Don't go further
      } else if (object@index != object@index2 && !(object@type %in% c("covar", "cor"))) {
        paste0("Parameter type must be 'covar' or 'cor' (index:", object@index, ", index2:", object@index2, ")")
      } else {
        character()
      }
    check4 <- 
      if (is.na(object@index) && is.na(object@index2)) {
        character() # Don't go further
      } else if (object@index == object@index2 && object@type %in% c("covar", "cor")) {
        paste0("Parameter type can't be 'covar' nor 'cor' (index:", object@index, ", index2:", object@index2, ")")
      } else {
        character()
      }
    return(c(check1, check2, check3, check4))
  }
)

#_______________________________________________________________________________
#----                                theta                                  ----
#_______________________________________________________________________________

#' 
#' Theta parameter class.
#' 
#' @slot unit parameter unit
#' @export
setClass(
  "theta",
  representation(
    unit = "character"
  ),
  contains = "single_array_parameter"
)

#' 
#' Create a THETA parameter.
#' 
#' @param name parameter name, e.g. CL (prefix THETA will be added automatically)
#' @param index parameter index
#' @param value parameter value
#' @param min minimum value for this parameter when parameter uncertainty is enabled
#' @param max maximum value for this parameter when parameter uncertainty is enabled
#' @param fix parameter was fixed in estimation, logical value
#' @param label parameter label, optional
#' @param unit parameter unit, optional
#' @param comment any comment, optional
#' @return a THETA parameter  
#' @export
Theta <- function(name=NA, index=NA, value=NA, min=NA, max=NA, fix=FALSE, label=NA, unit=NA, comment=NA) {
  return(new("theta", name=as.character(name), index=as.integer(index),
             value=as.numeric(value), min=as.numeric(min), max=as.numeric(max), fix=fix,
             label=as.character(label), unit=as.character(unit), comment=as.character(comment)))
}

#_______________________________________________________________________________
#----                                omega                                  ----
#_______________________________________________________________________________

#'
#' Omega parameter class.
#' 
#' @slot same logical value, tell if this omega is the same as the previous one
#' @export
setClass(
  "omega",
  representation(
    same = "logical"
  ),
  contains = "double_array_parameter",
  prototype = prototype(same=as.logical(NA), index2=as.integer(NA)),
  validity = function(object) {
    return(expectOne(object, "same"))
  }
)

#' 
#' Process double array arguments.
#' 
#' @param index parameter index
#' @param index2 second parameter index
#' @param type variance type
#' @return variance type
#' @keywords internal
#'
processDoubleArrayArguments <- function(index, index2, type) {
  # Only 1 NA: NOK
  if (xor(is.na(index), is.na(index2))) {
    stop("Please specify both OMEGA indexes")
  }
  # Both NA: OK
  if (is.null(type) && is.na(index) && is.na(index2)) {
    type <- "var"
  }
  if (is.null(type)) {
    type <- if (index==index2) {"var"} else {"covar"}
  }
  return(type)
}

#' 
#' Create an OMEGA parameter.
#' 
#' @param name parameter name, e.g. CL (prefix OMEGA will be added automatically)
#' @param index parameter index
#' @param index2 second parameter index
#' @param value parameter value
#' @param min minimum value for this parameter when parameter uncertainty is enabled
#' @param max maximum value for this parameter when parameter uncertainty is enabled
#' @param fix parameter was fixed in estimation, logical value
#' @param type variance type: 'var', 'sd', 'covar', 'cor', 'cv' or 'cv\%'
#' @param same NA by default, FALSE for first OMEGA followed by 'SAME' OMEGA's, TRUE for 'SAME' OMEGA's
#' @param label parameter label, optional
#' @param comment any comment, optional
#' @return an OMEGA parameter  
#' @export
Omega <- function(name=NA, index=NA, index2=NA, value=NA, min=NA, max=NA, fix=FALSE, type=NULL, same=NA, label=NA, comment=NA) {
  type <- processDoubleArrayArguments(index=index, index2=index2, type=type)
  return(new("omega", name=as.character(name), index=as.integer(index), index2=as.integer(index2),
             value=as.numeric(value), min=as.numeric(min), max=as.numeric(max), fix=fix, type=type, same=as.logical(same),
             label=as.character(label), comment=as.character(comment)))
}

#_______________________________________________________________________________
#----                                sigma                                  ----
#_______________________________________________________________________________

#' Sigma parameter class.
#' 
#' @export
setClass(
  "sigma",
  representation(
  ),
  contains = "double_array_parameter"
)

#' 
#' Create a SIGMA parameter.
#' 
#' @param name parameter name, e.g. CL (prefix SIGMA will be added automatically)
#' @param index parameter index
#' @param index2 second parameter index
#' @param value parameter value
#' @param min minimum value for this parameter when parameter uncertainty is enabled
#' @param max maximum value for this parameter when parameter uncertainty is enabled
#' @param fix parameter was fixed in estimation, logical value
#' @param type variance type: 'var', 'sd', 'covar', 'cv' or 'cv\%'
#' @param label parameter label, optional
#' @param comment any comment, optional
#' @return a SIGMA parameter  
#' @export
Sigma <- function(name=NA, index=NA, index2=NA, value=NA, min=NA, max=NA, fix=FALSE, type=NULL, label=NA, comment=NA) {
  type <- processDoubleArrayArguments(index=index, index2=index2, type=type)
  return(new("sigma", name=as.character(name), index=as.integer(index), index2=as.integer(index2),
             value=as.numeric(value), min=as.numeric(min), max=as.numeric(max), fix=fix, type=type,
             label=as.character(label), comment=as.character(comment)))
}

#_______________________________________________________________________________
#----                           as.data.frame                               ----
#_______________________________________________________________________________

#' As data frame method.
#'
#' @param x generic object
#' @param row.names row names
#' @param optional optional
#' @param ... extra arguments
#' @return data frame
#' @export
#' @rdname as.data.frame
as.data.frame <- function(x, row.names=NULL, optional=FALSE, ...) {
  base::as.data.frame(x, row.names=row.names, optional=optional, ...)
}

setGeneric("as.data.frame", function(x, row.names=NULL, optional=FALSE, ...) {
  standardGeneric("as.data.frame")
})

#' @rdname as.data.frame
setMethod("as.data.frame", signature("theta", "character", "logical"), function(x, row.names=NULL, optional=FALSE, ...) {
  return(data.frame(name=x@name, index=x@index, value=x@value, min=x@min, max=x@max, fix=x@fix, label=x@label, unit=x@unit, comment=x@comment))
})

#' @rdname as.data.frame
setMethod("as.data.frame", signature("omega", "character", "logical"), function(x, row.names=NULL, optional=FALSE, ...) {
  return(data.frame(name=x@name, index=x@index, index2=x@index2, value=x@value, min=x@min, max=x@max, fix=x@fix, type=x@type, same=x@same, label=x@label, comment=x@comment))
})

#' @rdname as.data.frame
setMethod("as.data.frame", signature("sigma", "character", "logical"), function(x, row.names=NULL, optional=FALSE, ...) {
  return(data.frame(name=x@name, index=x@index, index2=x@index2, value=x@value, min=x@min, max=x@max, fix=x@fix, type=x@type, label=x@label, comment=x@comment))
})

#_______________________________________________________________________________
#----                           exportToJSON                                ----
#_______________________________________________________________________________

#' @rdname exportToJSON
setMethod("exportToJSON", signature=c("theta"), definition=function(object, ...) {
  json <- mapS4SlotsToJSONProperties(object, optional=c("min", "max", "label", "comment", "unit"))
  assertthat::assert_that(!is.null(json$name), msg="All THETAs must be named")
  return(JSONElement(json))
})

#' @rdname exportToJSON
setMethod("exportToJSON", signature=c("omega"), definition=function(object, ...) {
  json <- mapS4SlotsToJSONProperties(object, add_type=FALSE, optional=c("min", "max", "label", "comment"), ignore="same")
  json$var_type <- json$type
  json$type <- "omega"
  if (json$index==json$index2) {
    assertthat::assert_that(!is.null(json$name), msg="All OMEGAs must be named")
  }
  return(JSONElement(json))
})

#' @rdname exportToJSON
setMethod("exportToJSON", signature=c("sigma"), definition=function(object, ...) {
  json <- mapS4SlotsToJSONProperties(object, add_type=FALSE, optional=c("min", "max", "label", "comment"))
  json$var_type <- json$type
  json$type <- "sigma"
  if (json$index==json$index2) {
    assertthat::assert_that(!is.null(json$name), msg="All SIGMAs must be named")
  }
  return(JSONElement(json))
})

#_______________________________________________________________________________
#----                               isDiag                                  ----
#_______________________________________________________________________________

#' Is diagonal.
#' 
#' @param object generic object
#' @return logical value
#' @export
#' @rdname isDiag
isDiag <- function(object) TRUE

setGeneric("isDiag", function(object) {
  standardGeneric("isDiag")
})

#' @rdname isDiag
setMethod("isDiag", signature(object = "double_array_parameter"), function(object) {
  return(object@index==object@index2)
})

#_______________________________________________________________________________
#----                            getNONMEMName                              ----
#_______________________________________________________________________________

#' Get NONMEM name.
#' 
#' @param object generic object
#' @return the NONMEM name associated with this object
#' @export
#' @rdname getNONMEMName
getNONMEMName <- function(object) {
  stop("No default function is provided")
}

setGeneric("getNONMEMName", function(object) {
  standardGeneric("getNONMEMName")
})

#' @rdname getNONMEMName
setMethod("getNONMEMName", signature=c("theta"), definition=function(object) {
  return(paste0("THETA(", object@index, ")"))
})

#' @rdname getNONMEMName
setMethod("getNONMEMName", signature=c("omega"), definition=function(object) {
  return(paste0("OMEGA(", object@index, ",", object@index2, ")"))
})

#' @rdname getNONMEMName
setMethod("getNONMEMName", signature=c("sigma"), definition=function(object) {
  return(paste0("SIGMA(", object@index, ",", object@index2, ")"))
})

#_______________________________________________________________________________
#----                              getName                                  ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature=c("theta"), definition=function(x) {
  if (is.na(x@name)) {
    return(paste0("THETA", "_", x@index))
  } else {
    return(paste0("THETA", "_", x@name))
  }
})

#' @rdname getName
setMethod("getName", signature=c("omega"), definition=function(x) {
  if (is.na(x@name)) {
    return(paste0("OMEGA", "_", x@index, "_", x@index2))
  } else {
    return(paste0("OMEGA", "_", x@name))
  }
})

#' @rdname getName
setMethod("getName", signature=c("sigma"), definition=function(x) {
  if (is.na(x@name)) {
    return(paste0("SIGMA", "_", x@index, "_", x@index2))
  } else {
    return(paste0("SIGMA", "_", x@name))
  }
})

#_______________________________________________________________________________
#----                         getNameInModel                                ----
#_______________________________________________________________________________

#' Get the name of the given parameter in the Campsis model.
#' 
#' @param x element to know the name
#' @return the name of this parameter
#' @export
#' @rdname getNameInModel 
getNameInModel <- function(x) {
  stop("No default function is provided")
}

setGeneric("getNameInModel", function(x) {
  standardGeneric("getNameInModel")
})

#' @rdname getNameInModel 
setMethod("getNameInModel", signature=c("theta"), definition=function(x) {
  if (is.na(x@name)) {
    return(paste0("THETA", "_", x@index))
  } else {
    return(paste0("THETA", "_", x@name))
  }
})

#' @rdname getNameInModel 
setMethod("getNameInModel", signature=c("omega"), definition=function(x) {
  if (is.na(x@name)) {
    if (x@index != x@index2) {
      stop("You should not call this method with different indexes!")
    }
    return(paste0("ETA", "_", x@index))
  } else {
    return(paste0("ETA", "_", x@name))
  }
})

#' @rdname getNameInModel 
setMethod("getNameInModel", signature=c("sigma"), definition=function(x) {
  if (is.na(x@name)) {
    if (x@index != x@index2) {
      stop("You should not call this method with different indexes!")
    }
    return(paste0("EPS", "_", x@index))
  } else {
    return(paste0("EPS", "_", x@name))
  }
})

#_______________________________________________________________________________
#----                           getUncertainty                              ----
#_______________________________________________________________________________

#' @param varcov variance covariance matrix
#' @importFrom tibble tibble
#' @rdname getUncertainty
setMethod("getUncertainty", signature=c("parameter"), definition=function(object, varcov, ...) {
  name <- object %>% getName()
  if (varcov %>% length() > 0) {
    standardisedParameter <- object %>% standardise(...)
    if (name %in% colnames(varcov)) {
      variance <- varcov[name, name]
      return(tibble::tibble(name=name, se=sqrt(variance), "rse%"=100*sqrt(variance)/abs(standardisedParameter@value)))
    }
  }
  return(tibble::tibble(name=name, se=as.numeric(NA), "rse%"=as.numeric(NA)))
})

#_______________________________________________________________________________
#----                           loadFromJSON                                ----
#_______________________________________________________________________________

#' @rdname loadFromJSON
setMethod("loadFromJSON", signature=c("theta", "json_element"), definition=function(object, json) {
  object <- mapJSONPropertiesToS4Slots(object, json)
  return(object)
})

#' @rdname loadFromJSON
setMethod("loadFromJSON", signature=c("omega", "json_element"), definition=function(object, json) {
  object <- mapJSONPropertiesToS4Slots(object, json, discard_type=FALSE)
  return(object)
})

#' @rdname loadFromJSON
setMethod("loadFromJSON", signature=c("sigma", "json_element"), definition=function(object, json) {
  object <- mapJSONPropertiesToS4Slots(object, json, discard_type=FALSE)
  return(object)
})

#_______________________________________________________________________________
#----                            standardise                                ----
#_______________________________________________________________________________

#' @rdname standardise
setMethod("standardise", signature=c("theta"), definition=function(object, ...) {
  return(object)
})

#' @rdname standardise
#' @param parameters the list of parameters, to be provided only if parameter type is 'cor'
setMethod("standardise", signature=c("double_array_parameter"), definition=function(object, parameters=NULL, ...) {
  type <- object@type
  index <- object@index
  index2 <- object@index2
  retValue <- object# Copy

  if (index == index2) {
    if (type == "var") {
      # Do nothing
    
    } else if (type == "sd") {
      retValue@value <- object@value ^ 2
    
    } else if (type == "covar") {
      stop(paste0("Type of parameter ", object %>% getName(), " can't be 'covar'"))
    
    } else if (type == "cv") {
      retValue@value <- log(object@value^2+1)
    
    } else if (type == "cv%") {
      retValue@value <- log((object@value/100)^2+1)
    } else {
      stop("Type should be one of: 'var', 'sd', 'cv' or 'cv%'")
    }
    retValue@type <- "var"
  } else {
    if (type == "covar") {
      # Do nothing
      
    } else if (type == "cor") {
      if (is.null(parameters)) {
        stop("Argument 'parameters' is needed to convert a covariance into a correlation")
      }
      # Retrieve both omega's on the diagonal related to index and index2
      # Make sure to standardise them to variances first
      omega1 <- parameters %>% getByIndex(Omega(index=object@index, index2=object@index)) %>% standardise()
      omega2 <- parameters %>% getByIndex(Omega(index=object@index2, index2=object@index2)) %>% standardise()
      retValue@value <- object@value*sqrt(omega1@value)*sqrt(omega2@value)
    } else {
      stop(paste0("Type of parameter ", object %>% getName(), " must be 'covar' or 'cor'"))
    }
    retValue@type <- "covar"
  }
  return(retValue)
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("parameter"), definition=function(object) {
  print(object %>% as.data.frame(row.names=character(), optional=FALSE))
})

library(testthat)

context("Test generic method addSuffix")

source(paste0("", "testUtils.R"))

test_that("Add suffix to parameters is working as expected", {
  
  parameters <- Parameters() %>%
    add(Theta(name="CL", index=1, fix=FALSE, value=4)) %>%
    add(Theta(name="V", index=2, fix=FALSE, value=40)) %>%
    add(Theta(name="KA", index=3, fix=FALSE, value=0.5)) %>%
    add(Omega(name="CL", index=1, index2=1, fix=FALSE, value=0.1)) %>%
    add(Omega(name="V", index=2, index2=2, fix=FALSE, value=0.1)) %>%
    add(Omega(name=NA, index=2, index2=1, fix=FALSE, value=0.5, type="cor"))
  parameters@varcov <- matrix(dimnames=list("THETA_CL", "THETA_CL"), data=0.1)
  
  updatedParameters <- parameters %>%
    addSuffix("A")
  
  expectedParameters <- Parameters() %>%
    add(Theta(name="CL_A", index=1, fix=FALSE, value=4)) %>%
    add(Theta(name="V_A", index=2, fix=FALSE, value=40)) %>%
    add(Theta(name="KA_A", index=3, fix=FALSE, value=0.5)) %>%
    add(Omega(name="CL_A", index=1, index2=1, fix=FALSE, value=0.1)) %>%
    add(Omega(name="V_A", index=2, index2=2, fix=FALSE, value=0.1)) %>%
    add(Omega(name=NA, index=2, index2=1, fix=FALSE, value=0.5, type="cor"))
  expectedParameters@varcov <- matrix(dimnames=list("THETA_CL_A", "THETA_CL_A"), data=0.1)
  
  expect_equal(updatedParameters, expectedParameters)
  
  # Separator can't be a dot for instance
  expect_error(parameters %>% addSuffix("A", separator="."), regexp="is not a valid separator")
  
  # Suffix can't contain a dot neither
  expect_error(parameters %>% addSuffix(".A"), regexp="is not a valid suffix")
})

test_that("Add suffix to a code record is working as expected", {
  
  main <- MainRecord() %>%
    add(Equation("KA", "THETA_KA * exp(ETA_KA)")) %>%
    add(Equation("KA2", "2 * KA")) %>%
    add(Equation("TMP", "0")) %>%
    add(IfStatement("KA2 > 10", Equation("TMP", "1")))
  
  updatedMain <- main %>%
    addSuffix("A")
  
  expectedMain <- MainRecord() %>%
    add(Equation("KA_A", "THETA_KA * exp(ETA_KA)")) %>%
    add(Equation("KA2_A", "2 * KA_A")) %>%
    add(Equation("TMP_A", "0")) %>%
    add(IfStatement("KA2_A > 10", Equation("TMP_A", "1")))
  
  expect_equal(updatedMain, expectedMain)
})

test_that("Add suffix to compartments is working as expected", {
  
  compartments <- Compartments() %>%
    add(Compartment(index=1, name="ABS")) %>%
    add(Compartment(index=2, name="CENTRAL")) %>%
    add(Compartment(index=3, name="OUTPUT"))
  
  updatedCompartments <- compartments %>%
    addSuffix("A")
  
  expectedCompartments <- Compartments() %>%
    add(Compartment(index=1, name="ABS_A")) %>%
    add(Compartment(index=2, name="CENTRAL_A")) %>%
    add(Compartment(index=3, name="OUTPUT_A"))
  
  expect_equal(updatedCompartments, expectedCompartments)
})

test_that("Add suffix to a Campsis model is working as expected", {
  
  modelA <- model_suite$testing$pk$`1cpt_zo_abs_lag` %>%
    addSuffix("1", separator="") %>%
    add(LineBreak())
  
  modelB <- model_suite$testing$pk$`1cpt_zo_abs_lag` %>%
    addSuffix("2", separator="")
  
  modelAB <- modelA %>%
    add(modelB)

  expectedModelPath <- paste0(testFolder, "custom/", "model_suffix_ab")
  # modelAB %>% write(expectedModelPath)
  
  expectedModel <- read.campsis(expectedModelPath)
  expect_equal(modelAB, expectedModel)
  # library(campsis)
  # ds <- Dataset(50) %>%
  #   add(Infusion(time=0, amount=1000, compartment=1)) %>%
  #   add(Infusion(time=0, amount=2000, compartment=3)) %>%
  #   add(Observations(seq(0,24)))
  # results <- simulate(model=modelAB, dataset=ds, dest="mrgsolve", seed=1)
  # shadedPlot(results, "CONC1")
  # shadedPlot(results, "CONC2")
})

library(testthat)

context("Test model library")

test_that("Templates can be loaded successfully", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  expect_true(model@parameters %>% contains(Theta(index=1, name="K")))
})

library(testthat)

context("Test parameter S4 objects")

test_that("Working THETA parameter", {
  
  theta1 <- Theta(name="CL", index=1, fix=TRUE, value=4)
  expect_equal(theta1@name, "CL")
  expect_equal(theta1@index, 1)
  expect_equal(theta1@fix, TRUE)
  expect_equal(theta1 %>% getName, "THETA_CL")
})

test_that("THETA parameter, default name", {
  theta1 <- Theta(index=1, fix=TRUE, value=4)
  expect_equal(theta1@name, as.character(NA))
})

test_that("Invalid OMEGA parameters raise an error", {
  expect_error(Omega(name="CL_V", index=1, index2=2, value=0.25, type="corr"),
               regexp="Type should be one of: 'var', 'sd', 'covar', 'cor', 'cv' or 'cv%'")
  
  expect_error(Omega(name="CL_V", index=1, index2=2, value=0.25, type="var"),
               regexp="Parameter type must be 'covar' or 'cor' \\(index:1, index2:2\\)")
})

test_that("Check incorrect length of parameter", {
  expect_error(Theta(name=c("CL", "CL"), index=1, fix=TRUE, value=4))
})

test_that("isDial method", {
  
  omega2_2 <- Omega(name="CL", index=2, index2=2, fix=TRUE, value=0.2)
  expect_true(isDiag(omega2_2))

  omega1_2 <- Omega(name="CL", index=1, index2=2, fix=TRUE, value=0.2)
  expect_false(isDiag(omega1_2))
})

test_that("getNONMEMName method", {
  
  theta1 <- Theta(name="CL", index=1)
  expect_equal(theta1 %>% getNONMEMName(), "THETA(1)")
  
  omega1_2 <- Omega(name="CL", index=1, index2=2)
  expect_equal(omega1_2 %>% getNONMEMName(), "OMEGA(1,2)")
  
  sigma1_2 <- Sigma(name="CL", index=1, index2=2)
  expect_equal(sigma1_2 %>% getNONMEMName(), "SIGMA(1,2)")
})

test_that("OMEGA index2 can't be NA", {
  # Missing index2 argument
  expect_error(Omega(index=1))
  
  # Missing index2 argument
  expect_error(Omega(index=1, index2=NA))
})

test_that("As.data.frame method", {
  theta1 <- Theta(name="CL", index=1, min=0, max=Inf)
  df <- theta1 %>% as.data.frame(row.names=character(), optional=F)
  
  expect_equal(data.frame(name="CL", index=1, value=as.numeric(NA), min=as.numeric(0), max=as.numeric(Inf), fix=FALSE,
                          label=as.character(NA), unit=as.character(NA), comment=as.character(NA)), df)
})

test_that("Standardise method works as expected", {
  
  # Type not valid
  expect_error(Omega(index=1, index2=1, value=117, type="cv%%"))
  
  # Value in CV%
  omega <- Omega(index=1, index2=1, value=117, type="cv%")
  omega_ <- omega %>% standardise()
  expect_equal(round(omega_@value, 2), 0.86)
  
  # Value in CV
  omega <- Omega(index=1, index2=1, value=1.17, type="cv")
  omega_ <- omega %>% standardise()
  expect_equal(round(omega_@value, 2), 0.86)
  
  # Value in sd
  omega <- Omega(index=1, index2=1, value=0.2, type="sd")
  omega_ <- omega %>% standardise()
  expect_equal(omega_@value, 0.04)
  
  # Co-variance (covar)
  omega <- Omega(index=1, index2=2, value=0.2, type="covar")
  omega_ <- omega %>% standardise()
  expect_equal(omega, omega_) # No change
  
  # Correlation (cor)
  omega <- Omega(index=1, index2=2, value=0.25, type="cor")
  expect_error(omega %>% standardise(), regexp="Argument 'parameters' is needed to convert a covariance into a correlation")
})

test_that("Malformed omega is not valid and can't be standardised", {
  omega1 <- Omega(index=1, index2=2, value=0, type="covar")
  
  # Force type to be unknown
  omega1@type <- "unknown"
  
  # Check omega is not valid
  expect_error(validObject(omega1), regexp="Parameter type must be 'covar' or 'cor'")
  
  # Check omega can't be standardised
  expect_error(omega1 %>% standardise(), regexp="Type of parameter OMEGA_1_2 must be 'covar' or 'cor'")
  
  omega2 <- Omega(index=2, index2=2, value=0, type="var")
  
  # Force type to be unknown
  omega2@type <- "unknown"
  
  # Check omega is not valid
  expect_error(validObject(omega2), regexp="Type should be one of: 'var', 'sd', 'covar', 'cor', 'cv' or 'cv%'")
  
  # Check omega can't be standardised
  expect_error(omega2 %>% standardise(), regexp="Type should be one of: 'var', 'sd', 'cv' or 'cv%'")
  
  # Force type to be covar
  omega2@type <- "covar"
  
  # Check omega can't be standardised
  expect_error(omega2 %>% standardise(), regexp="Type of parameter OMEGA_2_2 can't be 'covar'")
})

library(testthat)

context("Test read/write methods on Campsis model")

source(paste0("", "testUtils.R"))

advanFilename <- function(advan, trans, ext=".txt") {
  return(paste0("advan", advan, "_trans", trans, ext))
}

writePath <- function(modelName) {
  return(paste0(testFolder, "write/models/", modelName))
}

test_that("Write/Read ADVAN1 TRANS1", {
  modelName <- "advan1_trans1"
  model <- model_suite$testing$nonmem[[modelName]]
  
  # write
  model %>% write(file=writePath(modelName))
  
  # read
  model2 <- read.campsis(file=writePath(modelName))

  # Check equality  
  expect_equal(model, model2)
})

test_that("Write/Read ADVAN3 TRANS4 with variance-covariance matrix", {
  modelName <- "my_model1"
  model <- model_suite$testing$other[[modelName]]
  
  # write
  model %>% write(file=writePath(modelName))
  
  # read
  model2 <- read.campsis(file=writePath(modelName))
  
  # Check equality  
  expect_equal(model, model2)
})

test_that("Write/Read ADVAN4 TRANS4 with various compartment properties", {
  modelName <- "advan4_trans4"
  model <- model_suite$testing$nonmem[[modelName]]
  
  # Add a few properties
  model <- model %>% add(LagTime(1, "ALAG1"))
  model <- model %>% add(Bioavailability(1, "F1"))
  model <- model %>% add(InfusionDuration(1, "D1"))
  model <- model %>% add(InfusionRate(2, "R2"))
  model <- model %>% add(InitialCondition(2, "100"))
  
  # Sort properties
  model <- model %>% sort()
  
  # Write
  model %>% write(file=writePath(modelName))
  prop <- model@compartments@properties
  # Read
  model2 <- read.campsis(file=writePath(modelName))
  prop2 <- model2@compartments@properties
  
  # Check equality  
  expect_equal(model, model2)
})

test_that("Model parameters can be annotated and persisted (fields label, unit, comment)", {
  model <- model_suite$testing$pk$'1cpt_fo'
  modelName <- "model_1cpt_fo_annotated"
  
  model <- model %>%
    replace(Theta(name="KA", value=1, label="Absorption rate", unit="/h")) %>%
    replace(Omega(name="KA", value=25, type="cv%", label="IIV on absorption")) %>%
    replace(Sigma(name="RUV_FIX", value=1, fix=TRUE, label="Proportional error", comment="Fixed epsilon, multiplied by THETA_PROP_RUV in model code."))
  
  # Write
  model %>% write(file=writePath(modelName))
  
  # Read
  model2 <- read.campsis(file=writePath(modelName))
  
  # Check equality  
  expect_equal(model, model2)
  
})

test_that("Model parameters with min and max values can be persisted correctly", {
  model <- model_suite$testing$pk$'1cpt_fo'
  modelName <- "model_1cpt_fo_min_max"
  
  model <- model %>%
    setMinMax("theta", min=0, max=Inf) %>%
    setMinMax(Sigma("RUV_FIX"), min=1, max=1)
  
  # Write
  model %>% write(file=writePath(modelName))
  
  # Read
  model2 <- read.campsis(file=writePath(modelName))
  
  # Check equality  
  expect_equal(model, model2)
})

#_______________________________________________________________________________
#----                        compartments class                             ----
#_______________________________________________________________________________

#' 
#' Compartments class.
#' 
#' @slot properties compartment properties of the compartments defined in this class
#' @export
setClass(
  "compartments",
  representation(
    properties="compartment_properties"
  ),
  contains = "pmx_list",
  prototype = prototype(type="compartment", properties=new("compartment_properties"))
)

#' 
#' Create a list of compartments
#' 
#' @return an empty list of compartments  
#' @export
Compartments <- function() {
  return(new("compartments"))
}

#_______________________________________________________________________________
#----                                add                                    ----
#_______________________________________________________________________________

#' @rdname add
setMethod("add", signature = c("compartments", "compartment_property"), definition = function(object, x) {
  object@properties <- object@properties %>% add(x) 
  return(object)
})

#' @rdname add
setMethod("add", signature=c("compartments", "compartments"), definition=function(object, x) {
  return(object %>% appendCompartments(x))
})

#' Append compartments.
#' 
#' @param compartments1 base set of compartments
#' @param compartments2 extra set of compartments to be appended
#' @return the resulting set of compartments
#' @keywords internal
#' 
appendCompartments <- function(compartments1, compartments2) {
  
  cmtNames1 <- compartments1@list %>% purrr::map_chr(~.x %>% toString())
  cmtNames2 <- compartments2@list %>% purrr::map_chr(~.x %>% toString())
  cmtMax <- compartments1 %>% length()
  
  checkCollisionOnCmts <- cmtNames1 %in% cmtNames2
  if (any(checkCollisionOnCmts)) {
    stop(paste0("Model can't be appended because of duplicate compartment name(s): ", paste0(cmtNames1[checkCollisionOnCmts], collapse=", ")))
  }
  for (compartment in compartments2@list) {
    compartment@index <- compartment@index + cmtMax
    compartments1 <- compartments1 %>% add(compartment)
  }
  for (property in compartments2@properties@list) {
    property@compartment <- property@compartment + cmtMax
    compartments1 <- compartments1 %>% add(property)
  }
  return(compartments1 %>% sort())
}

#_______________________________________________________________________________
#----                               delete                                  ----
#_______________________________________________________________________________

#' @rdname delete
setMethod("delete", signature=c("compartments", "compartment_property"), definition=function(object, x) {
  object@properties <- object@properties %>% delete(x)
  return(object)
})

#_______________________________________________________________________________
#----                                 find                                  ----
#_______________________________________________________________________________

#' @rdname find
setMethod("find", signature=c("compartments", "compartment_property"), definition=function(object, x) {
  return(object@properties %>% find(x))
})

#_______________________________________________________________________________
#----                          getCompartmentIndex                          ----
#_______________________________________________________________________________

#' @rdname getCompartmentIndex
setMethod("getCompartmentIndex", signature=c("compartments", "character"), definition=function(object, name) {
  compartment <- object@list %>% purrr::detect(~.x@name == name)
  if (compartment %>% length() == 0) {
    stop(paste0("Compartment ", name, " not found."))
  }
  return(compartment@index)
})

#_______________________________________________________________________________
#----                               replace                                 ----
#_______________________________________________________________________________

#' @rdname replace
setMethod("replace", signature=c("compartments", "compartment_property"), definition=function(object, x) {
  object@properties <- object@properties %>% replace(x)
  return(object)
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("compartments"), definition=function(object) {
  cat("Compartments:\n")
  for (element in object@list) {
    show(element)
    cat("\n")
  }
})

#_______________________________________________________________________________
#----                                  sort                                 ----
#_______________________________________________________________________________

#' @rdname sort
setMethod("sort", signature=c("compartments"), definition=function(x, decreasing=FALSE, ...) {
  # Sort compartment properties
  x@properties <- x@properties %>% sort()
  return(x)
})

#' Has exact method allows to check the existence of a S4 method in Campsis
#' based on its signature.
#' 
#' @param generic generic function name
#' @param signature function signature
#' @param where where to search functions
#' @return logical value
#' @importFrom methods getGeneric findMethods isGeneric
#' @export
#' 
hasExactMethod <- function(generic, signature, where=topenv(parent.frame())) {
  if (!methods::isGeneric(generic, where = where))
    return(FALSE)
  gen <- methods::getGeneric(generic, where=where)
  ml <- methods::findMethods(gen)
  target <- paste(signature, collapse = "#")
  return(target %in% names(ml))
}

#' Map JSON properties to S4 slots.
#' 
#' @param object S4 object
#' @param json JSON element, json_element class
#' @param discard_type discard JSON property 'type'
#' @return a S4 object
#' @export
#' 
mapJSONPropertiesToS4Slots <- function(object, json, discard_type=TRUE) {
  json <- json@data
  properties <- names(json)
  if (discard_type) {
    properties <- properties[properties != "type"]
  }
  
  for (property in properties) {
    value <- json[[property]]
    isList <- is.list(value)
    
    if (isList && !is.null(value$type)) {
      # Recursion
      if (hasExactMethod(generic="loadFromJSON", signature=c(value$type, "json_element"))) {
        value <- loadFromJSON(object=new(value$type), json=JSONElement(value))
      } else {
        value <- mapJSONPropertiesToS4Slots(object=new(value$type),
                                            json=JSONElement(value), discard_type=TRUE)
      }
    } else {
      if (isList) {
        value <- unlist(value)
      }
      if (is.null(value)) {
        value <- character(0)
      }
    }
    slot(object, property) <- value
  }
  return(object)
}

#' Map S4 slots to JSON properties.
#' 
#' @param object S4 object
#' @param add_type add type as a property, TRUE by default
#' @param optional properties that are optional in JSON, character vector
#' @param ignore slots to be ignored
#' @return a JSON object ready to be serialised
#' @export
#' 
mapS4SlotsToJSONProperties <- function(object, add_type=TRUE, optional=NULL, ignore=NULL) {
  if (!isS4(object)) {
    stop("Input must be an S4 object.")
  }
  
  # Initialize list for JSON properties
  json <- list()
  
  # Optionally add the type field
  if (add_type) {
    json$type <- class(object)[[1]]
  }
  
  # Iterate over slots
  slotNames <- slotNames(object)
  for (property in slotNames[!slotNames %in% ignore]) {
    value <- slot(object, property)
    
    if (isS4(value)) {
      # Recursive call for nested S4
      json[[property]] <- mapS4SlotsToJSONProperties(value, add_type=TRUE)
      
    } else if (is.list(value)) {
      # Handle lists: check if elements are S4
      json[[property]] <- lapply(value, function(v) {
        if (isS4(v)) {
          mapS4SlotsToJSONProperties(v, add_type=TRUE)
        } else {
          v
        }
      })
      
    } else if (length(value) == 0) {
      # Map empty slot to NULL
      if (!property %in% optional) {
        json[[property]] <- NULL
      }

    } else {
      # Atomic vectors, scalars, etc.
      if (!(property %in% optional && is.na(value))) {
        json[[property]] <- value
      }
    }
  }
  
  return(json)
}

#' JSON to Campsis dataset.
#' 
#' @param object empty dataset
#' @param json json element
#' @return Campsis dataset
#' @importFrom jsonlite parse_json
#' @importFrom purrr keep imap map flatten_chr
#' @keywords internal
#' 
jsonToCampsisModel <- function(object, json) {

  json <- json@data
  model <- object
  
  # Parse model code
  text <- unlist(json$code)
  if (!is.null(text)) {
    model@model <- read.model(text=text)
  }
  
  # Parse parameters
  jsonParameters <- json$parameters
  jsonThetas <- jsonParameters %>%
    purrr::keep(~.x$type=="theta")
  jsonOmegasOnDiag <- jsonParameters %>%
    purrr::keep(~.x$type=="omega" && is.null(.x$name2))
  jsonOmegasOffDiag <- jsonParameters %>%
    purrr::keep(~.x$type=="omega" && !is.null(.x$name2))
  jsonSigmasOnDiag <- jsonParameters %>%
    purrr::keep(~.x$type=="sigma" && is.null(.x$name2))
  jsonSigmasOffDiag <- jsonParameters %>%
    purrr::keep(~.x$type=="sigma" && !is.null(.x$name2))

  thetas <- jsonThetas %>%
    purrr::imap(~jsonToParameter(x=.x, index=.y, index2=.y))
  omegas <- jsonOmegasOnDiag %>%
    purrr::imap(~jsonToParameter(x=.x, index=.y, index2=.y))
  sigmas <- jsonSigmasOnDiag %>%
    purrr::imap(~jsonToParameter(x=.x, index=.y, index2=.y))
  
  omegaNames <- omegas %>%
    purrr::map(~.x@name)
  sigmaNames <- sigmas %>%
    purrr::map(~.x@name)
  
  omegasOffDiag <- jsonOmegasOffDiag %>%
    purrr::map(~jsonToOffDiagParameter(json=.x, diag_names=omegaNames))
  sigmasOffDiag <- jsonSigmasOffDiag %>%
    purrr::map(~jsonToOffDiagParameter(json=.x, diag_names=sigmaNames))
  
  model@parameters@list <- c(thetas, omegas, omegasOffDiag, sigmas, sigmasOffDiag)
  
  # Update compartments
  model <- model %>%
    updateCompartments()
  
  # Sort model parameters
  model <- model %>%
    campsismod::sort()
  
  # Parse variance-covariance matrix
  varcov <- json$varcov
  if (length(varcov) > 0) {
    # Find all possible parameter names and initialize the matrix
    rowNames <- varcov %>%
      purrr::map(~c(findVarcovParameter(ref=.x$ref1, model=model) %>% getName(),
                    findVarcovParameter(ref=.x$ref2, model=model) %>% getName())) %>%
      purrr::flatten_chr() %>%
      unique()
    
    matrix <- matrix(0L, nrow=length(rowNames), ncol=length(rowNames))
    dimnames(matrix) <- list(rowNames, rowNames)
    
    # Fill in with values
    for (entry in varcov) {
      ref1Name <- findVarcovParameter(ref=entry$ref1, model=model) %>% getName()
      ref2Name <- findVarcovParameter(ref=entry$ref2, model=model) %>% getName()
      matrix[ref1Name, ref2Name] <- entry$cov
      matrix[ref2Name, ref1Name] <- entry$cov
    }
    
    model@parameters@varcov <- matrix
  }
  return(model)
}

findVarcovDoubleArrayParameter <- function(ref, model, type) {
  if (type=="omega") {
    paramRef = Omega()
  } else if (type=="sigma") {
    paramRef = Sigma()
  } else {
    stop("type must be 'omega' or 'sigma'")
  }
  if (is.null(ref$name2)) {
    paramRef@name <- ref$name
    retValue <- model %>% find(paramRef)
  } else {
    # First attempt
    paramRef@name <- paste0(ref$name, "_", ref$name2)
    retValue <- model %>% find(paramRef)
    # Second attempt
    if (is.null(retValue)) {
      paramRef@name <- paste0(ref$name2, "_", ref$name)
      retValue <- model %>% find(paramRef)
    }
  }
  return(retValue)
}

findVarcovParameter <- function(ref, model) {
  if (ref$type=="theta_ref") {
    retValue <- model %>% find(Theta(name=ref$name))
  } else if (ref$type=="omega_ref") {
    retValue <- findVarcovDoubleArrayParameter(ref=ref, model=model, type="omega")
  } else if (ref$type=="sigma_ref") {
    retValue <- findVarcovDoubleArrayParameter(ref=ref, model=model, type="sigma")
  }
  if (is.null(retValue)) {
    if (is.null(ref$name2)) {
      stop(sprintf("Parameter reference not found (type: %s, name: %s)", ref$type, ref$name))
    } else {
      stop(sprintf("Parameter reference not found (type: %s, name: %s, name2: %s)", ref$type, ref$name, ref$name2))
    }
  }
  return(retValue)
}

#' Convert JSON correlation parameter (OMEGA or SIGMA) into a Campsis parameter.
#' 
#' @param json JSON data
#' @param diag_names parameter names on the diagonal, character vector
#' @return the corresponding Campsis parameter
#' 
jsonToOffDiagParameter <- function(json, diag_names) {
  name <- json$name
  name2 <- json$name2
  index <- which(diag_names==name)
  index2 <- which(diag_names==name2)
  return(jsonToParameter(x=json, index=index, index2=index2))
}

#' Process JSON double array parameter.
#' 
#' @param x JSON data, OMEGA or SIGMA parameter
#' @return updated JSON data with updated 'name' field and removed 'name2' field
#' 
processJSONDoubleArrayParameter <- function(x) {
  if (!is.null(x$name2)) {
    x$name <- paste0(x$name, "_", x$name2)
    x$name2 <- NULL
  }
  return(x)
}

#' JSON to Campsis parameter.
#' 
#' @param x JSON data
#' @param index parameter index to add
#' @param index2 second parameter index to add for OMEGAs and SIGMAs
#' @return Campsis parameter
#' @export
#' 
jsonToParameter <- function(x, index=NULL, index2=NULL) {
  if (x$type=="theta") {
    if (is.null(index)) {
      theta <- Theta()
    } else {
      theta <- Theta(index=index)
    }
    x$type <- NULL
    return(loadFromJSON(object=theta, JSONElement(x)))
    
  } else if (x$type=="omega") {
    x <- processJSONDoubleArrayParameter(x)
    if (is.null(index)) {
      omega <- Omega()
    } else {
      omega <- Omega(index=index, index2=index2)
    }
    x$type <- x$var_type
    x$var_type <- NULL
    return(loadFromJSON(object=omega, JSONElement(x)))
    
  } else if (x$type=="sigma")  {
    x <- processJSONDoubleArrayParameter(x)
    if (is.null(index)) {
      sigma <- Sigma()
    } else {
      sigma <- Sigma(index=index, index2=index2)
    }
    x$type <- x$var_type
    x$var_type <- NULL
    return(loadFromJSON(object=sigma, JSONElement(x)))
    
  } else {
    stop("Unknown parameter type")
  }
}


#' Open JSON file.
#' 
#' @param json JSON in its string form or path to JSON file
#' @param schema JSON schema
#' 
#' @return parsed JSON object
#' @importFrom jsonlite parse_json
#' @importFrom jsonvalidate json_schema
#' @keywords internal
#' 
openJSON <- function(json, schema=NULL) {
  if (is.list(json)) {
    return(JSONElement(json)) # Don't go further if data is already parsed
  }
  assertthat::assert_that(length(json)==1, msg="Argument json must be a path or the JSON string")
  
  if (grepl(pattern="\\s*[\\[\\{]", x=json)) {
    rawJson <- json
  } else {
    rawJson <- suppressWarnings(paste0(readLines(json), collapse="\n"))
  }
  
  # Validate content against schema
  if (getCampsismodOption(name="VALIDATE_JSON", default=TRUE)) {
    obj <- jsonvalidate::json_schema$new(schema)
    obj$validate(rawJson, error=TRUE)
  }
  
  json_ <- jsonlite::parse_json(rawJson, simplifyVector=FALSE)
  
  return(JSONElement(json_))
} 

#_______________________________________________________________________________
#----                             ode class                                 ----
#_______________________________________________________________________________

#' 
#' ODE class. Any statement in the form d/dt(A_CMT) = B.
#' 
#' @export
setClass(
  "ode",
  representation(
  ),
  contains = "equation",
  validity = function(object) {
    if (startsWith(object@lhs, "A_")) {
      return(TRUE)
    } else {
      return("Derivative name must start with 'A_'")
    }
  }
)

#' 
#' Create a new ordinary differential equation (ODE).
#' 
#' @param lhs left-hand side variable corresponding to derivative name, must start with 'A_'
#' @param rhs right-hand side expression corresponding to derivative value
#' @param comment comment if any, single character string
#' @return an ODE
#' @export
Ode <- function(lhs, rhs="", comment=as.character(NA)) {
  return(new("ode", lhs=lhs, rhs=rhs, comment=comment))
}

#_______________________________________________________________________________
#----                            getName                                    ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature = c("ode"), definition = function(x) {
  return(paste0("ODE (", x@lhs, ")"))
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("ode"), definition=function(object, ...) {
  dest <- processExtraArg(args=list(...), name="dest", default="campsis")
  model <- processExtraArg(args=list(...), name="model", default=CampsisModel())
  
  if (dest=="campsis" || isRxODE(dest)) {
    retValue <- paste0("d/dt(", object@lhs, ")", "=", object@rhs)
  } else if (dest=="mrgsolve") {
    retValue <- paste0("dxdt_", object@lhs, "=", object@rhs, ";")
  } else if (dest=="NONMEM") {
    retValue <- paste0("DADT(", model %>% getCompartmentIndex(gsub("A_", "", object@lhs)), ")", "=", object@rhs)
  } else {
    UnsupportedDestException()
  }
  return(retValue %>% appendComment(object, dest))
})

library(testthat)

context("Test the very basic conversion to NONMEM")
dest <- "NONMEM"

test_that("All structural elements can be converted to NONMEM (this is still a basic conversion)", {
  equation <- Equation("KA", "THETA(1)*EXP(ETA(1))", comment="This is a comment")
  expect_equal(equation %>% toString(dest=dest), "KA=THETA(1)*EXP(ETA(1)) ; This is a comment")
  
  linebreak <- LineBreak()
  expect_equal(linebreak %>% toString(dest=dest), "")
  
  ifStatement <- IfStatement("COV.EQ.1", equation=Equation("COV_EFFECT", "0.1"), comment="This is a comment")
  expect_equal(ifStatement %>% toString(dest=dest), "IF (COV.EQ.1) COV_EFFECT=0.1 ; This is a comment")
  
  model <- model_suite$testing$nonmem$advan1_trans1
  central <- model %>% find(Ode("A_CENTRAL"))  
  expect_equal(central %>% toString(dest=dest, model=model), "DADT(1)=-K*A_CENTRAL")
  
  comment <- Comment("This is a comment")
  expect_equal(comment %>% toString(dest=dest), "; This is a comment")
  
  text <- UnknownStatement("THIS IS UNKNOWN TEXT")
  expect_equal(text %>% toString(dest=dest), "THIS IS UNKNOWN TEXT")
})

#_______________________________________________________________________________
#----                           omega_blocks class                          ----
#_______________________________________________________________________________

setClass(
  "omega_blocks",
  representation(
  ),
  contains="pmx_list",
  prototype=prototype(type="omega_block")
)

#' Create a list of OMEGA blocks.
#'
#' @export
OmegaBlocks <- function() {
  return(new("omega_blocks"))
}

#_______________________________________________________________________________
#----                                add                                    ----
#_______________________________________________________________________________

addOmega <- function(object, x) {
  block <- object %>% getOmegaBlock(x)
  newBlock <- FALSE
  if (is.null(block)) {
    block <- OmegaBlock()
    newBlock <- TRUE
  }
  block <- block %>% add(x)
  if (newBlock) {
    object <- object %>% add(block)
  } else {
    object <- object %>% campsismod::replace(block)
  }
  return(object)
}

#' @rdname add
setMethod("add", signature=c("omega_blocks", "omega_block"), definition=function(object, x) {
  x@block_index <- object %>% length() + 1L
  return(methods::callNextMethod(object, x))
})

#' @rdname add
setMethod("add", signature=c("omega_blocks", "parameters"), definition=function(object, x) {
  retValue <- OmegaBlocks()

  # Keeping only the OMEGA's and SIGMA's
  x@list <- x@list %>%
    purrr::keep(~is(.x, "double_array_parameter"))
  x <- x %>% sort() # Sort
  
  off_diag_omegas <- x # Init
  on_diag_omegas <- x # Init
  off_diag_omegas@list <- off_diag_omegas@list %>% purrr::keep(~!isDiag(.x))
  on_diag_omegas@list <- on_diag_omegas@list %>% purrr::keep(~isDiag(.x))

  for (omega in off_diag_omegas@list) {
    retValue <- retValue %>% addOmega(omega)
  }
  for (omega in on_diag_omegas@list) {
    retValue <- retValue %>% addOmega(omega)
  }
  
  # Sort all blocks (see method below)
  retValue <- retValue %>% sort()
  
  # Set up the start index
  cumulatedIndex <- 0L
  for (block in retValue@list) {
      block@start_index <- cumulatedIndex
      cumulatedIndex <- cumulatedIndex + block %>% length()
      block <- block %>% shiftOmegaIndexes()
      retValue <- retValue %>% replace(block)
  }
  return(retValue)
})

#_______________________________________________________________________________
#----                            getOmegaBlock                              ----
#_______________________________________________________________________________

#' Get the right block of OMEGA's.
#' 
#' @param object list of OMEGA blocks
#' @param x omega param
#' @return the corresponding OMEGA block or NULL if not found
#' @export
#' @rdname getOmegaBlock
getOmegaBlock <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("getOmegaBlock", function(object, x) {
  standardGeneric("getOmegaBlock")
})

#' @rdname getOmegaBlock
setMethod("getOmegaBlock", signature=c("omega_blocks", "double_array_parameter"), definition=function(object, x) {
  for (block in object@list) {
    indexes <- block %>% getOmegaIndexes()
    if (x@index %in% indexes || x@index2 %in% indexes) {
      return(block)
    }
  }
  return(NULL)
})

#_______________________________________________________________________________
#----                               sort                                    ----
#_______________________________________________________________________________

#' @rdname sort
setMethod("sort", signature=c("omega_blocks"), definition=function(x, decreasing=FALSE, ...) {
  temp <- x@list %>% purrr::map_int(.f=~.x %>% getOmegaIndexes() %>% min())
  x@list <- x@list[order(temp)]
  return(x)
})

#_______________________________________________________________________________
#----                        if_statement class                             ----
#_______________________________________________________________________________

#' 
#' If-statement class. Any statement in the form if (condition) A = B.
#' 
#' @slot condition IF statement condition
#' @slot equation any equation or ODE
#' @export
setClass(
  "if_statement",
  representation(
    condition = "character",
    equation = "equation"
  ),
  contains = "model_statement",
  validity = function(object) {
    return(expectOne(object, "condition"))
  }
)

#' 
#' Create a new IF-statement.
#' 
#' @param condition condition, single character string
#' @param equation equation if condition is met
#' @param comment comment if any, single character string
#' @return an IF-statement
#' @export
IfStatement <- function(condition, equation, comment=as.character(NA)) {
  return(new("if_statement", condition=condition, equation=equation, comment=comment))
}

#_______________________________________________________________________________
#----                            getName                                    ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature = c("if_statement"), definition = function(x) {
  return(paste0("IF (", x@condition, ") ", x@equation %>% getName()))
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("if_statement", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  object@condition <- object@condition %>% replaceAll(pattern=pattern, replacement=replacement, ...)
  object@equation <- object@equation %>% replaceAll(pattern=pattern, replacement=replacement, ...)
  return(object)
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("if_statement"), definition=function(object, ...) {
  dest <- processExtraArg(args=list(...), name="dest", default="campsis")
  if (dest=="campsis" || isRxODE(dest) || dest=="mrgsolve") {
    retValue <- paste0("if (", object@condition, ") ", object@equation %>% toString(dest=dest, init=FALSE))
  } else if (dest=="NONMEM") {
    retValue <- paste0("IF (", object@condition, ") ", object@equation %>% toString(dest=dest, init=FALSE))
  } else {
    UnsupportedDestException()
  }
  return(retValue %>% appendComment(object, dest))
})

getSlotLength <- function(object, slot) {
  return(length(eval(parse(text = paste0("object@", slot)))))
}

checkLength <- function(object, slot, expected=1) {
  lengthSlot <- getSlotLength(object, slot)
  error <- character()
  if (lengthSlot != expected) {
    error <- paste0(slot, " is length ", lengthSlot, ". Should be ", expected, ".")
  }
  return(error)
}

expectOne <- function(object, slot) {
  return(checkLength(object, slot, expected=1))
}

addError <- function(error, errors) {
  if (length(error)==0) {
    return(errors)
  } else {
    return(c(errors, error))
  }
}

expectOneForAll <- function(object, attrs) {
  errors <- character()
  
  for(attr in attrs) {
    errors <- addError(expectOne(object, attr), errors)
  }
  
  return(errors)
}

#_______________________________________________________________________________
#----                         code_records class                            ----
#_______________________________________________________________________________

setClass(
  "code_records",
  representation(
  ),
  contains = "pmx_list",
  prototype = prototype(type="code_record"),
  validity = function(object) {
    hasUnknownStatements <-
      object@list %>% purrr::map_lgl(~.x@statements@list %>%
                                       purrr::map_lgl(~is(.x, "unknown_statement")) %>% any()) %>% any()
    if (hasUnknownStatements) {
      warning(
        "Model code contains unknown statements. Conversion to rxode2 and mrgsolve may lead to errors."
      )
    }
    return(TRUE)
  }
)

#' 
#' Create a list of code records.
#' 
#' @return an empty list of code records  
#' @export
CodeRecords <- function() {
  return(new("code_records"))
}

#_______________________________________________________________________________
#----                                add                                    ----
#_______________________________________________________________________________

#' @rdname add
setMethod("add", signature=c("code_records", "code_records"), definition=function(object, x) {
  return(object %>% appendCodeRecords(x))
})

#' Append code records
#' 
#' @param records1 base set of code records
#' @param records2 extra set of code records to be appended
#' @return the resulting set of code records
#' @keywords internal
#' 
appendCodeRecords <- function(records1, records2) {
  for (record in (records2)@list) {
    baseRecord <- records1 %>% getByName(record %>% getName())
    if (baseRecord %>% length() == 0) {
      records1 <- records1 %>% add(record)
    } else {
      baseRecord <- baseRecord %>% add(record)
      records1 <- records1 %>% replace(baseRecord)
    }
  }
  return(records1 %>% sort())
}

findRecordByPosition <- function(object, pos) {
  if (is.null(pos)) {
    stop("pos is null")
  }
  for (record in object@list) {
    if (pos@by_index) {
      # Return first record
      return(record)
    } else if (pos@by_element) {
      # Element here should be a model statement
      # Check if current record contains the given element
      if (record %>% find(pos@element) %>% length() > 0) {
        return(record)
      }
    } else {
      stop("pos either by index or by element")
    }
  }
  stop("No code record found for specified position 'pos'")
}

#' @param pos position where x needs to be added in list
#' @rdname add
setMethod("add", signature=c("code_records", "model_statement"), definition=function(object, x, pos=NULL) {
  if (is.null(pos)) {
    if (is(x, "ode")) {
      recordName <- "ODE"
    } else {
      recordName <- "MAIN"
    }
    record <- object %>% getByName(recordName)
    if (isS4(record)) {
      # Existing record
      record <- record %>% add(x)
      object <- object %>% replace(record)
    } else {
      # New record
      record <- new(paste0(recordName %>% tolower(), "_record"))
      record <- record %>% add(x)
      object <- object %>% add(record) %>% sort()
    }
  } else {
    
    if (pos@by_element && is(pos@element, "code_record")) {
      # Special case (position by code record element)
      record <- object %>% find(pos@element)
      newRecordNeeded <- is.null(record)
      if (newRecordNeeded) {
        record <- pos@element
      }
      if (pos@after) {
        record <- record %>% add(x) # x is appended to the end
      } else {
        record <- record %>% add(x, Position(1, after=FALSE)) # x added at the beginning
      }
      if (newRecordNeeded) {
        object <- object %>% add(record)
      } else {
        object <- object %>% replace(record)
      }
      
    } else {
      # Usual case (position by index or model statement element)
      record <- object %>% findRecordByPosition(pos)
      record <- record %>% add(x, pos=pos)
      object <- object %>% replace(record)
    }
    
  }
  return(object)
})

#_______________________________________________________________________________
#----                         addPropertiesRecords                           ----
#_______________________________________________________________________________

addPropertiesRecords <- function(records, model) {
  properties <- model@compartments@properties
  
  for (name in getRecordNames()) {
    record <- new(tolower(paste0(name, "_record")))
    if (!is(record, "properties_record")) {
      next
    }
    subProperties <- properties %>% select(name) 
    if (subProperties %>% length() == 0) {
      next
    }
    for (subProperty in subProperties@list) {
      compartment <- model@compartments %>% find(Compartment(index=subProperty@compartment))
      equation <- Equation(compartment %>% toString(), subProperty@rhs, comment=subProperty@comment)
      record <- record %>% add(equation)
    }
    records <- records %>% add(record)
  }
  return(records %>% sort())
}

#_______________________________________________________________________________
#----                               delete                                  ----
#_______________________________________________________________________________

#' @rdname delete
setMethod("delete", signature=c("code_records", "model_statement"), definition=function(object, x) {
  copy <- object
  for (record in object@list) {
    if (!is.null(record %>% find(x))) {
      copy <- copy %>% replace(record %>% delete(x))
    }
  }
  return(copy)
})

#_______________________________________________________________________________
#----                                find                                   ----
#_______________________________________________________________________________

#' @rdname find
setMethod("find", signature=c("code_records", "model_statement"), definition=function(object, x) {
  for (record in object@list) {
    statement <- record %>% find(x)
    if (!is.null(statement)) {
      return(statement)
    }
  }
  return(NULL)
})

#_______________________________________________________________________________
#----                          getCompartments                              ----
#_______________________________________________________________________________

#' Add ODE compartment to compartments object.
#'
#' @param compartments compartments object
#' @param ode ODE
#' @return a compartments object
#' @importFrom purrr discard map_chr
#' @keywords internal
#' 
addODECompartment <- function(compartments, ode) {
  if (!is(ode, "ode")) {
    stop("ode is not an ODE")
  }
  name <- ode@lhs
  
  cmtIndex <- compartments %>% length() + 1
  if (startsWith(name, prefix="A_")) {
    name <- gsub("^A_", "", name)
    if (name == as.character(cmtIndex)) {
      name <- NA
    }
  } else {
    stop(paste0("Compartment ", name, " does not start with 'A_'"))
  }
  compartment <- Compartment(index=cmtIndex, name=name)
  names <- compartments@list %>% purrr::map_chr(~.x@name) %>% purrr::discard(is.na)
  
  # Add compartment only if not there yet
  if (is.na(name) || !(name %in% names)) {
    compartments <- compartments %>% add(compartment)
  }
  return(compartments)
}

#' Detect all compartments names from the code records.
#' Only for model instantiation. Not exported.
#'
#' @param records code records
#' @return a list of compartments
#' @keywords internal
#' 
getCompartments <- function(records) {
  assertthat::assert_that(is(records, "code_records"), msg="records class is not 'code_records'")
  odeRecord <- records %>% getByName("ODE")
  compartments <- Compartments()
  if (odeRecord %>% length() == 0) {
    return(compartments)
  }
  for (statement in odeRecord@statements@list) {
    if (is(statement, "ode")) {
      compartments <- compartments %>% addODECompartment(ode=statement)
    }
  }
  return(compartments)
}

#' Add properties to compartments object.
#'
#' @param compartments compartments object
#' @param records all records
#' @param name record name to look at
#' @param init empty characteristic, to be completed
#' @return updated compartments object
#' @keywords internal
#' 
addProperties <- function(compartments, records, name, init) {
  record <- records %>% getByName(name)
  if (record %>% length() == 0) {
    return(compartments)
  }
  # Filter on equations (line breaks and comments are accepted in properties record)
  for (equation in record@statements@list %>% purrr::keep(~is(.x, "equation"))) {
    cmtName <- equation@lhs
    compartment <- compartments@list %>% purrr::detect(~.x %>% toString() == cmtName)
    if (is.null(compartment)) {
      stop(paste0("Compartment undefined: '", cmtName, "' in record ", record %>% getName()))
    }
    property <- init
    property@compartment <- compartment@index
    property@rhs <- equation@rhs
    property@comment <- equation@comment 
    compartments <- compartments %>% add(property)
  }
  return(compartments)
}

#_______________________________________________________________________________
#----                             getRecordNames                            ----
#_______________________________________________________________________________

getRecordNames <- function() {
  return(c("MAIN", "ODE", "F", "LAG", "DURATION", "RATE", "INIT", "ERROR"))
}

#_______________________________________________________________________________
#----                              move                                     ----
#_______________________________________________________________________________

#' @rdname move
setMethod("move", signature=c("code_records", "model_statement", "pmx_position"), definition=function(object, x, to, ...) {
  # Find statement
  statement <- object %>%
    find(x)
  if (is.null(statement)) {
    stop(paste("Statement", x %>% getName(), "not found in model"))
  }
  
  # Delete statement
  object <- object %>%
    delete(x)
  
  # Add statement at right position
  object <- object %>%
    add(statement, pos=to)
  
  return(object)
})

#' @rdname move
setMethod("move", signature=c("code_records", "list", "pmx_position"), definition=function(object, x, to, ...) {
  for (statement in x) {
    object <- object %>%
      move(x=statement, to=to, ...)
  }
  return(object)
})

#' @rdname move
setMethod("move", signature=c("code_records", "model_statements", "pmx_position"), definition=function(object, x, to, ...) {
  object <- object %>%
    move(x=x@list, to=to, ...)
  return(object)
})

#' @rdname move
setMethod("move", signature=c("code_records", "code_record", "pmx_position"), definition=function(object, x, to, ...) {
  record <- object %>%
    find(x)
  if (is.null(record)) {
    stop(paste("Record", x %>% getName(), "not found in model"))
  }
  object <- object %>%
    move(x=record@statements, to=to, ...)
  return(object)
})

#_______________________________________________________________________________
#----                                read.model                             ----
#_______________________________________________________________________________

#' Remove all trailing line breaks.
#' 
#' @param x character vector
#' @return a character vector
#' @keywords internal
#' 
removeTrailingLineBreaks <- function(x) {
  lenX <- x %>% length()
  if (lenX > 0) {
    res <- rle(isEmptyLine(x))
    lastValue <- res$values[res$values %>% length()]
    if (lastValue) {
      len <- res$lengths[res$values %>% length()]
      x <- x[-((lenX-len+1):lenX)]
    }
  }
  return(x)
}

addContentToRecord <- function(record, content) {
  # In all cases, we remove trailing line breaks
  content <- content %>% removeTrailingLineBreaks()
  
  if (is(record, "properties_record")) {
    record@statements <- parseStatements(content)
    # Because properties records are transient and will not be part of the final model
    # We validate here the content
    methods::validObject(record, complete=TRUE)
  } else if (is(record, "statements_record")) {
    record@statements <- parseStatements(content)
  } else {
    stop("Record must be either a 'properties_record' or a 'statements_record'")
  }
  return(record)
}

#' Read model file.
#' 
#' @param file path to file 'model.campsis'
#' @param text model file as text, character (single or multiple lines)
#' @return records object
#' @export
read.model <- function(file=NULL, text=NULL) {
  if (!is.null(file)) {
    allLines <- readLines(con=file, warn=FALSE)
  } else if (!is.null(text)) {
    if (text %>% length() == 1) {
      allLines <- strsplit(text, split="(\r\n)|(\r)|(\n)")[[1]]
    } else {
      allLines <- text
    }
  } else {
    stop("Please provide argument 'file' or 'text'")
  }
  
  records <- CodeRecords()

  # Read all records
  lastLineIndexInPrevRecord <- 1
  for (index in seq_along(allLines)) {
    line <- allLines[index]
    if (isStrictRecordDelimiter(line)) {
      # Extract record delimiter
      recordDelimiter <- getRecordDelimiter(line)
      
      # Extract a possible comment
      comment <- as.character(NA)
      if (hasComment(line)) {
        comment <- extractRhs(line, split="#") %>% trim()
      }
      
      # Create empty record and add it to list
      record <-
        tryCatch({
          new(paste0(tolower(recordDelimiter), "_record"), comment=comment)
        },
        error = function(cond) {
          stop(paste0("Record delimiter '", recordDelimiter, "' is unknown"))
        })
      records@list <- c(records@list, record)
      
      # Add lines to previous record
      if (records %>% length() > 1) {
        content <- allLines[(lastLineIndexInPrevRecord + 1):(index-1)]
        prevRecordIndex <- records %>% length() - 1
        records@list[[prevRecordIndex]] <-
          addContentToRecord(records@list[[prevRecordIndex]], content)
      } else  {
        # If no last record is present, check if content is detected:
        # i.e. anything but blank line(s) or Campsis comment(s) before the first
        # record delimiter
        # If content is detected, throw an error
        if (lastLineIndexInPrevRecord==1 && index > 1) {
          content <- allLines[1:(index-1)]
          if (any(!grepl("^((\\s*)|(\\s*#.*))$", x=content))) {
            stop("Missing record delimiter at beginning of model")
          }
        }
      }
      lastLineIndexInPrevRecord <- index
    } else if (isRecordDelimiter(line)) {
      stop(paste0("Record delimiter '", line, "' is not valid"))
    }
  }
  # Filling in with lines of last record
  content <- allLines[(lastLineIndexInPrevRecord + 1):length(allLines)]
  lastRecordIndex <- records %>% length()
  if (lastRecordIndex==0) {
    stop("No record delimiter found in model")
  }
  records@list[[lastRecordIndex]] <-
    addContentToRecord(records@list[[lastRecordIndex]], content)
  
  return(records)
}

#_______________________________________________________________________________
#----                               replace                                 ----
#_______________________________________________________________________________

#' @rdname replace
setMethod("replace", signature=c("code_records", "model_statement"), definition=function(object, x) {
  copy <- object
  for (record in object@list) {
    if (record %>% find(x) %>% length() > 0) {
      copy <- copy %>% replace(record %>% replace(x))
    }
  }
  return(copy)
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("code_records", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  object@list <- object@list %>% purrr::map(~.x %>% replaceAll(pattern=pattern, replacement=replacement, ...))
  return(object)
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("code_records"), definition=function(object) {
  for (record in object@list) {
    show(record)
    cat("\n")
  }
})

#_______________________________________________________________________________
#----                                  sort                                 ----
#_______________________________________________________________________________

#' @rdname sort
setMethod("sort", signature=c("code_records"), definition=function(x, decreasing=FALSE, ...) {
  names <- x@list %>% purrr::map_chr(~.x %>% getName())

  # Reorder
  names <- factor(names, levels=getRecordNames(), labels=getRecordNames())
  order <- order(names)
  
  # Apply result to original list
  x@list <- x@list[order]
  return(x)
})

#_______________________________________________________________________________
#----                                 write                                 ----
#_______________________________________________________________________________

#' @rdname write
#' @importFrom utils write.table
setMethod("write", signature=c("code_records", "character"), definition=function(object, file, ...) {
  # The model is needed to get the compartment properties
  model <- processExtraArg(args=list(...), name="model")
  if (is.null(model)) {
    warning("model not provided, compartment properties will be lost")
    object <- object %>% sort()
  } else {
    # Add transient records and sort
    object <- object %>% addPropertiesRecords(model)
  }

  # Write code record
  code <- NULL
  for (record in object@list) {
    # Add record delimiter
    code <- code %>% append(writeRecordDelimiter(record))
    
    # Add all statements
    for (statement in record@statements@list) {
      code <- code %>% append(statement %>% toString())
    }
    code <- code %>% append("") # write.table will add a new line
  }
  utils::write.table(x=code, file=file, row.names=FALSE, col.names=FALSE, quote=FALSE)
})

#_______________________________________________________________________________
#----                     replication_settings interface                    ----
#_______________________________________________________________________________

#' 
#' Replication settings interface.
#' 
#' @export
setClass(
  "replication_settings",
  representation(
  )
)

#_______________________________________________________________________________
#----                       auto_replication_settings class                 ----
#_______________________________________________________________________________

#' 
#' Auto replication settings class.
#' 
#' @slot wishart logical, sample OMEGAs and SIGMAs from scaled inverse chi-squared or Wishart distributions
#' @slot odf the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the OMEGAs, integer vector
#' @slot sdf the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the SIGMAs, integer vector
#' @slot quiet logical, suppress info messages
#' @slot max_iterations number of iterations maximum to sample the parameters
#' @slot max_chunk_size maximum number of rows to sample at once, default value will be the number of replicates, unless specified.
#' @slot check_min_max logical, check for min/max values when sampling the parameters
#' @slot check_pos_def logical, check for positive definiteness when sampling the OMEGA/SIGMA parameters
#' @slot wishart_correction logical, FALSE is default, see https://github.com/metrumresearchgroup/simpar/issues/11
#' @export
setClass(
  "auto_replication_settings",
  representation(
    wishart="logical",
    odf="integer",
    sdf="integer",
    quiet="logical",
    max_iterations="integer",
    max_chunk_size="integer",
    check_min_max="logical",
    check_pos_def="logical",
    wishart_correction="logical"
  ),
  contains="replication_settings",
  prototype=prototype(wishart=FALSE, odf=as.integer(NA), sdf=as.integer(NA), quiet=as.logical(NA),
                      max_iterations=100L, max_chunk_size=as.integer(NA), check_min_max=TRUE, check_pos_def=FALSE,
                      wishart_correction=FALSE)
)

#'
#' Create auto replication settings.
#' 
#' @description
#' By default, all model parameters are sampled from a multivariate normal 
#' distribution, whose characteristics are specified by the variance-covariance matrix.
#' OMEGAs and SIGMAs can be sampled from scaled inverse chi-squared or Wishart distributions
#' by setting the \code{wishart} argument to \code{TRUE}. In that case, THETAs are still sampled
#' from a multivariate normal distribution, while OMEGAS and SIGMAs are sampled from
#' scaled inverse chi-squared (univariate OMEGA/SIGMA distribution) and Wishart (block of
#' OMEGAs/SIGMAs) distribution, respectively. When \code{wishart} is set to \code{TRUE}, the degrees
#' of freedom of the distribution must be specified, respectively, \code{odf} for the OMEGAs and
#' \code{sdf} for the SIGMAs.
#' 
#' @param wishart logical, sample OMEGAs and SIGMAs from scaled inverse chi-squared (univariate OMEGA distribution)
#'  or Wishart distribution (block of OMEGAs)
#' @param odf the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the OMEGAs,
#'  single integer value (the same degrees of freedom for all OMEGA blocks) or integer vector (one value per OMEGA block)
#' @param sdf the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the SIGMAs,
#'  single integer value (the same degrees of freedom for all SIGMA blocks) or integer vector (one value per SIGMA block)
#' @param checkMinMax logical, check for min/max values when sampling the parameters, default is TRUE
#' @param checkPosDef logical, check for positive definiteness when sampling the OMEGA/SIGMA parameters from the variance-covariance matrix (i.e. when \code{wishart=FALSE}), default is FALSE (requires extra time)
#' @param quiet logical, suppress info messages, default is NA. By default, messages will be printed out when the success rate of sampling the parameters is below 95\%.
#' @return replication settings
#' @export
AutoReplicationSettings <- function(wishart=FALSE, odf=NA, sdf=NA, checkMinMax=TRUE, checkPosDef=FALSE, quiet=NA) {
  return(new("auto_replication_settings", wishart=as.logical(wishart),
             odf=as.integer(odf), sdf=as.integer(sdf),
             check_min_max=as.logical(checkMinMax), check_pos_def=as.logical(checkPosDef),
             quiet=as.logical(quiet)))
}

#_______________________________________________________________________________
#----                     manual_replication_settings class                 ----
#_______________________________________________________________________________

#' 
#' Manual replication settings class.
#' 
#' @slot replicated_parameters data frame, 1 row per replicate
#' @export
setClass(
  "manual_replication_settings",
  representation(
    replicated_parameters="data.frame"
  ),
  contains="replication_settings"
)

#'
#' Create manual replication settings.
#' 
#' @details Use these settings to import custom replicated model parameters.
#'
#' @param data data frame with 1 row per replicate, must contain a column named 'REPLICATE'
#'  with unique integers from 1 to nrow(data), other columns are model parameters to use.
#' @return replication settings
#' @export
ManualReplicationSettings <- function(data) {
  colnames <- colnames(data)
  assertthat::assert_that("REPLICATE" %in% colnames, msg="REPLICATE column is missing")
  assertthat::assert_that(all(data$REPLICATE %in% seq_len(nrow(data))), msg="REPLICATE column must contain unique integers from 1 to nrow(data)")
  return(new("manual_replication_settings", replicated_parameters=tibble::as_tibble(data)))
}

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("auto_replication_settings"), definition=function(object) {
  if (identical(object, AutoReplicationSettings())) {
    cat("Replication settings: default (auto)")    
  } else {
    cat(sprintf("Replication settings: wishart=%s, odf=%s, sdf=%s", as.character(object@wishart),
                as.character(object@odf), as.character(object@sdf)))
  }
  cat("\n")
})

setMethod("show", signature=c("manual_replication_settings"), definition=function(object) {
  cat("Replication settings: manual") 
  cat("\n")
})

#_______________________________________________________________________________
#----                         pmx_position class                            ----
#_______________________________________________________________________________

#' 
#' PMX position class.
#' 
#' @export
setClass(
  "pmx_position",
  representation(
    undefined = "logical",
    by_index = "logical",
    by_element = "logical",
    after = "logical"
  ),
  prototype=prototype(after=TRUE),
  validity=function(object) {
    check1 <- expectOne(object, "undefined")
    check2 <- expectOne(object, "by_index")
    check3 <- expectOne(object, "by_element")
    check4 <- expectOne(object, "after")
    return(c(check1, check2, check3, check4))
  }
)

#_______________________________________________________________________________
#----                      pmx_position_by_index class                      ----
#_______________________________________________________________________________

#' 
#' PMX position by index class.
#' 
#' @export
setClass(
  "pmx_position_by_index",
  representation(
    index = "integer"
  ),
  contains="pmx_position",
  validity=function(object) {
    check <- expectOne(object, "index")
    return(check)
  }
)

#_______________________________________________________________________________
#----                     pmx_position_by_element class                     ----
#_______________________________________________________________________________

#' 
#' PMX position by element class.
#' 
#' @export
setClass(
  "pmx_position_by_element",
  representation(
    element = "pmx_element"
  ),
  contains="pmx_position"
)

#' 
#' Element position in list.
#' 
#' @param x either an integer position (useful to add an element in a code record at a specified position) or
#'  an model element (element can be a model statement or a code record)
#' @param after element to be added will be added after x (if after is TRUE) or before x (if after is FALSE)
#' @return a position object
#' @export
Position <- function(x, after=TRUE) {
  if (is(x, "pmx_element")) {
    retValue <- new("pmx_position_by_element", element=x, after=after, undefined=FALSE, by_index=FALSE, by_element=TRUE)
  } else if (is.numeric(x)) {
    retValue <- new("pmx_position_by_index", index=as.integer(x), after=after, undefined=FALSE, by_index=TRUE, by_element=FALSE)
  } else {
    stop("x can only be a PMX element or an integer position")
  }
  return(retValue)
}

#' Get the parameters block for mrgsolve.
#' 
#' @param model Campsis model
#' @param extra_params extra parameter names to be added. By default, they will be assigned a zero value.
#' @return character vector, 1 parameter per line. First one is header [PARAM].
#' @export
mrgsolveParam <- function(model, extra_params=character(0)) {
  params <- rxodeParams(model)
  retValue <- "[PARAM] @annotated"
  for (index in seq_along(params)) {
    param <- params[index]
    retValue <- retValue %>% append(paste0(names(param), " : ", as.numeric(param), " : ", names(param)))
  }
  for (param in extra_params) {
    retValue <- retValue %>% append(paste0(param, " : ", as.numeric(0), " : ", param))
  }
  return(retValue)
}

#' Get the compartment block for mrgsolve.
#' 
#' @param model Campsis model
#' @return character vector, each value is a line
#' @export
mrgsolveCompartment <- function(model) {
  compartments <- model@compartments
  retValue <- "[CMT] @annotated"
  for (compartment in compartments@list) {
    retValue <- retValue %>% append(paste0(compartment %>% toString(), " : ", compartment@name))
  }
  return(retValue)
}

#' Get the OMEGA/SIGMA matrix for mrgsolve.
#' 
#' @param model Campsis model
#' @param type either omega or sigma
#' @return named matrix or character(0) if matrix is empty
#' @export
mrgsolveMatrix <- function(model, type="omega") {
  matrix <- rxodeMatrix(model, type=type)
  if (nrow(matrix) == 0) {
    return(character(0))
  }
  if (type=="omega") {
    retValue <- "[OMEGA] @annotated @block"
  } else {
    retValue <- "[SIGMA] @annotated @block"
  }
  names <- row.names(matrix)
  for (rowIndex in seq_len(nrow(matrix))) {
    retValue <- retValue %>% append(paste0(names[rowIndex], " : ",
                                    paste0(matrix[rowIndex, seq_len(rowIndex)], collapse=" "), " : ",
                                    names[rowIndex]))
  }  
  return(retValue)
}

#' Get the MAIN block for mrgsolve.
#' 
#' @param model Campsis model
#' @return MAIN block
#' @export
mrgsolveMain <- function(model) {
  records <- model@model
  properties <- model@compartments@properties
  retValue <- "[MAIN]"
  record <- records %>% getByName("MAIN")
  retValue <- mrgsolveBlock(record, init="[MAIN]")
  if (properties %>% length() > 0) {
    for (property in properties@list) {
      compartmentIndex <- property@compartment
      compartment <- model@compartments %>% find(Compartment(index=compartmentIndex))
      equation <- paste0(property %>% toString(model=model, dest="mrgsolve"), ";")
      retValue <- retValue %>% append(equation)
    }
  }
  return(retValue)
}

#' Convert Campsis comment style to C/C++ code.
#' Only the first # is translated to //.
#' 
#' @param x any record line
#' @return same line with comments translated to C/C++
#' @keywords internal
convertAnyComment <- function(x) {
  return(sub(pattern="#", replacement="//", x=x))
}



#' Convert code record for mrgsolve.
#' 
#' @param record code record
#' @param init name of mrgsolve block
#' @param capture 'capture' instead of 'double'
#' @return translated record for mrgsolve
#' @export
mrgsolveBlock <- function(record, init=NULL, capture=FALSE) {
  retValue <- init
  if (is.null(record)) {
    return(retValue)
  }
  for (statement in record@statements@list) {
    retValue <-
      retValue %>% append(statement %>% toString(
        dest = "mrgsolve",
        init = !capture,
        capture = capture
      ))
  }
  return(retValue)
}

#' Get the ODE block for mrgsolve.
#' 
#' @param model Campsis model
#' @return ODE block
#' @export
mrgsolveOde <- function(model) {
  records <- model@model
  odeRecord <- records %>% getByName("ODE")
  
  # Automatically replace simulation time 't' (default in Campsis) by SOLVERTIME
  if (!is.null(odeRecord)) {
    odeRecord <- odeRecord %>% campsismod::replaceAll(pattern=VariablePattern("t"), replacement="SOLVERTIME")
  }
  retValue <- mrgsolveBlock(odeRecord, init="[ODE]")
  return(retValue)
}

#' Get the TABLE block for mrgsolve.
#'
#' @param model Campsis model
#' @return TABLE block if at least one line in error record, character(0) otherwise
#' @export
mrgsolveTable <- function(model) {
  records <- model@model
  errorRecord <- records %>% getByName("ERROR")
  if (is.null(errorRecord)) {
    return(character(0))
  }
  retValue <- mrgsolveBlock(errorRecord, init="[TABLE]", capture=TRUE)
  return(retValue)
}

#' Get the CAPTURE block for mrgsolve.
#'
#' @param outvars outvars in method simulate
#' @param model Campsis model
#' @return CAPTURE block or character(0) if no variable in outvars
#' @export
mrgsolveCapture <- function(outvars, model) {
  # Get rid of variables that are already in error block (and thus already captured in TABLE)
  outvars <- convertOutvarsToCapture(outvars, model=model)
  
  if (is.null(outvars) || outvars %>% length()==0) {
    return(character(0))
  } else {
    return(paste("[CAPTURE]", outvars))
  }
}

#' Convert outvars argument to capture. Variables that are already in error block
#' will be discarded.
#'
#' @param outvars character vector
#' @param model Campsis model
#' @importFrom purrr keep map_chr
#' @return all variables to capture
#' @keywords internal
convertOutvarsToCapture <- function(outvars, model) {
  # List all variables that are already exported into mrgsolve TABLE block by pmxmod
  error <- model@model %>% getByName("ERROR")
  list <- NULL
  if (!is.null(error)) {
    list <- error@statements@list %>% purrr::keep(~is(.x, "equation")) %>% purrr::map_chr(~.x@lhs)
    outvars <- outvars[!(outvars %in% list)]
  }
  return(outvars)
}

#_______________________________________________________________________________
#----                   compartment_property class                    ----
#_______________________________________________________________________________

#' 
#' Compartment property class.
#' 
#' @slot compartment related compartment index
#' @slot rhs right-hand side formula
#' @slot comment comment if any, single character string
#' @export
setClass(
  "compartment_property",
  representation(
    compartment = "integer",
    rhs = "character",
    comment = "character"
  ),
  contains="pmx_element",
  prototype=prototype(comment=as.character(NA), rhs=""),
  validity=function(object) {
    return(expectOneForAll(object, c("compartment", "rhs")))
  } 
)

#_______________________________________________________________________________
#----                            getPrefix                                  ----
#_______________________________________________________________________________

#' Get prefix.
#' 
#' @param object generic object
#' @param ... e.g. \code{dest='mrgsolve'}
#' @return the prefix of this object
#' @export
#' @rdname getPrefix
getPrefix <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("getPrefix", function(object, ...) {
  standardGeneric("getPrefix")
})

#_______________________________________________________________________________
#----                          getRecordName                                ----
#_______________________________________________________________________________

#' Get record name.
#' 
#' @param object generic object
#' @return the name of the record
#' @export
#' @rdname getRecordName
getRecordName <- function(object) {
  stop("No default function is provided")
}

setGeneric("getRecordName", function(object) {
  standardGeneric("getRecordName")
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("compartment_property", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  object@rhs <- object@rhs %>% replaceAll(pattern=pattern, replacement=replacement, ...)
  return(object)
})

#_______________________________________________________________________________
#----                               show                                    ----
#_______________________________________________________________________________

setMethod("show", signature=c("compartment_property"), definition=function(object) {
  cat(paste0(object %>% getName(), ": ", object@rhs))
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("compartment_property"), definition=function(object, ...) {
  model <- processExtraArg(args=list(...), name="model", mandatory=TRUE)
  dest <- processExtraArg(args=list(...), name="dest", mandatory=TRUE)
  
  compartmentIndex <- object@compartment
  compartment <- model@compartments %>% find(Compartment(index=compartmentIndex))
  
  if (isRxODE(dest)) {
    return(paste0(object %>% getPrefix(dest=dest), "(", compartment %>% toString(), ")=", object@rhs))
  } else if (dest=="mrgsolve") {
    return(paste0(object %>% getPrefix(dest=dest), "_", compartment %>% toString(), "=", object@rhs))
  } else if (dest=="campsis") {
    return(paste0(compartment %>% toString(), "=", object@rhs))
  } else {
    UnsupportedDestException()
  }
})

#_______________________________________________________________________________
#----                     code_record class (ABSTRACT)                      ----
#_______________________________________________________________________________

checkNonODERecord <- function(object) {
  hasODE <- object@statements@list %>% purrr::map_lgl(~is(.x, "ode")) %>% any()
  errors <- character(0)
  if (hasODE) {
    errors <- errors %>% append("ODE detected in non ODE record")
  }
  return(errors)
}

#' 
#' Code record class. See this code record as an abstract class.
#' 2 implementations are possible:
#' - properties record (lag, duration, rate & bioavailability properties)
#' - statements record (main, ode & error records)
#' 
#' @slot comment a comment, single character value
#' @slot statements model statements
#' 
setClass(
  "code_record",
  representation(
    comment = "character",
    statements = "model_statements"
  ),
  contains = "pmx_element",
  prototype = prototype(statements=ModelStatements(), comment=as.character(NA)),
  validity = function(object) {
    return(TRUE)
  }
)

#_______________________________________________________________________________
#----                      statements_record class                           ----
#_______________________________________________________________________________

#' 
#' Properties record class.
#' 
setClass(
  "properties_record",
  representation(
  ),
  contains = "code_record",
  validity = function(object) {
    hasODE <- object@statements@list %>% purrr::map_lgl(~is(.x, "ode")) %>% any()
    hasUnknownStatement <- object@statements@list %>% purrr::map_lgl(~is(.x, "unknown_statement")) %>% any()
    hasIfStatement <- object@statements@list %>% purrr::map_lgl(~is(.x, "if_statement")) %>% any()
    errors <- character(0)
    if (hasODE) {
      errors <- errors %>% append("ODE detected in properties record")
    }
    if (hasUnknownStatement) {
      errors <- errors %>% append("Unknown statement detected in properties record")
    }
    if (hasIfStatement) {
      errors <- errors %>% append("IF-statement detected in properties record")
    }
    return(errors)
  }
)

#_______________________________________________________________________________
#----                      statements_record class                          ----
#_______________________________________________________________________________

#' 
#' Statements record class.
#' 
#' @export
setClass(
  "statements_record",
  representation(
  ),
  contains = "code_record"
)

#_______________________________________________________________________________
#----                           MAIN record                                 ----
#_______________________________________________________________________________

#' 
#' Main record class.
#' 
#' @export
setClass(
  "main_record",
  representation(
  ),
  contains = "statements_record",
  validity = checkNonODERecord
)

#' 
#' Create MAIN code record.
#' 
#' @param code code record
#' @export
MainRecord <- function(code=character()) {
  return(new("main_record", statements=parseStatements(code)))
}

#_______________________________________________________________________________
#----                            ODE record                                 ----
#_______________________________________________________________________________

#' 
#' ODE record class.
#' 
#' @export
setClass(
  "ode_record",
  representation(
  ),
  contains = "statements_record"
)

#' 
#' Create ODE code record.
#' 
#' @param code code record
#' @return an ODE code record
#' @export
OdeRecord <- function(code=character()) {
  return(new("ode_record", statements=parseStatements(code)))
}

#_______________________________________________________________________________
#----                              F record                                 ----
#_______________________________________________________________________________

#' 
#' Bioavailability record class.
#' 
#' @export
setClass(
  "f_record",
  representation(
  ),
  contains = "properties_record"
)

#_______________________________________________________________________________
#----                              LAG record                               ----
#_______________________________________________________________________________

#' 
#' Lag record class.
#' 
#' @export
setClass(
  "lag_record",
  representation(
  ),
  contains = "properties_record"
)

#_______________________________________________________________________________
#----                           DURATION record                             ----
#_______________________________________________________________________________

#' 
#' (Infusion)-duration record class.
#' 
#' @export
setClass(
  "duration_record",
  representation(
  ),
  contains = "properties_record"
)

#_______________________________________________________________________________
#----                             RATE record                               ----
#_______________________________________________________________________________

#' 
#' (Infusion)-rate record class.
#' 
#' @export
setClass(
  "rate_record",
  representation(
  ),
  contains = "properties_record"
)

#_______________________________________________________________________________
#----                             INIT record                               ----
#_______________________________________________________________________________

#' 
#' Init record class.
#' 
#' @export
setClass(
  "init_record",
  representation(
  ),
  contains = "properties_record"
)

#_______________________________________________________________________________
#----                           ERROR record                                ----
#_______________________________________________________________________________


#' 
#' Error record class.
#' 
#' @export
setClass(
  "error_record",
  representation(
  ),
  contains = "statements_record",
  validity = checkNonODERecord
)

#' 
#' Create ERROR code record.
#' 
#' @param code code record
#' @return an ERROR code record
#' @export
ErrorRecord <- function(code=character()) {
  return(new("error_record", statements=parseStatements(code)))
}

#_______________________________________________________________________________
#----                                add                                    ----
#_______________________________________________________________________________

#' @param pos position where x needs to be added in list
#' @rdname add
setMethod("add", signature=c("code_record", "model_statement"), definition=function(object, x, pos=NULL) {
  object@statements <- object@statements %>% add(x, pos=pos)
  return(object)
})

#' @rdname add
setMethod("add", signature=c("code_record", "code_record"), definition=function(object, x) {
  object@statements <- object@statements %>% add(x@statements)
  return(object)
})

#_______________________________________________________________________________
#----                                contains                               ----
#_______________________________________________________________________________

#' @rdname contains
setMethod("contains", signature=c("statements_record", "model_statement"), definition=function(object, x) {
  return(!is.null(object@statements %>% find(x)))
})

#_______________________________________________________________________________
#----                               delete                                  ----
#_______________________________________________________________________________

#' @rdname delete
setMethod("delete", signature=c("statements_record", "model_statement"), definition=function(object, x) {
  object@statements <- object@statements %>% delete(x)
  return(object)
})

#' @rdname delete
setMethod("delete", signature=c("statements_record", "integer"), definition=function(object, x) {
  object@statements <- object@statements %>% delete(x)
  return(object)
})

#_______________________________________________________________________________
#----                                find                                   ----
#_______________________________________________________________________________

#' @rdname find
setMethod("find", signature=c("statements_record", "model_statement"), definition=function(object, x) {
  return(object@statements %>% find(x))
})

#_______________________________________________________________________________
#----                              getName                                  ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature=c("main_record"), definition=function(x) {
  return("MAIN")
})

#' @rdname getName
setMethod("getName", signature=c("ode_record"), definition=function(x) {
  return("ODE")
})

#' @rdname getName
setMethod("getName", signature=c("f_record"), definition=function(x) {
  return("F")
})

#' @rdname getName
setMethod("getName", signature=c("lag_record"), definition=function(x) {
  return("LAG")
})

#' @rdname getName
setMethod("getName", signature=c("duration_record"), definition=function(x) {
  return("DURATION")
})

#' @rdname getName
setMethod("getName", signature=c("rate_record"), definition=function(x) {
  return("RATE")
})

#' @rdname getName
setMethod("getName", signature=c("init_record"), definition=function(x) {
  return("INIT")
})

#' @rdname getName
setMethod("getName", signature=c("error_record"), definition=function(x) {
  return("ERROR")
})

#_______________________________________________________________________________
#----                             length                                    ----
#_______________________________________________________________________________

#' @rdname length
setMethod("length", signature=c("statements_record"), definition=function(x) {
  return(x@statements %>% length())
})

#_______________________________________________________________________________
#----                               replace                                 ----
#_______________________________________________________________________________

#' @rdname replace
setMethod("replace", signature=c("statements_record", "model_statement"), definition=function(object, x) {
  object@statements <- object@statements %>% replace(x)
  return(object)
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("code_record", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  object@statements@list <- object@statements@list %>% purrr::map(~.x %>% replaceAll(pattern=pattern, replacement=replacement, ...))
  return(object)
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("code_record"), definition=function(object) {
  cat(writeRecordDelimiter(object), "\n", sep="")
  show(object@statements)
})

#' 
#' Write record delimiter line.
#' 
#' @param object code record
#' @return a record delimiter line
#' @keywords internal
writeRecordDelimiter <- function(object) {
  recordDelimiter <- paste0("[", object %>% getName(), "]")
  recordDelimiter <- recordDelimiter %>% appendComment(object=object, dest="campsis")
  return(recordDelimiter)
}

library(testthat)

context("Check warnings are displayed in some particular cases")

source(paste0("", "testUtils.R"))

test_that("Warnings displayed when model contains unknown statements (1)", {
  modelName <- "unknown_statement_hello"
  expect_warning(read.campsis(paste0(testFolder, "custom/", modelName)), regexp="Model code contains unknown statements")
})

test_that("Warnings displayed when model contains unknown statements (2)", {
  modelName <- "unknown_statement_illegal_variable_name"
  expect_warning(read.campsis(paste0(testFolder, "custom/", modelName)), regexp="Model code contains unknown statements")
})

test_that("Warnings displayed when no theta, omega and sigma.csv files", {
  modelName <- "unknown_statement_hello"
  expect_warning(read.campsis(paste0(testFolder, "custom/", modelName)), regexp="No file 'theta.csv' could be found")
  expect_warning(read.campsis(paste0(testFolder, "custom/", modelName)), regexp="No file 'omega.csv' could be found")
  expect_warning(read.campsis(paste0(testFolder, "custom/", modelName)), regexp="No file 'sigma.csv' could be found")
  
  model <- suppressWarnings(read.campsis(paste0(testFolder, "custom/", modelName)))
  mrgmod <- model %>% export(dest="mrgsolve")
  expect_equal(mrgmod@param, "[PARAM] @annotated")
  expect_equal(mrgmod@omega, character(0))
  expect_equal(mrgmod@sigma, character(0))
})

test_that("Error displayed if model file has ODE's in non ODE record", {
  modelName <- "ode_in_non_ode_record"
  expect_error(suppressWarnings(read.campsis(paste0(testFolder, "custom/", modelName))), regexp="ODE detected in non ODE record")
})

test_that("Error displayed if model file has an IF-statement in a properties record", {
  modelName <- "if_in_properties_record"
  expect_error(suppressWarnings(read.campsis(paste0(testFolder, "custom/", modelName))), regexp="IF-statement detected in properties record")
})

test_that("Parsing a Campsis model with an incomplete final line shouldn't raise any warning", {
  # This model contains the following files:
  # Campsis model file with incomplete final line
  # File theta.csv with incomplete final line
  # Correct file omega.csv
  # Correct file sigma.csv
  
  modelName <- "incomplete_final_line/"
  model <- read.campsis(paste0(testFolder, "custom/", modelName))
  
  # Check MAIN and ODE record can be found
  expect_true(!is.null(model %>% find(MainRecord())))
  expect_true(!is.null(model %>% find(OdeRecord())))
  
  # Check both theta's K and V have been loaded correctly
  expect_equal(model@parameters %>% select("theta") %>% length(), 2)
})

library(testthat)

context("Test parsing of record delimiters")

source(paste0("", "testUtils.R"))

test_that("Parsing non standard record delimiters works as expected", {
  folder <- paste0(testFolder, "custom/", "non_standard_record_delimiters/")
  model <- readCampsisModelNoParams(folder)
  expect_equal(model %>% find(MainRecord()) %>% length(), 3)
  expect_equal(model %>% find(OdeRecord()) %>% length(), 3)
  expect_equal(model %>% find(ErrorRecord()) %>% length(), 1)
})

test_that("Parsing unknown record delimiters must raise an error", {
  folder <- paste0(testFolder, "custom/", "unknow_record_delimiters/")
  expect_error(read.campsis(folder), regexp="Record delimiter 'PK' is unknown")
})

test_that("Record delimiters may accept comments", {
  folder <- paste0(testFolder, "custom/", "record_delimiters_with_comment/")
  model <- readCampsisModelNoParams(folder)

  expect_equal(model %>% find(MainRecord()) %>% .@comment, "MAIN block")
  expect_equal(model %>% find(OdeRecord()) %>% .@comment, "ODE block")
  expect_equal(model %>% find(ErrorRecord()) %>% .@comment, "ERROR block")
  
  expect_true("[MAIN] # MAIN block" %in% capture.output(show(model)))
  expect_true("[ODE] # ODE block" %in% capture.output(show(model)))
  expect_true("[ERROR] # ERROR block" %in% capture.output(show(model)))
  
  # Please note that comments will be LOST for all properties record delimiters
  # This is because compartment properties are 'modeled' all-together in model@compartments@properties  
})

test_that("Malformed record delimiters throw an error", {
  folder <- paste0(testFolder, "custom/", "malformed_record_delimiter1/")
  expect_error(readCampsisModelNoParams(folder), regexp="Record delimiter '\\[MAIN\\] Hello' is not valid")
  
  folder <- paste0(testFolder, "custom/", "malformed_record_delimiter2/")
  expect_error(readCampsisModelNoParams(folder), regexp="Record delimiter '\\[ODE\\] d/dt\\(A_CENTRAL\\)=-K\\*A_CENTRAL' is not valid")
})

test_that("A record delimiter must always exist", {
  folder <- paste0(testFolder, "custom/", "no_record_delimiter_available1/")
  expect_error(readCampsisModelNoParams(folder), regexp="Missing record delimiter at beginning of model")
  
  folder <- paste0(testFolder, "custom/", "no_record_delimiter_available2/")
  expect_error(readCampsisModelNoParams(folder), regexp="No record delimiter found in model")
})

test_that("Line breaks can precede the first record delimiter", {
  folder <- paste0(testFolder, "custom/", "line_breaks_before_main/")
  model <- readCampsisModelNoParams(folder)
  expect_equal(model %>% find(MainRecord()) %>% length(), 2)
  expect_equal(model %>% find(OdeRecord()) %>% length(), 3)
})

test_that("Comments can precede the first record delimiter", {
  folder <- paste0(testFolder, "custom/", "comments_before_main/")
  model <- readCampsisModelNoParams(folder)
  expect_equal(model %>% find(MainRecord()) %>% length(), 2)
  expect_equal(model %>% find(OdeRecord()) %>% length(), 3)
})

#_______________________________________________________________________________
#----                     unknown_statement class                           ----
#_______________________________________________________________________________

#' 
#' Unknown statement class. Any statement not recognized by campsismod.
#' 
#' @export
setClass(
  "unknown_statement",
  representation(
    line = "character"
  ),
  contains = "model_statement",
  validity = function(object) {
    return(expectOne(object, "line"))
  }
)

#' 
#' Create a new ordinary differential equation (ODE).
#' 
#' @param line line which was not recognised
#' @param comment comment if any, single character string
#' @return an unknown statement
#' @export
UnknownStatement <- function(line, comment=as.character(NA)) {
  return(new("unknown_statement", line=line, comment=comment))
}

#_______________________________________________________________________________
#----                            getName                                    ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature = c("unknown_statement"), definition = function(x) {
  return(as.character(NA)) # unknown statement non-identifiable 
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("unknown_statement", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  object@line <- object@line %>% replaceAll(pattern=pattern, replacement=replacement, ...)
  return(object)
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("unknown_statement"), definition=function(object, ...) {
  args <- list(...)
  dest <- processExtraArg(args=args, name="dest", default="campsis")
  show <- processExtraArg(args=args, name="show", default=FALSE)

  if (dest=="campsis") {
    if (show) {
      retValue <- paste0("[UNKNOWN STATEMENT] ", object@line)
    } else {
      retValue <- object@line
    }
  } else if (isRxODE(dest) || dest=="mrgsolve" || dest=="NONMEM") {
    retValue <- object@line
  } else {
    UnsupportedDestException()
  }
  return(retValue %>% appendComment(object, dest))
})

#' Assert the given character vector is a single character string.
#' 
#' @param x single character string
#' @return no return value
#' @importFrom assertthat assert_that
#' @export
assertSingleCharacterString <- function(x) {
  assertthat::assert_that(is.character(x) && length(x)==1, msg="x must be a single character string")
}

#' Process extra arguments.
#' 
#' @param args arguments list
#' @param name argument name to retrieve
#' @param default default value if argument is not present
#' @param mandatory mandatory argument, logical value
#' @return requested argument value
#' @importFrom utils hasName
#' @export
processExtraArg <- function(args, name, default=NULL, mandatory=FALSE) {
  if (utils::hasName(args, name)) {
    retValue <- args[[name]]
  } else {
    if (is.null(default) && mandatory) {
      stop(paste0("Extra argument '", name, "' is mandatory."))
    }
    retValue <- default
  }
  return(retValue)
}

#' Say if line(s) in record is/are ODE or not.
#' 
#' @param x character vector
#' @return logical vector
#' @export
isODE <- function(x) {
  return(grepl(pattern="^d/dt\\s*\\(.*\\)\\s*=", x=trim(x), ignore.case=TRUE))
}

#' Return the variable pattern (string form).
#' 
#' @return pattern (regular expression)
#' @keywords internal
#' 
variablePatternStr <- function() {
  return("[a-zA-Z_][a-zA-Z0-9_]*")
}

#' Return the variable pattern (string form), without the first character.
#' 
#' @return pattern (regular expression)
#' @keywords internal
#' 
variablePatternNoStartStr <- function() {
  return("[a-zA-Z0-9_]*")
}

#' Say if line in record is an equation not.
#' 
#' @param x character value
#' @return logical value
#' @export
isEquation <- function(x) {
  assertSingleCharacterString(x)
  parts <- strsplit(x, split="=")[[1]]
  if (length(parts) == 1) {
    return(FALSE)
  }
  variable <- parts[1] %>% trim()
  return(grepl(pattern=paste0("^", variablePatternStr(), "$"), x=variable))
}

#' Return the IF-statement pattern (string form).
#' 
#' @return pattern (regular expression)
#' @keywords internal
ifStatementPatternStr <- function() {
  return(paste0("if\\s*\\(.*\\)\\s*", variablePatternStr(), "\\s*="))
}

#' Say if line in record is an IF-statement.
#' 
#' @param x character value
#' @return logical value
#' @export
isIfStatement <- function(x) {
  return(grepl(pattern=paste0("^", ifStatementPatternStr()), x=trim(x), ignore.case=TRUE))
}

#' Extract text between brackets.
#' 
#' @param x character value
#' @return text between brackets (trimmed)
#' @export
extractTextBetweenBrackets <- function(x) {
  assertSingleCharacterString(x)
  retValue <- gsub("[\\(\\)]", "", regmatches(x, gregexpr("\\(.*?\\)", x))[[1]])
  if (length(retValue) == 0) {
    stop(paste0("No parentheses found in ", x))
  }
  return(retValue[1] %>% trim())
}

#' Extract right-hand-side expression.
#' 
#' @param x character value
#' @param split character where to split
#' @return right-hand side expression
#' @export
extractRhs <- function(x, split="=") {
  assertSingleCharacterString(x)
  tmp <- strsplit(x=x, split=split)[[1]]
  # Remove lhs and collapse (in case of several =)
  rhs <- paste0(tmp[-1], collapse="=")
  return(rhs)
}

#' Extract left-hand-side expression.
#' 
#' @param x character value
#' @param split character where to split
#' @return left-hand-side expression, not trimmed
#' @export
extractLhs <- function(x, split="=") {
  assertSingleCharacterString(x)
  tmp <- strsplit(x=x, split=split)[[1]]
  lhs <- tmp[1]
  return(lhs)
}

#' Trim character vector. Remove all leading and trailing spaces.
#' 
#' @param x character vector
#' @return character vector without leading and trailing spaces
#' @importFrom assertthat assert_that
#' @export
trim <- function(x) {
  assertthat::assert_that(is.character(x), msg="x must be a character vector")
  return(gsub("^\\s+|\\s+$", "", x))
}

#' Check if string contains Campsis-style comments.
#' 
#' @param x character vector
#' @return logical value
#' @export
hasComment <- function(x) {
  return(grepl("#", x=x, fixed=TRUE))
}

#' Check if string is a Campsis comment (i.e. not an equation).
#' 
#' @param x character vector
#' @return logical value
#' @export
isComment <- function(x) {
  return(grepl("^\\s*#", x=x))
}

#' Check if string is an empty line.
#' 
#' @param x character vector
#' @return logical value
#' @export
isEmptyLine <- function(x) {
  return(grepl("^\\s*$", x=x))
}

#' Is strict record delimiter. A strict record delimiter is any line starting 
#' with [...] and followed by nothing but spaces or a possible comment.
#' 
#' @param line any line, single character value
#' @return a logical value
#' @export
isStrictRecordDelimiter <- function(line) {
  return(grepl("^\\s*\\[.*\\]((\\s*)|(\\s*#.*))$", line))
}

#' Is record delimiter. A record delimiter is any line starting with [...].
#' 
#' @param line any line, single character value
#' @return a logical value
#' @export
isRecordDelimiter <- function(line) {
  return(grepl("^\\s*\\[.*\\].*$", line))
}

#' Get record delimiter.
#' 
#' @param line any line, single character value
#' @return the record delimiter between brackets
#' @export
getRecordDelimiter <- function(line) {
  return(gsub("\\[(.*)\\](.*)","\\1", line) %>% trim())
}

#' Get record equation names
#' 
#' @param record any code record
#' @return a character vector with the equation names
#' @export
#' @keywords internal
getRecordEquationNames <- function(record) {
  retValue <- NULL
  for (statement in record@statements@list) {
    if (is(statement, "equation") && !(is(statement, "ode"))) {
      retValue <- c(retValue, statement@lhs)
    }
  }
  return(retValue)
}

#' Check is vector has NA's only.
#' 
#' @param x any vector
#' @return TRUE if all values are NA, FALSE otherwise
#' @export
#' @keywords internal
allNa <- function(x) {
  return(all(is.na(x)))
}

#' Remove given column(s) if it has only NA's.
#' 
#' @param x any data frame
#' @param column column name(s)
#' @return updated data frame
#' @importFrom dplyr any_of where
#' @export
#' @keywords internal
removeNaColumn <- function(x, column) {
  return(x %>% dplyr::select(!(dplyr::any_of(column) & dplyr::where(allNa))))
}

#' Check if the destination engine is RxODE or rxode2.
#' Note that rxode2 is the successor of RxODE.
#' 
#' @param dest destination engine
#' @return TRUE if RxODE or rxode2, FALSE otherwise
#' @export
#' @keywords internal
isRxODE <- function(dest) {
  return(dest %in% c("RxODE", "rxode2"))
}

#'
#' Get the Campsismod options (R options).
#'
#' @return global options for Campsismod
#' @export
#' @keywords internal
getCampsismodOptions <- function() {
  return(getOption("campsismod.options"))
}

#'
#' Get Campsismod option logic.
#'
#' @param name option to search
#' @param default default value if option not found
#' @return option value
#' @export
getCampsismodOption <- function(name, default) {
  option <- getCampsismodOptions()
  if (is.null(option)) {
    return(default)
  } else {
    value <- option[[name]]
    if (is.null(value)) {
      return(default)
    } else {
      return(value)
    }
  }
}

library(testthat)

context("Test all methods from utilities.R")

test_that("trim method is working well", {
  expect_equal(trim("  hello  "), "hello")
})

test_that("isODE method is working well", {
  expect_true(isODE("d/dt(A_DEPOT)=-KA*A_DEPOT"))
  expect_true(isODE("d/dt(A_OUTPUT)=K*A_CENTRAL"))
  expect_false(isODE("d/dtHELLO(A_OUTPUT)=K*A_CENTRAL"))
  expect_true(isODE("d/dt (A_OUTPUT) = K*A_CENTRAL"))
})

test_that("extractTextBetweenBrackets method is working well", {
  expect_equal(extractTextBetweenBrackets("d/dt(A_DEPOT)=-KA*A_DEPOT"), "A_DEPOT")
  expect_error(extractTextBetweenBrackets(c("d/dt(A_DEPOT)=-KA*A_DEPOT", "X=1")))
})

test_that("isEquation is working well", {
  expect_true(isEquation("V3=THETA_V3*VDBW"))
  expect_false(isEquation("if (OCC == 1) VIS1=1"))
  expect_false(isEquation("THETA_V3"))
})

test_that("isComment is working well", {
  expect_true(isComment("# HELLO"))
  expect_true(isComment("   # HELLO"))
  expect_true(isComment("\t# HELLO"))
  expect_false(isComment("A# HELLO"))
  expect_false(isComment(" A # HELLO"))
  expect_false(isComment("V3=THETA_V3*VDBW"))
  expect_false(isComment("V3=THETA_V3*VDBW # COMMENT")) # This is an equation that has a comment
})

test_that("hasComment is working well", {
  expect_true(hasComment("# HELLO"))
  expect_true(hasComment("   # HELLO"))
  expect_true(hasComment("\t# HELLO"))
  expect_true(hasComment("V3=THETA_V3*VDBW # COMMENT"))
  expect_false(hasComment("V3=THETA_V3*VDBW"))
})

test_that("isEmptyLine is working well", {
  expect_true(isEmptyLine(""))
  expect_true(isEmptyLine("  "))
  expect_true(isEmptyLine("\t"))
  expect_false(isEmptyLine("\tA"))
})

test_that("extractLhs and extractRhs with comment works well", {
  expect_equal(extractLhs("KA=THETA_KA*exp(ETA_KA) # Comment", split="#"), "KA=THETA_KA*exp(ETA_KA) ")
  expect_equal(extractRhs("KA=THETA_KA*exp(ETA_KA) # Comment", split="#"), " Comment")
})

test_that("isIfStatement works well", {
  line <- "  if (ID > 30) TVCL=THETA_7*pow(0.009*TBW, THETA_8)"
  expect_true(isIfStatement(line))
  
  line <- "  if (ID > (30)) TVCL=THETA_7*pow(0.009*TBW, THETA_8)"
  expect_true(isIfStatement(line))
  
  line <- " if(ID == 30) TVCL = THETA_7*pow(0.009*TBW, THETA_8)"
  expect_true(isIfStatement(line))
})

test_that("isStrictRecordDelimiter works well", {
  line <- "[MAIN]"
  expect_true(isStrictRecordDelimiter(line))
  
  line <- " [MAIN] "
  expect_true(isStrictRecordDelimiter(line))
  
  line <- " [  MAIN\t]"
  expect_true(isStrictRecordDelimiter(line))
  
  line <- "{MAIN]"
  expect_false(isStrictRecordDelimiter(line))
  
  line <- "[MAIN]# COMMENT1 "
  expect_true(isStrictRecordDelimiter(line))
  
  line <- "[MAIN] #COMMENT2 "
  expect_true(isStrictRecordDelimiter(line))
  
  line <- "[MAIN] A=1"
  expect_false(isStrictRecordDelimiter(line))
  expect_true(isRecordDelimiter(line))
})

test_that("getRecordDelimiter works well", {
  line <- "[MAIN]"
  expect_equal(getRecordDelimiter(line), "MAIN")
  
  line <- " [MAIN\t] "
  expect_equal(getRecordDelimiter(line), "MAIN")
  
  line <- " [  MAIN  ]"
  expect_equal(getRecordDelimiter(line), "MAIN")
  
  line <- "[MAIN] # MAIN block"
  expect_equal(getRecordDelimiter(line), "MAIN")
})

#' Parse statements code and return Campsis statements.
#' 
#' @param code character vector containing all statements (text form)
#' @return a list of Campsis statements
#' 
parseStatements <- function(code) {
  statements <- ModelStatements()
  
  for (index in seq_along(code)) {
    line <- code[index]
    hasComment <- hasComment(line) 
    comment <- as.character(NA)
    if (hasComment) {
      comment <- extractRhs(line, split="#") %>% trim()
      line_ <- extractLhs(line, split="#")
    } else {
      line_ <- line
    }
    
    if (isEmptyLine(line)) {
      statements <- statements %>% add(LineBreak())
      
    } else if (isComment(line)) {
      statements <- statements %>% add(Comment(comment))
      
    } else if (isODE(line_)) {
      lhs <- extractTextBetweenBrackets(line_)
      rhs <- extractRhs(line_) %>% trim()
      statements <- statements %>% add(Ode(lhs, rhs, comment=comment))
    
    } else if (isEquation(line_)) {
      lhs <- extractLhs(line_) %>% trim()
      rhs <- extractRhs(line_) %>% trim()
      statements <- statements %>% add(Equation(lhs, rhs, comment=comment))
      
    } else if (isIfStatement(line_)) {
      statements <- statements %>% add(parseIfStatement(line_, comment=comment))
      
    } else  {
      statements <- statements %>% add(UnknownStatement(line_, comment=comment))
    }
  }
  return(statements)
}

#' Parse IF-statement.
#' Assumption: \code{isIfStatement} method already called and returned TRUE.
#' 
#' @param line IF-statement as single character string value, comment omitted
#' @param comment any comment, NA by default
#' @return an IF statement object
#' 
parseIfStatement <- function(line, comment=as.character(NA)) {
  # Trim input
  line <- line %>% trim()
  
  # Lhs/rhs extraction
  tmp1 <- regexpr(pattern=paste0("^", ifStatementPatternStr()), line, ignore.case=TRUE)
  equalSymbolIndex <- attr(tmp1, "match.length")
  lhs <- substring(line, first=1, last=equalSymbolIndex - 1) %>% trim()
  rhs <- substring(line, first=equalSymbolIndex + 1, last=nchar(line)) %>% trim()
  
  # Identify first parenthesis
  tmp2 <- regexpr("^if\\s*\\(", lhs, ignore.case=TRUE)
  firstParenthesisIndex <- attr(tmp2, "match.length")
  
  # Identify variable start
  variableStartIndex <- regexpr(paste0(variablePatternStr(), "$"), lhs) %>% as.integer()
  
  # Identify condition
  conditionWithParentheses <- substring(lhs, first=firstParenthesisIndex, last=variableStartIndex-1) %>% trim()
  condition <- substring(conditionWithParentheses, first=2, last=nchar(conditionWithParentheses) - 1) %>% trim()
  
  # Identify variable
  variable <- substring(lhs, first=variableStartIndex, last=nchar(lhs)) %>% trim()
  
  # Return IF-statement
  return(IfStatement(condition=condition, equation=Equation(variable, rhs=rhs), comment=comment))
}

library(testthat)

context("Test all methods from model_parser.R")

test_that("parseIfStatement works well (1)", {
  line <- "if(ID > (30))TVCL=THETA_7*pow(0.009*TBW, THETA_8)"
  ifStatement <- parseStatements(line) %>% getByIndex(1)
  
  expect_equal(ifStatement@condition, "ID > (30)")
  expect_equal(ifStatement@equation, Equation("TVCL", "THETA_7*pow(0.009*TBW, THETA_8)"))
})

test_that("parseIfStatement works well (2)", {
  line <- "  if  (ID > (30)  )  TVCL   =   THETA_7*pow(0.009*TBW, THETA_8)  "
  ifStatement <- parseStatements(line) %>% getByIndex(1)
  
  expect_equal(ifStatement@condition, "ID > (30)")
  expect_equal(ifStatement@equation, Equation("TVCL", "THETA_7*pow(0.009*TBW, THETA_8)"))
})

test_that("parseIfStatement works well (3)", {
  line <- "  if  (ID == 30 )  TVCL   =   THETA_7*pow(0.009*TBW, THETA_8)  "
  ifStatement <- parseStatements(line) %>% getByIndex(1)

  expect_equal(ifStatement@condition, "ID == 30")
  expect_equal(ifStatement@equation, Equation("TVCL", "THETA_7*pow(0.009*TBW, THETA_8)"))
})

#' Magritt operator for piping.
#' 
#' @importFrom magrittr `%>%`
#' @export
#' @keywords internal
`%>%` <- magrittr::`%>%`

#' Colon-equals operator.
#' 
#' @importFrom rlang `:=`
#' @export
#' @keywords internal
`:=` <- rlang::`:=`

#' 
#' JSON element class.
#' 
#' @export
setClass(
  "json_element",
  representation(
    data="ANY"
  )
)

#' 
#' Instantiate a JSON element.
#' 
#' @param x JSON object representation in R
#' @return JSON element
#' @export
JSONElement <- function(x) {
  return(new("json_element", data=x))
}

#_______________________________________________________________________________
#----                          equation class                               ----
#_______________________________________________________________________________

#' 
#' Equation class. Any statement in the form A = B.
#' 
#' @slot lhs left-hand side expression
#' @slot rhs right-hand side expression
#' @export
setClass(
  "equation",
  representation(
    lhs = "character",
    rhs = "character"
  ),
  contains = "model_statement",
  prototype = prototype(rhs=""),
  validity = function(object) {
    return(expectOne(object, c("lhs", "rhs")))
  }
)

#' 
#' Create a new equation.
#' 
#' @param lhs left-hand side variable corresponding to the assigned variable name
#' @param rhs right-hand side expression corresponding to a formula
#' @param comment comment if any, single character string
#' @return an equation
#' @export
Equation <- function(lhs, rhs="", comment=as.character(NA)) {
  return(new("equation", lhs=lhs, rhs=rhs, comment=comment))
}

#_______________________________________________________________________________
#----                            getName                                    ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature = c("equation"), definition = function(x) {
  return(paste0("EQUATION (", x@lhs, ")"))
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("equation", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  object@lhs <- object@lhs %>% replaceAll(pattern=pattern, replacement=replacement, ...)
  object@rhs <- object@rhs %>% replaceAll(pattern=pattern, replacement=replacement, ...)
  return(object)
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("equation"), definition=function(object, ...) {
  args <- list(...)
  dest <- processExtraArg(args=args, name="dest", default="campsis")
  init <- processExtraArg(args=args, name="init", default=TRUE)
  capture <- processExtraArg(args=args, name="capture", default=FALSE)
  
  if (dest=="campsis" || isRxODE(dest) || dest=="NONMEM") {
    retValue <- paste0(object@lhs, "=", object@rhs)
  } else if (dest=="mrgsolve") {
    retValue <- paste0(object@lhs, "=", object@rhs, ";")
    if (init) {
      retValue <- paste0("double ", retValue)
    } else if (capture) {
      retValue <- paste0("capture ", retValue)
    }
  } else {
    UnsupportedDestException()
  }
  return(retValue %>% appendComment(object, dest))
})

#_______________________________________________________________________________
#----                         pmx_list class                             ----
#_______________________________________________________________________________

#' 
#' PMX list class.
#' 
#' @slot list effective list which will contain the elements
#' @slot type type of the elements this list may contain
#' @export
#' @keywords internal
setClass(
  "pmx_list",
  representation(
    list="list",
    type="character" # Interface / Main class type
  ),
  prototype=prototype(list=list()),
  validity=function(object) {
    check <- expectOne(object, "type")
    for (elem in object@list) {
      methods::validObject(elem, complete=TRUE) # TEST=FALSE (default) raises error
    }
    return(check)
  }
)

#_______________________________________________________________________________
#----                              add                                      ----
#_______________________________________________________________________________

#' Add element to list.
#' 
#' @param object list object
#' @param x element to add
#' @param ... extra arguments, unused by this generic list
#' @return modified list object
#' @export
#' @rdname add
add <- function(object, x, ...) {
  stop("No default function is provided")
}

setGeneric("add", function(object, x, ...) {
  standardGeneric("add")
})

#' @param pos position where x needs to be added in list
#' @rdname add
setMethod("add", signature=c("pmx_list", "pmx_element"), definition=function(object, x, pos=NULL) {
  if (methods::validObject(x)) {
    if (!is(x, object@type)) {
      stop(paste0("Element '", x %>% getName(), "' does not extend type '", object@type, "'."))
    
    } else if(object %>% contains(x)) {
      stop(paste0("Element '", x %>% getName(), "' is already present."))
    
    } else {
      if (is.null(pos)) {
        pos <- Position(object %>% length(), after=TRUE)
      }
      if (pos@by_index) {
        index <- pos@index
      } else if (pos@by_element) {
        index <- object %>% indexOf(pos@element)
      } else {
        stop("Element position can only by index or by position")
      }
      if (!pos@after) {
        index <- index - 1
      }
      object@list <- object@list %>% append(x, after=index)
    }
  }
  return(object)
})

#' @rdname add
setMethod("add", signature=c("pmx_list", "pmx_list"), definition=function(object, x) {
  return(object %>% add(x@list))
})

#' @rdname add
setMethod("add", signature=c("pmx_list", "list"), definition=function(object, x) {
  for (element in x) {
    object <- object %>% add(element)
  }
  return(object)
})

#_______________________________________________________________________________
#----                             replace                                   ----
#_______________________________________________________________________________

#' Replace element by another in list.
#' 
#' @param object list object
#' @param x element to replace
#' @return list object or an error if the element does not exist in the list
#' @export
#' @rdname replace
replace <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("replace", function(object, x) {
  standardGeneric("replace")
})

#' @rdname replace
setMethod("replace", signature=c("pmx_list", "pmx_element"), definition=function(object, x) {
  if (object %>% contains(x)) {
    index <- object %>% indexOf(x)
    object@list[[index]] <- x
  } else {
    stop(paste("Element", x %>% getName(), "does not exist."))
  }
  return(object)
})

#' @rdname replace
setMethod("replace", signature=c("pmx_list", "pmx_list"), definition=function(object, x) {
  return(object %>% replace(x@list))
})

#' @rdname replace
setMethod("replace", signature=c("pmx_list", "list"), definition=function(object, x) {
  for (element in x) {
    object <- object %>% replace(element)
  }
  return(object)
})

#_______________________________________________________________________________
#----                             indexOf                                   ----
#_______________________________________________________________________________

#' Get the index of an element in list.
#' 
#' @param object list object
#' @param x element to know the index
#' @return index of this element
#' @export
#' @rdname indexOf
indexOf <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("indexOf", function(object, x) {
  standardGeneric("indexOf")
})

#' @rdname indexOf
setMethod("indexOf", signature=c("pmx_list", "pmx_element"), definition=function(object, x) {
  logicalVector <- object@list %>% purrr::map_lgl(.f=function(.x) {
    retValue <- .x %>% getName() == x %>% getName()
    return(ifelse(is.na(retValue), FALSE, retValue))
  })
  index <- which(logicalVector)
  if (length(index) > 0) {
    index <- index[[1]]
  }
  return(index)
})

#_______________________________________________________________________________
#----                           getByName                                   ----
#_______________________________________________________________________________

#' Get an element from a list by name.
#' Never return more than 1 element.
#' 
#' @param object list object
#' @param name element name to search for
#' @return the element that was found or NULL if no element was found with the same name
#' @export
#' @rdname getByName
getByName <- function(object, name) {
  stop("No default function is provided")
}

setGeneric("getByName", function(object, name) {
  standardGeneric("getByName")
})

#' @rdname getByName
setMethod("getByName", signature=c("pmx_list", "character"), definition=function(object, name) {
  if (is.na(name)) {
    return(NULL)
  }
  return(object@list %>% purrr::detect(~(!is.na(.x %>% getName()) && .x %>% getName()==name)))
})

#_______________________________________________________________________________
#----                            contains                                   ----
#_______________________________________________________________________________

#' Check if an element exists in list.
#' 
#' @param object list object
#' @param x element to check if exists
#' @return logical value, TRUE or FALSE
#' @export
#' @rdname contains
#' @keywords internal
contains <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("contains", function(object, x) {
  standardGeneric("contains")
})

#' @rdname contains
setMethod("contains", signature=c("pmx_list", "pmx_element"), definition=function(object, x) {
  return(!is.null(object %>% find(x)))
})

#_______________________________________________________________________________
#----                             delete                                    ----
#_______________________________________________________________________________

#' Delete an element from this list.
#' 
#' @param object list object
#' @param x element to delete or element index
#' @return the updated list
#' @export
#' @rdname delete
delete <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("delete", function(object, x) {
  if (is.numeric(x)) {
    x <- as.integer(x)
  }
  if (is.integer(x) && x %>% length() > 1) {
    stop("Only 1 element can be deleted at a time")
  }
  standardGeneric("delete")
})

#' @rdname delete
setMethod("delete", signature=c("pmx_list", "pmx_element"), definition=function(object, x) {
  index <- object %>% indexOf(x)
  if (index %>% length() > 0) {
    object@list <- object@list[-index]
    return(object)
  } else {
    stop(paste("Element", x %>% getName(), "does not exist."))
  }
})

#' @rdname delete
setMethod("delete", signature=c("pmx_list", "integer"), definition=function(object, x) {
  if (x %>% length() != 1) {
    stop("x must be a single integer/numeric value")
  }
  if (x > 0 && x <= object %>% length()) {
    object@list <- object@list[-x]
    return(object)
  } else {
    stop(paste("No element exists at index", x))
  }
})

#_______________________________________________________________________________
#----                              find                                     ----
#_______________________________________________________________________________

#' Find an element in list.
#' 
#' @param object list object
#' @param x element to find, only key slots need to be filled in
#' @return the element from the list that has same name as x, or NULL if no element was found
#' @export
#' @rdname find
find <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("find", function(object, x) {
  standardGeneric("find")
})

#' @rdname find
setMethod("find", signature=c("pmx_list", "pmx_element"), definition=function(object, x) {
  return(object %>% getByName(x %>% getName()))
})

#_______________________________________________________________________________
#----                            getNames                                   ----
#_______________________________________________________________________________

#' Get element names from list.
#' 
#' @param object list object
#' @return character vector with all the element names of this list
#' @export
#' @rdname getNames
getNames <- function(object) {
  stop("No default function is provided")
}

setGeneric("getNames", function(object) {
  standardGeneric("getNames")
})

#' @rdname getNames
setMethod("getNames", signature=c("pmx_list"), definition=function(object) {
  return(object@list %>% purrr::map_chr(~.x %>% getName()))
})

#_______________________________________________________________________________
#----                             length                                    ----
#_______________________________________________________________________________

#' Return the length of this list.
#' 
#' @param x list object
#' @return the length of this list, integer value
#' @rdname length
#' @keywords internal
setMethod("length", signature=c("pmx_list"), definition=function(x) {
  return(length(x@list))
})

#_______________________________________________________________________________
#----                           getByIndex                                  ----
#_______________________________________________________________________________

#' Get element by index.
#' 
#' @param object list object
#' @param x element index
#' @return element from the list whose index matches with provided index
#' @export
#' @rdname getByIndex
getByIndex <- function(object, x) {
  stop("No default function is provided")
}

setGeneric("getByIndex", function(object, x) {
  standardGeneric("getByIndex")
})

#' @rdname getByIndex
setMethod("getByIndex", signature=c("pmx_list", "integer"), definition=function(object, x) {
  len <- object %>% length()
  assertthat::assert_that(len > 0, msg="x must be greater than 0")
  if (x > len) {
    stop(paste0("Can't find element at index ", x, " in list. List has a length of ", len, "."))
  }
  return(object@list[[x]])
})

#' @rdname getByIndex
setMethod("getByIndex", signature=c("pmx_list", "numeric"), definition=function(object, x) {
  return(getByIndex(object, x=as.integer(x)))
})


#_______________________________________________________________________________
#----                         sort (ABSTRACT)                               ----
#_______________________________________________________________________________

# Reuse base::sort(x, decreasing = FALSE, ...) definition

#' Sort the specified list.
#' 
#' @param x list object
#' @param decreasing increasing or decreasing order
#' @param ... extra arguments
#' @return same list but ordered
#' @export
#' @rdname sort
sort <- function(x, decreasing = FALSE, ...) {
  stop("No default function is provided")
}

setGeneric("sort", function(x, decreasing = FALSE, ...) {
  standardGeneric("sort")
})

#_______________________________________________________________________________
#----                       default (ABSTRACT)                              ----
#_______________________________________________________________________________

#' Get default element from list.
#' 
#' @param object list object
#' @param ... additional arguments
#' @return the default element from list
#' @export
default <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("default", function(object, ...) {
  standardGeneric("default")
})

library(testthat)

context("Test the generic list")

source(paste0("", "testUtils.R"))

test_that("GetByIndex method works well", {
  model <- model_suite$testing$nonmem$advan4_trans4
  
  depot <- model@compartments %>% getByIndex(1)
  expect_equal(depot@name, "DEPOT")
  
  expect_error(model@compartments %>% getByIndex(5), regexp="Can't find element at index 5")
})

test_that("Generic methods should throw an error when the call is incorrect", {
  msg <- "No default function is provided"
  expect_error(getPrefix(""), regexp=msg)
  expect_error(getRecordName(""), regexp=msg)
  expect_error(replace(""), regexp=msg)
  expect_error(indexOf(""), regexp=msg)
  expect_error(getByName(""), regexp=msg)
  expect_error(contains(""), regexp=msg)
  expect_error(find(""), regexp=msg)
  expect_error(getNames(""), regexp=msg)
  expect_error(getByIndex (""), regexp=msg)
  expect_error(sort(""), regexp=msg)
  expect_error(default(""), regexp=msg)
  expect_error(getName(""), regexp=msg)
})

library(testthat)

context("Test all methods that may be applied on compartments")

test_that("Method 'add' can't be used to add a compartment (add an ODE instead)", {
  model <- model_suite$testing$nonmem$advan4_trans4
  expect_error(model %>% add(Compartment(index=model@compartments %>% length() + 1, name="OUTPUT2")),
               regexp="No default function is provided")
})

test_that("Method 'contains' works with compartments", {
  model <- model_suite$testing$nonmem$advan4_trans4
  expect_true(model %>% contains(Compartment(1)))
  expect_false(model %>% contains(Compartment(5)))
})

test_that("Method 'delete' can't be used to delete a compartment (at least for now)", {
  model <- model_suite$testing$nonmem$advan4_trans4
  expect_error(model %>% delete(Compartment(index=1)), regexp="No default function is provided")
})

test_that("Method 'find' may return a compartment", {
  # Example 1
  model <- model_suite$testing$nonmem$advan4_trans4
  compartment <- model %>% find(Compartment(index=2))
  
  expect_equal(compartment@index, 2)
  expect_equal(compartment@name, "CENTRAL")
  
  # Example 2
  model <- model_suite$testing$nonmem$advan1_trans1
  compartments <- model@compartments
  compartment1 <- compartments %>% find(Compartment(index=1))
  compartment2 <- compartments %>% find(Compartment(index=2))
  
  expect_equal(compartments %>% length(), 2)
  expect_equal(compartment1@name, "CENTRAL")
  expect_equal(compartment2@name, "OUTPUT")
  
  expect_equal(capture_output(show(compartment1)), "A_CENTRAL (CMT=1)")
  expect_equal(capture_output(show(compartment2)), "A_OUTPUT (CMT=2)")
})

test_that("Method 'getCompartmentIndex' works as expected", {
  model <- model_suite$testing$nonmem$advan4_trans4
  index_depot <- model %>% getCompartmentIndex("DEPOT")
  index_central <- model %>% getCompartmentIndex("CENTRAL")
  
  expect_equal(index_depot, 1)
  expect_equal(index_central, 2)
  expect_error(model %>% getCompartmentIndex("XX"), regexp="Compartment XX not found")
})

test_that("Method 'replace' can be used to adapt the name of a compartment", {
  model <- model_suite$testing$nonmem$advan4_trans4 %>% add(Bioavailability(1, "0.75"))
  
  # Replace compartment and ODE
  model <- model %>% replace(Compartment(1, "ABS")) %>% 
    delete(Ode("A_DEPOT")) %>%
    add(Ode("A_ABS", "-KA*ABS"), Position(Ode("A_CENTRAL"), after=FALSE))
  
  # Replace occurrence in CENTRAL compartment
  central <- model %>% find(Ode("A_CENTRAL"))
  central@rhs <- gsub(pattern="A_DEPOT", replacement="A_ABS", x=central@rhs)
  model <- model %>% replace(central)
  
  # Check compartment has been correctly replaced
  compartment <- model %>% find(Compartment(1))
  expect_equal(compartment@name, "ABS")
  
  # Check 2 first ODE's
  abs <- model %>% find(Ode("A_ABS"))
  central <- model %>% find(Ode("A_CENTRAL"))
  expect_equal(abs@rhs, "-KA*ABS")
  expect_equal(central@rhs, "KA*A_ABS + Q*A_PERIPHERAL/V3 + (-CL/V2 - Q/V2)*A_CENTRAL")
  
  # Check bioavailability equation will be adapted too
  expect_equal(model %>% find(Bioavailability(1)) %>% toString(model=model, dest="campsis"), "A_ABS=0.75")
})

test_that("Method 'toString' works as expected on compartment", {
  cmt <- Compartment(index=1)
  expect_equal(cmt %>% toString(), "A_1")
  
  cmt <- Compartment(index=1, name="CENTRAL")
  expect_equal(cmt %>% toString(), "A_CENTRAL")
})

library(testthat)

context("Test utilities on the Campsis model")

source(paste0("", "testUtils.R"))

test_that("Method 'add' works as expected", {
  
  model <- model_suite$testing$nonmem$advan4_trans4
  
  # Add a parameter
  model <- model %>% add(Theta(name="XX", index=6, value=1))
  expect_equal(model@parameters %>% select("theta") %>% length(), 6)
  
  # Add a code record
  # TODO: update this
  # model <- model %>% add(PredRecord())
  # model@model <- model@model %>% sort()
  # expect_equal(model@model %>% length(), 4)
})

test_that("Method 'replace' works as expected", {
  
  model <- model_suite$testing$nonmem$advan4_trans4
  
  # Replace a parameter (1)
  model <- model %>% replace(Theta(name="KA", index=1, value=1.5))
  expect_equal((model %>% find(Theta("KA")))@value, 1.5)
  
  # Replace a parameter, index does not need to be provided (2)
  model <- model %>% replace(Theta(name="KA", value=1.5))
  expect_equal((model %>% find(Theta("KA")))@value, 1.5)
  
  # Replace a code record
  error <- ErrorRecord() # 0 statement
  model <- model %>% replace(error)
  expect_true(!is.null(model %>% find(ErrorRecord())))
  expect_equal(model %>% find(ErrorRecord()) %>% length(), 0)
  
  # Replace a model statement
  model <- model %>% replace(Equation("S2", "V2*1000"))
  equation <- model %>% find(Equation("S2"))
  expect_equal(equation@rhs, "V2*1000")
  
  # Add and replace a compartment property
  model <- model %>% add(Bioavailability(1, "0.75"))
  model <- model %>% replace(Bioavailability(1, "0.50"))
  expect_equal((model@compartments@properties %>% getByIndex(1))@rhs, "0.50")
})

test_that("Method 'delete' works as expected", {
  
  model <- model_suite$testing$nonmem$advan4_trans4
  
  # Delete a parameter
  updatedModel <- model %>% delete(Theta(name="KA"))
  expect_equal(updatedModel@parameters %>% length(), model@parameters %>% length() - 1)
  
  # Delete a code record
  updatedModel <- model %>% delete(ErrorRecord())
  expect_equal(updatedModel@model %>% length(), model@model %>% length() - 1)
  
  # Delete a model statement
  updatedModel <- model %>% delete(Equation("S2"))
  expect_equal(updatedModel %>% find(MainRecord()) %>% length(), model %>% find(MainRecord()) %>% length() - 1)
  
  # Add and delete a compartment property
  model <- model %>% add(Bioavailability(1, "0.75"))
  model <- model %>% add(Bioavailability(2, "0.75"))
  updatedModel <- model %>% delete(Bioavailability(1))
  expect_equal(updatedModel@compartments@properties %>% length(), model@compartments@properties %>% length() - 1)
  
  # Delete works also with an index
  main <- model %>% find(MainRecord())
  expect_true(main %>% contains(Equation("S2")))
  updatedMain <- main %>% delete(Equation("S2"))
  expect_false(updatedMain %>% contains(Equation("S2")))
})

test_that("Method 'add' on Campsis model, exceptions on parameters names", {
  model1 <- model_suite$testing$nonmem$advan4_trans4
  model2 <- model_suite$testing$nonmem$advan1_trans1
  expect_error(model1 %>% add(model2), regexp="Model can't be appended because of duplicate parameter name\\(s\\): EPS_PROP")
  
  model1 <- model_suite$testing$nonmem$advan1_trans1
  model2 <- model_suite$testing$nonmem$advan1_trans1
  expect_error(model1 %>% add(model2), regexp="Model can't be appended because of duplicate parameter name\\(s\\): THETA_K, THETA_V, ETA_K, ETA_V, EPS_PROP")
  
  # Unnamed correlations should not be an issue when merging models
  model1 <- model_suite$testing$nonmem$advan1_trans1 %>%
    addSuffix(suffix="1") %>%
    add(Omega(index=2, index2=1, value=0.5, type="cor"))
  model2 <- model_suite$testing$nonmem$advan1_trans1 %>%
    addSuffix(suffix="2") %>%
    add(Omega(index=2, index2=1, value=0.5, type="cor"))
  resultingModel <- model1 %>% add(model2)
  
  # Check correlations are still there at the correct omega indexes
  expect_equal(resultingModel@parameters %>% getByIndex(Omega(index=2, index2=1)) %>% .@type, "cor")
  expect_equal(resultingModel@parameters %>% getByIndex(Omega(index=4, index2=3)) %>% .@type, "cor") 
})

test_that("Method 'add' on Campsis model, exceptions on compartment names", {
  model1 <- model_suite$testing$nonmem$advan4_trans4
  model1@parameters@list <- model1@parameters@list[-(model1@parameters %>% length())] # Remove SIGMA PROP
  
  model2 <- model_suite$testing$nonmem$advan1_trans1
  model2@parameters@list <- model2@parameters@list[-(model2@parameters %>% length())] # Remove SIGMA PROP
  
  expect_error(model1 %>% add(model2), regexp="Element 'ODE \\(A_CENTRAL\\)' is already present.")
  #expect_error(model1 %>% add(model2), regexp="Model can't be appended because of duplicate compartment name\\(s\\): A_CENTRAL, A_OUTPUT")
})

test_that("Add effect compartment model to PK model using add method", {
  regFilename <- "pk_pd_appended"
  
  pk <- model_suite$testing$nonmem$advan4_trans4
  pk <- pk %>% add(Bioavailability(1, "0.75"))
  
  pd <- new("campsis_model") %>%
    add(Equation("KE0", "THETA_KE0*exp(ETA_KE0)")) %>%
    add(Ode("A_EFFECT", "KE0*(A_CENTRAL/S2 - A_EFFECT)"))
  
  expect_equal(pd@compartments %>% length(), 1)
  
  pd <- pd %>% add(Theta("KE0", value=0.5))
  pd <- pd %>% add(Omega("KE0", value=0.3, type="sd"))
  
  pd <- pd %>% add(InitialCondition(1, "0.40"))
  pd <- pd %>% add(Bioavailability(1, "0.50"))
  
  pkpd <- pk %>% add(pd)
  
  campsisNonRegTest(pkpd, regFilename)
})

test_that("Valid object method works depending on complete argument", {
  model <- model_suite$testing$nonmem$advan4_trans4
  model@parameters@list[[1]]@index <- c(1,1) %>% as.integer()
  
  expect_true(validObject(model))
  expect_error(validObject(model, complete=TRUE))
})

test_that("Method 'add' properly merges variance-covariance matrices", {
  modelA <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    addSuffix("A") %>%
    disable("VARCOV_OMEGA")
  
  modelB <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    addSuffix("B")
  
  model <- modelA %>%
    add(modelB)
  
  varcov <- model %>%
    getVarCov()
  
  varcovExpected <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    getVarCov()
  
  # Check variance-covariance content
  expect_equal(as.numeric(varcov[1:7, 1:7]), as.numeric(varcovExpected[1:7, 1:7]))
  expect_equal(as.numeric(varcov[8:17, 8:17]), as.numeric(varcovExpected))
  dimnames <- dimnames(varcov)
  expect_equal(dimnames[[1]], dimnames[[2]])
  
  expected <- c("THETA_METAB_CL_A", "THETA_DUR_A", "THETA_VC_A", "THETA_VP_A", "THETA_Q_A", "THETA_CL_A", "THETA_PROP_RUV_A",
                "THETA_METAB_CL_B", "THETA_DUR_B", "THETA_VC_B", "THETA_VP_B", "THETA_Q_B", "THETA_CL_B", "THETA_PROP_RUV_B",
                "OMEGA_DUR_B", "OMEGA_VC_B", "OMEGA_CL_B")
  expect_equal(expected, dimnames[[1]])
  
  # Add the empty model to the model with variance-covariance matrix
  expect_equal(as.numeric(modelB %>% add(CampsisModel()) %>% getVarCov()), as.numeric(varcovExpected))
  
  # Add the model with variance-covariance matrix to the empty model
  expect_equal(as.numeric(CampsisModel() %>% add(modelB) %>% getVarCov()), as.numeric(varcovExpected))
})

test_that("Method 'addRSE' works as expected", {
  model <- model_suite$testing$nonmem$advan4_trans4 %>%
    addRSE(Theta("CL"), 10) %>%
    addRSE(Theta("Q"), 8) %>% # Test #92 (addRSE can be called multiple times)
    addRSE(Theta("Q"), 10) %>%
    addRSE(Omega("KA"), 50)
  
  uncertainty <- getUncertainty(model) %>%
    dplyr::filter(!is.na(.data$`rse%`))
  
  expect_equal(tibble::tibble(name=c("THETA_CL", "THETA_Q", "OMEGA_KA"), se=c(0.5, 0.4, 0.0125), `rse%`=c(10, 10, 50)), uncertainty)
})

test_that("Method 'move' works as expected", {
  # Move CP from error block to ODE block
  model <- model_suite$testing$nonmem$advan4_trans4 %>%
    move(Equation("CP"), Position(OdeRecord()))
  
  # Check that CP is now in the ODE block
  expect_true(model %>% find(OdeRecord()) %>% contains(Equation("CP")))
  
  # Check that CP is not in the error block
  expect_false(model %>% find(ErrorRecord()) %>% contains(Equation("CP")))
  
  # Move the error block to the ODE block
  model <- model_suite$testing$nonmem$advan4_trans4 %>%
    move(ErrorRecord(), Position(OdeRecord()))
  
  # Check error block is empty
  expect_equal(model %>% find(ErrorRecord()) %>% length(), 0)
  expect_true(model %>% find(OdeRecord()) %>% contains(Equation("CP")))
  expect_true(model %>% find(OdeRecord()) %>% contains(Equation("OBS_CP")))
  expect_true(model %>% find(OdeRecord()) %>% contains(Equation("Y")))
})

library(testthat)

context("Test all methods from the parameters class")

source(paste0("", "testUtils.R"))

test_that("Write/Read THETA's", {
  
  file <- paste0(testFolder, "write/parameters/")
  
  # Read THETA's
  theta1 <- Theta(name="CL", index=1, value=5, fix=TRUE)
  theta2 <- Theta(name="KA", index=2, value=1, fix=TRUE)
  theta3 <- Theta(name="V", index=3,  value=80, fix=TRUE)
  
  thetas <- Parameters() %>% add(theta1) %>% add(theta2) %>% add(theta3)
  thetas %>% write(file=file)
  
  # Write THETA's
  thetas2 <- read.parameters(paste0(file, "theta.csv"), type="theta")
  
  # Check we can retrieve the exact same list
  expect_equal(thetas, thetas2)
})


test_that("Add/contains methods", {
  
  theta1 <- Theta(name="CL", index=1, value=5, fix=TRUE)
  theta2 <- Theta(name="KA", index=2, value=1, fix=TRUE)
  thetas <- Parameters() %>% add(theta1) %>% add(theta2)

  expect_true(thetas %>% contains(Theta(name="CL")))
  expect_false(thetas %>% contains(Theta(name="V")))
  
  thetas <- thetas %>% add(Theta(name="V"))
  expect_true(thetas %>% contains(Theta(name="V")))
})

test_that("GetByIndex, find & select method work well", {
  p <- Parameters()
  p <- p %>% add(Theta())
  p <- p %>% add(Theta())
  p <- p %>% add(Theta())
  p <- p %>% add(Omega())
  p <- p %>% add(Omega())
  p <- p %>% add(Sigma())

  # Search by index
  expect_equal(p %>% getByIndex(Theta(index=3)), Theta(index=3))
  expect_equal(p %>% getByIndex(Omega(index=1, index2=1)), Omega(index=1, index2=1))
  expect_equal(p %>% getByIndex(Sigma(index=1, index2=1)), Sigma(index=1, index2=1))
  
  # Or equivalently using find
  expect_equal(p %>% find(Theta(index=3)), Theta(index=3))
  expect_equal(p %>% find(Omega(index=1, index2=1)), Omega(index=1, index2=1))
  expect_equal(p %>% find(Sigma(index=1, index2=1)), Sigma(index=1, index2=1))
  
  p <- p %>% select("theta")
  expect_equal(p %>% length(), 3)
  
  # Check there is no conflict with dplyr::select
  df <- data.frame(ID=c(1,2,3), ID2=c(3,4,5))
  df <- df %>% select(ID)
  expect_false("ID2" %in% colnames(df))
})


test_that("Sort method", {
  
  sigma1 <- Sigma(index=1, index2=1, value=1)
  omega2 <- Omega(index=2, index2=2, value=1)
  omega1 <- Omega(index=1, index2=1, value=1)
  theta2 <- Theta(index=2, value=1)
  theta1 <- Theta(index=1, value=1)
  
  params <- Parameters() %>% add(c(sigma1, omega2, omega1, theta2, theta1))
  
  orderedParams <- params %>% sort()
  
  expectedParams <- Parameters() %>% add(c(theta1, theta2, omega1, omega2, sigma1))
  expect_equal(orderedParams, expectedParams)
})

test_that("Disable method (IIV/RUV)", {
  
  model <- model_suite$testing$nonmem$advan4_trans4
  
  model <- model %>% disable("IIV")
  expect_equal((model@parameters %>% find(Omega(name="KA")))@value, 0)
  expect_equal((model@parameters %>% find(Omega(name="CL")))@value, 0)
  expect_equal((model@parameters %>% find(Omega(name="V2")))@value, 0)
  expect_equal((model@parameters %>% find(Omega(name="V3")))@value, 0)
  expect_equal((model@parameters %>% find(Omega(name="Q")))@value, 0)
  expect_equal((model@parameters %>% find(Sigma(name="PROP")))@value, 0.025)
  
  model <- model %>% disable("RUV")
  expect_equal((model@parameters %>% find(Sigma(name="PROP")))@value, 0)
  
  expect_error(model %>% disable("SOMETHING"), regexp="Only these variabilities can be disabled")
})

test_that("Disable method (VARCOV)", {
  
  model <- model_suite$testing$other$my_model1
  expect_equal(model@parameters@varcov %>% length(), 49) # 7*7 matrix
  
  model <- model %>% disable("VARCOV")
  expect_equal(model@parameters@varcov %>% length(), 0)
})

test_that("Disable method (VARCOV_OMEGA & VARCOV_SIGMA)", {
  
  model <- model_suite$testing$other$my_model1
  expect_equal(model@parameters@varcov %>% length(), 49) # 7*7 matrix
  
  model <- model %>% disable("VARCOV_OMEGA")
  expect_equal(model@parameters@varcov %>% length(), 25) # 2 OMEGA's removed
  
  model <- model %>% disable("VARCOV_SIGMA")
  expect_equal(model@parameters@varcov %>% length(), 16) # 1 SIGMA removed
  
  # All at once
  model <- model_suite$testing$other$my_model1
  model <- model %>% disable(c("VARCOV_OMEGA", "VARCOV_SIGMA"))
  expect_equal(model@parameters@varcov %>% length(), 16)
})

test_that("Disable method (IOV)", {
  model <- read.campsis(paste0(testFolder, "custom/", "model1_omega_fixed"))
  model <- model %>% disable("IOV")
  
  omega4 <- model@parameters %>% getByIndex(Omega(index=4, index2=4))
  expect_true(omega4@value != 0)
  
  omega5 <- model@parameters %>% getByIndex(Omega(index=5, index2=5))
  expect_true(omega5@value == 0)
  
  omega6 <- model@parameters %>% getByIndex(Omega(index=6, index2=6))
  expect_true(omega6@value == 0)
})

test_that("Fix omega method is working", {
  
  # 'Unfix' OMEGA matrix
  original_model <- read.campsis(paste0(testFolder, "custom/", "model1_omega_fixed"))
  list <- original_model@parameters@list %>% purrr::map(.f=function(x) {
    if (is(x, "omega")) {
      if(isTRUE(x@same)) {
        x@value <- as.numeric(NA)
      }
      x@same <- as.logical(NA)
    }
    return(x)
  })
  model_not_fixed <- original_model
  model_not_fixed@parameters@list <- list
  
  # Fix OMEGA
  model_fixed <- model_not_fixed
  model_fixed@parameters <- model_fixed@parameters %>% fixOmega()
  
  expect_equal(original_model, model_fixed)
})

test_that("Name column is optional", {
  
  model <- read.campsis(paste0(testFolder, "custom/", "advan1_trans1_no_name"))
  names <- (model@parameters %>% select("theta"))@list %>% purrr::map_chr(.f=~.x@name)
  expect_true(all(is.na(names)))
})

test_that("THETA indexes validation", {
  # 2 THETAS OK
  theta1 <- Theta(index=1, value=0)
  theta2 <- Theta(index=2, value=0)
  thetas <- Parameters() %>% add(theta1) %>% add(theta2)
  expect_equal(thetas %>% length(), 2)
  expect_true(validObject(thetas))
  
  # Wrong starting index
  thetas <- Parameters() %>% add(Theta(index=2, value=0))
  expect_error(validObject(thetas))
  
  # NA value
  thetas <- Parameters() %>% add(Theta(index=1, value=NA))
  expect_error(validObject(thetas))
  
  # Index 2 missing
  theta1 <- Theta(index=1, value=0)
  theta2 <- Theta(index=3, value=0)
  thetas <- Parameters() %>% add(theta1) %>% add(theta2)
  expect_equal(thetas %>% length(), 2)
  expect_error(validObject(thetas))
})

test_that("OMEGA indexes validation", {
  # 2 OMEGA OK
  omega1 <- Omega(index=1, index2=1, value=0)
  omega2 <- Omega(index=2, index2=2, value=0)
  omegas <- Parameters() %>% add(omega1) %>% add(omega2)
  expect_equal(omegas %>% length(), 2)
  expect_true(validObject(omegas))
  
  # Wrong starting index
  omegas <- Parameters() %>% add(Omega(index=2, index2=2, value=0))
  expect_error(validObject(omegas))
  
  # NA value
  omegas <- Parameters() %>% add(Omega(index=1, index2=1, value=NA))
  expect_error(validObject(omegas))
  
  # Index 2 missing
  omega1 <- Omega(index=1, index2=1, value=0)
  omega2 <- Omega(index=3, index2=3, value=0)
  omegas <- Parameters() %>% add(omega1) %>% add(omega2)
  expect_equal(omegas %>% length(), 2)
  expect_error(validObject(omegas))
})

test_that("SIGMA indexes validation", {
  # 2 SIGMA OK
  sigma1 <- Sigma(index=1, index2=1, value=0)
  sigma2 <- Sigma(index=2, index2=2, value=0)
  sigmas <- Parameters() %>% add(sigma1) %>% add(sigma2)
  expect_equal(sigmas %>% length(), 2)
  expect_true(validObject(sigmas))
  
  # Wrong starting index
  sigmas <- Parameters() %>% add(Sigma(index=2, index2=2, value=0))
  expect_error(validObject(sigmas))
  
  # NA value
  sigmas <- Parameters() %>% add(Sigma(index=1, index2=1, value=NA))
  expect_error(validObject(sigmas))
  
  # Index 2 missing
  sigma1 <- Sigma(index=1, index2=1, value=0)
  sigma2 <- Sigma(index=3, index2=3, value=0)
  sigmas <- Parameters() %>% add(sigma1) %>% add(sigma2)
  expect_equal(sigmas %>% length(), 2)
  expect_error(validObject(sigmas))
})

test_that("Add parameters with NA indexes", {
  # Thetas
  parameters <- Parameters()
  parameters <- parameters %>% add(Theta(value=0))
  parameters <- parameters %>% add(Theta(value=0))
  parameters <- parameters %>% add(Theta(value=0))
  
  expect_equal(parameters %>% getByIndex(Theta(index=1)) %>% length(), 1)
  expect_equal(parameters %>% getByIndex(Theta(index=2)) %>% length(), 1)
  expect_equal(parameters %>% getByIndex(Theta(index=3)) %>% length(), 1)
  expect_true(validObject(parameters))
  
  # Omegas
  parameters <- Parameters()
  parameters <- parameters %>% add(Omega(value=0))
  parameters <- parameters %>% add(Omega(value=0))
  parameters <- parameters %>% add(Omega(value=0))
  
  expect_equal(parameters %>% getByIndex(Omega(index=1, index2=1)) %>% length(), 1)
  expect_equal(parameters %>% getByIndex(Omega(index=2, index2=2)) %>% length(), 1)
  expect_equal(parameters %>% getByIndex(Omega(index=3, index2=3)) %>% length(), 1)
  expect_true(validObject(parameters))
  
  # Sigmas
  parameters <- Parameters()
  parameters <- parameters %>% add(Sigma(value=0))
  parameters <- parameters %>% add(Sigma(value=0))
  parameters <- parameters %>% add(Sigma(value=0))
  
  expect_equal(parameters %>% getByIndex(Sigma(index=1, index2=1)) %>% length(), 1)
  expect_equal(parameters %>% getByIndex(Sigma(index=2, index2=2)) %>% length(), 1)
  expect_equal(parameters %>% getByIndex(Sigma(index=3, index2=3)) %>% length(), 1)
  expect_true(validObject(parameters))
})

test_that("Replace parameters without specifying the index", {
  # Thetas
  parameters <- Parameters()
  parameters <- parameters %>% add(Theta(name="X1", value=0))
  parameters <- parameters %>% add(Theta(name="X2", value=0))
  parameters <- parameters %>% add(Theta(name="X3", value=0))
  parameters <- parameters %>% replace(Theta(name="X1", value=10))
  
  expect_equal((parameters %>% find(Theta(name="X1")))@value, 10)
  
  # Omegas
  parameters <- Parameters()
  parameters <- parameters %>% add(Omega(name="X1", value=0))
  parameters <- parameters %>% add(Omega(name="X2", value=0))
  parameters <- parameters %>% add(Omega(name="X3", value=0))
  parameters <- parameters %>% replace(Omega(name="X1", value=10))
  
  expect_equal((parameters %>% find(Omega(name="X1")))@value, 10)
  
  # Sigmas
  parameters <- Parameters()
  parameters <- parameters %>% add(Sigma(name="X1", value=0))
  parameters <- parameters %>% add(Sigma(name="X2", value=0))
  parameters <- parameters %>% add(Sigma(name="X3", value=0))
  parameters <- parameters %>% replace(Sigma(name="X1", value=10))
  
  expect_equal((parameters %>% find(Sigma(name="X1")))@value, 10)
})

test_that("Add all method works well", {
  # Use c() operator and add
  parameters1 <- Parameters()
  parameters1 <- parameters1 %>% add(c(Theta(value=0), Theta(value=0), Theta(value=0)))
  
  expect_equal(parameters1 %>% length(), 3)
  
  # Use list() operator and add
  parameters2 <- Parameters()
  parameters2 <- parameters2 %>% add(list(Theta(value=0), Theta(value=0), Theta(value=0)))
  
  expect_equal(parameters2 %>% length(), 3)
  expect_equal(parameters1, parameters2)
})

test_that("Max index on empty list works and returns an integer", {
  parameters <- Parameters()
  index <- parameters %>% maxIndex()
  expect_true(is.integer(index))
  expect_equal(index, 0)
})

test_that("Parameters encoded with locale 'French' (semi-colon used as delimiter in CSV file) can be read properly", {
  
  # Model with parameters that were saved into CSV files on laptop with locale French
  # This model correspond to advan1_trans1
  model <- read.campsis(paste0(testFolder, "custom/", "csv_locale_french"))

  # Let's compare it with original model from model library
  expect_equal(model, model_suite$testing$nonmem$advan1_trans1)
})

test_that("Standardise method works as expected", {
  
  params1 <- Parameters() %>%
    add(Omega(name="CL", value=0.2, type="sd")) %>%
    add(Omega(name="V", value=0.3^2, type="var")) %>%
    add(Omega(name="CL_V", index=1, index2=2, value=0.25, type="cor"))
  
  params2 <- Parameters() %>%
    add(Omega(name="CL", value=0.2, type="sd")) %>%
    add(Omega(name="V", value=0.3^2, type="var")) %>%
    add(Omega(name="CL_V", index=1, index2=2, value=0.015, type="covar"))
  
  params_expected <- Parameters() %>%
    add(Omega(name="CL", value=0.2^2, type="var")) %>%
    add(Omega(name="V", value=0.3^2, type="var")) %>%
    add(Omega(name="CL_V", index=1, index2=2, value=0.015, type="covar"))
  
  expect_equal(params1 %>% standardise(), params_expected)
  expect_equal(params2 %>% standardise(), params_expected)
})

library(testthat)

context("Test the conversion of the simulation time when mrgsolve is used")

test_that("Time-dependent ODE's are correctly translated", {
  
  posMain <- Position(MainRecord())
  posOde <- Position(OdeRecord())
  weibullEquation <- Equation("WF", "DOSE * exp(-pow(t/TAU, SH)) * (SH/TAU) * pow(t/TAU, SH-1)")
  
  model <- CampsisModel() %>%
    add(Ode("A_ABS", "-WF*A_ABS"), pos=posOde) %>%
    add(Ode("A_CENTRAL", "WF*A_ABS - CL/VC*A_CENTRAL"), pos=posOde) %>%
    add(weibullEquation, pos=Position(OdeRecord(), after=FALSE)) %>% # WF will start the ODE block
    add(Equation("CP", "A_CENTRAL/VC"), pos=posOde) %>%
    add(Equation("TAU", "10"), pos=posMain) %>%
    add(Equation("SH", "5"), pos=posMain) %>%
    add(Equation("CL", "5"), pos=posMain) %>%
    add(Equation("VC", "100"), pos=posMain)
  
  ode <- model %>% find(OdeRecord())
  main <- model %>% find(MainRecord())
  expect_equal(ode@statements %>% getNames(), c("EQUATION (WF)", "ODE (A_ABS)", "ODE (A_CENTRAL)", "EQUATION (CP)"))
  expect_equal(main@statements %>% getNames(), c("EQUATION (TAU)", "EQUATION (SH)", "EQUATION (CL)", "EQUATION (VC)"))
  
  mrgmod <- model %>% export(dest="mrgsolve")
  expected <- c("[ODE]",
                "double WF=DOSE * exp(-pow(SOLVERTIME/TAU, SH)) * (SH/TAU) * pow(SOLVERTIME/TAU, SH-1);",
                "dxdt_A_ABS=-WF*A_ABS;",
                "dxdt_A_CENTRAL=WF*A_ABS - CL/VC*A_CENTRAL;",
                "double CP=A_CENTRAL/VC;")
})

library(testthat)

context("Test all methods from the code records class")

source(paste0("", "testUtils.R"))

test_that("Check records can be added correctly into the Campsis model", {
  
  model <- CampsisModel()
  model <- model %>% add(MainRecord())
  model <- model %>% add(ErrorRecord())
  
  expect_true(!is.null(model %>% find(MainRecord())))
  expect_true(!is.null(model %>% find(ErrorRecord())))
  expect_true(is.null(model %>% find(OdeRecord())))
})

test_that("Check non-record items can't be added to a list of code records", {
  records <- CodeRecords()
  expect_error(records %>% add(Theta(index=1)), regexp="Element 'THETA_1' does not extend type 'code_record'")
})

test_that("Write/Read methods", {
  
  model1 <- CodeRecords()
  model1 <- model1 %>% add(MainRecord(code=c("A=1", "B=2")))
  model1 <- model1 %>% add(ErrorRecord(code=c("C=3")))
  
  # Write model
  # Warning is well provided: model not provided, compartment characteristics will be lost
  expect_warning(model1 %>% write(file=paste0(testFolder, "write/records/records1.mod")))
  
  # Read model
  model2 <- read.model(file=paste0(testFolder, "write/records/records1.mod"))
  
  expect_equal(model1, model2)
})

test_that("Sort methods", {
  
  model <- CampsisModel()
  model <- model %>% add(OdeRecord())
  model <- model %>% add(ErrorRecord())
  model <- model %>% add(MainRecord())
  
  expectedModel <- CampsisModel()
  expectedModel <- expectedModel %>% add(MainRecord())
  expectedModel <- expectedModel %>% add(OdeRecord())
  expectedModel <- expectedModel %>% add(ErrorRecord())
  
  expect_equal(model %>% sort(), expectedModel)
})

test_that("Create very basic model on the fly", {
  model <- CampsisModel()
  
  # ODE must starts with 'A_'
  expect_error(model %>% add(Ode("CENTRAL", "-K*CENTRAL")), regexp="Derivative name must start with 'A_'")
  
  # Create minimalist model
  model <- model %>% add(Ode("A_CENTRAL", "-K*A_CENTRAL"))
  model <- model %>% add(Equation("THALF", "12"))
  model <- model %>% add(Equation("K", "log(2)/THALF"))
  
  expect_equal(model@model %>% getNames(), c("MAIN", "ODE"))
  
  # Check THALF and K can be found in MAIN code record
  expect_equal(model %>% find(MainRecord()) %>% find(Equation("THALF")), Equation("THALF", "12"))
  expect_equal(model %>% find(MainRecord()) %>% find(Equation("K")), Equation("K", "log(2)/THALF"))
  
  # Check A_CENTRAL ODE can be found in ODE record
  expect_equal(model %>% find(OdeRecord()) %>% find(Ode("A_CENTRAL")), Ode("A_CENTRAL", "-K*A_CENTRAL"))    
})

test_that("Equations can be removed", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  expect_equal(model %>% find(MainRecord()) %>% length(), 3) # 3 equations: K, V, S1
  
  model <- model %>% delete(Equation("S1"))
  expect_equal(model %>% find(MainRecord()) %>% length(), 2) # 2 equations: K, V
})

test_that("Equations can be replaced", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  expect_equal(model %>% find(MainRecord()) %>% length(), 3) # 3 equations: K, V, S1
  
  model <- model %>% replace(Equation("S1", "V/1000"))
  equation <- model %>% find(Equation("S1"))
  expect_equal(equation@rhs, "V/1000") # Equation well modified
})

test_that("Equations can be added at a specific position", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  
  model1 <- model %>% add(Equation("V2", "THETA_V2*exp(ETA_V2)"), Position(Equation("V")))
  model2 <- model %>% add(Equation("V2", "THETA_V2*exp(ETA_V2)"), Position(Equation("S1"), after=FALSE))
  
  expect_equal(model1, model2)
  
  model3 <- model %>% add(Equation("V2", "THETA_V2*exp(ETA_V2)"), Position(Equation("S1")))
  model4 <- model %>% add(Equation("V2", "THETA_V2*exp(ETA_V2)")) # Will be appended to MAIN block
  
  expect_equal(model3, model4)
  
  # By index
  model5 <- model %>% add(Equation("V2", "THETA_V2*exp(ETA_V2)"), Position(2))
  model6 <- model %>% add(Equation("V2", "THETA_V2*exp(ETA_V2)"), Position(3, after=FALSE))
  
  expect_equal(model5, model6)
  
  # Delete the added equation (by index vs by equation, works only at the level of the code record)
  main1 <- model1 %>% find(MainRecord())
  main2 <- model2 %>% find(MainRecord())
  expect_equal(main1 %>% delete(3L), main2 %>% delete(Equation("V2")))
})

test_that("Find method works well to search for a specific equation", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  equation <- model %>% find(Equation("V"))
  expect_equal(equation@rhs, "THETA_V*exp(ETA_V)")
  
  equation <- model %>% find(Equation("V2"))
  expect_true(is.null(equation))
})

test_that("Contains method can be used to check the existence of an equation", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  expect_true(model %>% contains(Equation("V")))
  expect_false(model %>% contains(Equation("V2")))
})

test_that("Add IF-statements at specific locations in the model", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  
  if1 <- IfStatement("COV==1", Equation("V", "V*1.0"))
  if2 <- IfStatement("COV==2", Equation("V", "V*1.1"))
  if3 <- IfStatement("COV==3", Equation("V", "V*1.2"))
  
  # Add before Equation S1
  model <- model %>% add(if1, Position(Equation("S1"), after=F))
  model <- model %>% add(if2, Position(Equation("S1"), after=F))
  
  # Or add after previous IF statement
  model <- model %>% add(if3, Position(if2, after=T))

  main <- model %>% find(MainRecord())
  expect_equal(main@statements %>% getByIndex(3), if1)
  expect_equal(main@statements %>% getByIndex(4), if2)
  expect_equal(main@statements %>% getByIndex(5), if3)
  expect_equal(main@statements %>% getByIndex(6), Equation("S1", "V"))
})

test_that("Add model statements into the given code record", {
  
  model <- model_suite$testing$nonmem$advan1_trans1
  
  cp <- Equation("CP", "A_CENTRAL/S1")
  
  model <- model %>%
    delete(Equation("F")) %>%
    delete(cp)
  
  # Add CP as first equation of ODE block
  model1 <- model %>% add(cp, pos=Position(OdeRecord(), after=FALSE))
  
  # Add CP as last equation of ODE block
  model2 <- model %>% add(cp, pos=Position(OdeRecord(), after=TRUE))
  
  # Retrieve respective ODE blocks and check where is CP
  ode1 <- model1 %>% find(OdeRecord())
  expect_equal(ode1@statements %>% getByIndex(1), cp)
  
  ode2 <- model2 %>% find(OdeRecord())
  expect_equal(ode2@statements %>% getByIndex(3), cp)
  
  # x in Position() must be either a PMX element or an integer
  expect_error(Position(data.frame(), after=FALSE), regexp="x can only be a PMX element or an integer position")
})

test_that("Method read.model can parse a model from a character vector", {
  
  expected <- model_suite$testing$pk$`1cpt_fo`
  
  input <- "
[MAIN]
TVKA=THETA_KA
TVVC=THETA_VC
TVCL=THETA_CL
TVPROP_RUV=THETA_PROP_RUV

KA=TVKA * exp(ETA_KA)
VC=TVVC * exp(ETA_VC)
CL=TVCL * exp(ETA_CL)
PROP_RUV=TVPROP_RUV

[ODE]
d/dt(A_ABS)=-KA*A_ABS
d/dt(A_CENTRAL)=KA*A_ABS - CL/VC*A_CENTRAL

[ERROR]
CONC=A_CENTRAL/VC
if (CONC <= 0.001) CONC=0.001
IPRED=log(CONC)
W=PROP_RUV
Y=IPRED + W*EPS_RUV_FIX
"
  # text can be a single string
  expect_equal(read.model(text=input), expected@model)
  
  # text can be a character vector
  expect_equal(read.model(text=(input %>% strsplit(split="\n"))[[1]]), expected@model)

})

library(testthat)

context("Test the conversion to rxode2")

source(paste0("", "testUtils.R"))

test_that("Export method works (2-cpt model, comments)", {
  regFilename = "2_cpt_model_abs_comments"
  model <- model_suite$testing$nonmem$advan4_trans4
  
  # Add some comments/line breaks programmatically
  main <- model@model %>% getByName("MAIN")
  main <- main %>% replace(Equation("KA", "THETA_KA*exp(ETA_KA)", comment="EQUATION COMMENT"))
  main <- main %>% add(LineBreak())
  main <- main %>% add(Comment("THIS IS A COMMENT"))
  
  model <- model %>% replace(main)
  
  model <- model %>% add(Bioavailability(compartment=1, rhs="0.75"))
  model <- model %>% add(LagTime(compartment=2, rhs="2"))
  model <- model %>% add(InfusionDuration(compartment=2, rhs="2"))
  model <- model %>% add(InfusionRate(compartment=1, rhs="2"))
  model <- model %>% add(InitialCondition(compartment=2, rhs="50"))
  
  # Code comparison
  rxmod <- model %>% export(dest="rxode2")
  rxodeNonRegTest(rxmod, regFilename)
  
  # Theta vector comparison
  expect_equal(c(THETA_KA=1, THETA_CL=5, THETA_V2=80, THETA_V3=20, THETA_Q=4), rxmod@theta)
  
  # Omega matrix comparison
  omega <- diag(rep(0.025, 5))
  rownames(omega) <- c("ETA_KA", "ETA_CL", "ETA_V2", "ETA_V3", "ETA_Q")
  colnames(omega) <- c("ETA_KA", "ETA_CL", "ETA_V2", "ETA_V3", "ETA_Q")
  expect_equal(omega, rxmod@omega)
  
  # Sigma matrix comparison
  sigma <- matrix(0.025)
  rownames(sigma) <- c("EPS_PROP")
  colnames(sigma) <- c("EPS_PROP")
  expect_equal(sigma, rxmod@sigma)
})

test_that("Export method works (2-cpt model, if-statements)", {
  regFilename = "2_cpt_model_abs_if_statements"
  
  model <- model_suite$testing$nonmem$advan4_trans4
  model <- model %>% delete(Equation("KA"))
  model <- model %>% add(Equation("KA", "0"))
  model <- model %>% add(IfStatement("OCC==1", Equation("KA", "THETA_KA*1.5*exp(ETA_KA)")))
  model <- model %>% add(IfStatement("OCC==2", Equation("KA", "THETA_KA*0.5*exp(ETA_KA)")))
  model <- model %>% add(IfStatement("OCC==3", Equation("KA", "THETA_KA*0.1*exp(ETA_KA)")))
  
  # Code comparison
  rxmod <- model %>% export(dest="rxode2")
  rxodeNonRegTest(rxmod, regFilename)
  
  # RxODE still accepted temporarily
  rxmod <- model %>% export(dest="RxODE")
  rxodeNonRegTest(rxmod, regFilename)
})

library(testthat)

context("Test the replication of the Campsis model")
source(paste0("", "testUtils.R"))

# options(campsismod.options=list(SKIP_PERFORMANCE_TESTS=TRUE))

test_that("Method 'replicate' allows to replicate a model based on its variance-covariance matrix", {
  
  set.seed(123)
  
  model <- model_suite$pk$`1cpt_fo` %>%
    setMinMax(Theta(name="KA"), min=0.9, max=1.1) %>%
    setMinMax(Theta(name="CL"), min=2.8, max=3.2) %>%
    addRSE(Theta(name="KA"), value=10) %>%
    addRSE(Theta(name="CL"), value=10)
  
  repModel <- model %>% replicate(1000)
  
  # Model can be printed out in the console!
  print(repModel)
  
  model1 <- repModel %>% export(dest=CampsisModel(), index=1)
  expect_equal(model1 %>% find(Theta(name="KA")) %>% .@value, 1.09958, tolerance=1e-4)
  expect_equal(model1 %>% find(Theta(name="CL")) %>% .@value, 2.831857, tolerance=1e-4)
  
  # The following test was moved from Campsis v1.6.0 to Campsismod v1.2.0
  
  set.seed(1)
  model <- model_suite$testing$other$my_model1
  repModel <- model %>% replicate(100)
  
  thetas <- repModel@replicated_parameters$THETA_CL
  var <- model@parameters@varcov["THETA_CL", "THETA_CL"]
  expect_equal(sd(thetas), sqrt(var), tolerance=1e-2)
  
  omegas <- repModel@replicated_parameters$OMEGA_CL
  var <- model@parameters@varcov["OMEGA_CL", "OMEGA_CL"]
  expect_equal(sd(omegas), sqrt(var), tolerance=1e-3)
  
  # Setting an impossible range will lead to an error
  set.seed(123)
  
  model <- model_suite$pk$`1cpt_fo` %>%
    setMinMax(Theta(name="KA"), min=1.3, max=2) %>%
    addRSE(Theta(name="KA"), value=10)
  
  expect_error(repModel <- model %>% replicate(1000), msg="Too many iterations")
  
  # Or may truncate the distribution as well if it is forced (here we increase the number of iterations)
  settings <- AutoReplicationSettings()
  settings@max_iterations <- 1000L
  repModel <- model %>% replicate(1000, settings=settings)
  # hist(repModel@replicated_parameters$THETA_KA)
  expect_equal(sum(repModel@replicated_parameters$THETA_KA >= 1.3), 1000)
  
  # Model can be printed out in the console!
  print(repModel)
})

getInverseWishartVar <- function(phiii, phijj, phiij, nu, p) {
  # See https://en.wikipedia.org/wiki/Inverse-Wishart_distribution
  num <- (nu-p+1)*phiij^2 + (nu-p-1)*phiii*phijj
  den <- (nu-p)*(nu-p-1)^2*(nu-p-3)
  return(num/den)
}

getInverseWishartVar2 <- function(phi, nu, p) {
  return((2*phi^2)/(nu-p-1)^2/(nu-p-3))
}

getInverseWishartMean <- function(phi, nu, p) {
  # See https://en.wikipedia.org/wiki/Inverse-Wishart_distribution
  return(phi/(nu-p-1))
}

getInverseChiSquaredVar <- function(tau2, nu) {
  # See https://en.wikipedia.org/wiki/Scaled_inverse_chi-squared_distribution
  return(2*nu^2*tau2^2/(nu-2)^2/(nu-4))
}

getInverseChiSquaredMean <- function(tau2, nu) {
  # See https://en.wikipedia.org/wiki/Scaled_inverse_chi-squared_distribution
  return(nu*tau2/(nu-2))
}

test_that("Sampling the OMEGAs and SIGMAs based on the scaled inverse chi-square or wishart distributions works as expected", {
  
  set.seed(123)
  
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    add(Omega(name="VC_CL", index=2, index2=3, value=0.8, type="cor")) %>%
    replace(Sigma(name="RUV_FIX", value=1, type="var", fix=FALSE)) # Unfix the RUV_FIX just for the test
  
  # Previously, issue when printing the model, see #88
  expect_no_error(show(model))
  
  settings <- AutoReplicationSettings(wishart=TRUE, odf=30, sdf=1000, quiet=FALSE)
  repModel <- model %>%
    replicate(30000, settings=settings)
  
  # Standardise model first
  model <- model %>%
    standardise()
  nu <- settings@odf
  nuSigma <- settings@sdf
  p <- 2
  
  # See https://en.wikipedia.org/wiki/Scaled_inverse_chi-squared_distribution
  
  # Check the generated values for OMEGA_DUR
  omegaDur <- model %>% find(Omega(name="DUR")) %>% .@value
  x <- repModel@replicated_parameters$OMEGA_DUR
  expect_equal(mean(x), getInverseChiSquaredMean(tau2=omegaDur, nu=nu), tolerance=0.0001)
  expect_equal(var(x), getInverseChiSquaredVar(tau2=omegaDur, nu=nu), tolerance=0.00001)
  
  # Check the generated values for OMEGA_VC
  omegaVc <- model %>% find(Omega(name="VC")) %>% .@value
  x <- repModel@replicated_parameters$OMEGA_VC
  expect_equal(mean(x), getInverseWishartMean(phi=omegaVc*nu, nu=nu, p=p), tolerance=0.0001)
  expect_equal(var(x), getInverseWishartVar2(phi=omegaVc*nu, nu=nu, p=p), tolerance=0.00001)
  
  # Check the generated values for OMEGA_CL
  omegaCl <- model %>% find(Omega(name="CL")) %>% .@value
  x <- repModel@replicated_parameters$OMEGA_CL
  expect_equal(mean(x), getInverseWishartMean(phi=omegaCl*nu, nu=nu, p=p), tolerance=0.0001)
  expect_equal(var(x), getInverseWishartVar2(phi=omegaCl*nu, nu=nu, p=p), tolerance=0.000001)
  
  # Check the generated values for OMEGA_VC (standardise parameter first!)
  omegaVcCl <- model %>% find(Omega(name="VC_CL")) %>% standardise(parameters=model@parameters) %>% .@value
  x <- repModel@replicated_parameters$OMEGA_VC_CL
  expect_equal(mean(x), getInverseWishartMean(phi=omegaVcCl*nu, nu=nu, p=p), tolerance=0.0001)
  expect_equal(var(x), getInverseWishartVar(phiii=omegaVc*nu, phijj=omegaCl*nu, phiij=omegaVcCl*nu, nu=nu, p=p), tolerance=0.00001)

  # Check the generated values for RUV_FIX
  sigmaFix <- model %>% find(Sigma(name="RUV_FIX")) %>% .@value
  x <- repModel@replicated_parameters$SIGMA_RUV_FIX
  expect_equal(mean(x), getInverseChiSquaredMean(tau2=sigmaFix, nu=nuSigma), tolerance=0.0005)
  expect_equal(var(x), getInverseChiSquaredVar(tau2=sigmaFix, nu=nuSigma), tolerance=0.00001)
  
  model1 <- repModel %>% export(dest=CampsisModel(), index=1)
  expect_equal(model1 %>% find(Omega(name="DUR")) %>% .@value, 0.01086832, tolerance=1e-4)
  expect_equal(model1 %>% find(Omega(name="VC")) %>% .@value, 0.07978391, tolerance=1e-4)
  expect_equal(model1 %>% find(Omega(name="CL")) %>% .@value, 0.05921943, tolerance=1e-4)
  expect_equal(model1 %>% find(Omega(name="VC_CL")) %>% .@value, 0.06040083, tolerance=1e-4)
  
  model2 <- repModel %>% export(dest=CampsisModel(), index=2)
  expect_equal(model2 %>% find(Omega(name="DUR")) %>% .@value, 0.01469925, tolerance=1e-4)
  expect_equal(model2 %>% find(Omega(name="VC")) %>% .@value, 0.06764799, tolerance=1e-4)
  expect_equal(model2 %>% find(Omega(name="CL")) %>% .@value, 0.04674834, tolerance=1e-4)
  expect_equal(model2 %>% find(Omega(name="VC_CL")) %>% .@value, 0.04431662, tolerance=1e-4)
  
  # Method replicate should also accept only 1 replicate (see issue #93)
  settings <- AutoReplicationSettings(wishart=TRUE, odf=30, sdf=1000, quiet=FALSE)
  
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    add(Omega(name="VC_CL", index=2, index2=3, value=0.8, type="cor")) %>%
    replace(Sigma(name="RUV_FIX", value=1, type="var", fix=FALSE)) %>%  # Unfix the RUV_FIX just for the test
    sort()
  
  repModel <- model %>%
    replicate(1, settings=settings)
  
  parameterNames <- model@parameters %>%
    getNames()
  
  expect_equal(c("REPLICATE", parameterNames), colnames(repModel@replicated_parameters))
})

test_that("Method 'setMinMax' can be used on THETAs, OMEGAs and SIGMAs to set limits", {
  
  set.seed(123)
  
  model1 <- model_suite$pk$`1cpt_fo` %>%
    setMinMax("theta", min=0, max=Inf) %>%
    addRSE(Theta(name="KA"), value=100) # Very large RSE on KA
  
  repModel1 <- model1 %>% replicate(1000, settings=AutoReplicationSettings(quiet=FALSE))
  
  # No THETA_KA should be negative
  expect_true(all(repModel1@replicated_parameters$THETA_KA >= 0))
  
  # Same model but no limit
  model2 <- model_suite$pk$`1cpt_fo` %>%
    addRSE(Theta(name="KA"), value=100) # Very large RSE on KA
  
  repModel2 <- model2 %>% replicate(1000, settings=AutoReplicationSettings(quiet=FALSE))
  
  # 841 values of THETA_KA are positive
  # 159 values of THETA_KA are negative
  expect_equal(sum(repModel2@replicated_parameters$THETA_KA >= 0), 841)
  expect_equal(sum(repModel2@replicated_parameters$THETA_KA < 0), 159)
})

test_that("Replicate a model that has IOV works as expected (+ check performances) ", {
  # The following test was moved from Campsis v1.6.0 to Campsismod v1.2.0
  set.seed(123)
  model <- model_suite$testing$nonmem$advan2_trans2

  iovOmegas <- list() %>%
    append(Omega(name="IOV_CL1", value=0.025, type="var", same=FALSE)) %>%
    append(Omega(name="IOV_CL2", value=0.025, type="var", same=TRUE)) %>%
    append(Omega(name="IOV_CL3", value=0.025, type="var", same=TRUE)) %>%
    append(Omega(name="IOV_CL4", value=0.025, type="var", same=TRUE)) %>%
    append(Omega(name="IOV_KA1", value=0.030, type="var", same=FALSE)) %>%
    append(Omega(name="IOV_KA2", value=0.030, type="var", same=TRUE)) %>%
    append(Omega(name="IOV_KA3", value=0.030, type="var", same=TRUE)) %>%
    append(Omega(name="IOV_KA4", value=0.030, type="var", same=TRUE))

  pk <- model %>%
    add(iovOmegas) %>%
    addRSE(Omega(name="IOV_CL1"), value=10) %>%
    addRSE(Omega(name="IOV_KA1"), value=20)
  
  replicates <- 1000
  repModel <- pk %>% replicate(replicates)
  
  # Check performances on exporting the model
  start <- Sys.time()
  models <- seq_len(replicates) %>%
    purrr::map(~repModel %>% export(dest=CampsisModel(), index=.x))
  end <- Sys.time()
  duration <- as.numeric(end - start)
  if (!skipPerformanceTests()) {
    expect_true(duration < 15) # (about 3 seconds on my machine)
  }
  
  # Check performances on exporting the OMEGA matrices
  start <- Sys.time()
  matrices <- models %>%
    purrr::map(~rxodeMatrix(.x))
  end <- Sys.time()
  duration <- as.numeric(end - start)
  if (!skipPerformanceTests()) {
    expect_true(duration < 5) # (about 1 seconds on my machine)
  }
  
  pk1 <- models[[1]]
  pk2 <- models[[2]]
  
  set <- iovOmegas %>% purrr::map_dbl(~pk %>% find(Omega(.x@name)) %>% .@value)
  set1 <- iovOmegas %>% purrr::map_dbl(~pk1 %>% find(Omega(.x@name)) %>% .@value)
  set2 <- iovOmegas %>% purrr::map_dbl(~pk2 %>% find(Omega(.x@name)) %>% .@value)
  
  expect_equal(set, c(0.025,0.025,0.025,0.025,0.03,0.03,0.03,0.03))
  expect_equal(round(set1, digits=4), c(0.0275,0.0275,0.0275,0.0275,0.0266,0.0266,0.0266,0.0266)) # Depends on seed
  expect_equal(round(set2, digits=4), c(0.0276,0.0276,0.0276,0.0276,0.0286,0.0286,0.0286,0.0286)) # Depends on seed
  
  # Check variance distribution on OMEGA IOV_CL_1
  iovCl1Omega <- models %>%
    purrr::map_dbl(~.x %>% find(Omega("IOV_CL1")) %>% .@value)
  expect_equal(sd(iovCl1Omega), 0.0025, tolerance=0.0005) # 10% RSE
  iovKa1Omega <- models %>%
    purrr::map_dbl(~.x %>% find(Omega("IOV_KA1")) %>% .@value)
  expect_equal(sd(iovKa1Omega), 0.0060, tolerance=0.0005) # 20% RSE
})
  
test_that("Method 'replicate' also allows to manually replicate a model based on a table (1 replicate/row)", {
  model <- model_suite$testing$nonmem$advan2_trans2
  data <- data.frame(REPLICATE=seq_len(5), SIGMA_PROP=c(0.1, 0.2, 0.3, 0.4, 0.5))
  
  expect_error(model %>% replicate(n=6, settings=ManualReplicationSettings(data)))

  repModel <- model %>% replicate(n=3, settings=ManualReplicationSettings(data))
  expect_equal(repModel@replicated_parameters$SIGMA_PROP, c(0.1, 0.2, 0.3))
  
  sigma1 <- repModel %>%
    export(dest=CampsisModel(), index=1) %>%
    find(Sigma("PROP"))
  expect_equal(sigma1@value, 0.1)
  
  sigma2 <- repModel %>%
    export(dest=CampsisModel(), index=2) %>%
    find(Sigma("PROP"))
  expect_equal(sigma2@value, 0.2)
  # Etc.
})

test_that("OMEGA's are correctly converted to block of OMEGA's (unfixed omegas, 1 correlation)", {
  
  parameters <- Parameters() %>%
    add(Theta(name="CL", value=5)) %>%
    add(Theta(name="VC", value=80)) %>%
    add(Theta(name="DUR", value=1)) %>%
    add(Theta(name="PROP_RUV", value=0.15)) %>%
    add(Omega(name="CL", value=30, type="cv%")) %>%
    add(Omega(name="VC", value=30, type="cv%")) %>%
    add(Omega(name="DUR", value=30, type="cv%")) %>%
    add(Omega(index=1, index2=2, value=0.5, type="cor"))
  
  # Create first block manually
  expectedBlock1 <- OmegaBlock() %>%
    add(parameters %>% find(Omega(name="CL"))) %>%
    add(parameters %>% find(Omega(name="VC"))) %>%
    add(parameters %>% find(Omega(index=1, index2=2)))
  expectedBlock1@start_index <- 0L
  expectedBlock1@block_index <- 1L
  
  expect_equal(expectedBlock1 %>% getOmegaIndexes(), c(1, 2))
  
  # Create all blocks automatically
  blocks <- OmegaBlocks() %>% add(parameters)
  
  expect_equal(blocks %>% length(), 2)
  
  block1 <- blocks %>% getByIndex(1)
  block2 <- blocks %>% getByIndex(2)
  expect_equal(block1, expectedBlock1)
  
  expect_true(block1 %>% length()==2)
  expect_true(block2 %>% length()==1)
  expect_true(block1 %>% hasOffDiagonalOmegas())
  expect_true(!block2 %>% hasOffDiagonalOmegas())
  
  expect_true(block1@start_index==0)
  expect_true(block2@start_index==2)
  
  expect_equal(capture_output(show(block1)), "BLOCK(2) - OMEGA_CL / OMEGA_VC")
  expect_equal(capture_output(show(block2)), "BLOCK(1) - OMEGA_DUR")
  
  # Generic functions not called correctly
  msg <- "No default function is provided"
  expect_error(getOmegaIndexes(""), regexp=msg)
  expect_error(hasOffDiagonalOmegas(""), regexp=msg)
  expect_error(shiftOmegaIndexes(""), regexp=msg)
})

test_that("Method 'show' called on replication settings works as expected", {
  manualSettings <- ManualReplicationSettings(data.frame(REPLICATE=1:5, THETA_KA=1:5))
  expect_equal(capture_output(show(manualSettings)), "Replication settings: manual")
  
  defaultSettings <- AutoReplicationSettings()
  expect_equal(capture_output(show(defaultSettings)), "Replication settings: default (auto)")
  
  adjustedSettings <- AutoReplicationSettings(wishart=TRUE, odf=30, sdf=1000)
  expect_equal(capture_output(show(adjustedSettings)), "Replication settings: wishart=TRUE, odf=30, sdf=1000")
})

test_that("Campsismod gives identical results to the simpar package", {
  
  set.seed(123)
  
  # Generate parameters with Campsismod
  model <- CampsisModel() %>%
    add(Theta(value=5)) %>%
    add(Omega(value=0.2, type="sd")) %>%
    add(Omega(value=0.4, type="sd")) %>%
    add(Omega(index=2, index2=1, value=0.1^2, type="covar")) %>%
    add(Sigma(value=0.1, type="sd")) %>%
    addRSE(Theta(index=1), value=4)
  
  res <- model %>%
    replicate(n=10000, settings=AutoReplicationSettings(wishart=TRUE, odf=30, sdf=1000)) %>%
    .@replicated_parameters
  
  expect_equal(mean(res$THETA_1), 5, tolerance=0.001)
  expect_equal(sd(res$THETA_1), 0.2, tolerance=0.001)
  
  expect_equal(mean(res$OMEGA_1_1), 0.044, tolerance=0.001)
  expect_equal(sd(res$OMEGA_1_1), 0.012, tolerance=0.001)
  
  expect_equal(mean(res$OMEGA_2_1), 0.011, tolerance=0.001)
  expect_equal(sd(res$OMEGA_2_1), 0.017, tolerance=0.001)
  
  expect_equal(mean(res$OMEGA_2_2), 0.177, tolerance=0.001)
  expect_equal(sd(res$OMEGA_2_2), 0.049, tolerance=0.001)
  
  expect_equal(mean(res$SIGMA_1_1), 0.010, tolerance=0.001)
  expect_equal(sd(res$SIGMA_1_1), 0.00045, tolerance=0.00001)
  
  # Generate parameters with simpar
  # theta <- c(5)
  # covar <- diag(0.2^2, 1)
  # omega <- matrix(c(0.2^2,0.1^2,0.1^2,0.4^2), nrow=2)
  # sigma <- matrix(0.1^2)
  # 
  # res <- simpar::simpar(10000, theta, covar, omega, sigma, odf=30, sdf=1000) %>%
  #   tibble::as_tibble()
  # 
  # mean(res$TH.1) # 5.000818
  # sd(res$TH.1) # 0.1994202
  # 
  # mean(res$OM1.1) # 0.04449487
  # sd(res$OM1.1) # 0.0128251
  # 
  # mean(res$OM2.1) # 0.01109515
  # sd(res$OM2.1) # 0.0176825
  # 
  # mean(res$OM2.2) # 0.1771933
  # sd(res$OM2.2) # 0.0497492
  # 
  # mean(res$SG1.1) # 0.01000736
  # sd(res$SG1.1) # 0.0004495882
})

test_that("Checking for positive definiteness works as expected", {
  set.seed(123)
  
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    add(Omega(name="VC_CL", index=2, index2=3, value=0.8, type="cor")) %>%
    replace(Sigma(name="RUV_FIX", value=1, type="var", fix=FALSE)) # Unfix the RUV_FIX just for the test
  
  settingsA <- AutoReplicationSettings(wishart=FALSE, quiet=FALSE, checkPosDef=TRUE)
  repModelA <- model %>%
    replicate(20, settings=settingsA)
  
  set.seed(123)
  
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl` %>%
    add(Omega(name="VC_CL", index=2, index2=3, value=0.8, type="cor")) %>%
    replace(Sigma(name="RUV_FIX", value=1, type="var", fix=FALSE)) # Unfix the RUV_FIX just for the test
  
  settingsB <- AutoReplicationSettings(wishart=FALSE, quiet=FALSE, checkPosDef=FALSE)
  repModelB <- model %>%
    replicate(20, settings=settingsB)
  
  # Compare manually
  # repModelA@replicated_parameters
  # repModelB@replicated_parameters
  
  # Omega matrix in replicate 9 of repModelB is not positive definite
  model9 <- repModelB %>% export(dest=CampsisModel(), index=9)
  expect_false(isMatrixPositiveDefinite(rxodeMatrix(model9, type="omega")))
  
  # Omega matrix in replicate 9 of repModelA is positive definite
  model9 <- repModelA %>% export(dest=CampsisModel(), index=9)
  expect_true(isMatrixPositiveDefinite(rxodeMatrix(model9, type="omega")))
})

library(testthat)

context("Test the auto-detect NONMEM function")

test_that("Bioavailability is well recognised", {
  model <- model_suite$testing$nonmem$advan4_trans4
  model <- model %>% add(Equation("F1", "0.75"))
  model <- model %>% autoDetectNONMEM()
  property <- model %>% find(Bioavailability(1))
  expect_equal(property, Bioavailability(1, "F1"))
})

test_that("Infusion duration is well recognised", {
  model <- model_suite$testing$nonmem$advan4_trans4
  model <- model %>% add(Equation("D2", "1"))
  model <- model %>% autoDetectNONMEM()
  property <- model %>% find(InfusionDuration(2))
  expect_equal(property, InfusionDuration(2, "D2"))
})

test_that("Infusion rate is well recognised", {
  model <- model_suite$testing$nonmem$advan4_trans4
  model <- model %>% add(Equation("R2", "500"))
  model <- model %>% autoDetectNONMEM()
  property <- model %>% find(InfusionRate(2))
  expect_equal(property, InfusionRate(2, "R2"))
})

test_that("Infusion rate is well recognised", {
  model <- model_suite$testing$nonmem$advan4_trans4
  model <- model %>% add(Equation("ALAG1", "2"))
  model <- model %>% autoDetectNONMEM()
  property <- model %>% find(LagTime(1))
  expect_equal(property, LagTime(1, "ALAG1"))
})

setClass(
  "mrgsolve_model",
  representation(
    param = "character",
    cmt = "character",
    main = "character",
    ode = "character",
    omega = "character",
    sigma = "character",
    table = "character",
    capture = "character"
  )
)

#_______________________________________________________________________________
#----                                export                                 ----
#_______________________________________________________________________________

#' @param outvars additional variables to capture
#' @param extra_params extra parameter names to be added. By default, they will be assigned a zero value.
#' @rdname export
setMethod("export", signature=c("campsis_model", "mrgsolve_type"), definition=function(object, dest, outvars=NULL, extra_params=character(0)) {
  return(
    new(
      "mrgsolve_model",
      param = mrgsolveParam(object, extra_params=extra_params),
      cmt = mrgsolveCompartment(object),
      main = mrgsolveMain(object),
      ode = mrgsolveOde(object),
      omega = mrgsolveMatrix(object, type="omega"),
      sigma = mrgsolveMatrix(object, type="sigma"),
      table = mrgsolveTable(object),
      capture = mrgsolveCapture(outvars, model=object)
    )
  )
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("mrgsolve_model"), definition=function(object, ...) {
  cpp <- NULL
  if (!is.null(object@param)) {
    cpp <- cpp %>% append(object@param)
    cpp <- cpp %>% append("")
  }
  cpp <- cpp %>% append(object@cmt)
  cpp <- cpp %>% append("")
  if (!is.null(object@omega)) {
    cpp <- cpp %>% append(object@omega)
    cpp <- cpp %>% append("")
  }
  if (!is.null(object@sigma)) {
    cpp <- cpp %>% append(object@sigma)
    cpp <- cpp %>% append("")
  }
  cpp <- cpp %>% append(object@main)
  cpp <- cpp %>% append("")
  cpp <- cpp %>% append(object@ode)
  cpp <- cpp %>% append("")
  cpp <- cpp %>% append(object@table)
  if (!is.null(object@capture)) {
    cpp <- cpp %>% append("")
    cpp <- cpp %>% append(object@capture)
  }
  cpp <- paste0(cpp, collapse="\n")
  return(cpp)
})

#_______________________________________________________________________________
#----                             addSuffix                                 ----
#_______________________________________________________________________________

#' Generic function to add a suffix to various objects like parameters, code records,
#' compartment names or a model (all previous objects at the same time).
#' This makes it an extremely powerful function to combine 2 models or more (using function 'add'),
#' that have similar equation, parameter or compartment names.
#' 
#' @param object generic object
#' @param suffix suffix to be appended, single character value
#' @param separator separator to use before the suffix, default is the underscore
#' @param ... extra arguments like 'model' if the changes need to be reflected in the model 
#' @return updated object of the same class as the provided object, unless 'model' was specified, in that case the model is returned
#' @export
#' @rdname addSuffix
addSuffix <- function(object, suffix, separator, ...) {
  stop("No default function is provided")
}

setGeneric("addSuffix", function(object, suffix, separator=NULL, ...) {
  if (!is.character(suffix) && length(suffix) != 1) {
    stop("suffix must be a single character value")
  }
  if (!(grepl(pattern=paste0("^", variablePatternNoStartStr(), "$"), x=suffix))) {
    stop(paste0("suffix '", suffix, "' is not a valid suffix"))
  }
  if (is.null(separator)) {
    separator <- "_" # Default value
  } else {
    if (!is.character(separator) && length(separator) != 1) {
      stop("separator must be a single character value")
    }
    if (!(grepl(pattern=paste0("^", variablePatternNoStartStr(), "$"), x=separator))) {
      stop(paste0("separator '", separator, "' is not a valid separator"))
    }
  }
  standardGeneric("addSuffix")
})

#' @rdname addSuffix
#' @importFrom assertthat are_equal
setMethod("addSuffix", signature=c("parameters", "character", "character"), definition=function(object, suffix, separator, ...) {
  args <- list(...)
  model <- args[["model"]]
  
  # Create a new list of parameters
  retValue <- Parameters()
  
  # Add suffix to all parameters and update model code accordingly
  for (parameter in object@list) {
    name <- parameter@name
    if (is.na(name) && is(parameter, "double_array_parameter") && !parameter %>% isDiag()) {
      # Do nothing except adding the parameter
      retValue <- retValue %>% add(parameter)
    } else if (is.na(name)) {
      stop(paste0("Please give a name to ", parameter %>% getNameInModel()))
    } else {
      oldName <- parameter %>% getNameInModel()
      parameter@name <- paste0(name, separator, suffix)
      newName <- parameter %>% getNameInModel()
      retValue <- retValue %>% add(parameter)
      if (!is.null(model)) {
        model <- model %>% replaceAll(pattern=oldName, replacement=newName)
      }
    }
  }
  
  # Update variance-covariance matrix
  varcov <- object@varcov
  if (length(varcov) > 0) {
    varcovDimNames <- dimnames(varcov)
    assertthat::are_equal(varcovDimNames[[1]], varcovDimNames[[2]])
    dimnames(varcov) <- list(paste0(varcovDimNames[[1]], separator, suffix), paste0(varcovDimNames[[2]], separator, suffix))
  }
  retValue@varcov <- varcov
  
  # Return updated parameters or updated model if model was provided
  if (is.null(model)) {
    return(retValue)
  } else {
    model@parameters <- retValue
    return(model)
  }
})

#' @rdname addSuffix
setMethod("addSuffix", signature=c("code_records", "character", "character"), definition=function(object, suffix, separator, ...) {
  args <- list(...)
  model <- args[["model"]]
  
  # Collect all equation names
  equationNames <- object@list %>%
    purrr::map(~getRecordEquationNames(.x)) %>%
    purrr::flatten_chr()
  
  # Get rid of duplicates
  equationNames <- equationNames[!duplicated(equationNames)]
  
  # List of code records
  retValue <- object
  
  # Add suffix
  for (equationName in equationNames) {
    replacementStr <- paste0(equationName, separator, suffix)
    retValue <- retValue %>% 
      replaceAll(pattern=VariablePattern(equationName), replacement=replacementStr)
    
    # Update properties as well (rhs)
    if (!is.null(model)) {
      for (index in seq_len(model@compartments@properties %>% length())) {
        model@compartments@properties@list[[index]]@rhs <- model@compartments@properties@list[[index]]@rhs %>%
          replaceAll(pattern=VariablePattern(equationName), replacement=replacementStr)
      }
    }
  }
  
  # Return updated code record list or updated model if model was provided
  if (is.null(model)) {
    return(retValue)
  } else {
    model@model <- retValue
    return(model)
  }
})

#' @rdname addSuffix
setMethod("addSuffix", signature=c("code_record", "character", "character"), definition=function(object, suffix, separator, ...) {
  retValue <- addSuffix(object=CodeRecords() %>% add(object), suffix=suffix, separator=separator, ...)
  if (is(retValue, "code_records")) {
    retValue <- retValue@list[[1]]
  }
  return(retValue)
})

#' @rdname addSuffix
setMethod("addSuffix", signature=c("compartments", "character", "character"), definition=function(object, suffix, separator, ...) {
  args <- list(...)
  model <- args[["model"]]
  
  # List of compartments
  retValue <- object
  
  # Update compartments
  for (index in seq_len(retValue %>% length())) {
    compartment <- retValue@list[[index]]
    oldName <- compartment %>% toString()
    replacementName <- paste0(oldName, separator, suffix)
    compartment@name <- gsub(pattern="^A_", replacement="", x=replacementName)
    retValue <- retValue %>%
      replace(compartment) # Replacement done by index for compartments
    if (!is.null(model)) {
      model <- model %>% 
        replaceAll(pattern=oldName, replacement=replacementName)
    }
  }
  
  # Return updated code record list or updated model if model was provided
  if (is.null(model)) {
    return(retValue)
  } else {
    model@compartments <- retValue
    return(model)
  }
})

#' @rdname addSuffix
setMethod("addSuffix", signature=c("campsis_model", "character", "character"), definition=function(object, suffix, separator, ...) {
  model <- object
  
  # Add suffix to parameters
  parameters <- model@parameters
  model <- parameters %>% addSuffix(suffix=suffix, separator=separator, model=model)
  
  # Add suffix to equations
  records <- model@model
  model <- records %>% addSuffix(suffix=suffix, separator=separator, model=model)
  
  # Add suffix to ODE compartments
  compartments <- model@compartments
  model <- compartments %>% addSuffix(suffix=suffix, separator=separator, model=model)
  
  # Save properties
  properties <- model@compartments@properties
  
  # Update compartment (properties will be lost)
  model <- model %>% updateCompartments()
  
  # Re-assign properties
  model@compartments@properties <- properties
  
  return(model)
})

setClass(
  "rxode_model",
  representation(
    code = "character",
    theta = "numeric",
    omega = "matrix",
    sigma = "matrix"
  )
)

#_______________________________________________________________________________
#----                                export                                 ----
#_______________________________________________________________________________

#' @rdname export
setMethod("export", signature=c("campsis_model", "rxode_type"), definition=function(object, dest, ...) {
  return(
    new(
      "rxode_model",
      code = rxodeCode(object),
      theta = rxodeParams(object),
      omega = rxodeMatrix(object, type = "omega"),
      sigma = rxodeMatrix(object, type = "sigma")
    )
  )
})

#_______________________________________________________________________________
#----                      replicated_campsis_model class                   ----
#_______________________________________________________________________________

#' 
#' Replicated Campsis model class.
#' 
#' @export
setClass(
  "replicated_campsis_model",
  representation(
    original_model = "campsis_model",     # Original Campsis model (sorted and standardised)
    replicated_parameters = "data.frame" # Replicated parameters
  ),
  prototype = prototype(original_model=CampsisModel(),
                        replicated_parameters=tibble::tibble())
)

#_______________________________________________________________________________
#----                           replicate                                   ----
#_______________________________________________________________________________

#' @rdname replicate
#' @importFrom methods validObject
#' @importFrom dplyr left_join
setMethod("replicate", signature = c("campsis_model", "integer", "auto_replication_settings"), definition = function(object, n, settings) {
  
  # Validate original Campsis model before sampling parameter uncertainty
  methods::validObject(object, complete=TRUE)

  # Sort and standardise model first
  object <- object %>%
    sort() %>%
    standardise()
  
  # Initialize a new replicated Campsis model
  retValue <- new("replicated_campsis_model", original_model=object)
  
  # Disable OMEGAs and SIGMAs in variance-covariance if Wishart is used
  if (settings@wishart) {
    object <- object %>%
      disable(c("VARCOV_OMEGA", "VARCOV_SIGMA"))
  }

  # Get variance-covariance matrix
  varcov <- object %>% getVarCov()

  if (varcov %>% length() == 0) {
    # No variance-covariance matrix is detected
    table <- tibble::tibble(REPLICATE=seq_len(n))
  } else {
    # Sample parameters in variance-covariance matrix from a multivariate normal distribution
    table <- sampleFromMultivariateNormalDistribution(parameters=object@parameters, n=n, settings=settings)
  }
  
  # Sample parameters (possibly OMEGA and SIGMA) from inverse chi-squared or Wishart distribution
  if (settings@wishart) {
    omegas <- object@parameters %>% select("omega")
    if (omegas %>% length() > 0) {
      sampledOmegas <- sampleFromInverseChiSquaredOrWishart(parameters=omegas, n=n, settings=settings)
      table <- table %>%
        dplyr::left_join(sampledOmegas, by="REPLICATE")
    }
    sigmas <- object@parameters %>% select("sigma")
    if (sigmas %>% length() > 0) {
      sampledSigmas <- sampleFromInverseChiSquaredOrWishart(parameters=sigmas, n=n, settings=settings)
      table <- table %>%
        dplyr::left_join(sampledSigmas, by="REPLICATE")
    }
  }
  
  retValue@replicated_parameters <- table
  return(retValue)
})

#' @rdname replicate
#' @importFrom methods validObject
setMethod("replicate", signature = c("campsis_model", "integer", "manual_replication_settings"), definition = function(object, n, settings) {
  
  # Validate original Campsis model
  methods::validObject(object, complete=TRUE)
  
  # Sort and standardise model first
  object <- object %>%
    sort() %>%
    standardise()
  
  # In the future, do additional checks on the data coming from the manual replication settings
  data <- settings@replicated_parameters
  if (nrow(data) < n) {
    stop("The number of rows in the data frame must be greater than or equal to the number of replicates 'n'.")
  }
  
  # Initialize a new replicated Campsis model
  retValue <- new("replicated_campsis_model", original_model=object, replicated_parameters=data[1:n,])
  return(retValue)
})

#_______________________________________________________________________________
#----                                export                                 ----
#_______________________________________________________________________________

#' @param index index of the replicated Campsis model to export
#' @rdname export
setMethod("export", signature=c("replicated_campsis_model", "campsis_model"), definition=function(object, dest=CampsisModel(), index, ...) {
  # If table is empty, return the original model
  if (nrow(object@replicated_parameters) == 0) {
    return(object@original_model)  
  }
  
  # Get the index of the last replicate
  maxIndex <- object@replicated_parameters %>%
    dplyr::pull("REPLICATE") %>%
    max()
  
  # Check the user-given index is in range
  if (index < 1 || index > maxIndex) {
    stop(sprintf("Index must be in the range [1, %i]", maxIndex))
  }
  
  # Find row
  row <- object@replicated_parameters %>%
    dplyr::filter(.data$REPLICATE==index) %>%
    dplyr::select(-c("REPLICATE"))
  
  retValue <- updateParameters(model=object@original_model, row=row)
  return(retValue)
})

#' Update model parameters based on the parameters issued from the model replication.
#' 
#' @param model Campsis model
#' @param row a data frame row containing the new parameter values
#' @return updated Campsis model
#' @importFrom assertthat assert_that
#' @keywords internal
#' 
updateParameters <- function(model, row) {
  assertthat::assert_that(nrow(row) == 1, msg="Only one row is expected.")
  sampledParameterNames <- names(row)
  sampledParameterValues <- as.numeric(row)
  parameters <- model@parameters
  parameterNames <- parameters %>% getNames()

  for (index in seq_along(sampledParameterNames)) {
    sampledParameterName <- sampledParameterNames[index]
    sampledParameterValue <- sampledParameterValues[index]
    parameterIndex <- which(sampledParameterName==parameterNames)
    if (length(index)==0) {
      stop(sprintf("Parameter %s not found", sampledParameterName))
    }
    model@parameters@list[[parameterIndex]]@value <- sampledParameterValue
  }
  
  # Update OMEGA's according that are same
  model <- updateOMEGAs(model)
  
  # Reset varcov
  model@parameters@varcov <- matrix(numeric(0), nrow=0, ncol=0)
  
  return(model)
}

#'
#' Update OMEGAs that are same. Same OMEGAs are written as follows:
#' OMEGA1 same is FALSE (first one, estimated)
#' OMEGA2 same is TRUE
#' OMEGA3 same is TRUE, etc.
#' OMEGA2 and OMEGA3 will take the same value as OMEGA1.
#' 
#' @param model Campsis model
#' @return updated Campsis model
#' @importFrom purrr map_lgl
#' @keywords internal
#' 
updateOMEGAs <- function(model) {
  isOmega <- model@parameters@list %>%
    purrr::map_lgl(~is(.x, "omega"))
  
  omegaIndexes <- which(isOmega)
  omegas <- model@parameters@list[omegaIndexes]
  
  sameVector <- omegas %>%
    purrr::map_lgl(~.x@same)
  
  # 'SAME' vector only contains NA's, do nothing
  allNAOmegas <- all(is.na(sameVector))
  if (allNAOmegas) {
    return(model)
  }
  
  # Apply run length encoding on 'SAME' vector
  rleRes <- rle(sameVector)
  lengths <- rleRes$lengths
  values <- rleRes$values
  falseIndexes <- which(values==FALSE)
  trueIndexes <- which(values==TRUE)
  
  for (falseIndex in falseIndexes) {
    # Check FALSE is followed by TRUE
    if ((falseIndex + 1) %in% trueIndexes) {
      falseOmegaIndex <- sum(lengths[1:falseIndex])
      trueOmegaIndexes <- falseOmegaIndex + seq_len(lengths[falseIndex + 1])
      
      # Estimated value
      falseOmegaValue <- omegas[[falseOmegaIndex]]@value
      
      # Replace where 'same' is TRUE
      for (trueOmegaIndex in trueOmegaIndexes) {
        omegas[[trueOmegaIndex]]@value <- falseOmegaValue
      }
    }
  }
  
  # Replace OMEGA's in the model
  model@parameters@list[omegaIndexes] <- omegas
    
  return(model)
}

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

#' @importFrom ggplot2 aes ggplot geom_vline geom_density theme_bw facet_wrap
#' @importFrom dplyr mutate
#' @importFrom tidyr pivot_longer
#' @importFrom purrr map_dbl
#' @importFrom assertthat assert_that
#' @importFrom tibble tibble
setMethod("show", signature=c("replicated_campsis_model"), definition=function(object) {
  
  model <- object@original_model
  parameters <- model@parameters
  parameterNames <- parameters %>% getNames()
  
  samplingData <- object@replicated_parameters %>%
    tidyr::pivot_longer(cols=-c("REPLICATE"), names_to="Parameter", values_to="Value")
  
  sampledParameterNames <- unique(samplingData$Parameter)
  
  assertthat::assert_that(all(sampledParameterNames %in% parameterNames),
                          msg="Some of the sampled parameters cannot be found in the model parameters.")
  
  sampledParameters <- parameters
  correspondingIndexes <- match(x=sampledParameterNames, table=parameterNames)
  sampledParameters@list <- sampledParameters@list[correspondingIndexes]

  parameterInfo <- tibble::tibble(Parameter=sampledParameters %>% getNames(),
                                  Value=sampledParameters@list %>% purrr::map_dbl(~.x@value)) %>%
    dplyr::mutate(Parameter=factor(.data$Parameter, levels=sampledParameterNames)) # Natural order
  
  samplingData <- samplingData %>%
    dplyr::mutate(Parameter=factor(.data$Parameter, levels=sampledParameterNames)) # Natural order
  
  plot <- ggplot2::ggplot(data=samplingData, mapping=ggplot2::aes(x=.data$Value, group=.data$Parameter)) +
    ggplot2::geom_vline(data=parameterInfo, mapping=ggplot2::aes(xintercept=.data$Value, group=.data$Parameter), linetype="dashed", colour="black") +
    ggplot2::geom_density() +
    ggplot2::theme_bw() +
    ggplot2::facet_wrap(~Parameter, scales="free")
  print(plot)
})

#' Campsis model suite.
#'
#' A library of models of all kinds, ready to be simulated in Campsis. These model
#' templates are sorted into the following categories: pharmacokinetic (PK), pharmacodynamic (PD), 
#' target-mediated drug disposition (TMDD), NONMEM, literature  and other (custom models).
#'
#' @format A list with all the models:
#' \describe{
#'    \item{pk}{extensive list of pharmacokinetic (PK) model templates}
#'    \item{pd}{list of pharmacodynamic (PD) model templates, to be plugged into any pharmacokinetic (PK) model}
#'    \item{tmdd}{extensive list of target-mediated drug disposition (TMDD) model templates}
#'    \item{nonmem}{list of model templates translated from standard NONMEM control streams}
#'    \item{literature}{a couple of models coming from the literature}
#'    \item{other}{a couple of custom models}
#' }
#' @source \url{https://calvagone.github.io/campsis.doc/}
#' @source \url{https://www.iconplc.com/solutions/technologies/nonmem/}
"model_suite"

#_______________________________________________________________________________
#----                         campsis_model class                           ----
#_______________________________________________________________________________

#' Campsis model class.
#' 
#' @slot model a list of code records
#' @slot parameters model parameters
#' @slot compartments model compartments
#' @export
setClass(
  "campsis_model",
  representation(
    model = "code_records",
    parameters = "parameters",
    compartments = "compartments"
  )
)

#' Create a new Campsis model.
#' 
#' @param json path to JSON Campsis model file or JSON content in string form
#' @return a Campsis model, empty
#' @export
CampsisModel <- function(json=NULL) {
  if (is.null(json)) {
    model <- new("campsis_model")
  } else {
    schema <- system.file("extdata", "campsismod.schema.json", package="campsismod")
    model <-  loadFromJSON(CampsisModel(), openJSON(json=json, schema=schema))
  }
  return(model)
}

#_______________________________________________________________________________
#----                                add                                    ----
#_______________________________________________________________________________

#' @rdname add
setMethod("add", signature=c("campsis_model", "compartment_property"), definition=function(object, x) {
  compartment <- object@compartments %>% find(Compartment(index=x@compartment))
  if (is.null(compartment)) {
    stop(paste0("Unable to find compartment ", x@compartment, " in Campsis model"))
  }
  
  # Add characteristic (delegate to add method in compartments class)
  object@compartments <- object@compartments %>% add(x) 
  
  return(object)
})

#' @rdname add
setMethod("add", signature=c("campsis_model", "parameter"), definition=function(object, x) {
  object@parameters <- object@parameters %>% add(x)
  return(object)
})

#' @rdname add
setMethod("add", signature=c("campsis_model", "code_record"), definition=function(object, x) {
  object@model <- object@model %>% add(x)
  return(object)
})

#' @param pos position where x needs to be added in list
#' @rdname add
setMethod("add", signature=c("campsis_model", "model_statement"), definition=function(object, x, pos=NULL) {
  if (is(x, "ode")) {
    object@compartments <- object@compartments %>% addODECompartment(ode=x)
  }
  object@model <- object@model %>% add(x, pos=pos)
  return(object)
})

#' @rdname add
setMethod("add", signature=c("campsis_model", "campsis_model"), definition=function(object, x) {
  object <- object %>% appendModel(x)
  return(object)
})

#' @rdname add
setMethod("add", signature=c("campsis_model", "list"), definition=function(object, x) {
  for (elem in x) {
    object <- object %>% add(elem)
  }
  return(object)
})

#' Append model (or simply add).
#' 
#' @param model1 base model
#' @param model2 model to append
#' @return the resulting Campsis model
#' @keywords internal
#' 
appendModel <- function(model1, model2) {
  # Append compartments (and included properties)
  model1@parameters <- model1@parameters %>% add(model2@parameters)
  
  # Append code records
  model1@model <- model1@model %>% add(model2@model)
  
  # Append compartments (and included properties)
  model1@compartments <- model1@compartments %>% add(model2@compartments)
  return(model1)
}

#_______________________________________________________________________________
#----                             addRSE                                    ----
#_______________________________________________________________________________

#' @rdname addRSE
setMethod("addRSE", signature=c("campsis_model", "parameter", "numeric"), definition=function(object, parameter, value, ...) {
  object@parameters <- object@parameters %>%
    addRSE(parameter=parameter, value=value, ...)
  return(object)
})

#_______________________________________________________________________________
#----                         autoDetectNONMEM                              ----
#_______________________________________________________________________________

#' @rdname autoDetectNONMEM
setMethod("autoDetectNONMEM", signature=c("campsis_model"), definition=function(object, ...) {
  main <- object@model %>% getByName("MAIN")
  numberOfCmts <- object@compartments %>% length()
  
  for (cmtIndex in seq_len(numberOfCmts)) {
    # Search for bioavailability
    fVar <- paste0("F", cmtIndex)
    if (main %>% find(Equation(fVar)) %>% length() > 0) {
      object <- object %>% add(Bioavailability(cmtIndex, rhs=fVar))
    }
    # Search for infusion duration
    dVar <- paste0("D", cmtIndex)
    if (main %>% find(Equation(dVar)) %>% length() > 0) {
      object <- object %>% add(InfusionDuration(cmtIndex, rhs=dVar))
    }
    # Search for infusion rate
    rVar <- paste0("R", cmtIndex)
    if (main %>% find(Equation(rVar)) %>% length() > 0) {
      object <- object %>% add(InfusionRate(cmtIndex, rhs=rVar))
    }
    # Search for lag time
    alagVar <- paste0("ALAG", cmtIndex)
    if (main %>% find(Equation(alagVar)) %>% length() > 0) {
      object <- object %>% add(LagTime(cmtIndex, rhs=alagVar))
    }
  }
  
  return(object)
})

#_______________________________________________________________________________
#----                               contains                                ----
#_______________________________________________________________________________

#' @rdname contains
setMethod("contains", signature=c("campsis_model", "pmx_element"), definition=function(object, x) {
  return(!is.null(object %>% find(x)))
})

#_______________________________________________________________________________
#----                               delete                                  ----
#_______________________________________________________________________________

#' @rdname delete
setMethod("delete", signature=c("campsis_model", "compartment_property"), definition=function(object, x) {
  object@compartments <- object@compartments %>% delete(x)
  return(object)
})

#' @rdname delete
setMethod("delete", signature=c("campsis_model", "parameter"), definition=function(object, x) {
  object@parameters <- object@parameters %>% delete(x)
  return(object)
})

#' @rdname delete
setMethod("delete", signature=c("campsis_model", "code_record"), definition=function(object, x) {
  object@model <- object@model %>% delete(x)
  return(object)
})

#' @rdname delete
setMethod("delete", signature=c("campsis_model", "model_statement"), definition=function(object, x) {
  object@model <- object@model %>% delete(x)
  return(object)
})

#_______________________________________________________________________________
#----                              disable                                  ----
#_______________________________________________________________________________

#' @rdname disable
setMethod("disable", signature=c("campsis_model", "character"), definition=function(object, x, ...) {
  object@parameters <- object@parameters %>% disable(x=x, ...)
  return(object)
})

#_______________________________________________________________________________
#----                           export_type                                 ----
#_______________________________________________________________________________

#' RxODE/rxode2 export type class.
#' 
#' @export
setClass(
  "rxode_type",
  representation(
  ),
  contains="export_type" 
)

#' Mrgsolve export type class.
#' 
#' @export
setClass(
  "mrgsolve_type",
  representation(
  ),
  contains="export_type" 
)

#_______________________________________________________________________________
#----                                export                                 ----
#_______________________________________________________________________________

#' @rdname export
setMethod("export", signature=c("campsis_model", "character"), definition=function(object, dest, ...) {
  if (isRxODE(dest)) {
    return(object %>% export(new("rxode_type")))
  } else if (dest=="mrgsolve") {
    return(object %>% export(new("mrgsolve_type"), ...))
  } else {
    stop("Only rxode2 (previously RxODE) and mrgsolve are currently supported")
  }
})

#_______________________________________________________________________________
#----                                find                                   ----
#_______________________________________________________________________________

#' @rdname find
setMethod("find", signature=c("campsis_model", "compartment"), definition=function(object, x) {
  return(object@compartments %>% find(x))
})

#' @rdname find
setMethod("find", signature=c("campsis_model", "compartment_property"), definition=function(object, x) {
  return(object@compartments %>% find(x))
})

#' @rdname find
setMethod("find", signature=c("campsis_model", "parameter"), definition=function(object, x) {
  return(object@parameters %>% find(x))
})

#' @rdname find
setMethod("find", signature=c("campsis_model", "code_record"), definition=function(object, x) {
  return(object@model %>% find(x))
})

#' @rdname find
setMethod("find", signature=c("campsis_model", "model_statement"), definition=function(object, x) {
  return(object@model %>% find(x))
})

#_______________________________________________________________________________
#----                           exportToJSON                                ----
#_______________________________________________________________________________

#' @rdname exportToJSON
#' @importFrom utils capture.output
setMethod("exportToJSON", signature=c("campsis_model"), definition=function(object, ...) {
  # Delete error record if empty
  errorRecord <- object %>%
    find(ErrorRecord())
  if (!is.null(errorRecord) && length(errorRecord)==0) {
    object <- object %>%
      delete(ErrorRecord())
  }
  lines <- capture.output(show(object@model %>% addPropertiesRecords(model=object)))
  json <- list()
  json$code <- lines
  json$parameters <- exportToJSON(object@parameters)@data
  if (length(object@parameters@varcov) > 0) {
    json$varcov <- varcovToJSON(parameters=object@parameters)
  }
  return(JSONElement(json))
})

#_______________________________________________________________________________
#----                          getCompartmentIndex                          ----
#_______________________________________________________________________________

#' @rdname getCompartmentIndex
setMethod("getCompartmentIndex", signature=c("campsis_model", "character"), definition=function(object, name) {
  return(object@compartments %>% getCompartmentIndex(name=name))
})

#_______________________________________________________________________________
#----                          getUncertainty                               ----
#_______________________________________________________________________________

#' @rdname getUncertainty
setMethod("getUncertainty", signature=c("campsis_model"), definition=function(object, ...) {
  return(object@parameters %>% getUncertainty(...))
})

#_______________________________________________________________________________
#----                             getVarCov                                 ----
#_______________________________________________________________________________

#' @rdname getVarCov
setMethod("getVarCov", signature=c("campsis_model"), definition=function(object) {
  return(object@parameters %>% getVarCov())
})

#_______________________________________________________________________________
#----                           loadFromJSON                                ----
#_______________________________________________________________________________

#' @rdname loadFromJSON
setMethod("loadFromJSON", signature=c("campsis_model", "json_element"), definition=function(object, json) {
  object <- jsonToCampsisModel(object=object, json=json)
  return(object)
})

#' @rdname loadFromJSON
setMethod("loadFromJSON", signature=c("campsis_model", "character"), definition=function(object, json) {
  schema <- system.file("extdata", "campsismod.schema.json", package="campsismod")
  return(loadFromJSON(object=object, json=openJSON(json=json, schema=schema)))
})

#_______________________________________________________________________________
#----                              move                                     ----
#_______________________________________________________________________________

#' @rdname move
setMethod("move", signature=c("campsis_model", "ANY", "pmx_position"), definition=function(object, x, to, ...) {
  object@model <- object@model %>% move(x=x, to=to, ...)
  return(object)
})

#_______________________________________________________________________________
#----                                 read                                  ----
#_______________________________________________________________________________

#' Read a Campsis model.
#' 
#' @param file path to folder (old format) or path to JSON file (new format)
#' @return Campsis model
#' @export
read.campsis <- function(file) {
  if (dir.exists(file)) {
    folder <- file
    
    # model.campsis and model.pmx are both accepted
    modelPath <- file.path(folder, "model.campsis")
    if (!file.exists(modelPath)) {
      modelPath <- file.path(folder, "model.pmx")
      warning("Please rename your model file to 'model.campsis'")
      if (!file.exists(modelPath)) {
        stop(paste0("Model file couln't be found."))
      }
    }
    
    # Construct Campsis model
    records <- read.model(file=modelPath)
    parameters <- read.allparameters(folder=folder)
    model <- new("campsis_model", model=records, parameters=parameters, compartments=Compartments())
    model <- model %>%
      updateCompartments()
    
  } else if (file.exists(file)) {
    if (!endsWith(file, ".json")) {
      stop("Only JSON model files can be opened.")
    }
    model <- loadFromJSON(CampsisModel(), paste0(readLines(file), collapse="\n"))
  } else {
    stop("file is not a JSON model or a valid Campsis model folder")
  }
  
  # Validate the whole model
  methods::validObject(model, complete=TRUE)
  
  return(model)
}


#' Read a Campsis model (deprecated).
#' 
#' @param file path to folder
#' @return a Campsis model
#' @export
#' @keywords internal
read.pmxmod <- function(file) {
  .Deprecated("read.campsis")
  return(read.campsis(file=file))
}

#' Update compartments list from the persisted records.
#' Exported especially for package \code{campsistrans}.
#' However, this method should not be called.
#' 
#' @param model Campsis model
#' @return an updated Campsis model, with an updated compartments list
#' @export
updateCompartments <- function(model) {
  if (!is(model, "campsis_model")) {
    stop("model is not a Campsis model")   
  }
  records <- model@model
  
  # Get list of compartments
  compartments <- records %>% getCompartments()
  
  # Extract characteristics
  compartments <- compartments %>% addProperties(records, "F", init=Bioavailability(0, rhs=""))
  compartments <- compartments %>% addProperties(records, "LAG", init=LagTime(0, rhs=""))
  compartments <- compartments %>% addProperties(records, "DURATION", init=InfusionDuration(0, rhs=""))
  compartments <- compartments %>% addProperties(records, "RATE", init=InfusionRate(0, rhs=""))
  compartments <- compartments %>% addProperties(records, "INIT", init=InitialCondition(0, rhs=""))
  
  # Remove properties records because information is found in properties
  records@list <- records@list %>% purrr::keep(~!is(.x, "properties_record"))

  model@model <- records
  model@compartments <- compartments
  return(model)
}

#_______________________________________________________________________________
#----                               replace                                 ----
#_______________________________________________________________________________

#' @rdname replace
setMethod("replace", signature=c("campsis_model", "compartment"), definition=function(object, x) {
  object@compartments <- object@compartments %>% replace(x)
  return(object)
})

#' @rdname replace
setMethod("replace", signature=c("campsis_model", "compartment_property"), definition=function(object, x) {
  object@compartments <- object@compartments %>% replace(x)
  return(object)
})

#' @rdname replace
setMethod("replace", signature=c("campsis_model", "parameter"), definition=function(object, x) {
  object@parameters <- object@parameters %>% replace(x)
  return(object)
})

#' @rdname replace
setMethod("replace", signature=c("campsis_model", "code_record"), definition=function(object, x) {
  object@model <- object@model %>% replace(x)
  return(object)
})

#' @rdname replace
setMethod("replace", signature=c("campsis_model", "model_statement"), definition=function(object, x) {
  object@model <- object@model %>% replace(x)
  return(object)
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("campsis_model", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  # Call 'replaceAll' on each code record
  object@model@list <- object@model@list %>%
    purrr::map(~.x %>% replaceAll(pattern=pattern, replacement=replacement, ...))
  
  # Call 'replaceAll' on each compartment property
  object@compartments@properties@list <- object@compartments@properties@list %>%
    purrr::map(~.x %>% replaceAll(pattern=pattern, replacement=replacement, ...))
  
  return(object)
})

#' @rdname replaceAll
setMethod("replaceAll", signature=c("campsis_model", "character", "character"), definition=function(object, pattern, replacement, ...) {
  # Call 'replaceAll' on each code record
  object@model@list <- object@model@list %>%
    purrr::map(~.x %>% replaceAll(pattern=VariablePattern(pattern), replacement=replacement, ...))
  
  # Call 'replaceAll' on each compartment property
  object@compartments@properties@list <- object@compartments@properties@list %>%
    purrr::map(~.x %>% replaceAll(pattern=VariablePattern(pattern), replacement=replacement, ...))
  
  return(object)
})

#_______________________________________________________________________________
#----                            setMinMax                                  ----
#_______________________________________________________________________________

#' @rdname setMinMax
setMethod("setMinMax", signature=c("campsis_model", "parameter", "numeric", "numeric"), definition=function(object, parameter, min, max, ...) {
  object@parameters <- object@parameters %>%
    setMinMax(parameter=parameter, min=min, max=max, ...)
  return(object)
})

#' @rdname setMinMax
setMethod("setMinMax", signature=c("campsis_model", "character", "numeric", "numeric"), definition=function(object, parameter, min, max, ...) {
  object@parameters <- object@parameters %>%
    setMinMax(parameter=parameter, min=min, max=max, ...)
  return(object)
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("campsis_model"), definition=function(object) {
  show(object@model %>% addPropertiesRecords(model=object))
  cat("\n")
  show(object@parameters)
  cat("\n")
  show(object@compartments)
})

#_______________________________________________________________________________
#----                                  sort                                 ----
#_______________________________________________________________________________

#' @rdname sort
setMethod("sort", signature=c("campsis_model"), definition=function(x, decreasing=FALSE, ...) {
  # Sort code records
  x@model <- x@model %>% sort()
  
  # Sort compartments (properties will be sorted correctly)
  x@compartments <- x@compartments %>% sort()
  
  # Sort parameters
  x@parameters <- x@parameters %>% sort()
  
  return(x)
})

#_______________________________________________________________________________
#----                            standardise                                ----
#_______________________________________________________________________________

#' @rdname standardise
setMethod("standardise", signature=c("campsis_model"), definition=function(object, ...) {
  object@parameters <- object@parameters %>% standardise(...)
  return(object)
})

#_______________________________________________________________________________
#----                                 write                                 ----
#_______________________________________________________________________________

#' @rdname write
setMethod("write", signature=c("campsis_model", "character"), definition=function(object, file, ...) {
  if (endsWith(file, ".json")) {
    return(exportToJSON(object) %>% campsismod::write(file=file))
  }
  
  zip <- processExtraArg(args=list(...), name="zip", default=FALSE)
  records <- object@model
  parameters <- object@parameters

  if (zip) {
    
  } else {
    if (dir.exists(file)) {
      # do nothing
    } else {
      dir.create(file)
    }
    records %>% write(file=file.path(file, "model.campsis"), model=object)
    parameters %>% write(file=file)
  }
  return(TRUE)
})

#' @rdname write
#' @param digits significant digits in JSON file, default is 12
setMethod("write", signature=c("json_element", "character"), definition=function(object, file, digits=12, ...) {
  jsonlite::write_json(object@data, path=file, pretty=TRUE, auto_unbox=TRUE, digits=digits)
  return(TRUE)
})

#_______________________________________________________________________________
#----                           omega_block class                           ----
#_______________________________________________________________________________

setClass(
  "omega_block",
  representation(
    block_index="integer",
    start_index="integer",
    on_diag_omegas="parameters",
    off_diag_omegas="parameters"
  ),
  contains="pmx_element"
)

#' Create a block of OMEGA's.
#'
#' @export
OmegaBlock <- function() {
  return(new("omega_block"))
}

#_______________________________________________________________________________
#----                                add                                    ----
#_______________________________________________________________________________

#' @rdname add
setMethod("add", signature=c("omega_block", "double_array_parameter"), definition=function(object, x) {
  if (x %>% isDiag()) {
    object@on_diag_omegas <- object@on_diag_omegas %>% add(x)
  } else {
    object@off_diag_omegas <- object@off_diag_omegas %>% add(x)
  }
  return(object)
})

#_______________________________________________________________________________
#----                              getName                                  ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature=c("omega_block"), definition=function(x) {
  return(paste0("Omega block ", x@block_index))
})

#_______________________________________________________________________________
#----                           getOmegaIndexes                             ----
#_______________________________________________________________________________

#' Get the indexes of the omegas.
#' 
#' @param object omega block
#' @return a list of integers
#' @export
#' @rdname getOmegaIndexes
getOmegaIndexes <- function(object) {
  stop("No default function is provided")
}

setGeneric("getOmegaIndexes", function(object) {
  standardGeneric("getOmegaIndexes")
})

#' @rdname getOmegaIndexes
setMethod("getOmegaIndexes", signature=c("omega_block"), definition=function(object) {
  # Careful: don't use parameters1 %>% add(parameters2) because it will shift indexes in parameters2
  omegas <- object@on_diag_omegas@list %>% append(object@off_diag_omegas@list)
  retValue <- omegas %>% purrr::map(.f=~c(.x@index, .x@index2)) %>% purrr::flatten_int() %>% unique()
  return(retValue)
})

#_______________________________________________________________________________
#----                       hasOffDiagonalOmegas                            ----
#_______________________________________________________________________________

#' Has off-diagonal omegas.
#' 
#' @param object omega block
#' @return TRUE or FALSE
#' @export
#' @rdname hasOffDiagonalOmegas
hasOffDiagonalOmegas <- function(object) {
  stop("No default function is provided")
}

setGeneric("hasOffDiagonalOmegas", function(object) {
  standardGeneric("hasOffDiagonalOmegas")
})

#' @rdname hasOffDiagonalOmegas
setMethod("hasOffDiagonalOmegas", signature=c("omega_block"), definition=function(object) {
  return(object@off_diag_omegas %>% length() > 0)
})

#_______________________________________________________________________________
#----                         shiftOmegaIndexes                             ----
#_______________________________________________________________________________

#' Shift OMEGA indexes.
#' 
#' @param object omega block
#' @return same block but shifted
#' @export
#' @rdname shiftOmegaIndexes
shiftOmegaIndexes <- function(object) {
  stop("No default function is provided")
}

setGeneric("shiftOmegaIndexes", function(object) {
  standardGeneric("shiftOmegaIndexes")
})

shiftOmega <- function(omega, x) {
  omega@index <- omega@index + x
  omega@index2 <- omega@index2 + x
  return(omega)
}

#' @rdname shiftOmegaIndexes
setMethod("shiftOmegaIndexes", signature=c("omega_block"), definition=function(object) {
  object@on_diag_omegas@list <- object@on_diag_omegas@list %>% purrr::map(.f=~shiftOmega(.x, -object@start_index))
  object@off_diag_omegas@list <- object@off_diag_omegas@list %>% purrr::map(.f=~shiftOmega(.x, -object@start_index))
  return(object)
})

#_______________________________________________________________________________
#----                             length                                    ----
#_______________________________________________________________________________

#' Return the number of OMEGA's on the diagonal.
#' 
#' @param x omega block
#' @return a number
#' @rdname length
setMethod("length", signature=c("omega_block"), definition=function(x) {
  return(length(x@on_diag_omegas))
})

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

getBlockLabel <- function(object) {
  omegaNames <- object@on_diag_omegas@list %>%
    purrr::map_chr(.f=~.x %>% getName())
  retValue <- sprintf("BLOCK(%i) - %s", length(omegaNames), omegaNames %>% paste(collapse=" / "))
  return(retValue)
}

setMethod("show", signature=c("omega_block"), definition=function(object) {
  cat(getBlockLabel(object))
})

#' Min/max default values for the given parameter.
#' 
#' @param parameter Campsis parameter
#' @return a tibble with the min and max values
#' @importFrom tibble tibble
#' @keywords internal
minMaxDefault <- function(parameter) {
  min <- parameter@min
  max <- parameter@max
  minNA <- is.na(min)
  maxNA <- is.na(max)
  
  if (is(parameter, "theta")) {
    return(tibble::tibble(min=ifelse(minNA, -Inf, min), max=ifelse(maxNA, Inf, max)))
  } else if (is(parameter, "double_array_parameter")) {
    if (parameter %>% isDiag()) {
      return(tibble::tibble(min=ifelse(minNA, 0, min), max=ifelse(maxNA, Inf, max)))
    } else {
      return(tibble::tibble(min=ifelse(minNA, -Inf, min), max=ifelse(maxNA, Inf, max)))
    }
  }
  stop("Unknown parameter type")
}

#' Sample from a multivariate normal distribution.
#' 
#' @param parameters model parameters (all), already standardised
#' @param n number of rows to sample
#' @param settings replication settings
#' @return a data frame with the sampled parameters
#' @keywords internal
#' 
sampleFromMultivariateNormalDistribution <- function(parameters, n, settings) {
  varcov <- parameters@varcov
  parameterNames <- parameters %>% getNames()
  varcovParameterNames <- colnames(varcov)
  assertthat::assert_that(all(varcovParameterNames %in% parameterNames),
                          msg="Some of the variance-covariance matrix parameters are not in the model parameters.")
  varcovParameters <- parameters
  correspondingIndexes <- match(x=varcovParameterNames, table=parameterNames)
  varcovParameters@list <- varcovParameters@list[correspondingIndexes]
  
  # Retrieve variance-covariance matrix names
  varcovNames <- colnames(varcov)
  
  # Retrieve min, max
  minMax <- varcovParameters@list %>%
    purrr::map_df(~minMaxDefault(.x)) %>%
    dplyr::mutate(name=varcovNames)
  
  # Retrieve mean
  mean <- varcovParameters@list %>%
    purrr::map_dbl(~.x@value)
  
  # Parameters used to assess positive definiteness
  if (settings@wishart) {
    parameters <- NULL  
  } else {
    # Only keep OMEGA/SIGMA parameters
    parameters@list <- parameters@list %>%
      purrr::keep(~is(.x, "double_array_parameter"))
  }
  
  # Sample parameters from the variance-covariance matrix
  msg <- getSamplingMessageTemplate(what="parameters", from="variance-covariance matrix")
  table <- sampleGeneric(fun=sampleFromMultivariateNormalDistributionCore,
                         args=list(mean=mean, varcov=varcov), n=n, minMax=minMax, msg=msg, settings=settings, parameters=parameters)
  
  return(table)
}

#' Sample parameters from inverse scaled chi-squared or wishart distribution(s).
#' 
#' @param parameters subset of Campsis parameters (OMEGAs or SIGMAs)
#' @param n number of rows to sample
#' @param settings replication settings
#' @return a data frame with the sampled parameters
#' @importFrom assertthat assert_that
#' @importFrom dplyr bind_cols
#' @importFrom tibble tibble
#' @keywords internal
#' 
sampleFromInverseChiSquaredOrWishart <- function(parameters, n, settings) {
  assertthat::assert_that(parameters %>% length() > 0)
  
  # Type of parameter
  type <- class(parameters@list[[1]]) %>% as.character()
  
  if (type=="omega") {
    dfs <- settings@odf
  } else if (type=="sigma") {
    dfs <- settings@sdf
  } else {
    stop("Should be either omega or sigma")
  }
  
  # Assign default names if not provided since the blocks re-index the OMEGA's
  parameters@list <- parameters@list %>% purrr::map(.f=function(x) {
    if (is.na(x@name)) {
      x@name <- paste0(x@index, "_", x@index2)
    }
    return(x)
  })

  # Detect blocks
  blocks <- OmegaBlocks() %>%
    add(parameters)

  retValue <- tibble::tibble(REPLICATE=seq_len(n))
  
  # Iterate over blocks
  for (blockIndex in seq_along(blocks@list)) {
    block <- blocks@list[[blockIndex]]
    blockLabel <- getBlockLabel(block)
    if (length(dfs)==1) {
      df <- dfs
    } else {
      assertthat::assert_that(length(dfs)==length(blocks@list),
                              msg="The number of blocks must correspond to the number of degrees of freedom")
      df <- dfs[blockIndex]
      if (isFALSE(settings@quiet)) {
        cat(sprintf("Sampling %s with %i degrees of freedom\n", blockLabel, df))
      }
    }
    if (isBlockFixed(block)) {
      next
    }
    if (length(block) == 1) {
      onDiagElements <- block@on_diag_omegas
      assertthat::assert_that(!hasOffDiagonalOmegas(block))
      elem <- onDiagElements@list[[1]]
      variable <- elem %>% getName()
      minMax <- minMaxDefault(elem) %>%
        dplyr::mutate(name=variable)
      msg <- getSamplingMessageTemplate(what=blockLabel, from="scaled inverse chi-squared distribution")
      tmp <- sampleGeneric(fun=sampleFromInverseChiSquaredCore, args=list(df=df, scale=elem@value, variable=variable), n=n, minMax=minMax, msg=msg, settings=settings)
      retValue <- dplyr::bind_cols(retValue, tmp[, -1])
    } else {
      params <- Parameters()
      params@list <- c(block@on_diag_omegas@list, block@off_diag_omegas@list)
      mat <- rxodeMatrix(params, type=type)

      size <- params@list %>% purrr::keep(~isDiag(.x)) %>% length()
      mappingMat <- getMappingMatrix(parameters=params, type=type)
      allColnames <- as.vector(mappingMat)

      minMax <- params@list %>%
        purrr::map_df(~minMaxDefault(.x)) %>%
        dplyr::mutate(name=params@list %>% purrr::map_chr(~.x %>% getName()))
      
      msg <- getSamplingMessageTemplate(what=blockLabel, from="scaled inverse Wishart distribution")
      wishartCorrection <- settings@wishart_correction
      tmp <- sampleGeneric(fun=sampleFromInverseWishartCore, args=list(df=df, mat=mat, allColnames=allColnames, wishartCorrection=wishartCorrection),
                           n=n, minMax=minMax, msg=msg, settings=settings)
      retValue <- dplyr::bind_cols(retValue, tmp[, -1])
    }
  }
  return(retValue)
}

#' Say if the block is fixed (i.e. all parameters are fixed).
#' 
#' @param block block of OMEGAs or SIGMAs
#' @return logical value
#' @importFrom purrr map_lgl
#' @keywords internal
isBlockFixed <- function(block) {
  list <- c(block@on_diag_omegas@list, block@off_diag_omegas@list)
  retValue <- list %>%
    purrr::map_lgl(~.x@fix)
  return(all(retValue))
}

#' Sample from multivariate normal distribution (core method).
#' 
#' @param n numbers of rows to sample
#' @param mean mean values to give to the multivariate normal distribution
#' @param varcov variance-covariance matrix
#' @return a data frame with the sampled parameters
#' @importFrom MASS mvrnorm
#' @importFrom tibble as_tibble
#' @keywords internal
sampleFromMultivariateNormalDistributionCore <- function(n, mean, varcov) {
  retValue <- MASS::mvrnorm(n=n, mu=mean, Sigma=varcov) %>%
    tibble::as_tibble()
  # If n=1, mvrnorm returns an unnamed vector
  if (n==1) {
    retValue <- data.frame(t(retValue)) %>%
      tibble::as_tibble()
    colnames(retValue) <- colnames(varcov)
  }
  return(retValue)
}

#' Sample from scaled inverse chi-squared distribution (core method).
#' 
#' @param n numbers of rows to sample
#' @param df degree of freedom
#' @param scale scale parameter
#' @param variable variable name
#' @return a data frame with the unique sampled parameter
#' @importFrom LaplacesDemon rinvchisq
#' @importFrom tibble tibble
#' @keywords internal
sampleFromInverseChiSquaredCore <- function(n, df, scale, variable) {
  table <- tibble::tibble(!!variable:=LaplacesDemon::rinvchisq(n=n, df=df, scale=scale))
  return(table)
}

#' Sample from scaled inverse Wishart distribution (core method).
#' 
#' @param n numbers of rows to sample
#' @param df degree of freedom
#' @param mat scaling matrix
#' @param allColnames all column names as they are going to appear when as.vector is called
#' @param wishartCorrection apply correction to the degrees of freedom, logical
#' @return a data frame with the sampled parameters
#' @importFrom LaplacesDemon rinvwishart
#' @keywords internal
sampleFromInverseWishartCore <- function(n, df, mat, allColnames, wishartCorrection) {
  indexesToKeep <- which(allColnames != "") # See getMappingMatrix method
  
  # If the correction is applied, the degrees of freedom are reduced by the number of rows + 1
  # See https://github.com/metrumresearchgroup/simpar/issues/11
  correction <- ifelse(wishartCorrection, -nrow(mat) + 1, 0)
  
  table <- seq_len(n) %>%
    purrr::map_df(~data.frame(t(as.vector(LaplacesDemon::rinvwishart(nu=df, S=mat*(df + correction)))))[, indexesToKeep])
  colnames(table) <- allColnames[indexesToKeep]
  return(table)
}

#' Generic function for parameter sampling according to the minimum and maximum values.
#' This function will sample parameters a first time and check if some parameters are out of range.
#' Based on the success rate, it will sample more parameters to reach the desired number of rows.
#' 
#' @param fun function to call to sample parameters
#' @param args arguments to pass to the function
#' @param n number of rows to sample
#' @param minMax a data frame with min, max values for each parameter
#' @param msg message template
#' @param settings replication settings
#' @param parameters double array parameters to check for positive definiteness
#' @return tibble with the sampled parameters (1 parameter per column + REPLICATE column)
#' @keywords internal
sampleGeneric <- function(fun, args, n, minMax, msg, settings, parameters=NULL) {
  # First call to method
  tempTable <- do.call(what=fun, args=args %>% append(list(n=n))) %>%
    dplyr::mutate(REPLICATE=seq_len(n)) %>%
    dplyr::mutate(VALID=NA)
  table <- flagSampledParameterRows(table=tempTable, minMax=minMax,
                                       settings=settings, parameters=parameters)
  
  # Re-sample if more parameters are needed due to constraints
  iterations <- 0
  
  while(sum(table$VALID) < n) {
    # Compute success rate
    successRate <- sum(table$VALID) / nrow(table)
    
    # Missing valid items
    missing <- n - sum(table$VALID)
    # print(sprintf("%i valid replicate missing", missing))
    
    # Optimize next value
    nextN <- round(missing/successRate)
    if (nextN < 1) {
      nextN <- 1 # At least 1 sample required
    }
    
    # Maximum number of rows to sample is the maximum chunk size
    # If not specified, maxN will be the number of replicates 'n'
    maxN <- settings@max_chunk_size
    if (is.na(maxN)) {
      maxN <- n
    }
    
    if (is.infinite(nextN) || nextN > maxN) {
      nextN <- maxN # At most 'maxN' samples
    }
    
    # Increment and check
    iterations <- iterations + 1
    if (iterations > settings@max_iterations) {
      stop("Too many iterations, please check your min and max constraints.")
    }
    
    if (isFALSE(settings@quiet)) {
      cat(sprintf("Sampling %i extra row(s)\n", nextN))
    }
    
    shift <- max(table$REPLICATE)
    tempTable <- do.call(what=fun, args=args %>% append(list(n=nextN)))  %>%
      dplyr::mutate(REPLICATE=seq_len(nextN) + shift) %>%
      dplyr::mutate(VALID=NA)
    table <- dplyr::bind_rows(table, flagSampledParameterRows(table=tempTable, minMax=minMax,
                                                                 settings=settings, parameters=parameters))
  }
  
  # Discard extra rows
  lastValidIndex <- which(table$VALID)[n]
  table <- table[1:lastValidIndex,]
  
  # Computing the success rate
  successRate <- sum(table$VALID) / nrow(table)
  
  # Show message logic
  quiet <- settings@quiet
  if (is.na(quiet)) {
    if (successRate < 0.95) {
      showMessage <- TRUE
    } else {
      showMessage <- FALSE
    }
  } else {
    showMessage <- !quiet
  }
  if (showMessage) {
    cat(sprintf(msg, successRate*100), "\n")
  }
  
  # Clean table before returning it
  table <- table %>%
    dplyr::relocate(c("REPLICATE")) %>%
    dplyr::filter(.data$VALID) %>%
    dplyr::select(-c("VALID")) %>%
    dplyr::mutate(REPLICATE=seq_len(n))
  
  return(table)
}

getSamplingMessageTemplate <- function(what, from) {
  return(sprintf("Success rate when sampling %s from %s: %%.1f%%%%", what, from))
}

#' Return a matrix filled in with OMEGA/SIGMA names to be mapped with the values.
#' Nonexistent parameters are filled in with the empty string.
#' 
#' @param parameters subset of parameters
#' @param type type of parameter to map (omega or sigma)
#' @return a matrix with the names of the OMEGA/SIGMA parameters
#' @keywords internal
#' 
getMappingMatrix <- function(parameters, type) {
  size <- parameters@list %>% purrr::keep(~isDiag(.x)) %>% length()
  retValue <- matrix(rep(0, size*size), nrow=size)
  for (i in 1:size) {
    for (j in 1:size) {
      if (type=="omega") {
        refParameter <- Omega(index=i, index2=j)
      } else {
        refParameter <- Sigma(index=i, index2=j)
      }
      parameter <- parameters %>% getByIndex(refParameter)
      if (length(parameter) == 0) {
        retValue[i, j] <- ""
      } else {
        retValue[i, j] <- parameter %>%
          getName()
      }
    }
  }
  return(retValue)
}

#' Flag all parameter rows that have at least one parameter out of the specified range
#' or that have a non positive definite OMEGA or SIGMA matrix.
#' 
#' @param table a data frame returned by \code{sampleMore}
#' @param minMax a data frame with min, max values for each parameter
#' @param settings replication settings
#' @param parameters double array parameters to check for positive definiteness
#' @importFrom dplyr bind_rows filter mutate
#' @importFrom purrr map flatten_int
#' @keywords internal
flagSampledParameterRows <- function(table, minMax, settings, parameters) {
  parameterNames <- colnames(table)
  parameterNames <- parameterNames[!parameterNames %in% c("REPLICATE", "VALID")]
  
  alreadyChecked <- table %>%
    dplyr::filter(!is.na(.data$VALID))
  
  toBeChecked <- table %>%
    dplyr::filter(is.na(.data$VALID)) %>%
    dplyr::mutate(VALID=TRUE) # Default
  
  if (settings@check_min_max) {
    # For each column, check min and max
    # Return the indexes where at least on parameter is out of range
    indexesToDiscard <- parameterNames %>% purrr::map(.f=function(.x) {
      values <- toBeChecked[[.x]]
      limits <- minMax %>% dplyr::filter(.data$name==.x)
      assertthat::assert_that(nrow(limits) == 1)
      return(which(values < limits$min | values > limits$max))
    }) %>% purrr::flatten_int() %>% unique() %>% base::sort()
    
    outOfRange <- rep(FALSE, nrow(toBeChecked))
    outOfRange[indexesToDiscard] <- TRUE
    
    before <- sum(toBeChecked$VALID)
    toBeChecked <- toBeChecked %>% 
      dplyr::mutate(VALID=!outOfRange)
    after <- sum(toBeChecked$VALID)
    difference <- before - after
    if (difference > 0 && isFALSE(settings@quiet)) {
      cat(sprintf("Invalidating %i row(s) that did not succeed the 'min-max' check\n", difference))
    }
  }

  if (settings@check_pos_def && !is.null(parameters)) {
    before <- sum(toBeChecked$VALID)

    toBeChecked <- toBeChecked %>%
      dplyr::left_join(checkMatrixIsPositiveDefinite(table=toBeChecked, parameters=parameters), by="REPLICATE")
    
    toBeChecked <- toBeChecked %>%
      dplyr::mutate(VALID=.data$VALID & .data$POSITIVE_DEFINITE) %>%
      dplyr::select(-c("POSITIVE_DEFINITE"))
    
    after <- sum(toBeChecked$VALID)
    difference <- before - after
    if (difference > 0 && isFALSE(settings@quiet)) {
      cat(sprintf("Invalidating %i row(s) that did not succeed the 'definite positiveness' check\n", difference))
    }
  }
  
  retValue <- dplyr::bind_rows(alreadyChecked, toBeChecked)
  
  return(retValue)
}

#' Check OMEGA/SIGMA matrix for positive definiteness.
#' 
#' @param table data frame with the sampled parameters to check
#' @param parameters double array parameters to check for positive definiteness
#' @importFrom dplyr across group_split select starts_with
#' @importFrom purrr map_df flatten_int
#' @importFrom tibble tibble
#' @keywords internal
checkMatrixIsPositiveDefinite <- function(table, parameters) {
  # Remove THETA_ columns
  table <- table %>%
    dplyr::select(-dplyr::starts_with("THETA_"))
  
  retValue <- table %>%
    dplyr::group_split(dplyr::across("REPLICATE")) %>%
    purrr::map_df(.f=function(row) {
      replicate <- row$REPLICATE
      valid <- row$VALID
      if (isFALSE(valid)) {
        # If the row is not valid, we don't need to check the matrix
        # isFALSE is called because valid could be NA
        return(tibble::tibble(REPLICATE=replicate, POSITIVE_DEFINITE=FALSE))
      }
      row <- row %>%
        dplyr::select(-c("REPLICATE", "VALID"))
      model <- CampsisModel()
      model@parameters <- parameters
      model <- updateParameters(model=model, row=row)
      omegaMatrix <- rxodeMatrix(model=model, type="omega")
      sigmaMatrix <- rxodeMatrix(model=model, type="sigma")
      omegaMatrixOK <- ifelse(length(omegaMatrix) == 0, TRUE, isMatrixPositiveDefinite(omegaMatrix))
      sigmaMatrixOK <- ifelse(length(sigmaMatrix) == 0, TRUE, isMatrixPositiveDefinite(sigmaMatrix))
      return(tibble::tibble(REPLICATE=replicate, POSITIVE_DEFINITE=omegaMatrixOK && sigmaMatrixOK))
    })
  return(retValue)
}

#' Is matrix positive definite. Same check as \code{mvtnorm} does.
#' 
#' @param matrix matrix to check
#' @param tol tolerance when checking the eigenvalues
#' @export
isMatrixPositiveDefinite <- function(matrix, tol=1e-06) {
  eS <- eigen(matrix, symmetric=TRUE)
  ev <- eS$values
  if (!all(ev >= -tol * abs(ev[1L]))) {
    return(FALSE)
  }
  return(TRUE)
}

#_______________________________________________________________________________
#----                               addRSE                                  ----
#_______________________________________________________________________________

#' Add relative standard error (RSE) to the specified parameter.
#' 
#' @param object model or parameters object
#' @param parameter parameter object (Theta, Omega or Sigma)
#' @param value RSE value, in percent
#' @param ... extra arguments, unused
#' @return updated object
#' @export
#' @rdname addRSE
addRSE <- function(object, parameter, value, ...) {
  stop("No default function is provided")
}

setGeneric("addRSE", function(object, parameter, value, ...) {
  standardGeneric("addRSE")
})
#_______________________________________________________________________________
#----                         autoDetectNONMEM                              ----
#_______________________________________________________________________________

#' Auto-detect special variables from NONMEM as compartment properties.
#' Bioavailabilities, infusion durations/rates and lag times will be automatically
#' detected.
#' 
#' @param object object that has NONMEM special variables to be identified
#' @param ... extra arguments, unused
#' @return updated object
#' @export
#' @rdname autoDetectNONMEM
autoDetectNONMEM <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("autoDetectNONMEM", function(object, ...) {
  standardGeneric("autoDetectNONMEM")
})

#_______________________________________________________________________________
#----                              disable                                  ----
#_______________________________________________________________________________

#' Disable.
#' 
#' @param object generic object
#' @param x what needs to be disabled
#' @param ... extra arguments needed for disabling
#' @return object with some disabled features
#' @export
#' @rdname disable
disable <- function(object, x, ...) {
  stop("No default function is provided")
}

setGeneric("disable", function(object, x, ...) {
  standardGeneric("disable")
})

#_______________________________________________________________________________
#----                           export_type                                 ----
#_______________________________________________________________________________

#' Export type class.
#' 
#' @export
setClass(
  "export_type",
  representation(
  )
)

#_______________________________________________________________________________
#----                                export                                 ----
#_______________________________________________________________________________

#' Export function.
#' 
#' @param object generic object
#' @param dest destination
#' @param ... optional arguments
#' @return specific object depending on given destination
#' @export
#' @rdname export
export <- function(object, dest, ...) {
  stop("No default function is provided")
}

setGeneric("export", function(object, dest, ...) {
  standardGeneric("export")
})

#_______________________________________________________________________________
#----                          getCompartmentIndex                          ----
#_______________________________________________________________________________

#' Get the compartment index for the specified compartment name.
#' 
#' @param object generic object that contains compartments information
#' @param name compartment name
#' @return the corresponding compartment index
#' @export
#' @rdname getCompartmentIndex
getCompartmentIndex <- function(object, name) {
  stop("No default function is provided")
}

setGeneric("getCompartmentIndex", function(object, name) {
  standardGeneric("getCompartmentIndex")
})

#_______________________________________________________________________________
#----                           getUncertainty                              ----
#_______________________________________________________________________________

#' Get uncertainty on the parameters.
#' 
#' @param object generic object
#' @param ... extra arguments
#' @return data frame with standard error (se) and relative standard error (rse%) columns 
#' @export
#' @rdname getUncertainty
getUncertainty <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("getUncertainty", function(object, ...) {
  standardGeneric("getUncertainty")
})

#_______________________________________________________________________________
#----                           exportToJSON                                ----
#_______________________________________________________________________________

#' Export the given object to a JSON object, ready to be written to files.
#' 
#' @param object any object
#' @param ... extra arguments, unused
#' @return the loaded S4 object
#' @export
#' @rdname exportToJSON
exportToJSON <- function(object, ...) {
  stop(sprintf("No default function is provided for 'object': %s", class(object)))
}

setGeneric("exportToJSON", function(object, ...) {
  standardGeneric("exportToJSON")
})

#_______________________________________________________________________________
#----                             getVarCov                                 ----
#_______________________________________________________________________________

#' Get variance-covariance matrix.
#' 
#' @param object generic object
#' @return a variance-covariance matrix (data frame) or NULL if no matrix present
#' @export
#' @rdname getVarCov
getVarCov <- function(object) {
  stop("No default function is provided")
}

setGeneric("getVarCov", function(object) {
  standardGeneric("getVarCov")
})

#_______________________________________________________________________________
#----                           loadFromJSON                                ----
#_______________________________________________________________________________

#' Fill-in S4 object from the JSON content.
#' 
#' @param object pre-initiated S4 object
#' @param json JSON (usually a list)
#' @return the loaded S4 object
#' @export
#' @rdname loadFromJSON
loadFromJSON <- function(object, json) {
  stop(sprintf("No default function is provided for 'object': %s, 'json': %s", class(object), class(json)))
}

setGeneric("loadFromJSON", function(object, json) {
  standardGeneric("loadFromJSON")
})

#_______________________________________________________________________________
#----                              move                                     ----
#_______________________________________________________________________________

#' Move element 'x' from object to a certain place.
#' 
#' @param object generic object (e.g. model, code records, etc.)
#' @param x element to move
#' @param to destination (e.g. a position)
#' @param ... extra arguments, unused
#' @return updated object
#' @export
#' @rdname move
move <- function(object, x, to, ...) {
  stop("No default function is provided")
}

setGeneric("move", function(object, x, to, ...) {
  standardGeneric("move")
})

#_______________________________________________________________________________
#----                                 read                                  ----
#_______________________________________________________________________________

#' Generic read method to read data from a file or a folder.
#' 
#' @param file path to the file or folder to be read
#' @param ... extra arguments
#' @return the object representation of the data contained in the file
#' @export
read <- function(file, ...) {
  stop("No default function is provided")
}

setGeneric("read", function(file, ...) {
  standardGeneric("read")
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' Replace all occurrences in object.
#' 
#' @param object generic object (e.g. model, code_record(s), etc.)
#' @param pattern pattern to be replaced
#' @param replacement replacement string
#' @param ... extra arguments
#' @return the same object with all occurrences replaced
#' @export
replaceAll <- function(object, pattern, replacement, ...) {
  stop("No default function is provided")
}

setGeneric("replaceAll", function(object, pattern, replacement, ...) {
  standardGeneric("replaceAll")
})

#_______________________________________________________________________________
#----                           replicate                                   ----
#_______________________________________________________________________________

#' Replicate generic object.
#' 
#' @param object generic object
#' @param n number of replicates required
#' @param settings settings for replication
#' @param ... extra arguments
#' @return object replicated n times
#' @export
#' @rdname replicate
replicate <- function(object, n, settings, ...) {
  stop("No default function is provided")
}

setGeneric("replicate", function(object, n, settings=NULL, ...) {
  n <- as.integer(n)
  if (is.null(settings)) {
    settings <- AutoReplicationSettings()
  }
  standardGeneric("replicate")
})

#_______________________________________________________________________________
#----                                 select                                ----
#_______________________________________________________________________________

#' Get a subset of an object.
#' 
#' @param object generic object
#' @param ... arguments to select
#' @return subset of an object
#' @export
#' @rdname select
select <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("select", function(object, ...) {
  standardGeneric("select")
})

#' @rdname select
setMethod("select", signature=c("data.frame"), definition=function(object, ...) {
  return(return(dplyr::select(.data=object, ...)))
})

#_______________________________________________________________________________
#----                            setMinMax                                  ----
#_______________________________________________________________________________

#' Set the minimum and maximum value on a model parameter.
#' 
#' @param object model or parameters object
#' @param parameter parameter object (Theta, Omega or Sigma)
#' @param min minimum value for this parameter when parameter uncertainty is enabled
#' @param max maximum value for this parameter when parameter uncertainty is enabled
#' @param ... extra arguments, unused
#' @return updated object
#' @export
#' @rdname setMinMax
setMinMax <- function(object, parameter, min, max, ...) {
  stop("No default function is provided")
}

setGeneric("setMinMax", function(object, parameter, min, max, ...) {
  min <- as.numeric(min)
  max <- as.numeric(max)
  standardGeneric("setMinMax")
})

#_______________________________________________________________________________
#----                            standardise                                ----
#_______________________________________________________________________________

#' Standardise.
#' 
#' @param object generic object
#' @param ... extra arguments needed for standardisation
#' @return standardised object
#' @export
#' @rdname standardise
standardise <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("standardise", function(object, ...) {
  standardGeneric("standardise")
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' ToString generic method.
#' 
#' @param object generic object
#' @param ... extra arguments needed for toString conversion
#' @return character value/vector
#' @export
#' @rdname toString
toString <- function(object, ...) {
  stop("No default function is provided")
}

setGeneric("toString", function(object, ...) {
  standardGeneric("toString")
})

#_______________________________________________________________________________
#----                                 write                                 ----
#_______________________________________________________________________________

#' Write generic object to files.
#' 
#' @param object generic object
#' @param file path of the output file or directory
#' @param ... extra arguments
#' @return logical value, TRUE for success, FALSE for failure
#' @export
#' @rdname write
write <- function(object, file, ...) {
  stop("No default function is provided")
}

setGeneric("write", function(object, file, ...) {
  standardGeneric("write")
})

library(testthat)

context("Showing uncertainty on parameters works as expected")

test_that("Showing uncertainty in the console works as expected", {

  # This model has a variance-covariance matrix
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl`
  
  expect_true(any(grepl(pattern="Variance-covariance matrix available", x=capture.output(show(model)))))
  expect_true(any(grepl(pattern="se\\s+rse%", x=capture.output(show(model@parameters %>% select("theta"))))))
  expect_true(any(grepl(pattern="se\\s+rse%", x=capture.output(show(model@parameters %>% select("omega"))))))
  expect_false(any(grepl(pattern="se\\s+rse%", x=capture.output(show(model@parameters %>% select("sigma"))))))
  
  # This model doesn't have a variance-covariance matrix
  model <- model_suite$testing$pk$`2cpt_zo`
  
  expect_true(any(grepl(pattern="No variance-covariance matrix", x=capture.output(show(model)))))
  expect_false(any(grepl(pattern="se\\s+rse%", x=capture.output(show(model)))))
  expect_false(any(grepl(pattern="se\\s+rse%", x=capture.output(show(model@parameters %>% select("theta"))))))
})

test_that("Method getUncertainty works as expected", {
  
  # This model has a variance-covariance matrix
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl`
  
  uncertainty <- model %>% getUncertainty() %>% dplyr::mutate_if(.predicate=is.numeric, .funs=~signif(.x, digits=2))
  expected <- tibble::tibble(name=c("THETA_METAB_CL", "THETA_DUR", "THETA_VC", "THETA_VP", "THETA_Q", "THETA_CL", "THETA_PROP_RUV", "OMEGA_DUR", "OMEGA_VC", "OMEGA_CL", "SIGMA_RUV_FIX"),
                             se=c(0.054, 0.022, 0.032, 0.032, 0.02, 0.036, 0.019, 0.0045, 0.0094, 0.0083, NA),
                             'rse%'=c(17, 5.3, 0.72, 0.6, 1.4, 2.2, 0.96, 33, 19, 19, NA))
  expect_equal(uncertainty, expected)
  
  # This model doesn't have a variance-covariance matrix
  model <- model_suite$testing$pk$`2cpt_zo`
  
  uncertainty <- model %>% getUncertainty()
  expected <- tibble::tibble(name=character(0),
                             se=numeric(0),
                             'rse%'=numeric(0))
  expect_equal(uncertainty, expected)
})

library(testthat)

context("Test various methods of the campsismod package that are not tested elsewhere")
source(paste0("", "testUtils.R"))

# options(campsismod.options=list(SKIP_PERFORMANCE_TESTS=TRUE))

test_that("Methods 'getName', 'getPrefix', 'show' and 'toString' work as expected on initial conditions", {
  model <- model_suite$testing$pk$`1cpt_fo`
  init <- InitialCondition(1, "1000")
  expect_equal(getName(init), "INIT (CMT=1)")
  expect_equal(getPrefix(init), "")
  expect_equal(capture_output(show(init)), "INIT (CMT=1): 1000")
  expect_equal(toString(init, dest="rxode2", model=model), "A_ABS(0)=1000")
  expect_equal(toString(init, dest="mrgsolve", model=model), "A_ABS_0=1000")
  expect_error(toString(init, dest="other", model=model))
})

test_that("Methods 'show' and 'toString' work as expected on compartment properties", {
  model <- model_suite$testing$pk$`1cpt_fo`
  property <- Bioavailability(1, rhs="F1")
  expect_equal(capture_output(show(property)), "BIOAVAILABILITY (CMT=1): F1")
  expect_equal(toString(property, dest="rxode2", model=model), "f(A_ABS)=F1")
  expect_equal(toString(property, dest="mrgsolve", model=model), "F_A_ABS=F1")
  expect_error(toString(property, dest="other", model=model))
})

test_that("Generic methods should throw an error when the call is incorrect", {
  msg <- "No default function is provided"
  expect_error(addRSE(""), regexp=msg)
  expect_error(autoDetectNONMEM(""), regexp=msg)
  expect_error(disable(""), regexp=msg)
  expect_error(export(""), regexp=msg)
  expect_error(getCompartmentIndex(""), regexp=msg)
  expect_error(getUncertainty(""), regexp=msg)
  expect_error(getVarCov(""), regexp=msg)
  expect_error(move(""), regexp=msg)
  expect_error(read(""), regexp=msg)
  expect_error(replaceAll(""), regexp=msg)
  expect_error(replicate("", ""), regexp=msg)
  expect_error(select (""), regexp=msg)
  expect_error(setMinMax("", "", "", ""), regexp=msg) # Strange, I need to pass the 4 args
  expect_error(standardise(""), regexp=msg)
  expect_error(toString (""), regexp=msg)
  expect_error(write(""), regexp=msg)
  
  expect_error(getPrefix(""), regexp=msg)
  expect_error(getRecordName(""), regexp=msg)
})

test_that("Method 'toString' of unknown statements works as expected", {
  statement <- UnknownStatement("HELLO")
  expect_equal(toString(statement, dest="campsis", show=TRUE), "[UNKNOWN STATEMENT] HELLO")
  expect_equal(toString(statement, dest="campsis", show=FALSE), "HELLO")
  expect_error(toString(statement, dest="other"), regexp="Only rxode2 \\(previously RxODE\\), mrgsolve or campsis are supported")
})

test_that("Export function on a replicated Campsis model should be fast", {
  model <- CampsisModel()
  noOfOmegas <- 11
  set.seed(123)
  replicates <- 1000
  
  # Add a huge OMEGA matrix
  for (index1 in 1:noOfOmegas) {
    for (index2 in 1:noOfOmegas) {
      if (index2 > index1) {
        next
      }
      model <- model %>%
        add(Omega(index=index1, index2=index2, value=ifelse(index1==index2, runif(1), 0)))
    }
  }
  
  repModel <- model %>%
    replicate(replicates, settings=AutoReplicationSettings(wishart=TRUE, odf=100, sdf=1000))
  
  noOfColumns <- noOfOmegas * (noOfOmegas - 1) / 2 + noOfOmegas # Lower triangle count + diagonal
  expect_equal(ncol(repModel@replicated_parameters) - 1, noOfColumns)
  
  # Check performances on exporting the Campsis model
  start <- Sys.time()
  models <- seq_len(replicates) %>%
    purrr::map(~repModel %>% export(dest=CampsisModel(), index=.x))
  end <- Sys.time()
  duration <- as.numeric(end - start)
  if (!skipPerformanceTests()) {
    expect_true(duration < 30) # 30 seconds (about 6 seconds on my machine)
  }
  
  
  # Check performances on exporting the OMEGA matrices
  start <- Sys.time()
  matrices <- models %>%
    purrr::map(~rxodeMatrix(.x))
  end <- Sys.time()
  duration <- as.numeric(end - start)
  
  if (!skipPerformanceTests()) {
    expect_true(duration < 15) # (about 3 seconds on my machine)
  }
  
  matrix1 <- matrices[[1]]
  expect_true(isSymmetric(matrix1))
  expect_equal(colnames(matrix1), paste0("ETA_", 1:noOfOmegas))
  
  # Check error message is given is OMEGA is missing
  model1 <- models[[1]] %>%
    delete(Omega(index=1, index2=1))
  expect_error(rxodeMatrix(model1), regexp="At least one OMEGA is missing")
})

# setwd("C:/prj/campsismod/")
# roxygen2::roxygenise()
# setwd("C:/prj/campsismod/tests/")
# testFolder <<- "C:/prj/campsismod/tests/testthat/"

testFolder <- ""
overwriteNonRegressionFiles <- FALSE

toFile <- function(code, path) {
  fileConn <- file(path)
  writeLines(code, fileConn)
  close(fileConn)
}

campsisNonRegPath <- function(regFilename) {
  return(paste0(testFolder, "non_regression/campsis/", regFilename))
}

mrgsolveNonRegPath <- function(regFilename) {
  return(paste0(testFolder, "non_regression/mrgsolve/", regFilename, ".txt"))
}

rxodeNonRegPath <- function(regFilename) {
  return(paste0(testFolder, "non_regression/rxode/", regFilename, ".txt"))
}

campsisNonRegTest <- function(model, regFilename) {
  if (overwriteNonRegressionFiles) {
    model %>% write(file=campsisNonRegPath(regFilename))
  }
  expectedModel <- read.campsis(file=campsisNonRegPath(regFilename))
  expect_equal(model, expectedModel)
}

mrgsolveNonRegTest <- function(mrgmod, regFilename) {
  mrgmodCode <- mrgmod %>% toString()
  if (overwriteNonRegressionFiles) {
    toFile(mrgmodCode, mrgsolveNonRegPath(regFilename))
  }
  expectedMrgmodCode <- readLines(con=mrgsolveNonRegPath(regFilename)) %>% paste0(collapse="\n")
  expect_equal(mrgmodCode, expectedMrgmodCode)
}

rxodeNonRegTest <- function(rxmod, regFilename) {
  rxmodCode <- rxmod@code %>% paste0(collapse="\n")
  if (overwriteNonRegressionFiles) {
    toFile(rxmodCode, rxodeNonRegPath(regFilename))
  }
  expectedRxmodCode <- readLines(con=rxodeNonRegPath(regFilename)) %>% paste0(collapse="\n")
  expect_equal(rxmodCode, expectedRxmodCode)
}

readCampsisModelNoParams <- function(file) {
  # Only first warning is actually checked
  model <- expect_warning(read.campsis(file),
                          regexp="No file '(theta|omega|sigma)\\.csv' could be found")
  return(model)
}

onCran <- function() {
  # Copied from testthat:::on_cran() 
  return(!interactive() && !envVarIsTrue("NOT_CRAN"))
}

envVarIsTrue <- function(x) {
  return(isTRUE(as.logical(Sys.getenv(x, "false"))))
}

skipPerformanceTests <- function() {
  # On CRAN, default value is TRUE
  # FALSE otherwise
  return(getCampsismodOption(name="SKIP_PERFORMANCE_TESTS", default=onCran()))
}

library(testthat)

context("Test the JSON interface")

testFolder <-  file.path(getwd(), test_path())

test_that("Import '1-cpt fo' Campsis model in JSON format", {
  
  # Import the 1-cpt PK model from JSON
  model <- CampsisModel(json=file.path(testFolder, "json_examples", "1cpt_fo_model.json"))
  expect_equal(model_suite$pk$`1cpt_fo`, model)
  
  # Export to JSON and re-import
  json <- model %>%
    exportToJSON()
  model <- CampsisModel(json=json@data)
  expect_equal(model_suite$pk$`1cpt_fo`, model)
  
  # Import an empty Campsis model from JSON
  model <- CampsisModel(json="{\"code\":[], \"parameters\":[]}")
  expect_equal(CampsisModel(), model)
  
  # Import the 1-cpt PK model with a correlation between CL and VC from JSON
  model <- loadFromJSON(CampsisModel(),
                        file.path(testFolder, "json_examples", "1cpt_fo_model_cl_vc_cor.json"))
  expectedModel <- model_suite$pk$`1cpt_fo` %>%
    add(Omega(name="CL_VC", index=3, index2=2, value=0.75, type="cor")) %>%
    campsismod::sort()
  expect_equal(expectedModel, model)
})


test_that("Export/Re-import '2cpt_zo_allo_metab_effect_on_cl' Campsis model to/from JSON", {
  model <- model_suite$testing$other$`2cpt_zo_allo_metab_effect_on_cl`
  
  tmp <- tempfile(fileext=".json") # gsub("\\\\", "/", normalizePath(tmp))
  model %>%
    exportToJSON() %>%
    write(file=tmp)
  
  model2 <- CampsisModel(json=paste0(readLines(tmp), collapse="\n"))
  expect_equal(model, model2)
})

test_that("Export/Re-import basic test model assembled on the fly to/from JSON", {
  model <- model_suite$pk$`1cpt_fo` %>%
    add(Omega(name="CL_VC", index=3, index2=2, value=0.5, type="cor")) %>% # Add correlation between CL and VC
    standardise() %>%
    addRSE(Theta(name="VC"), value=10) %>% # 10% RSE on VC
    addRSE(Theta(name="CL"), value=10) %>% # 10% RSE on CL
    addRSE(Omega(name="VC"), value=50) %>% # 50% RSE on OMEGA_VC
    addRSE(Omega(name="CL"), value=50) %>% # 50% RSE on OMEGA_CL
    addRSE(Omega(name="CL_VC"), value=100) %>% # 100% RSE on correlation
    addRSE(Sigma(name="PROP_RUV"), value=5) %>% # 5% RSE on proportional error
    campsismod::sort() # Correlation will be put at the right place in the list
  
  tmp <- tempfile(fileext=".json") # gsub("\\\\", "/", normalizePath(tmp))
  json <- model %>%
    exportToJSON()
  model %>%
    write(file=tmp)
  jsonVarcov <- json@data$varcov
  expect_equal(length(jsonVarcov), 6) # We expect exactly 6 values (since zeroes are omitted)
  
  model2 <- read.campsis(tmp)
  expect_equal(model, model2)
  
  # Test error catching
  tmp <- tempfile(fileext=".json") # gsub("\\\\", "/", normalizePath(tmp))
  json@data$varcov[[1]]$ref1$name <- "UNKNOWN"
  json %>%
    write(file=tmp)
  expect_error(read.campsis(tmp), regexp="Parameter reference not found")
})

#'
#' A empty method only used to declare some of the global function definitions
#' contained in the package 'methods', which are used intensively.
#'
#' @return no return value
#' @importFrom methods callNextMethod is new show slot slot<- slotNames validObject
#' @keywords internal
methodsGlobalFunctions <- function() {
  
}

#' @importFrom rlang .data :=
#' @keywords internal
rlangGlobal <- function() {
  
}

#' Get code for rxode2
#' 
#' @param model Campsis model
#' @return corresponding model code for rxode2
#' @export
rxodeCode <- function(model) {
  records <- model@model
  properties <- model@compartments@properties
  propertiesCode <- NULL
  if (properties %>% length() > 0) {
    for (property in properties@list) {
      compartmentIndex <- property@compartment
      compartment <- model@compartments %>% find(Compartment(index=compartmentIndex))
      equation <- property %>% toString(model=model, dest="rxode2")
      propertiesCode <- propertiesCode %>% append(equation)
    }
  }
  
  # All records to character vector
  code <- NULL
  for (record in records@list) {
    for (statement in record@statements@list) {
      code <- code %>% append(statement %>% toString(dest="rxode2"))
    }
    if (is(record, "ode_record")) {
      code <- code %>% append(propertiesCode)
    }
  }
  return(code)
}

#' Get the parameters vector for rxode2.
#' 
#' @param model Campsis model
#' @return named vector with THETA values
#' @export
rxodeParams <- function(model) {
  type <- "theta"
  params <- model@parameters
  if (params %>% length() == 0) {
    retValue <- numeric(0)
    names(retValue) <- character(0)
    return(retValue) # Must be named numeric, otherwise rxode2 complains
  }
  maxIndex <- params %>% select("theta") %>% maxIndex()
  
  # Careful, as.numeric(NA) is important...
  # If values are all integers, rxode2 gives a strange error message:
  # Error in rxSolveSEXP(object, .ctl, .nms, .xtra, params, events, inits,  : 
  # when specifying 'thetaMat', 'omega', or 'sigma' the parameters cannot be a 'data.frame'/'matrix'
  
  retValue <- rep(as.numeric(NA), maxIndex)
  names <- rep("", maxIndex)
  
  for (i in seq_len(maxIndex)) {
    param <- params %>% getByIndex(Theta(index=i))
    if (length(param) == 0) {
      stop(paste0("Missing param ", i, "in ", type, " vector"))
    } else {
      retValue[i] <- param@value
      names[i] <- param %>% getNameInModel()
    }
  }
  names(retValue) <- names
  
  return(retValue)
}

#' Get the OMEGA/SIGMA matrix for rxode2.
#' 
#' @param model Campsis model or Campsis parameters
#' @param type either omega or sigma
#' @return omega/sigma named matrix
#' @export
rxodeMatrix <- function(model, type="omega") {
  
  if (is(model, "campsis_model")) {
    subset <- model@parameters %>%
      select(type)
  } else if (is(model, "parameters")) {
    subset <- model %>%
      select(type)
  } else {
    stop("model must be either a Campsis model or a parameters object")
  }

  if (subset %>% length()==0) {
    return(matrix(data=numeric(0), nrow=0, ncol=0))
  }
  
  # Standardise parameters
  subset <- subset %>%
    standardise()
  
  # Retrieve max index
  maxIndex <- subset %>%
    maxIndex()
  matrix <- matrix(0L, nrow=maxIndex, ncol=maxIndex)
  names <- rep("", maxIndex)
  
  # Fill in matrix
  for (elem in subset@list) {
    matrix[elem@index, elem@index2] <- elem@value
    matrix[elem@index2, elem@index] <- elem@value
    if (elem@index==elem@index2) {
      names[elem@index] <- elem %>% getNameInModel()
    }
  }
  
  assertthat::assert_that(all(names != ""),
                          msg=sprintf("At least one %s is missing.", toupper(type)))

  rownames(matrix) <- names
  colnames(matrix) <- names
  return(matrix)
}

#_______________________________________________________________________________
#----                       model_statement class                           ----
#_______________________________________________________________________________

#' 
#' Model statement class. Any statement in a code record.
#' 
#' @slot comment a comment associated to this model statement
#' @export
setClass(
  "model_statement",
  representation(
    comment = "character"
  ),
  contains = "pmx_element",
  prototype = prototype(comment=as.character(NA)),
  validity = function(object) {
    return(expectOne(object, c("comment")))
  }
)

#' 
#' Append comment.
#' 
#' @param str single character string
#' @param object model statement
#' @param dest destination engine, string
#' @return single character string followed by comment
#' @keywords internal
appendComment <- function(str, object, dest) {
  comment <- object@comment
  if (is.na(comment)) {
    return(str)
  } else {
    if (dest=="mrgsolve") {
      symbol <- "//"
    } else if (dest=="NONMEM") {
      symbol <- ";"
    } else {
      symbol <- "#" # Both campsis and rxode2
    }
    if (str=="") {
      return(paste0(symbol, " ", comment))
    } else {
      return(paste0(str, " ", symbol, " ", comment))
    }
  }
}

#' 
#' Unknown destination engine exception.
#'
#' @keywords internal
#' @return no return value
#' 
UnsupportedDestException <- function() {
  stop("Only rxode2 (previously RxODE), mrgsolve or campsis are supported")
}

#_______________________________________________________________________________
#----                                  show                                 ----
#_______________________________________________________________________________

setMethod("show", signature=c("model_statement"), definition=function(object) {
  cat(object %>% toString(show=TRUE))
})

#_______________________________________________________________________________
#----                             replaceAll                                ----
#_______________________________________________________________________________

#' @rdname replaceAll
setMethod("replaceAll", signature=c("model_statement", "pattern", "character"), definition=function(object, pattern, replacement, ...) {
  return(object)
})

#_______________________________________________________________________________
#----               compartment_initial_condition class                     ----
#_______________________________________________________________________________

#' 
#' Compartment initial condition class.
#' 
#' @export
setClass(
  "compartment_initial_condition",
  representation(
  ),
  contains = "compartment_property",
  validity=function(object) {
    return(TRUE)
  }
)

#'
#' Create an initial condition.
#'
#' @param compartment compartment index
#' @param rhs right-hand side part of the equation
#' @return an initial condition property
#' @export
InitialCondition <- function(compartment, rhs="") {
  return(new("compartment_initial_condition", compartment=as.integer(compartment), rhs=rhs))
}

#_______________________________________________________________________________
#----                            getName                                    ----
#_______________________________________________________________________________

#' @rdname getName
setMethod("getName", signature=c("compartment_initial_condition"), definition=function(x) {
  return(paste0("INIT (", "CMT=", x@compartment, ")"))
})

#_______________________________________________________________________________
#----                             getPrefix                                ----
#_______________________________________________________________________________

#' @rdname getPrefix
setMethod("getPrefix", signature = c("compartment_initial_condition"), definition = function(object, ...) {
  return("")
})

#_______________________________________________________________________________
#----                           getRecordName                               ----
#_______________________________________________________________________________

#' @rdname getRecordName
setMethod("getRecordName", signature = c("compartment_initial_condition"), definition = function(object) {
  return("INIT")
})

#_______________________________________________________________________________
#----                               show                                    ----
#_______________________________________________________________________________

setMethod("show", signature=c("compartment_initial_condition"), definition=function(object) {
  cat(paste0(object %>% getName(), ": ", object@rhs))
})

#_______________________________________________________________________________
#----                             toString                                  ----
#_______________________________________________________________________________

#' @rdname toString
setMethod("toString", signature=c("compartment_initial_condition"), definition=function(object, ...) {
  model <- processExtraArg(args=list(...), name="model", mandatory=TRUE)
  dest <- processExtraArg(args=list(...), name="dest", mandatory=TRUE)
  
  compartmentIndex <- object@compartment
  compartment <- model@compartments %>% find(Compartment(index=compartmentIndex))
  
  if (isRxODE(dest)) {
    return(paste0(compartment %>% toString(), "(0)=", object@rhs))
  } else if (dest=="mrgsolve") {
    return(paste0(compartment %>% toString(), "_0=", object@rhs))
  } else {
    stop("Only rxode2 (previously RxODE) and mrgsolve are currently supported")
  }
})

---
title: "Create a minimalist model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Create a minimalist model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

### Prerequisite

For this exercise, we'll need the `campsismod` package. This package can be loaded as follows:

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
```

### Create a minimalist model in the Notepad++ editor

Assume a very simple 1-compartment PK model with first-order eliminate rate `K`. Say this parameter has a typical value of log(2)/12≈0.06 (where 12 is the elimination half life) and has 15% CV. Let's also initiate the central compartment to 1000.

This can be translated into the following CAMPSIS model (
```{r echo=FALSE}
xfun::embed_file('resources/campsis_npp_plugin.xml', text="download Notepad++ plugin for CAMPSIS")
```
):

![Minimalist model edited in Notepad++](resources/notepad_plugin_example.png)
             
Let's now create our `theta.csv` with our single parameter `K` as follows:

![Edition of THETA's in Excel](resources/minimal_model_theta.png)
           
And finally, let's also create our `omega.csv` to include inter-individual variability on `K`:

![Edition of OMEGA's in Excel](resources/minimal_model_omega.png)
          
This model can now be loaded by `campsismod`...

```{r, eval=TRUE}
model <- read.campsis("resources/minimalist_model/")
```

Let's simulated this model in CAMPSIS:

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(25) %>% add(Observations(seq(0,24,by=0.5)))
results <- model %>% simulate(dataset=dataset, seed=1)
spaghettiPlot(results, "A_CENTRAL")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/minimalist_example_sim1.png", width=7, height=3, dpi=100)
```

![](resources/minimalist_example_sim1.png)

### Create the same model programmatically

The same model can be created programmatically. First, let's create an empty CAMPSIS model.

```{r}
model <- CampsisModel()
```

Then, let's define the equation of our model parameter `K`.

```{r}
model <- model %>% add(Equation("K", "THETA_K*exp(ETA_K)"))
```

We can add an ordinary differential equation as follows:

```{r}
model <- model %>% add(Ode("A_CENTRAL", "-K*A_CENTRAL"))
```

We can init the central compartment as well on the fly:

```{r}
model <- model %>% add(InitialCondition(compartment=1, "1000"))
```

Finally, let's define our `THETA_K` and `ETA_K`:

```{r}
model <- model %>% add(Theta("K", value=0.06))
model <- model %>% add(Omega("K", value=15, type="cv%"))
```

This model can simulated by CAMPSIS as well. Powerful, isn't it?

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(25) %>% add(Observations(seq(0,24,by=0.5)))
results <- model %>% simulate(dataset=dataset, seed=2)
spaghettiPlot(results, "A_CENTRAL")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/minimalist_example_sim2.png", width=7, height=3, dpi=100)
```

![](resources/minimalist_example_sim2.png)

---
title: "Insight on the structural model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Insight on the structural model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

This vignette attends to give a brief insight on the different statements that may be used to build a structural model, compatible with Campsis.  
The number of model statement types proposed in `campsismod` is rather limited, with the underlying idea of keeping our model simple and the translation to rxode2 and mrgsolve as clean as possible.
However, in a near future, new types of model statements will likely be supported.

### Prerequisite

The examples below require the package `campismod`.

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
```

### Equation

Equations are described by 3 fields:

- lhs: the left-hand side string, which corresponds to the variable name
- rhs: the right-hand side string, which corresponds to the formula
- comment: a possible comment

Example:

```{r}
equation <- Equation("KA", "THETA_KA * exp(ETA_KA)", comment="This is my KA parameter")
```

The equivalent text form is:

```{r}
equation
```

### Ordinary differential equation (ODE)

Similarly, Ordinary differential equations (ODE's) are described by 3 fields as well:

- lhs: the left-hand side string, which corresponds to the variable name, must start with 'A_'
- rhs: the right-hand side string, which corresponds to the derivative formula
- comment: a possible comment

Example:

```{r}
ode <- Ode("A_DEPOT", "-KA * A_DEPOT", comment="This is my depot compartment")
```

The equivalent text form is:

```{r}
ode
```

### Line break

Line breaks can be added to a model to add clear separations between blocks of equations. It does not have any field.

Example:

```{r}
linebreak <- LineBreak()
```

The equivalent text form is obviously a separation line.

### Comment

Comments can be specified at any place in the model. They have a unique field:

- x: the comment you'd like to add

Example:

```{r}
comment <- Comment("This is my first comment")
```

The equivalent text form is:

```{r}
comment
```

### If-statement

If-statements allow a variable to take different values according to different specified conditions.

- condition: the given condition
- equation: an equation (see definition above)
- comment: a possible comment

Example:

```{r}
ifStatement <- IfStatement("COV==1", Equation("COV_EFFECT", "0.2"), comment="This is an if statement")
```

The equivalent text form is:

```{r}
ifStatement
```

A common use of the if-statements is to add covariate effects into the model. Here is an example:

```{r}
main <- MainRecord()
main <- main %>%
  add(Equation("COV_EFFECT", "0")) %>% # Initialisation
  add(IfStatement("COV==1", Equation("COV_EFFECT", "0.1"))) %>%  # Covariate value is 1 in dataset
  add(IfStatement("COV==2", Equation("COV_EFFECT", "0.2"))) %>%  # Covariate value is 2 in dataset
  add(IfStatement("COV==3", Equation("COV_EFFECT", "0.3")))      # Covariate value is 3 in dataset
```

The equivalent text would then be:

```{r}
main
```

Please note that mrgsolve require all extra variables (like COV_EFFECT in the previous example) to be initialised to a predefined value, which makes a lot of sense in general.

### Additional remarks

Package `campsismod` does not check the equations from a grammar point of view. This work is delegated to the simulation engine (rxode2 or mrgsolve) and C compiler.

In general, respecting the few rules listed below will give you a successful compatibility with both engines:

- All formula's and if-statement conditions must be written in C code
- Functions included in the Math library of C are compatible (e.g. cos, sin, etc)
- Use the pow function for writing exponents (^ is accepted in rxode2, not in mrgsolve)
- Scientific notation for numbers is accepted (e.g. 2.51e-01)
- Initialise all extra variables (see example in the section above)

---
title: "Insight on the structural model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Insight on the structural model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

This vignette attends to give a brief insight on the different statements that may be used to build a structural model, compatible with CAMPSIS.  
The number of model statement types proposed in `campsismod` is rather limited, with the underlying idea of keeping our model simple and the translation to RxODE and mrgsolve as clean as possible.
However, in a near future, new types of model statements will likely be supported.

### Prerequisite

The examples below require the package `campismod`.

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
```

### Equation

Equations are described by 3 fields:

- lhs: the left-hand side string, which corresponds to the variable name
- rhs: the right-hand side string, which corresponds to the formula
- comment: a possible comment

Example:

```{r}
equation <- Equation("KA", "THETA_KA * exp(ETA_KA)", comment="This is my KA parameter")
```

The equivalent text form is:

```{r}
equation
```

### Ordinary differential equation (ODE)

Similarly, Ordinary differential equations (ODE's) are described by 3 fields as well:

- lhs: the left-hand side string, which corresponds to the variable name, must start with 'A_'
- rhs: the right-hand side string, which corresponds to the derivative formula
- comment: a possible comment

Example:

```{r}
ode <- Ode("A_DEPOT", "-KA * A_DEPOT", comment="This is my depot compartment")
```

The equivalent text form is:

```{r}
ode
```

### Line break

Line breaks can be added to a model to add clear separations between blocks of equations. It does not have any field.

Example:

```{r}
linebreak <- LineBreak()
```

The equivalent text form is obviously a separation line.

### Comment

Comments can be specified at any place in the model. They have a unique field:

- x: the comment you'd like to add

Example:

```{r}
comment <- Comment("This is my first comment")
```

The equivalent text form is:

```{r}
comment
```

### If-statement

If-statements allow a variable to take different values according to different specified conditions.

- condition: the given condition
- equation: an equation (see definition above)
- comment: a possible comment

Example:

```{r}
ifStatement <- IfStatement("COV==1", Equation("COV_EFFECT", "0.2"), comment="This is an if statement")
```

The equivalent text form is:

```{r}
ifStatement
```

A common use of the if-statements is to add covariate effects into the model. Here is an example:

```{r}
main <- MainRecord()
main <- main %>%
  add(Equation("COV_EFFECT", "0")) %>% # Initialisation
  add(IfStatement("COV==1", Equation("COV_EFFECT", "0.1"))) %>%  # Covariate value is 1 in dataset
  add(IfStatement("COV==2", Equation("COV_EFFECT", "0.2"))) %>%  # Covariate value is 2 in dataset
  add(IfStatement("COV==3", Equation("COV_EFFECT", "0.3")))      # Covariate value is 3 in dataset
```

The equivalent text would then be:

```{r}
main
```

Please note that mrgsolve require all extra variables (like COV_EFFECT in the previous example) to be initialised to a predefined value, which makes a lot of sense in general.

### Additional remarks

Package `campsismod` does not check the equations from a grammar point of view. This work is delegated to the simulation engine (RxODE or mrgsolve) and C compiler.

In general, respecting the few rules listed below will give you a successful compatibility with both engines:

- All formula's and if-statement conditions must be written in C code
- Functions included in the Math library of C are compatible (e.g. cos, sin, etc)
- Use the pow function for writing exponents (^ is accepted in RxODE, not in mrgsolve)
- Scientific notation for numbers is accepted (e.g. 2.51e-01)
- Initialise all extra variables (see example in the section above)

---
title: "Adapt your model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Adapt your model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

### Prerequisite

For this vignette, please load the `campsismod` package and load the minimalist model that we have created in the previous vignette.

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
model <- read.campsis("resources/minimalist_model/")
```

### Define the concentration

A concentration is defined by the amount of drug present in the central compartment divided by the apparent volume of distribution.  
Let's now define the volume as a fixed constant of `100` in the model. This can be achieved as follows:

```{r}
model <- model %>% add(Equation("V", "100"))
model
```

By default, this new equation has been appended to the model parameters, in the `MAIN` code record.  

Now, we would like to define the plasma concentration. This can be done by adding an equation into the `ODE` code record. To do that, we'll use the argument `pos` to define the position of this equation. It can be inserted, for instance, right after the ordinary differential equation:

```{r}
model <- model %>% add(Equation("CP", "A_CENTRAL/V"), pos=Position(Ode("A_CENTRAL")))
model
```

### Define an error model

Say we want to add a proportional error model on the concentration with a coefficient of variation of 20%.  
We therefore need to add a new parameter SIGMA into the model:

```{r}
model <- model %>% add(Sigma("PROP", value=20, type="cv%"))
```

Then, we need to add a new `ERROR` code record with the appropriate equations:

```{r}
error <- ErrorRecord()
error <- error %>% add(Equation("OBS_CP", "CP*(1 + EPS_PROP)"))
model <- model %>% add(error)
model
```

### Simulate the observed concentration

Let's now simulate a few individuals and show `OBS_CP`.

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(3) %>% add(Observations(seq(0,24,by=3)))
results <- model %>% simulate(dataset=dataset, seed=0)
spaghettiPlot(results, "OBS_CP")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/minimalist_example_obs_cp.png", width=7, height=3, dpi=100)
```

![](resources/minimalist_example_obs_cp.png)

### A couple of useful functions in action

We can check the existence of an equation (or any other type of model statement), by calling the function `contains`.

```{r}
model %>% contains(Equation("CP"))

model %>% contains(Ode("A_CENTRAL"))
```

In the same way, we can retrieve any model statement easily using the function `find`:

```{r}
model %>% find(Equation("CP"))

model %>% find(Ode("A_CENTRAL"))
```

For instance, right-hand side formula of equation `CP` can be retrieved as follows:

```{r}
(model %>% find(Equation("CP")))@rhs
```

Any model statement may be replaced using the function `replace`:

```{r}
model %>% replace(Equation("V", "50")) # Previous value of 100 is overridden
```

Finally, model statements can also be deleted forever (making the model broken in the following case):

```{r}
model %>% delete(Equation("V"))
```

---
title: "Adapt your parameters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Adapt your parameters}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

### Prerequisite

The examples below require the package `campismod`.

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
```

### Retrieve your parameters

The examples below are illustrated based on the reference 2-compartment PK model that you find in the built-in library. This model has 5 parameters. All these parameters have inter-individual variability defined.

```{r}
model <- model_suite$pk$`2cpt_fo`
model
```

To retrieve a parameter by its name, just call the method `find` as follows:

```{r}
model %>% find(Theta("CL"))
model %>% find(Omega("KA"))
model %>% find(Sigma("RUV_FIX"))
```

These parameters can also alternatively be retrieved by their index(es). Use the specific method `getByIndex` created for that purpose:

```{r}
model@parameters %>% getByIndex(Theta(index=5))
model@parameters %>% getByIndex(Omega(index=1, index2=1))
model@parameters %>% getByIndex(Sigma(index=1, index2=1))
```

### Access parameter values

Accessing parameter values is straightforward. Parameters have a slot `value` that may be accessed.

```{r}
thetaCL <- model %>% find(Theta("CL"))
thetaCL@value
```

For OMEGA and SIGMA parameters, be careful; the interpretation of this value depends on the `type` of the parameter. It may be `var` (for a variance), `covar` (for a covariance), `sd` (for standard deviation), `cv` (value expressed as coefficient of variation), `cv%` (value expressed as coefficient of variation in percentage).  
For a quick access to the value as variance or covariance, the method `standardise` can be called first on the parameter itself. This is especially useful for values expressed in CV or in standard deviation.

```{r}
theta <- Omega(name="TEST", index=1, index2=1, value=15, type="cv%")
theta@value # 15 is returned

theta_standardised <- theta %>% standardise() # Conversion to variance
theta_standardised 
theta_standardised@value
```

### Replace parameters

Parameters can be replaced easily. Here are a few examples:

```{r}
model <- model %>% replace(Theta("KA", value=2)) # Previous value for KA was 1
model <- model %>% replace(Omega("CL", value=20, type="cv%")) # Previous value was a 25% CV
model
```

### Delete parameters

Parameters can be deleted. Please note that it does not do anything to the equations. Also, the indexes won't be re-adjusted. Here are a few examples:

```{r}
model <- model %>% delete(Theta("KA"))
model <- model %>% delete(Omega("CL"))
model
```

As expected, this model will not be valid anymore:
```{r}
tryCatch({validObject(model, complete=TRUE)}, error=function(msg) {
  print(msg)
})
```

---
title: "Get started with CAMPSIS model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Get started with CAMPSIS model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

### Load example from model library

First import the `campsismod` package. This step is not required if you have already loaded the `campsis` package.

```{r, message=FALSE}
library(campsismod)
```

Load 2-compartment PK model from built-in model library and show content.

```{r}
model <- model_suite$pk$`2cpt_fo`
show(model)
```

### Write CAMPSIS model

A CAMPSIS model can be persisted on your local drive as follows:

```{r}
model %>% write(file="path_to_model_folder")
list.files("path_to_model_folder")
```

As shown, the output directory will contain the model (all code records) and 1 `csv` file per type of parameter (THETA, OMEGA and SIGMA). 

### Read CAMPSIS model

To read a CAMPSIS model from your local drive, use the `read.campsis` function.
The exact same model should be retrieved.

```{r}
model <- read.campsis(file="path_to_model_folder")
show(model)
```

The `MAIN` record is the part of the model where your model parameters are defined. The `ODE` record is where your ordinary differential equations (ODE) go, as well as any equation depending on the simulation time. The `ERROR` record is the place where the error model is defined. The model parameters are then shown, followed by the all the compartments.
 
### Export CAMPSIS model to RxODE

`campsismod` has powerful export capabilities to `RxODE` and `mrgsolve`, the 2 simulation engines supported by `campsis`.
The following code exports the model to `RxODE`. Please note that this step is implicit in CAMPSIS when you call the `simulate` method with your preferred simulation engine.

```{r}
rxmod <- model %>% export(dest="RxODE")
rxmod
```

### Export CAMPSIS model to mrgsolve

The following code exports the model to `mrgsolve` (text form).

```{r}
mrgmod <- model %>% export(dest="mrgsolve")
mrgmod
```

---
title: "Play with compartment properties"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Play with compartment properties}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

### Prerequisite

For this vignette, please load the `campsismod` package and load the minimalist model that we have created in the first vignette.

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
model <- read.campsis("resources/minimalist_model/")
```

### Create new compartment properties

Let's invent a very basic scenario: we would like to infuse `1000` into the central compartment with a fixed rate of `100` and a fixed lag time of `2`.    
First, we're going to delete the initial condition that we had in the minimalist model. This is done as follows:

```{r}
model_ <- model %>% delete(InitialCondition(compartment=1))
model_
```

This is strictly equal as doing (if you prefer working with compartment names):

```{r}
model <- model %>% delete(InitialCondition(compartment= model %>% getCompartmentIndex("CENTRAL")))
```

We can now add a fixed rate for all infusions that go into the central compartment:

```{r}
model <- model %>% add(InfusionRate(compartment=1, "100"))
```

Finally, let's now add a constant lag time:

```{r}
model <- model %>% add(LagTime(compartment=1, "2"))
```

OK, this is how our model looks like now:

```{r}
model
```

### Simulate our model

Let's now simulate a few individuals and show `A_CENTRAL`, i.e., the amount of drug in the central compartment.

First, we need to define an infusion of `1000` in a Campsis dataset, as well as the observations times.

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(5) %>% 
  add(Infusion(time=0, amount=1000)) %>%
  add(Observations(seq(0,36,by=0.5)))
```

Then, we can run the simulation.

```{r, eval=EXPORT_PNG}
results <- model %>% simulate(dataset=dataset, seed=1)
spaghettiPlot(results, "A_CENTRAL")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/minimalist_example_cmt_properties.png", width=7, height=3, dpi=100)
```

![](resources/minimalist_example_cmt_properties.png)

### A couple of useful functions in action

As previously, let's demonstrate the use of a couple of interesting functions:  

Check the existence of a compartment:

```{r}
model %>% contains(Compartment(1)) 
# Or equivalenty:
model %>% contains(Compartment(model %>% getCompartmentIndex("CENTRAL")))
```

Check the existence of a property:

```{r}
model %>% contains(InfusionRate(1))
model %>% contains(InfusionDuration(1)) 
```

Find a compartment:

```{r}
model %>% find(Compartment(1)) 
```

Find a compartment property:

```{r}
model %>% find(InfusionRate(1)) 
```

Replace a compartment property:

```{r}
model %>% replace(InfusionRate(1, "200")) # Previous value of 100 is overridden
```

Interestingly, the name of a compartment can be replaced as follows:

```{r}
model %>% replace(Compartment(1, name="CENT")) %>%
  delete(Ode("A_CENTRAL")) %>%
  add(Ode("A_CENT", "-K*A_CENT"))
```

---
title: "Append a model to another"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Append a model to another}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

This vignette shows how a model can be appended to another. This is particularly useful when appending a PD model to a existing PK model. In this vignette, we'll demonstrate how an effect compartment model can be appended to a 2-compartment model.

### Prerequisite

The examples below require the package `campismod`.

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
```

### Load your base PK model 

The following code will load our reference 2-compartment PK model.

```{r}
pk_model <- model_suite$pk$`2cpt_fo`
```

### Load an effect-compartment model

The effect-compartment model can be loaded from the model library as follows:

```{r echo=TRUE, warning=FALSE, message=FALSE}
pd_model <- model_suite$pd$effect_cmt_model
pd_model
```

This PD model has a variable `PK_CONC`, that needs to be linked with the PK concentration.  
Therefore, we need to adapt it as follows:

```{r}
pd_model <- pd_model %>% replace(Equation("PK_CONC", "A_CENTRAL/VC"))
```

### Append PD model to PK model

Appending the PD model to the PK model is done using the `add` function:

```{r}
pkpd_model <- pk_model %>% add(pd_model)
pkpd_model
```

### Simulate our PK/PD model

Let's now simulate our PK/PD model:

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(25) %>% 
  add(Bolus(time=0, amount=1000, compartment=1, ii=12, addl=2)) %>%
  add(Observations(times=0:36))
results <- pkpd_model %>% simulate(dataset=dataset, seed=1)
shadedPlot(results, "CONC")
```

```{r, eval=EXPORT_PNG}
shadedPlot(results, "CONC")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/pkpd_model_concentration.png", width=7, height=3, dpi=100)
```

![PK concentration](resources/pkpd_model_concentration.png)

```{r, eval=EXPORT_PNG}
shadedPlot(results, "EFFECT")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/pkpd_model_effect.png", width=7, height=3, dpi=100)
```

![Showing the delayed effect](resources/pkpd_model_effect.png)

---
title: "Get started with Campsis model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Get started with Campsis model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

### Load example from model library

First import the `campsismod` package. This step is not required if you have already loaded the `campsis` package.

```{r, message=FALSE}
library(campsismod)
```

Load 2-compartment PK model from built-in model library and show content.

```{r}
model <- model_suite$pk$`2cpt_fo`
show(model)
```

### Write Campsis model

A Campsis model can be persisted on your local drive as follows:

```{r}
model %>% write(file="path_to_model_folder")
list.files("path_to_model_folder")
```

As shown, the output directory will contain the model (all code records) and 1 `csv` file per type of parameter (THETA, OMEGA and SIGMA). 

### Read Campsis model

To read a Campsis model from your local drive, use the `read.campsis` function.
The exact same model should be retrieved.

```{r}
model <- read.campsis(file="path_to_model_folder")
show(model)
```

The `MAIN` record is the part of the model where your model parameters are defined. The `ODE` record is where your ordinary differential equations (ODE) go, as well as any equation depending on the simulation time. The `ERROR` record is the place where the error model is defined. The model parameters are then shown, followed by the all the compartments.
 
### Export Campsis model to rxode2

`campsismod` has powerful export capabilities to `rxode2` and `mrgsolve`, the 2 simulation engines supported by `campsis`.
The following code exports the model to `rxode2`. Please note that this step is implicit in Campsis when you call the `simulate` method with your preferred simulation engine.

```{r}
rxmod <- model %>% export(dest="rxode2")
rxmod
```

### Export Campsis model to mrgsolve

The following code exports the model to `mrgsolve` (text form).

```{r}
mrgmod <- model %>% export(dest="mrgsolve")
mrgmod
```

---
title: "Create a minimalist model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Create a minimalist model}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, echo=FALSE, warning=FALSE, message=FALSE, results='hide'}
EXPORT_PNG <- FALSE
```

### Prerequisite

For this exercise, we'll need the `campsismod` package. This package can be loaded as follows:

```{r echo=TRUE, warning=FALSE, message=FALSE}
library(campsismod)
```

### Create a minimalist model in the Notepad++ editor

Assume a very simple 1-compartment PK model with first-order eliminate rate `K`. Say this parameter has a typical value of log(2)/12≈0.06 (where 12 is the elimination half life) and has 15% CV. Let's also initiate the central compartment to 1000.

This can be translated into the following Campsis model (
```{r echo=FALSE}
xfun::embed_file('resources/campsis_npp_plugin.xml', text="download Notepad++ plugin for Campsis")
```
):

![Minimalist model edited in Notepad++](resources/notepad_plugin_example.png)
             
Let's now create our `theta.csv` with our single parameter `K` as follows:

![Edition of THETA's in Excel](resources/minimal_model_theta.png)
           
And finally, let's also create our `omega.csv` to include inter-individual variability on `K`:

![Edition of OMEGA's in Excel](resources/minimal_model_omega.png)
          
This model can now be loaded by `campsismod`...

```{r, eval=TRUE}
model <- read.campsis("resources/minimalist_model/")
```

Let's simulated this model in Campsis:

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(25) %>% add(Observations(seq(0,24,by=0.5)))
results <- model %>% simulate(dataset=dataset, seed=1)
spaghettiPlot(results, "A_CENTRAL")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/minimalist_example_sim1.png", width=7, height=3, dpi=100)
```

![](resources/minimalist_example_sim1.png)

### Create the same model programmatically

The same model can be created programmatically. First, let's create an empty Campsis model.

```{r}
model <- CampsisModel()
```

Then, let's define the equation of our model parameter `K`.

```{r}
model <- model %>% add(Equation("K", "THETA_K*exp(ETA_K)"))
```

We can add an ordinary differential equation as follows:

```{r}
model <- model %>% add(Ode("A_CENTRAL", "-K*A_CENTRAL"))
```

We can init the central compartment as well on the fly:

```{r}
model <- model %>% add(InitialCondition(compartment=1, "1000"))
```

Finally, let's define our `THETA_K` and `ETA_K`:

```{r}
model <- model %>% add(Theta("K", value=0.06))
model <- model %>% add(Omega("K", value=15, type="cv%"))
```

This model can simulated by Campsis as well. Powerful, isn't it?

```{r, eval=EXPORT_PNG}
library(campsis)
dataset <- Dataset(25) %>% add(Observations(seq(0,24,by=0.5)))
results <- model %>% simulate(dataset=dataset, seed=2)
spaghettiPlot(results, "A_CENTRAL")
```

```{r, eval=EXPORT_PNG, echo=FALSE, results='hide'}
ggplot2::ggsave(filename="resources/minimalist_example_sim2.png", width=7, height=3, dpi=100)
```

![](resources/minimalist_example_sim2.png)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_equation.R
\name{Equation}
\alias{Equation}
\title{Create a new equation.}
\usage{
Equation(lhs, rhs = "", comment = as.character(NA))
}
\arguments{
\item{lhs}{left-hand side variable corresponding to the assigned variable name}

\item{rhs}{right-hand side expression corresponding to a formula}

\item{comment}{comment if any, single character string}
}
\value{
an equation
}
\description{
Create a new equation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\docType{class}
\name{mrgsolve_type-class}
\alias{mrgsolve_type-class}
\title{Mrgsolve export type class.}
\description{
Mrgsolve export type class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_element_position.R
\name{Position}
\alias{Position}
\title{Element position in list.}
\usage{
Position(x, after = TRUE)
}
\arguments{
\item{x}{either an integer position (useful to add an element in a code record at a specified position) or
an model element (element can be a model statement or a code record)}

\item{after}{element to be added will be added after x (if after is TRUE) or before x (if after is FALSE)}
}
\value{
a position object
}
\description{
Element position in list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_equation.R
\docType{class}
\name{equation-class}
\alias{equation-class}
\title{Equation class. Any statement in the form A = B.}
\description{
Equation class. Any statement in the form A = B.
}
\section{Slots}{

\describe{
\item{\code{lhs}}{left-hand side expression}

\item{\code{rhs}}{right-hand side expression}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_properties.R
\docType{class}
\name{compartment_properties-class}
\alias{compartment_properties-class}
\title{Compartment properties class.}
\description{
Compartment properties class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_lag_time.R
\docType{class}
\name{compartment_lag_time-class}
\alias{compartment_lag_time-class}
\title{Compartment lag time class.}
\description{
Compartment lag time class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\name{ErrorRecord}
\alias{ErrorRecord}
\title{Create ERROR code record.}
\usage{
ErrorRecord(code = character())
}
\arguments{
\item{code}{code record}
}
\value{
an ERROR code record
}
\description{
Create ERROR code record.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_property.R,
%   R/compartment_bioavailability.R, R/compartment_lag_time.R,
%   R/compartment_infusion_duration.R, R/compartment_infusion_rate.R,
%   R/compartment_initial_condition.R
\name{getRecordName}
\alias{getRecordName}
\alias{getRecordName,compartment_bioavailability-method}
\alias{getRecordName,compartment_lag_time-method}
\alias{getRecordName,compartment_infusion_duration-method}
\alias{getRecordName,compartment_infusion_rate-method}
\alias{getRecordName,compartment_initial_condition-method}
\title{Get record name.}
\usage{
getRecordName(object)

\S4method{getRecordName}{compartment_bioavailability}(object)

\S4method{getRecordName}{compartment_lag_time}(object)

\S4method{getRecordName}{compartment_infusion_duration}(object)

\S4method{getRecordName}{compartment_infusion_rate}(object)

\S4method{getRecordName}{compartment_initial_condition}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
the name of the record
}
\description{
Get record name.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{removeNaColumn}
\alias{removeNaColumn}
\title{Remove given column(s) if it has only NA's.}
\usage{
removeNaColumn(x, column)
}
\arguments{
\item{x}{any data frame}

\item{column}{column name(s)}
}
\value{
updated data frame
}
\description{
Remove given column(s) if it has only NA's.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{read.parameters}
\alias{read.parameters}
\title{Read parameters file.}
\usage{
read.parameters(file, type)
}
\arguments{
\item{file}{path to CSV file}

\item{type}{parameter type: 'theta', 'omega' or 'sigma'}
}
\value{
parameters sub list
}
\description{
Read parameters file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameters.R, R/campsis_model.R
\name{addRSE}
\alias{addRSE}
\alias{addRSE,parameters,parameter,numeric-method}
\alias{addRSE,campsis_model,parameter,numeric-method}
\title{Add relative standard error (RSE) to the specified parameter.}
\usage{
addRSE(object, parameter, value, ...)

\S4method{addRSE}{parameters,parameter,numeric}(object, parameter, value, ...)

\S4method{addRSE}{campsis_model,parameter,numeric}(object, parameter, value, ...)
}
\arguments{
\item{object}{model or parameters object}

\item{parameter}{parameter object (Theta, Omega or Sigma)}

\item{value}{RSE value, in percent}

\item{...}{extra arguments, unused}
}
\value{
updated object
}
\description{
Add relative standard error (RSE) to the specified parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{assertSingleCharacterString}
\alias{assertSingleCharacterString}
\title{Assert the given character vector is a single character string.}
\usage{
assertSingleCharacterString(x)
}
\arguments{
\item{x}{single character string}
}
\value{
no return value
}
\description{
Assert the given character vector is a single character string.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_parser.R
\name{parseStatements}
\alias{parseStatements}
\title{Parse statements code and return Campsis statements.}
\usage{
parseStatements(code)
}
\arguments{
\item{code}{character vector containing all statements (text form)}
}
\value{
a list of Campsis statements
}
\description{
Parse statements code and return Campsis statements.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pattern.R
\name{Pattern}
\alias{Pattern}
\title{Create a pattern.}
\usage{
Pattern(x)
}
\arguments{
\item{x}{regular expression}
}
\value{
a pattern
}
\description{
Create a pattern.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{processDoubleArrayArguments}
\alias{processDoubleArrayArguments}
\title{Process double array arguments.}
\usage{
processDoubleArrayArguments(index, index2, type)
}
\arguments{
\item{index}{parameter index}

\item{index2}{second parameter index}

\item{type}{variance type}
}
\value{
variance type
}
\description{
Process double array arguments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{getMappingMatrix}
\alias{getMappingMatrix}
\title{Return a matrix filled in with OMEGA/SIGMA names to be mapped with the values.
Nonexistent parameters are filled in with the empty string.}
\usage{
getMappingMatrix(parameters, type)
}
\arguments{
\item{parameters}{subset of parameters}

\item{type}{type of parameter to map (omega or sigma)}
}
\value{
a matrix with the names of the OMEGA/SIGMA parameters
}
\description{
Return a matrix filled in with OMEGA/SIGMA names to be mapped with the values.
Nonexistent parameters are filled in with the empty string.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_initial_condition.R
\name{InitialCondition}
\alias{InitialCondition}
\title{Create an initial condition.}
\usage{
InitialCondition(compartment, rhs = "")
}
\arguments{
\item{compartment}{compartment index}

\item{rhs}{right-hand side part of the equation}
}
\value{
an initial condition property
}
\description{
Create an initial condition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{Sigma}
\alias{Sigma}
\title{Create a SIGMA parameter.}
\usage{
Sigma(
  name = NA,
  index = NA,
  index2 = NA,
  value = NA,
  min = NA,
  max = NA,
  fix = FALSE,
  type = NULL,
  label = NA,
  comment = NA
)
}
\arguments{
\item{name}{parameter name, e.g. CL (prefix SIGMA will be added automatically)}

\item{index}{parameter index}

\item{index2}{second parameter index}

\item{value}{parameter value}

\item{min}{minimum value for this parameter when parameter uncertainty is enabled}

\item{max}{maximum value for this parameter when parameter uncertainty is enabled}

\item{fix}{parameter was fixed in estimation, logical value}

\item{type}{variance type: 'var', 'sd', 'covar', 'cv' or 'cv\%'}

\item{label}{parameter label, optional}

\item{comment}{any comment, optional}
}
\value{
a SIGMA parameter
}
\description{
Create a SIGMA parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{isRxODE}
\alias{isRxODE}
\title{Check if the destination engine is RxODE or rxode2.
Note that rxode2 is the successor of RxODE.}
\usage{
isRxODE(dest)
}
\arguments{
\item{dest}{destination engine}
}
\value{
TRUE if RxODE or rxode2, FALSE otherwise
}
\description{
Check if the destination engine is RxODE or rxode2.
Note that rxode2 is the successor of RxODE.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_infusion_rate.R
\docType{class}
\name{compartment_infusion_rate-class}
\alias{compartment_infusion_rate-class}
\title{Compartment infusion rate class.}
\description{
Compartment infusion rate class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{allNa}
\alias{allNa}
\title{Check is vector has NA's only.}
\usage{
allNa(x)
}
\arguments{
\item{x}{any vector}
}
\value{
TRUE if all values are NA, FALSE otherwise
}
\description{
Check is vector has NA's only.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/campsis_model.R,
%   R/replicated_campsis_model.R, R/rxode_model.R, R/mrgsolve_model.R
\name{export}
\alias{export}
\alias{export,campsis_model,character-method}
\alias{export,replicated_campsis_model,campsis_model-method}
\alias{export,campsis_model,rxode_type-method}
\alias{export,campsis_model,mrgsolve_type-method}
\title{Export function.}
\usage{
export(object, dest, ...)

\S4method{export}{campsis_model,character}(object, dest, ...)

\S4method{export}{replicated_campsis_model,campsis_model}(object, dest = CampsisModel(), index, ...)

\S4method{export}{campsis_model,rxode_type}(object, dest, ...)

\S4method{export}{campsis_model,mrgsolve_type}(object, dest, outvars = NULL, extra_params = character(0))
}
\arguments{
\item{object}{generic object}

\item{dest}{destination}

\item{...}{optional arguments}

\item{index}{index of the replicated Campsis model to export}

\item{outvars}{additional variables to capture}

\item{extra_params}{extra parameter names to be added. By default, they will be assigned a zero value.}
}
\value{
specific object depending on given destination
}
\description{
Export function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode_conversion.R
\name{rxodeMatrix}
\alias{rxodeMatrix}
\title{Get the OMEGA/SIGMA matrix for rxode2.}
\usage{
rxodeMatrix(model, type = "omega")
}
\arguments{
\item{model}{Campsis model or Campsis parameters}

\item{type}{either omega or sigma}
}
\value{
omega/sigma named matrix
}
\description{
Get the OMEGA/SIGMA matrix for rxode2.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/compartments.R,
%   R/parameters.R, R/code_record.R, R/code_records.R, R/campsis_model.R
\name{replace}
\alias{replace}
\alias{replace,pmx_list,pmx_element-method}
\alias{replace,pmx_list,pmx_list-method}
\alias{replace,pmx_list,list-method}
\alias{replace,compartments,compartment_property-method}
\alias{replace,parameters,single_array_parameter-method}
\alias{replace,parameters,double_array_parameter-method}
\alias{replace,statements_record,model_statement-method}
\alias{replace,code_records,model_statement-method}
\alias{replace,campsis_model,compartment-method}
\alias{replace,campsis_model,compartment_property-method}
\alias{replace,campsis_model,parameter-method}
\alias{replace,campsis_model,code_record-method}
\alias{replace,campsis_model,model_statement-method}
\title{Replace element by another in list.}
\usage{
replace(object, x)

\S4method{replace}{pmx_list,pmx_element}(object, x)

\S4method{replace}{pmx_list,pmx_list}(object, x)

\S4method{replace}{pmx_list,list}(object, x)

\S4method{replace}{compartments,compartment_property}(object, x)

\S4method{replace}{parameters,single_array_parameter}(object, x)

\S4method{replace}{parameters,double_array_parameter}(object, x)

\S4method{replace}{statements_record,model_statement}(object, x)

\S4method{replace}{code_records,model_statement}(object, x)

\S4method{replace}{campsis_model,compartment}(object, x)

\S4method{replace}{campsis_model,compartment_property}(object, x)

\S4method{replace}{campsis_model,parameter}(object, x)

\S4method{replace}{campsis_model,code_record}(object, x)

\S4method{replace}{campsis_model,model_statement}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element to replace}
}
\value{
list object or an error if the element does not exist in the list
}
\description{
Replace element by another in list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R
\docType{class}
\name{pmx_list-class}
\alias{pmx_list-class}
\title{PMX list class.}
\description{
PMX list class.
}
\section{Slots}{

\describe{
\item{\code{list}}{effective list which will contain the elements}

\item{\code{type}}{type of the elements this list may contain}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveOde}
\alias{mrgsolveOde}
\title{Get the ODE block for mrgsolve.}
\usage{
mrgsolveOde(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
ODE block
}
\description{
Get the ODE block for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{jsonToOffDiagParameter}
\alias{jsonToOffDiagParameter}
\title{Convert JSON correlation parameter (OMEGA or SIGMA) into a Campsis parameter.}
\usage{
jsonToOffDiagParameter(json, diag_names)
}
\arguments{
\item{json}{JSON data}

\item{diag_names}{parameter names on the diagonal, character vector}
}
\value{
the corresponding Campsis parameter
}
\description{
Convert JSON correlation parameter (OMEGA or SIGMA) into a Campsis parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{fixOmega}
\alias{fixOmega}
\alias{fixOmega,parameters-method}
\title{Fix omega matrix for SAME OMEGA parameters that have NA values due to imperfections in Pharmpy import.}
\usage{
fixOmega(object)

\S4method{fixOmega}{parameters}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
the parameter that matches
}
\description{
Fix omega matrix for SAME OMEGA parameters that have NA values due to imperfections in Pharmpy import.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{getCampsismodOption}
\alias{getCampsismodOption}
\title{Get Campsismod option logic.}
\usage{
getCampsismodOption(name, default)
}
\arguments{
\item{name}{option to search}

\item{default}{default value if option not found}
}
\value{
option value
}
\description{
Get Campsismod option logic.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\name{updateCompartments}
\alias{updateCompartments}
\title{Update compartments list from the persisted records.
Exported especially for package \code{campsistrans}.
However, this method should not be called.}
\usage{
updateCompartments(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
an updated Campsis model, with an updated compartments list
}
\description{
Update compartments list from the persisted records.
Exported especially for package \code{campsistrans}.
However, this method should not be called.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replication_settings.R
\docType{class}
\name{auto_replication_settings-class}
\alias{auto_replication_settings-class}
\title{Auto replication settings class.}
\description{
Auto replication settings class.
}
\section{Slots}{

\describe{
\item{\code{wishart}}{logical, sample OMEGAs and SIGMAs from scaled inverse chi-squared or Wishart distributions}

\item{\code{odf}}{the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the OMEGAs, integer vector}

\item{\code{sdf}}{the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the SIGMAs, integer vector}

\item{\code{quiet}}{logical, suppress info messages}

\item{\code{max_iterations}}{number of iterations maximum to sample the parameters}

\item{\code{max_chunk_size}}{maximum number of rows to sample at once, default value will be the number of replicates, unless specified.}

\item{\code{check_min_max}}{logical, check for min/max values when sampling the parameters}

\item{\code{check_pos_def}}{logical, check for positive definiteness when sampling the OMEGA/SIGMA parameters}

\item{\code{wishart_correction}}{logical, FALSE is default, see https://github.com/metrumresearchgroup/simpar/issues/11}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{isStrictRecordDelimiter}
\alias{isStrictRecordDelimiter}
\title{Is strict record delimiter. A strict record delimiter is any line starting 
with [...] and followed by nothing but spaces or a possible comment.}
\usage{
isStrictRecordDelimiter(line)
}
\arguments{
\item{line}{any line, single character value}
}
\value{
a logical value
}
\description{
Is strict record delimiter. A strict record delimiter is any line starting 
with [...] and followed by nothing but spaces or a possible comment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{Omega}
\alias{Omega}
\title{Create an OMEGA parameter.}
\usage{
Omega(
  name = NA,
  index = NA,
  index2 = NA,
  value = NA,
  min = NA,
  max = NA,
  fix = FALSE,
  type = NULL,
  same = NA,
  label = NA,
  comment = NA
)
}
\arguments{
\item{name}{parameter name, e.g. CL (prefix OMEGA will be added automatically)}

\item{index}{parameter index}

\item{index2}{second parameter index}

\item{value}{parameter value}

\item{min}{minimum value for this parameter when parameter uncertainty is enabled}

\item{max}{maximum value for this parameter when parameter uncertainty is enabled}

\item{fix}{parameter was fixed in estimation, logical value}

\item{type}{variance type: 'var', 'sd', 'covar', 'cor', 'cv' or 'cv\%'}

\item{same}{NA by default, FALSE for first OMEGA followed by 'SAME' OMEGA's, TRUE for 'SAME' OMEGA's}

\item{label}{parameter label, optional}

\item{comment}{any comment, optional}
}
\value{
an OMEGA parameter
}
\description{
Create an OMEGA parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartments.R
\name{Compartments}
\alias{Compartments}
\title{Create a list of compartments}
\usage{
Compartments()
}
\value{
an empty list of compartments
}
\description{
Create a list of compartments
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replication_settings.R
\name{AutoReplicationSettings}
\alias{AutoReplicationSettings}
\title{Create auto replication settings.}
\usage{
AutoReplicationSettings(
  wishart = FALSE,
  odf = NA,
  sdf = NA,
  checkMinMax = TRUE,
  checkPosDef = FALSE,
  quiet = NA
)
}
\arguments{
\item{wishart}{logical, sample OMEGAs and SIGMAs from scaled inverse chi-squared (univariate OMEGA distribution)
or Wishart distribution (block of OMEGAs)}

\item{odf}{the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the OMEGAs,
single integer value (the same degrees of freedom for all OMEGA blocks) or integer vector (one value per OMEGA block)}

\item{sdf}{the degrees of freedom for the scaled inverse chi-squared/Wishart distribution with regards to the SIGMAs,
single integer value (the same degrees of freedom for all SIGMA blocks) or integer vector (one value per SIGMA block)}

\item{checkMinMax}{logical, check for min/max values when sampling the parameters, default is TRUE}

\item{checkPosDef}{logical, check for positive definiteness when sampling the OMEGA/SIGMA parameters from the variance-covariance matrix (i.e. when \code{wishart=FALSE}), default is FALSE (requires extra time)}

\item{quiet}{logical, suppress info messages, default is NA. By default, messages will be printed out when the success rate of sampling the parameters is below 95\%.}
}
\value{
replication settings
}
\description{
By default, all model parameters are sampled from a multivariate normal 
distribution, whose characteristics are specified by the variance-covariance matrix.
OMEGAs and SIGMAs can be sampled from scaled inverse chi-squared or Wishart distributions
by setting the \code{wishart} argument to \code{TRUE}. In that case, THETAs are still sampled
from a multivariate normal distribution, while OMEGAS and SIGMAs are sampled from
scaled inverse chi-squared (univariate OMEGA/SIGMA distribution) and Wishart (block of
OMEGAs/SIGMAs) distribution, respectively. When \code{wishart} is set to \code{TRUE}, the degrees
of freedom of the distribution must be specified, respectively, \code{odf} for the OMEGAs and
\code{sdf} for the SIGMAs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveCompartment}
\alias{mrgsolveCompartment}
\title{Get the compartment block for mrgsolve.}
\usage{
mrgsolveCompartment(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
character vector, each value is a line
}
\description{
Get the compartment block for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{sampleFromMultivariateNormalDistribution}
\alias{sampleFromMultivariateNormalDistribution}
\title{Sample from a multivariate normal distribution.}
\usage{
sampleFromMultivariateNormalDistribution(parameters, n, settings)
}
\arguments{
\item{parameters}{model parameters (all), already standardised}

\item{n}{number of rows to sample}

\item{settings}{replication settings}
}
\value{
a data frame with the sampled parameters
}
\description{
Sample from a multivariate normal distribution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{getCompartments}
\alias{getCompartments}
\title{Detect all compartments names from the code records.
Only for model instantiation. Not exported.}
\usage{
getCompartments(records)
}
\arguments{
\item{records}{code records}
}
\value{
a list of compartments
}
\description{
Detect all compartments names from the code records.
Only for model instantiation. Not exported.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_if_statement.R
\docType{class}
\name{if_statement-class}
\alias{if_statement-class}
\title{If-statement class. Any statement in the form if (condition) A = B.}
\description{
If-statement class. Any statement in the form if (condition) A = B.
}
\section{Slots}{

\describe{
\item{\code{condition}}{IF statement condition}

\item{\code{equation}}{any equation or ODE}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{read.allparameters}
\alias{read.allparameters}
\title{Read all parameters files at once.}
\usage{
read.allparameters(folder)
}
\arguments{
\item{folder}{path to folder or path to zipped project}
}
\value{
parameters object
}
\description{
Read all parameters files at once.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{model_suite}
\alias{model_suite}
\title{Campsis model suite.}
\format{
A list with all the models:
\describe{
   \item{pk}{extensive list of pharmacokinetic (PK) model templates}
   \item{pd}{list of pharmacodynamic (PD) model templates, to be plugged into any pharmacokinetic (PK) model}
   \item{tmdd}{extensive list of target-mediated drug disposition (TMDD) model templates}
   \item{nonmem}{list of model templates translated from standard NONMEM control streams}
   \item{literature}{a couple of models coming from the literature}
   \item{other}{a couple of custom models}
}
}
\source{
\url{https://calvagone.github.io/campsis.doc/}

\url{https://www.iconplc.com/solutions/technologies/nonmem/}
}
\usage{
model_suite
}
\description{
A library of models of all kinds, ready to be simulated in Campsis. These model
templates are sorted into the following categories: pharmacokinetic (PK), pharmacodynamic (PD), 
target-mediated drug disposition (TMDD), NONMEM, literature  and other (custom models).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{jsonToCampsisModel}
\alias{jsonToCampsisModel}
\title{JSON to Campsis dataset.}
\usage{
jsonToCampsisModel(object, json)
}
\arguments{
\item{object}{empty dataset}

\item{json}{json element}
}
\value{
Campsis dataset
}
\description{
JSON to Campsis dataset.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveParam}
\alias{mrgsolveParam}
\title{Get the parameters block for mrgsolve.}
\usage{
mrgsolveParam(model, extra_params = character(0))
}
\arguments{
\item{model}{Campsis model}

\item{extra_params}{extra parameter names to be added. By default, they will be assigned a zero value.}
}
\value{
character vector, 1 parameter per line. First one is header [PARAM].
}
\description{
Get the parameters block for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameter.R, R/parameters.R,
%   R/campsis_model.R
\name{exportToJSON}
\alias{exportToJSON}
\alias{exportToJSON,theta-method}
\alias{exportToJSON,omega-method}
\alias{exportToJSON,sigma-method}
\alias{exportToJSON,parameters-method}
\alias{exportToJSON,campsis_model-method}
\title{Export the given object to a JSON object, ready to be written to files.}
\usage{
exportToJSON(object, ...)

\S4method{exportToJSON}{theta}(object, ...)

\S4method{exportToJSON}{omega}(object, ...)

\S4method{exportToJSON}{sigma}(object, ...)

\S4method{exportToJSON}{parameters}(object, ...)

\S4method{exportToJSON}{campsis_model}(object, ...)
}
\arguments{
\item{object}{any object}

\item{...}{extra arguments, unused}
}
\value{
the loaded S4 object
}
\description{
Export the given object to a JSON object, ready to be written to files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{isMatrixPositiveDefinite}
\alias{isMatrixPositiveDefinite}
\title{Is matrix positive definite. Same check as \code{mvtnorm} does.}
\usage{
isMatrixPositiveDefinite(matrix, tol = 1e-06)
}
\arguments{
\item{matrix}{matrix to check}

\item{tol}{tolerance when checking the eigenvalues}
}
\description{
Is matrix positive definite. Same check as \code{mvtnorm} does.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{ifStatementPatternStr}
\alias{ifStatementPatternStr}
\title{Return the IF-statement pattern (string form).}
\usage{
ifStatementPatternStr()
}
\value{
pattern (regular expression)
}
\description{
Return the IF-statement pattern (string form).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameters.R, R/campsis_model.R
\name{setMinMax}
\alias{setMinMax}
\alias{setMinMax,parameters,parameter,numeric,numeric-method}
\alias{setMinMax,parameters,character,numeric,numeric-method}
\alias{setMinMax,campsis_model,parameter,numeric,numeric-method}
\alias{setMinMax,campsis_model,character,numeric,numeric-method}
\title{Set the minimum and maximum value on a model parameter.}
\usage{
setMinMax(object, parameter, min, max, ...)

\S4method{setMinMax}{parameters,parameter,numeric,numeric}(object, parameter, min, max, ...)

\S4method{setMinMax}{parameters,character,numeric,numeric}(object, parameter, min, max, ...)

\S4method{setMinMax}{campsis_model,parameter,numeric,numeric}(object, parameter, min, max, ...)

\S4method{setMinMax}{campsis_model,character,numeric,numeric}(object, parameter, min, max, ...)
}
\arguments{
\item{object}{model or parameters object}

\item{parameter}{parameter object (Theta, Omega or Sigma)}

\item{min}{minimum value for this parameter when parameter uncertainty is enabled}

\item{max}{maximum value for this parameter when parameter uncertainty is enabled}

\item{...}{extra arguments, unused}
}
\value{
updated object
}
\description{
Set the minimum and maximum value on a model parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omega_block.R
\name{getOmegaIndexes}
\alias{getOmegaIndexes}
\alias{getOmegaIndexes,omega_block-method}
\title{Get the indexes of the omegas.}
\usage{
getOmegaIndexes(object)

\S4method{getOmegaIndexes}{omega_block}(object)
}
\arguments{
\item{object}{omega block}
}
\value{
a list of integers
}
\description{
Get the indexes of the omegas.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replication_settings.R
\docType{class}
\name{replication_settings-class}
\alias{replication_settings-class}
\title{Replication settings interface.}
\description{
Replication settings interface.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\docType{class}
\name{rate_record-class}
\alias{rate_record-class}
\title{(Infusion)-rate record class.}
\description{
(Infusion)-rate record class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_element.R
\docType{class}
\name{json_element-class}
\alias{json_element-class}
\title{JSON element class.}
\description{
JSON element class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/special_operators.R
\name{\%>\%}
\alias{\%>\%}
\title{Magritt operator for piping.}
\usage{
lhs \%>\% rhs
}
\description{
Magritt operator for piping.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{addProperties}
\alias{addProperties}
\title{Add properties to compartments object.}
\usage{
addProperties(compartments, records, name, init)
}
\arguments{
\item{compartments}{compartments object}

\item{records}{all records}

\item{name}{record name to look at}

\item{init}{empty characteristic, to be completed}
}
\value{
updated compartments object
}
\description{
Add properties to compartments object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartments.R
\name{appendCompartments}
\alias{appendCompartments}
\title{Append compartments.}
\usage{
appendCompartments(compartments1, compartments2)
}
\arguments{
\item{compartments1}{base set of compartments}

\item{compartments2}{extra set of compartments to be appended}
}
\value{
the resulting set of compartments
}
\description{
Append compartments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omega_blocks.R
\name{OmegaBlocks}
\alias{OmegaBlocks}
\title{Create a list of OMEGA blocks.}
\usage{
OmegaBlocks()
}
\description{
Create a list of OMEGA blocks.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/model_unknown_statement.R,
%   R/model_comment.R, R/model_line_break.R, R/model_equation.R, R/model_ode.R,
%   R/model_if_statement.R, R/compartment.R, R/compartment_property.R,
%   R/compartment_initial_condition.R, R/mrgsolve_model.R
\name{toString}
\alias{toString}
\alias{toString,unknown_statement-method}
\alias{toString,comment-method}
\alias{toString,line_break-method}
\alias{toString,equation-method}
\alias{toString,ode-method}
\alias{toString,if_statement-method}
\alias{toString,compartment-method}
\alias{toString,compartment_property-method}
\alias{toString,compartment_initial_condition-method}
\alias{toString,mrgsolve_model-method}
\title{ToString generic method.}
\usage{
toString(object, ...)

\S4method{toString}{unknown_statement}(object, ...)

\S4method{toString}{comment}(object, ...)

\S4method{toString}{line_break}(object, ...)

\S4method{toString}{equation}(object, ...)

\S4method{toString}{ode}(object, ...)

\S4method{toString}{if_statement}(object, ...)

\S4method{toString}{compartment}(object, ...)

\S4method{toString}{compartment_property}(object, ...)

\S4method{toString}{compartment_initial_condition}(object, ...)

\S4method{toString}{mrgsolve_model}(object, ...)
}
\arguments{
\item{object}{generic object}

\item{...}{extra arguments needed for toString conversion}
}
\value{
character value/vector
}
\description{
ToString generic method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_infusion_rate.R
\name{InfusionRate}
\alias{InfusionRate}
\title{Create an infusion rate.}
\usage{
InfusionRate(compartment, rhs = "")
}
\arguments{
\item{compartment}{compartment index}

\item{rhs}{right-hand side part of the equation}
}
\value{
an infusion rate property
}
\description{
Create an infusion rate.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\name{writeRecordDelimiter}
\alias{writeRecordDelimiter}
\title{Write record delimiter line.}
\usage{
writeRecordDelimiter(object)
}
\arguments{
\item{object}{code record}
}
\value{
a record delimiter line
}
\description{
Write record delimiter line.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_line_break.R
\name{LineBreak}
\alias{LineBreak}
\title{Create a new line break.}
\usage{
LineBreak()
}
\value{
a line break
}
\description{
Create a new line break.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replication_settings.R
\name{ManualReplicationSettings}
\alias{ManualReplicationSettings}
\title{Create manual replication settings.}
\usage{
ManualReplicationSettings(data)
}
\arguments{
\item{data}{data frame with 1 row per replicate, must contain a column named 'REPLICATE'
with unique integers from 1 to nrow(data), other columns are model parameters to use.}
}
\value{
replication settings
}
\description{
Create manual replication settings.
}
\details{
Use these settings to import custom replicated model parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\name{read.campsis}
\alias{read.campsis}
\title{Read a Campsis model.}
\usage{
read.campsis(file)
}
\arguments{
\item{file}{path to folder (old format) or path to JSON file (new format)}
}
\value{
Campsis model
}
\description{
Read a Campsis model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveTable}
\alias{mrgsolveTable}
\title{Get the TABLE block for mrgsolve.}
\usage{
mrgsolveTable(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
TABLE block if at least one line in error record, character(0) otherwise
}
\description{
Get the TABLE block for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{isBlockFixed}
\alias{isBlockFixed}
\title{Say if the block is fixed (i.e. all parameters are fixed).}
\usage{
isBlockFixed(block)
}
\arguments{
\item{block}{block of OMEGAs or SIGMAs}
}
\value{
logical value
}
\description{
Say if the block is fixed (i.e. all parameters are fixed).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pattern.R
\name{VariablePattern}
\alias{VariablePattern}
\title{Create a variable pattern.}
\usage{
VariablePattern(x)
}
\arguments{
\item{x}{variable name}
}
\value{
a variable pattern
}
\description{
Create a variable pattern.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{jsonToParameter}
\alias{jsonToParameter}
\title{JSON to Campsis parameter.}
\usage{
jsonToParameter(x, index = NULL, index2 = NULL)
}
\arguments{
\item{x}{JSON data}

\item{index}{parameter index to add}

\item{index2}{second parameter index to add for OMEGAs and SIGMAs}
}
\value{
Campsis parameter
}
\description{
JSON to Campsis parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{minIndex}
\alias{minIndex}
\alias{minIndex,parameters-method}
\title{Min index.}
\usage{
minIndex(object)

\S4method{minIndex}{parameters}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
min index
}
\description{
Min index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_unknown_statement.R
\docType{class}
\name{unknown_statement-class}
\alias{unknown_statement-class}
\title{Unknown statement class. Any statement not recognized by campsismod.}
\description{
Unknown statement class. Any statement not recognized by campsismod.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{sampleFromInverseChiSquaredOrWishart}
\alias{sampleFromInverseChiSquaredOrWishart}
\title{Sample parameters from inverse scaled chi-squared or wishart distribution(s).}
\usage{
sampleFromInverseChiSquaredOrWishart(parameters, n, settings)
}
\arguments{
\item{parameters}{subset of Campsis parameters (OMEGAs or SIGMAs)}

\item{n}{number of rows to sample}

\item{settings}{replication settings}
}
\value{
a data frame with the sampled parameters
}
\description{
Sample parameters from inverse scaled chi-squared or wishart distribution(s).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{Parameters}
\alias{Parameters}
\title{Create a list of parameters.}
\usage{
Parameters()
}
\value{
an empty list of parameters
}
\description{
Create a list of parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pattern.R
\docType{class}
\name{pattern-class}
\alias{pattern-class}
\title{Pattern class.}
\description{
Pattern class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_bioavailability.R
\name{Bioavailability}
\alias{Bioavailability}
\title{Create a bioavailability for the specified compartment.}
\usage{
Bioavailability(compartment, rhs = "")
}
\arguments{
\item{compartment}{compartment index}

\item{rhs}{right-hand side part of the equation}
}
\value{
a bioavailability property
}
\description{
Create a bioavailability for the specified compartment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{writeVarcov}
\alias{writeVarcov}
\title{Write variance-covariance matrix.}
\usage{
writeVarcov(object, file)
}
\arguments{
\item{object}{matrix}

\item{file}{filename}
}
\value{
TRUE if success
}
\description{
Write variance-covariance matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{sampleFromInverseChiSquaredCore}
\alias{sampleFromInverseChiSquaredCore}
\title{Sample from scaled inverse chi-squared distribution (core method).}
\usage{
sampleFromInverseChiSquaredCore(n, df, scale, variable)
}
\arguments{
\item{n}{numbers of rows to sample}

\item{df}{degree of freedom}

\item{scale}{scale parameter}

\item{variable}{variable name}
}
\value{
a data frame with the unique sampled parameter
}
\description{
Sample from scaled inverse chi-squared distribution (core method).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{processExtraArg}
\alias{processExtraArg}
\title{Process extra arguments.}
\usage{
processExtraArg(args, name, default = NULL, mandatory = FALSE)
}
\arguments{
\item{args}{arguments list}

\item{name}{argument name to retrieve}

\item{default}{default value if argument is not present}

\item{mandatory}{mandatory argument, logical value}
}
\value{
requested argument value
}
\description{
Process extra arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\docType{class}
\name{duration_record-class}
\alias{duration_record-class}
\title{(Infusion)-duration record class.}
\description{
(Infusion)-duration record class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment.R
\docType{class}
\name{compartment-class}
\alias{compartment-class}
\title{Compartment class.}
\description{
Compartment class.
}
\section{Slots}{

\describe{
\item{\code{name}}{compartment name (without prefix)}

\item{\code{index}}{compartment index}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{openJSON}
\alias{openJSON}
\title{Open JSON file.}
\usage{
openJSON(json, schema = NULL)
}
\arguments{
\item{json}{JSON in its string form or path to JSON file}

\item{schema}{JSON schema}
}
\value{
parsed JSON object
}
\description{
Open JSON file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/compartment_properties.R,
%   R/parameters.R
\name{select}
\alias{select}
\alias{select,data.frame-method}
\alias{select,compartment_properties-method}
\alias{select,parameters-method}
\title{Get a subset of an object.}
\usage{
select(object, ...)

\S4method{select}{data.frame}(object, ...)

\S4method{select}{compartment_properties}(object, ...)

\S4method{select}{parameters}(object, ...)
}
\arguments{
\item{object}{generic object}

\item{...}{arguments to select}
}
\value{
subset of an object
}
\description{
Get a subset of an object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global.R
\name{methodsGlobalFunctions}
\alias{methodsGlobalFunctions}
\title{A empty method only used to declare some of the global function definitions
contained in the package 'methods', which are used intensively.}
\usage{
methodsGlobalFunctions()
}
\value{
no return value
}
\description{
A empty method only used to declare some of the global function definitions
contained in the package 'methods', which are used intensively.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{matchDoubleArrayParameter}
\alias{matchDoubleArrayParameter}
\title{Match double array parameter from list based on the name instead of the index.
If a match is found, its indexes are automatically copied.}
\usage{
matchDoubleArrayParameter(object, x)
}
\arguments{
\item{object}{parameters}

\item{x}{double array parameter to match}
}
\value{
the same parameter is no match was found or the same parameter with updated indexes if a match was found
}
\description{
Match double array parameter from list based on the name instead of the index.
If a match is found, its indexes are automatically copied.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{flagSampledParameterRows}
\alias{flagSampledParameterRows}
\title{Flag all parameter rows that have at least one parameter out of the specified range
or that have a non positive definite OMEGA or SIGMA matrix.}
\usage{
flagSampledParameterRows(table, minMax, settings, parameters)
}
\arguments{
\item{table}{a data frame returned by \code{sampleMore}}

\item{minMax}{a data frame with min, max values for each parameter}

\item{settings}{replication settings}

\item{parameters}{double array parameters to check for positive definiteness}
}
\description{
Flag all parameter rows that have at least one parameter out of the specified range
or that have a non positive definite OMEGA or SIGMA matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{isIfStatement}
\alias{isIfStatement}
\title{Say if line in record is an IF-statement.}
\usage{
isIfStatement(x)
}
\arguments{
\item{x}{character value}
}
\value{
logical value
}
\description{
Say if line in record is an IF-statement.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{mapS4SlotsToJSONProperties}
\alias{mapS4SlotsToJSONProperties}
\title{Map S4 slots to JSON properties.}
\usage{
mapS4SlotsToJSONProperties(
  object,
  add_type = TRUE,
  optional = NULL,
  ignore = NULL
)
}
\arguments{
\item{object}{S4 object}

\item{add_type}{add type as a property, TRUE by default}

\item{optional}{properties that are optional in JSON, character vector}

\item{ignore}{slots to be ignored}
}
\value{
a JSON object ready to be serialised
}
\description{
Map S4 slots to JSON properties.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{removeTrailingLineBreaks}
\alias{removeTrailingLineBreaks}
\title{Remove all trailing line breaks.}
\usage{
removeTrailingLineBreaks(x)
}
\arguments{
\item{x}{character vector}
}
\value{
a character vector
}
\description{
Remove all trailing line breaks.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartments.R
\docType{class}
\name{compartments-class}
\alias{compartments-class}
\title{Compartments class.}
\description{
Compartments class.
}
\section{Slots}{

\describe{
\item{\code{properties}}{compartment properties of the compartments defined in this class}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{isDiag}
\alias{isDiag}
\alias{isDiag,double_array_parameter-method}
\title{Is diagonal.}
\usage{
isDiag(object)

\S4method{isDiag}{double_array_parameter}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
logical value
}
\description{
Is diagonal.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_infusion_duration.R
\name{InfusionDuration}
\alias{InfusionDuration}
\title{Create an infusion duration.}
\usage{
InfusionDuration(compartment, rhs = "")
}
\arguments{
\item{compartment}{compartment index}

\item{rhs}{right-hand side part of the equation}
}
\value{
an infusion duration property
}
\description{
Create an infusion duration.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omega_blocks.R
\name{getOmegaBlock}
\alias{getOmegaBlock}
\alias{getOmegaBlock,omega_blocks,double_array_parameter-method}
\title{Get the right block of OMEGA's.}
\usage{
getOmegaBlock(object, x)

\S4method{getOmegaBlock}{omega_blocks,double_array_parameter}(object, x)
}
\arguments{
\item{object}{list of OMEGA blocks}

\item{x}{omega param}
}
\value{
the corresponding OMEGA block or NULL if not found
}
\description{
Get the right block of OMEGA's.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{hasExactMethod}
\alias{hasExactMethod}
\title{Has exact method allows to check the existence of a S4 method in Campsis
based on its signature.}
\usage{
hasExactMethod(generic, signature, where = topenv(parent.frame()))
}
\arguments{
\item{generic}{generic function name}

\item{signature}{function signature}

\item{where}{where to search functions}
}
\value{
logical value
}
\description{
Has exact method allows to check the existence of a S4 method in Campsis
based on its signature.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameter.R, R/parameters.R,
%   R/campsis_model.R
\name{getUncertainty}
\alias{getUncertainty}
\alias{getUncertainty,parameter-method}
\alias{getUncertainty,parameters-method}
\alias{getUncertainty,campsis_model-method}
\title{Get uncertainty on the parameters.}
\usage{
getUncertainty(object, ...)

\S4method{getUncertainty}{parameter}(object, varcov, ...)

\S4method{getUncertainty}{parameters}(object, ...)

\S4method{getUncertainty}{campsis_model}(object, ...)
}
\arguments{
\item{object}{generic object}

\item{...}{extra arguments}

\item{varcov}{variance covariance matrix}
}
\value{
data frame with standard error (se) and relative standard error (rse%) columns
}
\description{
Get uncertainty on the parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_statement.R
\name{UnsupportedDestException}
\alias{UnsupportedDestException}
\title{Unknown destination engine exception.}
\usage{
UnsupportedDestException()
}
\value{
no return value
}
\description{
Unknown destination engine exception.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{isRecordDelimiter}
\alias{isRecordDelimiter}
\title{Is record delimiter. A record delimiter is any line starting with [...].}
\usage{
isRecordDelimiter(line)
}
\arguments{
\item{line}{any line, single character value}
}
\value{
a logical value
}
\description{
Is record delimiter. A record delimiter is any line starting with [...].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{sampleGeneric}
\alias{sampleGeneric}
\title{Generic function for parameter sampling according to the minimum and maximum values.
This function will sample parameters a first time and check if some parameters are out of range.
Based on the success rate, it will sample more parameters to reach the desired number of rows.}
\usage{
sampleGeneric(fun, args, n, minMax, msg, settings, parameters = NULL)
}
\arguments{
\item{fun}{function to call to sample parameters}

\item{args}{arguments to pass to the function}

\item{n}{number of rows to sample}

\item{minMax}{a data frame with min, max values for each parameter}

\item{msg}{message template}

\item{settings}{replication settings}

\item{parameters}{double array parameters to check for positive definiteness}
}
\value{
tibble with the sampled parameters (1 parameter per column + REPLICATE column)
}
\description{
Generic function for parameter sampling according to the minimum and maximum values.
This function will sample parameters a first time and check if some parameters are out of range.
Based on the success rate, it will sample more parameters to reach the desired number of rows.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_ode.R
\docType{class}
\name{ode-class}
\alias{ode-class}
\title{ODE class. Any statement in the form d/dt(A_CMT) = B.}
\description{
ODE class. Any statement in the form d/dt(A_CMT) = B.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/compartments.R, R/campsis_model.R
\name{getCompartmentIndex}
\alias{getCompartmentIndex}
\alias{getCompartmentIndex,compartments,character-method}
\alias{getCompartmentIndex,campsis_model,character-method}
\title{Get the compartment index for the specified compartment name.}
\usage{
getCompartmentIndex(object, name)

\S4method{getCompartmentIndex}{compartments,character}(object, name)

\S4method{getCompartmentIndex}{campsis_model,character}(object, name)
}
\arguments{
\item{object}{generic object that contains compartments information}

\item{name}{compartment name}
}
\value{
the corresponding compartment index
}
\description{
Get the compartment index for the specified compartment name.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_property.R,
%   R/compartment_bioavailability.R, R/compartment_lag_time.R,
%   R/compartment_infusion_duration.R, R/compartment_infusion_rate.R,
%   R/compartment_initial_condition.R
\name{getPrefix}
\alias{getPrefix}
\alias{getPrefix,compartment_bioavailability-method}
\alias{getPrefix,compartment_lag_time-method}
\alias{getPrefix,compartment_infusion_duration-method}
\alias{getPrefix,compartment_infusion_rate-method}
\alias{getPrefix,compartment_initial_condition-method}
\title{Get prefix.}
\usage{
getPrefix(object, ...)

\S4method{getPrefix}{compartment_bioavailability}(object, ...)

\S4method{getPrefix}{compartment_lag_time}(object, ...)

\S4method{getPrefix}{compartment_infusion_duration}(object, ...)

\S4method{getPrefix}{compartment_infusion_rate}(object, ...)

\S4method{getPrefix}{compartment_initial_condition}(object, ...)
}
\arguments{
\item{object}{generic object}

\item{...}{e.g. \code{dest='mrgsolve'}}
}
\value{
the prefix of this object
}
\description{
Get prefix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omega_block.R
\name{shiftOmegaIndexes}
\alias{shiftOmegaIndexes}
\alias{shiftOmegaIndexes,omega_block-method}
\title{Shift OMEGA indexes.}
\usage{
shiftOmegaIndexes(object)

\S4method{shiftOmegaIndexes}{omega_block}(object)
}
\arguments{
\item{object}{omega block}
}
\value{
same block but shifted
}
\description{
Shift OMEGA indexes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\docType{class}
\name{code_record-class}
\alias{code_record-class}
\title{Code record class. See this code record as an abstract class.
2 implementations are possible:
- properties record (lag, duration, rate & bioavailability properties)
- statements record (main, ode & error records)}
\description{
Code record class. See this code record as an abstract class.
2 implementations are possible:
- properties record (lag, duration, rate & bioavailability properties)
- statements record (main, ode & error records)
}
\section{Slots}{

\describe{
\item{\code{comment}}{a comment, single character value}

\item{\code{statements}}{model statements}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameters.R, R/code_records.R,
%   R/campsis_model.R
\name{write}
\alias{write}
\alias{write,parameters,character-method}
\alias{write,code_records,character-method}
\alias{write,campsis_model,character-method}
\alias{write,json_element,character-method}
\title{Write generic object to files.}
\usage{
write(object, file, ...)

\S4method{write}{parameters,character}(object, file, ...)

\S4method{write}{code_records,character}(object, file, ...)

\S4method{write}{campsis_model,character}(object, file, ...)

\S4method{write}{json_element,character}(object, file, digits = 12, ...)
}
\arguments{
\item{object}{generic object}

\item{file}{path of the output file or directory}

\item{...}{extra arguments}

\item{digits}{significant digits in JSON file, default is 12}
}
\value{
logical value, TRUE for success, FALSE for failure
}
\description{
Write generic object to files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R
\name{default}
\alias{default}
\title{Get default element from list.}
\usage{
default(object, ...)
}
\arguments{
\item{object}{list object}

\item{...}{additional arguments}
}
\value{
the default element from list
}
\description{
Get default element from list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{convertAnyComment}
\alias{convertAnyComment}
\title{Convert Campsis comment style to C/C++ code.
Only the first # is translated to //.}
\usage{
convertAnyComment(x)
}
\arguments{
\item{x}{any record line}
}
\value{
same line with comments translated to C/C++
}
\description{
Convert Campsis comment style to C/C++ code.
Only the first # is translated to //.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R
\name{getByName}
\alias{getByName}
\alias{getByName,pmx_list,character-method}
\title{Get an element from a list by name.
Never return more than 1 element.}
\usage{
getByName(object, name)

\S4method{getByName}{pmx_list,character}(object, name)
}
\arguments{
\item{object}{list object}

\item{name}{element name to search for}
}
\value{
the element that was found or NULL if no element was found with the same name
}
\description{
Get an element from a list by name.
Never return more than 1 element.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R
\name{getNames}
\alias{getNames}
\alias{getNames,pmx_list-method}
\title{Get element names from list.}
\usage{
getNames(object)

\S4method{getNames}{pmx_list}(object)
}
\arguments{
\item{object}{list object}
}
\value{
character vector with all the element names of this list
}
\description{
Get element names from list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_property.R
\docType{class}
\name{compartment_property-class}
\alias{compartment_property-class}
\title{Compartment property class.}
\description{
Compartment property class.
}
\section{Slots}{

\describe{
\item{\code{compartment}}{related compartment index}

\item{\code{rhs}}{right-hand side formula}

\item{\code{comment}}{comment if any, single character string}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{extractTextBetweenBrackets}
\alias{extractTextBetweenBrackets}
\title{Extract text between brackets.}
\usage{
extractTextBetweenBrackets(x)
}
\arguments{
\item{x}{character value}
}
\value{
text between brackets (trimmed)
}
\description{
Extract text between brackets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_lag_time.R
\name{LagTime}
\alias{LagTime}
\title{Create a lag time for the specified compartment.}
\usage{
LagTime(compartment, rhs = "")
}
\arguments{
\item{compartment}{compartment index}

\item{rhs}{right-hand side part of the equation}
}
\value{
a lag time property
}
\description{
Create a lag time for the specified compartment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameter.R, R/parameters.R,
%   R/campsis_model.R
\name{standardise}
\alias{standardise}
\alias{standardise,theta-method}
\alias{standardise,double_array_parameter-method}
\alias{standardise,parameters-method}
\alias{standardise,campsis_model-method}
\title{Standardise.}
\usage{
standardise(object, ...)

\S4method{standardise}{theta}(object, ...)

\S4method{standardise}{double_array_parameter}(object, parameters = NULL, ...)

\S4method{standardise}{parameters}(object, ...)

\S4method{standardise}{campsis_model}(object, ...)
}
\arguments{
\item{object}{generic object}

\item{...}{extra arguments needed for standardisation}

\item{parameters}{the list of parameters, to be provided only if parameter type is 'cor'}
}
\value{
standardised object
}
\description{
Standardise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replication_settings.R
\docType{class}
\name{manual_replication_settings-class}
\alias{manual_replication_settings-class}
\title{Manual replication settings class.}
\description{
Manual replication settings class.
}
\section{Slots}{

\describe{
\item{\code{replicated_parameters}}{data frame, 1 row per replicate}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_add_suffix.R
\name{addSuffix}
\alias{addSuffix}
\alias{addSuffix,parameters,character,character-method}
\alias{addSuffix,code_records,character,character-method}
\alias{addSuffix,code_record,character,character-method}
\alias{addSuffix,compartments,character,character-method}
\alias{addSuffix,campsis_model,character,character-method}
\title{Generic function to add a suffix to various objects like parameters, code records,
compartment names or a model (all previous objects at the same time).
This makes it an extremely powerful function to combine 2 models or more (using function 'add'),
that have similar equation, parameter or compartment names.}
\usage{
addSuffix(object, suffix, separator = NULL, ...)

\S4method{addSuffix}{parameters,character,character}(object, suffix, separator = NULL, ...)

\S4method{addSuffix}{code_records,character,character}(object, suffix, separator = NULL, ...)

\S4method{addSuffix}{code_record,character,character}(object, suffix, separator = NULL, ...)

\S4method{addSuffix}{compartments,character,character}(object, suffix, separator = NULL, ...)

\S4method{addSuffix}{campsis_model,character,character}(object, suffix, separator = NULL, ...)
}
\arguments{
\item{object}{generic object}

\item{suffix}{suffix to be appended, single character value}

\item{separator}{separator to use before the suffix, default is the underscore}

\item{...}{extra arguments like 'model' if the changes need to be reflected in the model}
}
\value{
updated object of the same class as the provided object, unless 'model' was specified, in that case the model is returned
}
\description{
Generic function to add a suffix to various objects like parameters, code records,
compartment names or a model (all previous objects at the same time).
This makes it an extremely powerful function to combine 2 models or more (using function 'add'),
that have similar equation, parameter or compartment names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/code_record.R,
%   R/campsis_model.R
\name{contains}
\alias{contains}
\alias{contains,pmx_list,pmx_element-method}
\alias{contains,statements_record,model_statement-method}
\alias{contains,campsis_model,pmx_element-method}
\title{Check if an element exists in list.}
\usage{
contains(object, x)

\S4method{contains}{pmx_list,pmx_element}(object, x)

\S4method{contains}{statements_record,model_statement}(object, x)

\S4method{contains}{campsis_model,pmx_element}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element to check if exists}
}
\value{
logical value, TRUE or FALSE
}
\description{
Check if an element exists in list.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\docType{class}
\name{ode_record-class}
\alias{ode_record-class}
\title{ODE record class.}
\description{
ODE record class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{sampleFromInverseWishartCore}
\alias{sampleFromInverseWishartCore}
\title{Sample from scaled inverse Wishart distribution (core method).}
\usage{
sampleFromInverseWishartCore(n, df, mat, allColnames, wishartCorrection)
}
\arguments{
\item{n}{numbers of rows to sample}

\item{df}{degree of freedom}

\item{mat}{scaling matrix}

\item{allColnames}{all column names as they are going to appear when as.vector is called}

\item{wishartCorrection}{apply correction to the degrees of freedom, logical}
}
\value{
a data frame with the sampled parameters
}
\description{
Sample from scaled inverse Wishart distribution (core method).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/campsis_model.R
\name{autoDetectNONMEM}
\alias{autoDetectNONMEM}
\alias{autoDetectNONMEM,campsis_model-method}
\title{Auto-detect special variables from NONMEM as compartment properties.
Bioavailabilities, infusion durations/rates and lag times will be automatically
detected.}
\usage{
autoDetectNONMEM(object, ...)

\S4method{autoDetectNONMEM}{campsis_model}(object, ...)
}
\arguments{
\item{object}{object that has NONMEM special variables to be identified}

\item{...}{extra arguments, unused}
}
\value{
updated object
}
\description{
Auto-detect special variables from NONMEM as compartment properties.
Bioavailabilities, infusion durations/rates and lag times will be automatically
detected.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode_conversion.R
\name{rxodeCode}
\alias{rxodeCode}
\title{Get code for rxode2}
\usage{
rxodeCode(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
corresponding model code for rxode2
}
\description{
Get code for rxode2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_bioavailability.R
\docType{class}
\name{compartment_bioavailability-class}
\alias{compartment_bioavailability-class}
\title{Compartment bioavailability class.}
\description{
Compartment bioavailability class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\docType{class}
\name{omega-class}
\alias{omega-class}
\title{Omega parameter class.}
\description{
Omega parameter class.
}
\section{Slots}{

\describe{
\item{\code{same}}{logical value, tell if this omega is the same as the previous one}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_initial_condition.R
\docType{class}
\name{compartment_initial_condition-class}
\alias{compartment_initial_condition-class}
\title{Compartment initial condition class.}
\description{
Compartment initial condition class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\docType{class}
\name{sigma-class}
\alias{sigma-class}
\title{Sigma parameter class.}
\description{
Sigma parameter class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{getNameInModel}
\alias{getNameInModel}
\alias{getNameInModel,theta-method}
\alias{getNameInModel,omega-method}
\alias{getNameInModel,sigma-method}
\title{Get the name of the given parameter in the Campsis model.}
\usage{
getNameInModel(x)

\S4method{getNameInModel}{theta}(x)

\S4method{getNameInModel}{omega}(x)

\S4method{getNameInModel}{sigma}(x)
}
\arguments{
\item{x}{element to know the name}
}
\value{
the name of this parameter
}
\description{
Get the name of the given parameter in the Campsis model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replicated_campsis_model.R
\name{updateOMEGAs}
\alias{updateOMEGAs}
\title{Update OMEGAs that are same. Same OMEGAs are written as follows:
OMEGA1 same is FALSE (first one, estimated)
OMEGA2 same is TRUE
OMEGA3 same is TRUE, etc.
OMEGA2 and OMEGA3 will take the same value as OMEGA1.}
\usage{
updateOMEGAs(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
updated Campsis model
}
\description{
Update OMEGAs that are same. Same OMEGAs are written as follows:
OMEGA1 same is FALSE (first one, estimated)
OMEGA2 same is TRUE
OMEGA3 same is TRUE, etc.
OMEGA2 and OMEGA3 will take the same value as OMEGA1.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameters.R, R/campsis_model.R
\name{disable}
\alias{disable}
\alias{disable,parameters,character-method}
\alias{disable,campsis_model,character-method}
\title{Disable.}
\usage{
disable(object, x, ...)

\S4method{disable}{parameters,character}(object, x, ...)

\S4method{disable}{campsis_model,character}(object, x, ...)
}
\arguments{
\item{object}{generic object}

\item{x}{what needs to be disabled}

\item{...}{extra arguments needed for disabling}
}
\value{
object with some disabled features
}
\description{
Disable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_ode.R
\name{Ode}
\alias{Ode}
\title{Create a new ordinary differential equation (ODE).}
\usage{
Ode(lhs, rhs = "", comment = as.character(NA))
}
\arguments{
\item{lhs}{left-hand side variable corresponding to derivative name, must start with 'A_'}

\item{rhs}{right-hand side expression corresponding to derivative value}

\item{comment}{comment if any, single character string}
}
\value{
an ODE
}
\description{
Create a new ordinary differential equation (ODE).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_statement.R
\docType{class}
\name{model_statement-class}
\alias{model_statement-class}
\title{Model statement class. Any statement in a code record.}
\description{
Model statement class. Any statement in a code record.
}
\section{Slots}{

\describe{
\item{\code{comment}}{a comment associated to this model statement}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{variablePatternStr}
\alias{variablePatternStr}
\title{Return the variable pattern (string form).}
\usage{
variablePatternStr()
}
\value{
pattern (regular expression)
}
\description{
Return the variable pattern (string form).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{sampleFromMultivariateNormalDistributionCore}
\alias{sampleFromMultivariateNormalDistributionCore}
\title{Sample from multivariate normal distribution (core method).}
\usage{
sampleFromMultivariateNormalDistributionCore(n, mean, varcov)
}
\arguments{
\item{n}{numbers of rows to sample}

\item{mean}{mean values to give to the multivariate normal distribution}

\item{varcov}{variance-covariance matrix}
}
\value{
a data frame with the sampled parameters
}
\description{
Sample from multivariate normal distribution (core method).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{appendParameters}
\alias{appendParameters}
\title{Append parameters.}
\usage{
appendParameters(params1, params2)
}
\arguments{
\item{params1}{base set of parameters}

\item{params2}{extra set of parameters to be appended}
}
\value{
the resulting set of parameters
}
\description{
Append parameters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{minMaxDefault}
\alias{minMaxDefault}
\title{Min/max default values for the given parameter.}
\usage{
minMaxDefault(parameter)
}
\arguments{
\item{parameter}{Campsis parameter}
}
\value{
a tibble with the min and max values
}
\description{
Min/max default values for the given parameter.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\docType{class}
\name{theta-class}
\alias{theta-class}
\title{Theta parameter class.}
\description{
Theta parameter class.
}
\section{Slots}{

\describe{
\item{\code{unit}}{parameter unit}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_unknown_statement.R
\name{UnknownStatement}
\alias{UnknownStatement}
\title{Create a new ordinary differential equation (ODE).}
\usage{
UnknownStatement(line, comment = as.character(NA))
}
\arguments{
\item{line}{line which was not recognised}

\item{comment}{comment if any, single character string}
}
\value{
an unknown statement
}
\description{
Create a new ordinary differential equation (ODE).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{read.varcov}
\alias{read.varcov}
\title{Read variance-covariance file.}
\usage{
read.varcov(file)
}
\arguments{
\item{file}{path to CSV file}
}
\value{
variance-covariance matrix
}
\description{
Read variance-covariance file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{maxIndex}
\alias{maxIndex}
\alias{maxIndex,parameters-method}
\title{Max index.}
\usage{
maxIndex(object)

\S4method{maxIndex}{parameters}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
max index
}
\description{
Max index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\docType{class}
\name{single_array_parameter-class}
\alias{single_array_parameter-class}
\title{Single-array parameter class. This parameter has a single index value.}
\description{
Single-array parameter class. This parameter has a single index value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{getNONMEMName}
\alias{getNONMEMName}
\alias{getNONMEMName,theta-method}
\alias{getNONMEMName,omega-method}
\alias{getNONMEMName,sigma-method}
\title{Get NONMEM name.}
\usage{
getNONMEMName(object)

\S4method{getNONMEMName}{theta}(object)

\S4method{getNONMEMName}{omega}(object)

\S4method{getNONMEMName}{sigma}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
the NONMEM name associated with this object
}
\description{
Get NONMEM name.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment_infusion_duration.R
\docType{class}
\name{compartment_infusion_duration-class}
\alias{compartment_infusion_duration-class}
\title{Compartment infusion duration class.}
\description{
Compartment infusion duration class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{getRecordEquationNames}
\alias{getRecordEquationNames}
\title{Get record equation names}
\usage{
getRecordEquationNames(record)
}
\arguments{
\item{record}{any code record}
}
\value{
a character vector with the equation names
}
\description{
Get record equation names
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_element_list.R
\docType{class}
\name{pmx_element-class}
\alias{pmx_element-class}
\title{PMX element class.}
\description{
PMX element class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter_uncertainty_utils.R
\name{checkMatrixIsPositiveDefinite}
\alias{checkMatrixIsPositiveDefinite}
\title{Check OMEGA/SIGMA matrix for positive definiteness.}
\usage{
checkMatrixIsPositiveDefinite(table, parameters)
}
\arguments{
\item{table}{data frame with the sampled parameters to check}

\item{parameters}{double array parameters to check for positive definiteness}
}
\description{
Check OMEGA/SIGMA matrix for positive definiteness.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{mapJSONPropertiesToS4Slots}
\alias{mapJSONPropertiesToS4Slots}
\title{Map JSON properties to S4 slots.}
\usage{
mapJSONPropertiesToS4Slots(object, json, discard_type = TRUE)
}
\arguments{
\item{object}{S4 object}

\item{json}{JSON element, json_element class}

\item{discard_type}{discard JSON property 'type'}
}
\value{
a S4 object
}
\description{
Map JSON properties to S4 slots.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\name{CampsisModel}
\alias{CampsisModel}
\title{Create a new Campsis model.}
\usage{
CampsisModel(json = NULL)
}
\arguments{
\item{json}{path to JSON Campsis model file or JSON content in string form}
}
\value{
a Campsis model, empty
}
\description{
Create a new Campsis model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_parser.R
\name{parseIfStatement}
\alias{parseIfStatement}
\title{Parse IF-statement.
Assumption: \code{isIfStatement} method already called and returned TRUE.}
\usage{
parseIfStatement(line, comment = as.character(NA))
}
\arguments{
\item{line}{IF-statement as single character string value, comment omitted}

\item{comment}{any comment, NA by default}
}
\value{
an IF statement object
}
\description{
Parse IF-statement.
Assumption: \code{isIfStatement} method already called and returned TRUE.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\docType{class}
\name{parameters-class}
\alias{parameters-class}
\title{Parameters class.}
\description{
Parameters class.
}
\section{Slots}{

\describe{
\item{\code{varcov}}{associated variance-covariance matrix}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveCapture}
\alias{mrgsolveCapture}
\title{Get the CAPTURE block for mrgsolve.}
\usage{
mrgsolveCapture(outvars, model)
}
\arguments{
\item{outvars}{outvars in method simulate}

\item{model}{Campsis model}
}
\value{
CAPTURE block or character(0) if no variable in outvars
}
\description{
Get the CAPTURE block for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{isComment}
\alias{isComment}
\title{Check if string is a Campsis comment (i.e. not an equation).}
\usage{
isComment(x)
}
\arguments{
\item{x}{character vector}
}
\value{
logical value
}
\description{
Check if string is a Campsis comment (i.e. not an equation).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{read.model}
\alias{read.model}
\title{Read model file.}
\usage{
read.model(file = NULL, text = NULL)
}
\arguments{
\item{file}{path to file 'model.campsis'}

\item{text}{model file as text, character (single or multiple lines)}
}
\value{
records object
}
\description{
Read model file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{isODE}
\alias{isODE}
\title{Say if line(s) in record is/are ODE or not.}
\usage{
isODE(x)
}
\arguments{
\item{x}{character vector}
}
\value{
logical vector
}
\description{
Say if line(s) in record is/are ODE or not.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/compartment_properties.R,
%   R/compartments.R, R/parameters.R, R/code_records.R, R/campsis_model.R,
%   R/omega_blocks.R
\name{sort}
\alias{sort}
\alias{sort,compartment_properties-method}
\alias{sort,compartments-method}
\alias{sort,parameters-method}
\alias{sort,code_records-method}
\alias{sort,campsis_model-method}
\alias{sort,omega_blocks-method}
\title{Sort the specified list.}
\usage{
sort(x, decreasing = FALSE, ...)

\S4method{sort}{compartment_properties}(x, decreasing = FALSE, ...)

\S4method{sort}{compartments}(x, decreasing = FALSE, ...)

\S4method{sort}{parameters}(x, decreasing = FALSE, ...)

\S4method{sort}{code_records}(x, decreasing = FALSE, ...)

\S4method{sort}{campsis_model}(x, decreasing = FALSE, ...)

\S4method{sort}{omega_blocks}(x, decreasing = FALSE, ...)
}
\arguments{
\item{x}{list object}

\item{decreasing}{increasing or decreasing order}

\item{...}{extra arguments}
}
\value{
same list but ordered
}
\description{
Sort the specified list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{getRecordDelimiter}
\alias{getRecordDelimiter}
\title{Get record delimiter.}
\usage{
getRecordDelimiter(line)
}
\arguments{
\item{line}{any line, single character value}
}
\value{
the record delimiter between brackets
}
\description{
Get record delimiter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\name{read.pmxmod}
\alias{read.pmxmod}
\title{Read a Campsis model (deprecated).}
\usage{
read.pmxmod(file)
}
\arguments{
\item{file}{path to folder}
}
\value{
a Campsis model
}
\description{
Read a Campsis model (deprecated).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omega_block.R
\name{hasOffDiagonalOmegas}
\alias{hasOffDiagonalOmegas}
\alias{hasOffDiagonalOmegas,omega_block-method}
\title{Has off-diagonal omegas.}
\usage{
hasOffDiagonalOmegas(object)

\S4method{hasOffDiagonalOmegas}{omega_block}(object)
}
\arguments{
\item{object}{omega block}
}
\value{
TRUE or FALSE
}
\description{
Has off-diagonal omegas.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\docType{class}
\name{parameter-class}
\alias{parameter-class}
\title{Parameter class. Any parameter in a pharmacometric model.}
\description{
Parameter class. Any parameter in a pharmacometric model.
}
\section{Slots}{

\describe{
\item{\code{name}}{parameter name, optional (although recommended)}

\item{\code{index}}{parameter index, integer}

\item{\code{value}}{parameter value (e.g. the estimated value from a modelling tool)}

\item{\code{min}}{minimum value for this parameter when parameter uncertainty is enabled}

\item{\code{max}}{maximum value for this parameter when parameter uncertainty is enabled}

\item{\code{fix}}{logical value, say if parameter was fixed in the modelling phase}

\item{\code{label}}{parameter label, any string}

\item{\code{comment}}{any comment on this parameter, any string}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveMatrix}
\alias{mrgsolveMatrix}
\title{Get the OMEGA/SIGMA matrix for mrgsolve.}
\usage{
mrgsolveMatrix(model, type = "omega")
}
\arguments{
\item{model}{Campsis model}

\item{type}{either omega or sigma}
}
\value{
named matrix or character(0) if matrix is empty
}
\description{
Get the OMEGA/SIGMA matrix for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_record.R
\name{MainRecord}
\alias{MainRecord}
\title{Create MAIN code record.}
\usage{
MainRecord(code = character())
}
\arguments{
\item{code}{code record}
}
\description{
Create MAIN code record.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\docType{class}
\name{double_array_parameter-class}
\alias{double_array_parameter-class}
\title{Double-array parameter class. This parameter has 2 indexes. 
It can thus be used to define correlations between parameters.}
\description{
Double-array parameter class. This parameter has 2 indexes. 
It can thus be used to define correlations between parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/compartments.R,
%   R/parameters.R, R/code_record.R, R/code_records.R, R/campsis_model.R
\name{delete}
\alias{delete}
\alias{delete,pmx_list,pmx_element-method}
\alias{delete,pmx_list,integer-method}
\alias{delete,compartments,compartment_property-method}
\alias{delete,parameters,single_array_parameter-method}
\alias{delete,parameters,double_array_parameter-method}
\alias{delete,statements_record,model_statement-method}
\alias{delete,statements_record,integer-method}
\alias{delete,code_records,model_statement-method}
\alias{delete,campsis_model,compartment_property-method}
\alias{delete,campsis_model,parameter-method}
\alias{delete,campsis_model,code_record-method}
\alias{delete,campsis_model,model_statement-method}
\title{Delete an element from this list.}
\usage{
delete(object, x)

\S4method{delete}{pmx_list,pmx_element}(object, x)

\S4method{delete}{pmx_list,integer}(object, x)

\S4method{delete}{compartments,compartment_property}(object, x)

\S4method{delete}{parameters,single_array_parameter}(object, x)

\S4method{delete}{parameters,double_array_parameter}(object, x)

\S4method{delete}{statements_record,model_statement}(object, x)

\S4method{delete}{statements_record,integer}(object, x)

\S4method{delete}{code_records,model_statement}(object, x)

\S4method{delete}{campsis_model,compartment_property}(object, x)

\S4method{delete}{campsis_model,parameter}(object, x)

\S4method{delete}{campsis_model,code_record}(object, x)

\S4method{delete}{campsis_model,model_statement}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element to delete or element index}
}
\value{
the updated list
}
\description{
Delete an element from this list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_statement.R
\name{appendComment}
\alias{appendComment}
\title{Append comment.}
\usage{
appendComment(str, object, dest)
}
\arguments{
\item{str}{single character string}

\item{object}{model statement}

\item{dest}{destination engine, string}
}
\value{
single character string followed by comment
}
\description{
Append comment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveBlock}
\alias{mrgsolveBlock}
\title{Convert code record for mrgsolve.}
\usage{
mrgsolveBlock(record, init = NULL, capture = FALSE)
}
\arguments{
\item{record}{code record}

\item{init}{name of mrgsolve block}

\item{capture}{'capture' instead of 'double'}
}
\value{
translated record for mrgsolve
}
\description{
Convert code record for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\name{appendModel}
\alias{appendModel}
\title{Append model (or simply add).}
\usage{
appendModel(model1, model2)
}
\arguments{
\item{model1}{base model}

\item{model2}{model to append}
}
\value{
the resulting Campsis model
}
\description{
Append model (or simply add).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_element_position.R
\docType{class}
\name{pmx_position-class}
\alias{pmx_position-class}
\title{PMX position class.}
\description{
PMX position class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/replicated_campsis_model.R
\name{replicate}
\alias{replicate}
\alias{replicate,campsis_model,integer,auto_replication_settings-method}
\alias{replicate,campsis_model,integer,manual_replication_settings-method}
\title{Replicate generic object.}
\usage{
replicate(object, n, settings = NULL, ...)

\S4method{replicate}{campsis_model,integer,auto_replication_settings}(object, n, settings)

\S4method{replicate}{campsis_model,integer,manual_replication_settings}(object, n, settings)
}
\arguments{
\item{object}{generic object}

\item{n}{number of replicates required}

\item{settings}{settings for replication}

\item{...}{extra arguments}
}
\value{
object replicated n times
}
\description{
Replicate generic object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartment.R
\name{Compartment}
\alias{Compartment}
\title{Create a compartment.}
\usage{
Compartment(index, name = NA)
}
\arguments{
\item{index}{compartment index}

\item{name}{compartment name (without prefix)}
}
\value{
an empty list of compartments
}
\description{
Create a compartment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{addODECompartment}
\alias{addODECompartment}
\title{Add ODE compartment to compartments object.}
\usage{
addODECompartment(compartments, ode)
}
\arguments{
\item{compartments}{compartments object}

\item{ode}{ODE}
}
\value{
a compartments object
}
\description{
Add ODE compartment to compartments object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_statements.R
\docType{class}
\name{model_statements-class}
\alias{model_statements-class}
\title{Model statements class. A list of statements.}
\description{
Model statements class. A list of statements.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/code_records.R, R/campsis_model.R
\name{move}
\alias{move}
\alias{move,code_records,model_statement,pmx_position-method}
\alias{move,code_records,list,pmx_position-method}
\alias{move,code_records,model_statements,pmx_position-method}
\alias{move,code_records,code_record,pmx_position-method}
\alias{move,campsis_model,ANY,pmx_position-method}
\title{Move element 'x' from object to a certain place.}
\usage{
move(object, x, to, ...)

\S4method{move}{code_records,model_statement,pmx_position}(object, x, to, ...)

\S4method{move}{code_records,list,pmx_position}(object, x, to, ...)

\S4method{move}{code_records,model_statements,pmx_position}(object, x, to, ...)

\S4method{move}{code_records,code_record,pmx_position}(object, x, to, ...)

\S4method{move}{campsis_model,ANY,pmx_position}(object, x, to, ...)
}
\arguments{
\item{object}{generic object (e.g. model, code records, etc.)}

\item{x}{element to move}

\item{to}{destination (e.g. a position)}

\item{...}{extra arguments, unused}
}
\value{
updated object
}
\description{
Move element 'x' from object to a certain place.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pattern.R
\docType{class}
\name{variable_pattern-class}
\alias{variable_pattern-class}
\title{Variable pattern class.}
\description{
Variable pattern class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R
\docType{class}
\name{export_type-class}
\alias{export_type-class}
\title{Export type class.}
\description{
Export type class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/compartment_properties.R,
%   R/parameters.R
\name{getByIndex}
\alias{getByIndex}
\alias{getByIndex,pmx_list,integer-method}
\alias{getByIndex,pmx_list,numeric-method}
\alias{getByIndex,compartment_properties,compartment_property-method}
\alias{getByIndex,parameters,parameter-method}
\title{Get element by index.}
\usage{
getByIndex(object, x)

\S4method{getByIndex}{pmx_list,integer}(object, x)

\S4method{getByIndex}{pmx_list,numeric}(object, x)

\S4method{getByIndex}{compartment_properties,compartment_property}(object, x)

\S4method{getByIndex}{parameters,parameter}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element index}
}
\value{
element from the list whose index matches with provided index
}
\description{
Get element by index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{variablePatternNoStartStr}
\alias{variablePatternNoStartStr}
\title{Return the variable pattern (string form), without the first character.}
\usage{
variablePatternNoStartStr()
}
\value{
pattern (regular expression)
}
\description{
Return the variable pattern (string form), without the first character.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{trim}
\alias{trim}
\title{Trim character vector. Remove all leading and trailing spaces.}
\usage{
trim(x)
}
\arguments{
\item{x}{character vector}
}
\value{
character vector without leading and trailing spaces
}
\description{
Trim character vector. Remove all leading and trailing spaces.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omega_block.R
\name{OmegaBlock}
\alias{OmegaBlock}
\title{Create a block of OMEGA's.}
\usage{
OmegaBlock()
}
\description{
Create a block of OMEGA's.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/pattern.R, R/model_statement.R,
%   R/model_unknown_statement.R, R/model_equation.R, R/model_if_statement.R,
%   R/compartment_property.R, R/code_record.R, R/code_records.R,
%   R/campsis_model.R
\name{replaceAll}
\alias{replaceAll}
\alias{replaceAll,character,variable_pattern,character-method}
\alias{replaceAll,character,pattern,character-method}
\alias{replaceAll,model_statement,pattern,character-method}
\alias{replaceAll,unknown_statement,pattern,character-method}
\alias{replaceAll,equation,pattern,character-method}
\alias{replaceAll,if_statement,pattern,character-method}
\alias{replaceAll,compartment_property,pattern,character-method}
\alias{replaceAll,code_record,pattern,character-method}
\alias{replaceAll,code_records,pattern,character-method}
\alias{replaceAll,campsis_model,pattern,character-method}
\alias{replaceAll,campsis_model,character,character-method}
\title{Replace all occurrences in object.}
\usage{
replaceAll(object, pattern, replacement, ...)

\S4method{replaceAll}{character,variable_pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{character,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{model_statement,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{unknown_statement,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{equation,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{if_statement,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{compartment_property,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{code_record,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{code_records,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{campsis_model,pattern,character}(object, pattern, replacement, ...)

\S4method{replaceAll}{campsis_model,character,character}(object, pattern, replacement, ...)
}
\arguments{
\item{object}{generic object (e.g. model, code_record(s), etc.)}

\item{pattern}{pattern to be replaced}

\item{replacement}{replacement string}

\item{...}{extra arguments}
}
\value{
the same object with all occurrences replaced
}
\description{
Replace all occurrences in object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_element_list.R,
%   R/model_unknown_statement.R, R/model_comment.R, R/model_line_break.R,
%   R/model_equation.R, R/model_ode.R, R/model_if_statement.R,
%   R/compartment.R, R/compartment_bioavailability.R, R/compartment_lag_time.R,
%   R/compartment_infusion_duration.R, R/compartment_infusion_rate.R,
%   R/compartment_initial_condition.R, R/parameter.R, R/code_record.R,
%   R/omega_block.R
\name{getName}
\alias{getName}
\alias{getName,unknown_statement-method}
\alias{getName,comment-method}
\alias{getName,line_break-method}
\alias{getName,equation-method}
\alias{getName,ode-method}
\alias{getName,if_statement-method}
\alias{getName,compartment-method}
\alias{getName,compartment_bioavailability-method}
\alias{getName,compartment_lag_time-method}
\alias{getName,compartment_infusion_duration-method}
\alias{getName,compartment_infusion_rate-method}
\alias{getName,compartment_initial_condition-method}
\alias{getName,theta-method}
\alias{getName,omega-method}
\alias{getName,sigma-method}
\alias{getName,main_record-method}
\alias{getName,ode_record-method}
\alias{getName,f_record-method}
\alias{getName,lag_record-method}
\alias{getName,duration_record-method}
\alias{getName,rate_record-method}
\alias{getName,init_record-method}
\alias{getName,error_record-method}
\alias{getName,omega_block-method}
\title{Get element name.}
\usage{
getName(x)

\S4method{getName}{unknown_statement}(x)

\S4method{getName}{comment}(x)

\S4method{getName}{line_break}(x)

\S4method{getName}{equation}(x)

\S4method{getName}{ode}(x)

\S4method{getName}{if_statement}(x)

\S4method{getName}{compartment}(x)

\S4method{getName}{compartment_bioavailability}(x)

\S4method{getName}{compartment_lag_time}(x)

\S4method{getName}{compartment_infusion_duration}(x)

\S4method{getName}{compartment_infusion_rate}(x)

\S4method{getName}{compartment_initial_condition}(x)

\S4method{getName}{theta}(x)

\S4method{getName}{omega}(x)

\S4method{getName}{sigma}(x)

\S4method{getName}{main_record}(x)

\S4method{getName}{ode_record}(x)

\S4method{getName}{f_record}(x)

\S4method{getName}{lag_record}(x)

\S4method{getName}{duration_record}(x)

\S4method{getName}{rate_record}(x)

\S4method{getName}{init_record}(x)

\S4method{getName}{error_record}(x)

\S4method{getName}{omega_block}(x)
}
\arguments{
\item{x}{element to know the name}
}
\value{
the name of this element
}
\description{
Get element name.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{appendCodeRecords}
\alias{appendCodeRecords}
\title{Append code records}
\usage{
appendCodeRecords(records1, records2)
}
\arguments{
\item{records1}{base set of code records}

\item{records2}{extra set of code records to be appended}
}
\value{
the resulting set of code records
}
\description{
Append code records
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/code_record.R,
%   R/omega_block.R
\name{length,pmx_list-method}
\alias{length,pmx_list-method}
\alias{length,statements_record-method}
\alias{length,omega_block-method}
\title{Return the length of this list.}
\usage{
\S4method{length}{pmx_list}(x)

\S4method{length}{statements_record}(x)

\S4method{length}{omega_block}(x)
}
\arguments{
\item{x}{omega block}
}
\value{
the length of this list, integer value

a number
}
\description{
Return the length of this list.

Return the number of OMEGA's on the diagonal.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{getCampsismodOptions}
\alias{getCampsismodOptions}
\title{Get the Campsismod options (R options).}
\usage{
getCampsismodOptions()
}
\value{
global options for Campsismod
}
\description{
Get the Campsismod options (R options).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replicated_campsis_model.R
\docType{class}
\name{replicated_campsis_model-class}
\alias{replicated_campsis_model-class}
\title{Replicated Campsis model class.}
\description{
Replicated Campsis model class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{mrgsolveMain}
\alias{mrgsolveMain}
\title{Get the MAIN block for mrgsolve.}
\usage{
mrgsolveMain(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
MAIN block
}
\description{
Get the MAIN block for mrgsolve.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/special_operators.R
\name{:=}
\alias{:=}
\title{Colon-equals operator.}
\usage{
`:=`(x, y)
}
\description{
Colon-equals operator.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_element.R
\name{JSONElement}
\alias{JSONElement}
\title{Instantiate a JSON element.}
\usage{
JSONElement(x)
}
\arguments{
\item{x}{JSON object representation in R}
}
\value{
JSON element
}
\description{
Instantiate a JSON element.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_interface.R
\name{processJSONDoubleArrayParameter}
\alias{processJSONDoubleArrayParameter}
\title{Process JSON double array parameter.}
\usage{
processJSONDoubleArrayParameter(x)
}
\arguments{
\item{x}{JSON data, OMEGA or SIGMA parameter}
}
\value{
updated JSON data with updated 'name' field and removed 'name2' field
}
\description{
Process JSON double array parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameters.R, R/campsis_model.R
\name{getVarCov}
\alias{getVarCov}
\alias{getVarCov,parameters-method}
\alias{getVarCov,campsis_model-method}
\title{Get variance-covariance matrix.}
\usage{
getVarCov(object)

\S4method{getVarCov}{parameters}(object)

\S4method{getVarCov}{campsis_model}(object)
}
\arguments{
\item{object}{generic object}
}
\value{
a variance-covariance matrix (data frame) or NULL if no matrix present
}
\description{
Get variance-covariance matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_records.R
\name{CodeRecords}
\alias{CodeRecords}
\title{Create a list of code records.}
\usage{
CodeRecords()
}
\value{
an empty list of code records
}
\description{
Create a list of code records.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_element_position.R
\docType{class}
\name{pmx_position_by_element-class}
\alias{pmx_position_by_element-class}
\title{PMX position by element class.}
\description{
PMX position by element class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{as.data.frame}
\alias{as.data.frame}
\alias{as.data.frame,theta,character,logical-method}
\alias{as.data.frame,omega,character,logical-method}
\alias{as.data.frame,sigma,character,logical-method}
\title{As data frame method.}
\usage{
as.data.frame(x, row.names = NULL, optional = FALSE, ...)

\S4method{as.data.frame}{theta,character,logical}(x, row.names = NULL, optional = FALSE, ...)

\S4method{as.data.frame}{omega,character,logical}(x, row.names = NULL, optional = FALSE, ...)

\S4method{as.data.frame}{sigma,character,logical}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{generic object}

\item{row.names}{row names}

\item{optional}{optional}

\item{...}{extra arguments}
}
\value{
data frame
}
\description{
As data frame method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/campsis_model.R
\docType{class}
\name{campsis_model-class}
\alias{campsis_model-class}
\title{Campsis model class.}
\description{
Campsis model class.
}
\section{Slots}{

\describe{
\item{\code{model}}{a list of code records}

\item{\code{parameters}}{model parameters}

\item{\code{compartments}}{model compartments}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_line_break.R
\docType{class}
\name{line_break-class}
\alias{line_break-class}
\title{Line-break class. A linebreak in the model.}
\description{
Line-break class. A linebreak in the model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/compartments.R,
%   R/code_record.R, R/code_records.R, R/campsis_model.R
\name{find}
\alias{find}
\alias{find,pmx_list,pmx_element-method}
\alias{find,compartments,compartment_property-method}
\alias{find,statements_record,model_statement-method}
\alias{find,code_records,model_statement-method}
\alias{find,campsis_model,compartment-method}
\alias{find,campsis_model,compartment_property-method}
\alias{find,campsis_model,parameter-method}
\alias{find,campsis_model,code_record-method}
\alias{find,campsis_model,model_statement-method}
\title{Find an element in list.}
\usage{
find(object, x)

\S4method{find}{pmx_list,pmx_element}(object, x)

\S4method{find}{compartments,compartment_property}(object, x)

\S4method{find}{statements_record,model_statement}(object, x)

\S4method{find}{code_records,model_statement}(object, x)

\S4method{find}{campsis_model,compartment}(object, x)

\S4method{find}{campsis_model,compartment_property}(object, x)

\S4method{find}{campsis_model,parameter}(object, x)

\S4method{find}{campsis_model,code_record}(object, x)

\S4method{find}{campsis_model,model_statement}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element to find, only key slots need to be filled in}
}
\value{
the element from the list that has same name as x, or NULL if no element was found
}
\description{
Find an element in list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R
\name{read}
\alias{read}
\title{Generic read method to read data from a file or a folder.}
\usage{
read(file, ...)
}
\arguments{
\item{file}{path to the file or folder to be read}

\item{...}{extra arguments}
}
\value{
the object representation of the data contained in the file
}
\description{
Generic read method to read data from a file or a folder.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_statements.R
\name{ModelStatements}
\alias{ModelStatements}
\title{Create an empty list of model statements.}
\usage{
ModelStatements()
}
\value{
a model statements object
}
\description{
Create an empty list of model statements.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{writeParameters}
\alias{writeParameters}
\title{Write subset of parameters (theta, omega or sigma).}
\usage{
writeParameters(object, file, ...)
}
\arguments{
\item{object}{subset of parameters}

\item{file}{filename}

\item{...}{extra arguments, like \code{defaultDf} for empty parameters list}
}
\value{
TRUE if success
}
\description{
Write subset of parameters (theta, omega or sigma).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{extractRhs}
\alias{extractRhs}
\title{Extract right-hand-side expression.}
\usage{
extractRhs(x, split = "=")
}
\arguments{
\item{x}{character value}

\item{split}{character where to split}
}
\value{
right-hand side expression
}
\description{
Extract right-hand-side expression.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_comment.R
\name{Comment}
\alias{Comment}
\title{Create a new comment.}
\usage{
Comment(x)
}
\arguments{
\item{x}{comment, single character string}
}
\value{
a comment
}
\description{
Create a new comment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mrgsolve_conversion.R
\name{convertOutvarsToCapture}
\alias{convertOutvarsToCapture}
\title{Convert outvars argument to capture. Variables that are already in error block
will be discarded.}
\usage{
convertOutvarsToCapture(outvars, model)
}
\arguments{
\item{outvars}{character vector}

\item{model}{Campsis model}
}
\value{
all variables to capture
}
\description{
Convert outvars argument to capture. Variables that are already in error block
will be discarded.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R
\name{indexOf}
\alias{indexOf}
\alias{indexOf,pmx_list,pmx_element-method}
\title{Get the index of an element in list.}
\usage{
indexOf(object, x)

\S4method{indexOf}{pmx_list,pmx_element}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element to know the index}
}
\value{
index of this element
}
\description{
Get the index of an element in list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_list.R, R/compartments.R,
%   R/parameters.R, R/code_record.R, R/code_records.R, R/campsis_model.R,
%   R/omega_block.R, R/omega_blocks.R
\name{add}
\alias{add}
\alias{add,pmx_list,pmx_element-method}
\alias{add,pmx_list,pmx_list-method}
\alias{add,pmx_list,list-method}
\alias{add,compartments,compartment_property-method}
\alias{add,compartments,compartments-method}
\alias{add,parameters,single_array_parameter-method}
\alias{add,parameters,double_array_parameter-method}
\alias{add,parameters,parameters-method}
\alias{add,code_record,model_statement-method}
\alias{add,code_record,code_record-method}
\alias{add,code_records,code_records-method}
\alias{add,code_records,model_statement-method}
\alias{add,campsis_model,compartment_property-method}
\alias{add,campsis_model,parameter-method}
\alias{add,campsis_model,code_record-method}
\alias{add,campsis_model,model_statement-method}
\alias{add,campsis_model,campsis_model-method}
\alias{add,campsis_model,list-method}
\alias{add,omega_block,double_array_parameter-method}
\alias{add,omega_blocks,omega_block-method}
\alias{add,omega_blocks,parameters-method}
\title{Add element to list.}
\usage{
add(object, x, ...)

\S4method{add}{pmx_list,pmx_element}(object, x, pos = NULL)

\S4method{add}{pmx_list,pmx_list}(object, x)

\S4method{add}{pmx_list,list}(object, x)

\S4method{add}{compartments,compartment_property}(object, x)

\S4method{add}{compartments,compartments}(object, x)

\S4method{add}{parameters,single_array_parameter}(object, x)

\S4method{add}{parameters,double_array_parameter}(object, x)

\S4method{add}{parameters,parameters}(object, x)

\S4method{add}{code_record,model_statement}(object, x, pos = NULL)

\S4method{add}{code_record,code_record}(object, x)

\S4method{add}{code_records,code_records}(object, x)

\S4method{add}{code_records,model_statement}(object, x, pos = NULL)

\S4method{add}{campsis_model,compartment_property}(object, x)

\S4method{add}{campsis_model,parameter}(object, x)

\S4method{add}{campsis_model,code_record}(object, x)

\S4method{add}{campsis_model,model_statement}(object, x, pos = NULL)

\S4method{add}{campsis_model,campsis_model}(object, x)

\S4method{add}{campsis_model,list}(object, x)

\S4method{add}{omega_block,double_array_parameter}(object, x)

\S4method{add}{omega_blocks,omega_block}(object, x)

\S4method{add}{omega_blocks,parameters}(object, x)
}
\arguments{
\item{object}{list object}

\item{x}{element to add}

\item{...}{extra arguments, unused by this generic list}

\item{pos}{position where x needs to be added in list}
}
\value{
modified list object
}
\description{
Add element to list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameter.R
\name{Theta}
\alias{Theta}
\title{Create a THETA parameter.}
\usage{
Theta(
  name = NA,
  index = NA,
  value = NA,
  min = NA,
  max = NA,
  fix = FALSE,
  label = NA,
  unit = NA,
  comment = NA
)
}
\arguments{
\item{name}{parameter name, e.g. CL (prefix THETA will be added automatically)}

\item{index}{parameter index}

\item{value}{parameter value}

\item{min}{minimum value for this parameter when parameter uncertainty is enabled}

\item{max}{maximum value for this parameter when parameter uncertainty is enabled}

\item{fix}{parameter was fixed in estimation, logical value}

\item{label}{parameter label, optional}

\item{unit}{parameter unit, optional}

\item{comment}{any comment, optional}
}
\value{
a THETA parameter
}
\description{
Create a THETA parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replicated_campsis_model.R
\name{updateParameters}
\alias{updateParameters}
\title{Update model parameters based on the parameters issued from the model replication.}
\usage{
updateParameters(model, row)
}
\arguments{
\item{model}{Campsis model}

\item{row}{a data frame row containing the new parameter values}
}
\value{
updated Campsis model
}
\description{
Update model parameters based on the parameters issued from the model replication.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rxode_conversion.R
\name{rxodeParams}
\alias{rxodeParams}
\title{Get the parameters vector for rxode2.}
\usage{
rxodeParams(model)
}
\arguments{
\item{model}{Campsis model}
}
\value{
named vector with THETA values
}
\description{
Get the parameters vector for rxode2.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_comment.R
\docType{class}
\name{comment-class}
\alias{comment-class}
\title{Comment class. A statement starting with #.}
\description{
Comment class. A statement starting with #.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R, R/parameter.R, R/campsis_model.R
\name{loadFromJSON}
\alias{loadFromJSON}
\alias{loadFromJSON,theta,json_element-method}
\alias{loadFromJSON,omega,json_element-method}
\alias{loadFromJSON,sigma,json_element-method}
\alias{loadFromJSON,campsis_model,json_element-method}
\alias{loadFromJSON,campsis_model,character-method}
\title{Fill-in S4 object from the JSON content.}
\usage{
loadFromJSON(object, json)

\S4method{loadFromJSON}{theta,json_element}(object, json)

\S4method{loadFromJSON}{omega,json_element}(object, json)

\S4method{loadFromJSON}{sigma,json_element}(object, json)

\S4method{loadFromJSON}{campsis_model,json_element}(object, json)

\S4method{loadFromJSON}{campsis_model,character}(object, json)
}
\arguments{
\item{object}{pre-initiated S4 object}

\item{json}{JSON (usually a list)}
}
\value{
the loaded S4 object
}
\description{
Fill-in S4 object from the JSON content.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_if_statement.R
\name{IfStatement}
\alias{IfStatement}
\title{Create a new IF-statement.}
\usage{
IfStatement(condition, equation, comment = as.character(NA))
}
\arguments{
\item{condition}{condition, single character string}

\item{equation}{equation if condition is met}

\item{comment}{comment if any, single character string}
}
\value{
an IF-statement
}
\description{
Create a new IF-statement.
}
